(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Fuse operator\<close>   

theory Fuse
imports Semantics 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)
text \<open>
This theory introduces the fuse operator.
 \<close>

subsection \<open>Definitions\<close>

primrec fuse :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval"
 where  fuse_St   : "fuse \<langle>x\<rangle> ys = ys"
      | fuse_Cons : "fuse (x \<odot> xs) ys = x \<odot> (fuse xs ys)"

primrec lfuse :: "'a interval interval \<Rightarrow> 'a interval"
where   lfuse_St   : "lfuse \<langle>xs\<rangle> = xs"
      | lfuse_Cons : "lfuse (x \<odot> xs) = fuse x (lfuse xs)"


primrec lastfirst :: "'a interval interval \<Rightarrow> bool"
 where "lastfirst \<langle>xs\<rangle> = True"
      | "lastfirst (xs \<odot> xxs) = 
         ( ((intlast xs) = (intfirst (intfirst xxs))) \<and> (lastfirst xxs))"


subsection \<open>Lemmas\<close>

lemma interval_fuse_intlen :
  assumes  "intlast xs = intfirst ys"
  shows    " intlen (fuse xs ys) = (intlen xs) + (intlen ys)"
using assms by (induct xs) simp_all

lemma interval_fuse_intlen_a:
 "intlen(fuse xs ys) = intlen xs + intlen ys"
proof
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case by simp
qed

lemma interval_fuse_nth:
assumes "i\<le> intlen (fuse xs ys)"
        "intlast xs = intfirst ys"
shows   "(i \<le> intlen xs \<longrightarrow> nth (fuse xs ys) i = nth xs i)
         \<and> 
         (intlen xs \<le> i \<and> i\<le> intlen (fuse xs ys) \<longrightarrow> nth (fuse xs ys) i = nth ys (i - intlen xs))
        "
using assms
proof
  (induct xs arbitrary: i)
  case (St x)
  then show ?case by simp 
  next
  case (Cons x1a xs)
  then show ?case 
  using less_Suc_eq_0_disj less_Suc_eq_le by fastforce
qed

lemma interval_fuse_nth_a:
 assumes " j \<le> intlen ys"
         " intlast xs = intfirst ys "
 shows   " nth (fuse xs ys) (intlen xs +j) = (nth ys j)"
using assms 
by (simp add: interval_fuse_intlen_a interval_fuse_nth)

lemma interval_fuse_leftneutral : 
    "fuse (St (intfirst xs)) xs = xs"
by simp

lemma interval_fuse_rightneutral : 
    "fuse xs (St (intlast xs)) = xs"
by (induct xs) simp_all

lemma interval_intfirst_fuse :
 assumes " intlast xs = intfirst ys"
 shows   " intfirst (fuse xs ys) = intfirst xs"
using assms by (induct xs) simp_all

lemma interval_intlast_fuse :
 assumes "intlast xs = intfirst ys"
 shows   " intlast (fuse xs ys) = intlast ys"
using assms by (induct xs) simp_all

lemma interval_FusionAssoc :
 assumes " (intlast xs) = (intfirst ys)"
         " (intlast ys) = (intfirst zs)"
 shows   " (fuse xs (fuse ys zs)) = (fuse (fuse xs ys) zs)"
using assms by (induct xs) simp_all


lemma interval_intlast_intfirst:
    " (intlast (prefix i xs)) = (intfirst (suffix i xs))"
using interval_intlast_intfirst by blast
  
lemma interval_prefix_fuse :
  assumes  " intlast xs = intfirst ys "
  shows    " (prefix (intlen xs) (fuse xs ys)) = xs"
using assms by (induct xs arbitrary: ys) simp_all

lemma interval_suffix_fuse :
 assumes   "intlast xs = intfirst ys "
 shows     "(suffix (intlen xs) (fuse xs ys)) = ys"
using assms by (induct xs arbitrary: ys) simp_all
 
lemma interval_fuse_prefix_suffix_intlen :
assumes "n \<le> intlen xs"
shows   "intlen (fuse (prefix n xs) (suffix n xs)) = intlen xs "
using assms 
by (metis interval_fuse_intlen_a interval_prefix_length_good interval_suffix_length_good 
    le_add_diff_inverse)

lemma interval_fuse_prefix_suffix_nth :
assumes "n \<le> intlen xs"
        "i \<le> intlen xs"
shows   "   nth (fuse (prefix n xs) (suffix n xs)) i = nth xs i "
using assms 
by (metis interval_fuse_prefix_suffix_intlen interval_fuse_nth interval_intlast_intfirst 
          interval_nth_prefix interval_nth_suffix interval_prefix_length_good le_cases 
          nat_add_left_cancel_le ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma interval_fuse_prefix_suffix:
assumes "n\<le> intlen xs"
shows   "fuse (prefix n xs) (suffix n xs) = xs"
using assms 
by (simp add: interval_fuse_prefix_suffix_intlen interval_fuse_prefix_suffix_nth interval_eq_nth_eq)

lemma interval_chop_fuse_1 : 
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1  \<Turnstile> f) \<and> (\<sigma>2  \<Turnstile> g) \<and> 
       (intlast \<sigma>1 = intfirst \<sigma>2)) =
      (\<exists> i. 0 \<le> i \<and> i\<le> intlen \<sigma> \<and> ( prefix i \<sigma>  \<Turnstile> f) \<and> (suffix i \<sigma>  \<Turnstile> g)) "
by (metis interval_fuse_intlen interval_fuse_prefix_suffix interval_intlast_intfirst 
          interval_intlen_gr_zero interval_prefix_fuse interval_suffix_fuse le_add_same_cancel1)

lemma interval_chop_fuse_2 : 
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1 \<in> X) \<and> (\<sigma>2 \<in> Y) \<and> 
       (intlast \<sigma>1 = intfirst \<sigma>2)) =
      (\<exists> i\<le> intlen \<sigma>. (prefix i \<sigma>)\<in>X \<and> (suffix i \<sigma>)\<in>Y) "
by (metis interval_fuse_intlen interval_fuse_prefix_suffix interval_intlast_intfirst 
          interval_prefix_fuse interval_suffix_fuse le_add1)
  
lemma interval_chop_fuse:
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1  \<Turnstile> f) \<and> (\<sigma>2  \<Turnstile> g) \<and> 
       (intlast \<sigma>1 = intfirst \<sigma>2)) =
      (\<sigma>  \<Turnstile> f;g) "
by (metis chop_defs interval_fuse_intlen_a interval_fuse_prefix_suffix interval_intlast_intfirst 
          interval_prefix_fuse interval_suffix_fuse le_add1)

lemma interval_sub_fuse:
 assumes " k \<le> n "
         " n \<le> m"
         " m \<le> intlen xs"
 shows   "fuse (sub k n xs) (sub n m xs) = (sub k m xs)"
proof -
 have 1: " intlast(sub k n xs) = (nth xs n)"
    using assms interval_intlast_sub le_trans less_imp_le_nat by blast
 have 2: " intfirst(sub n m xs) = (nth xs n)"
    using assms interval_intfirst_sub less_imp_le_nat by blast
 have 3: " intlen(fuse (sub k n xs) (sub n m xs)) = intlen(sub k m xs)"
    by (metis Nat.add_diff_assoc2 assms(1) assms(2) assms(3) interval_fuse_intlen_a 
        interval_intlen_sub le_trans ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 4: " (\<forall> i. i \<le> intlen(sub k m xs) \<longrightarrow>
              (nth (fuse (sub k n xs) (sub n m xs)) i) = (nth xs (k+i)))"
    proof 
      fix i
      show "i \<le> intlen (sub k m xs) \<longrightarrow> nth (fuse (sub k n xs) (sub n m xs)) i = nth xs (k + i)  "
      proof -
       have 41: "intlen (sub k m xs) = (m-k) "
          using assms interval_intlen_sub le_trans less_imp_le_nat by metis
       have 42: "i \<le> intlen (sub k m xs) \<longrightarrow> nth (fuse (sub k n xs) (sub n m xs)) i =
                  (if i \<le> intlen(sub k n xs) then (nth (sub k n xs) i) 
                                             else (nth (sub n m xs) (i-intlen(sub k n xs))))"
       by (metis "1" "2" "3" interval_fuse_nth le_cases)
       have 43: " i \<le> (m-k) \<longrightarrow> nth (fuse (sub k n xs) (sub n m xs)) i = 
                  (if i\<le> (n-k) then (nth xs (k+i)) else (nth xs (n+(i-(n-k)))))"
            using "42" Nat.le_diff_conv2 assms(1) assms(2) assms(3) by auto
       have 44: "i \<le> (m-k) \<longrightarrow> nth (fuse (sub k n xs) (sub n m xs)) i = 
                  (nth xs (k+i))  "
              by (simp add: "43" add.commute assms less_imp_le_nat)
        show ?thesis 
        by (simp add: "41" "44")
     qed
   qed
  have 5: "(\<forall> i. i \<le> intlen(sub k m xs) \<longrightarrow> (nth (sub k m xs) i) = (nth xs (k+i))) "
      using assms(1) assms(2) assms(3) by auto
  show ?thesis 
  by (simp add: "3" "4" "5" interval_eq_nth_eq)
qed

lemma interval_sub_fuse_idx:
 assumes "index_sequence 0 l"
         " nth l (intlen l) = intlen \<sigma>"
         " (Suc i) < intlen l"
 shows   " fuse (sub (nth l i) (nth l (Suc i)) \<sigma>) (sub (nth l (Suc i)) (nth l (intlen l)) \<sigma>) = 
           (sub (nth l i) (nth l (intlen l)) \<sigma>)"
proof -
 have 1: " intlast(sub (nth l i) (nth l (Suc i)) \<sigma>) = (nth \<sigma> (nth l (Suc i)))"
    by (metis assms interval_idx_less_equal interval_idx_less_last_1 interval_intlast_sub le_add2 
         less_imp_le_nat plus_1_eq_Suc)
 have 2: " intfirst(sub (nth l (Suc i)) (nth l (intlen l)) \<sigma> ) = (nth \<sigma> (nth l (Suc i))) "
    by (metis assms dual_order.strict_iff_order eq_imp_le interval_idx_less_last_1 
      interval_intfirst_sub)
 have 3: " (nth l i) < (nth l (Suc i)) " 
    using Suc_lessD assms interval_idx_less_than less_imp_le_nat by blast
 have 4: "(nth l (Suc i)) < (nth l (intlen l)) " 
    using assms interval_idx_less_last_1 by blast
 show ?thesis using "3" "4" assms by (simp add: interval_sub_fuse)
qed


lemma interval_idx_fuse_intfirst_intlast:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "nth l1 (intlen l1) = cp"
         "nth l2 (intlen l2) = intlen \<sigma> -cp"
         "cp \<le> intlen \<sigma>"
         "l = fuse l1 (map (shift cp) l2) "
 shows   "intlast l1 = intfirst (map (shift cp) l2)"
using assms 
 by (metis Interval.shift_def add.left_neutral index_sequence_def 
          interval_nth_intlen_intlast interval_nth_map interval_nth_zero_intfirst)

lemma interval_idx_fuse_nth_cp:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2"
         "nth l1 (intlen l1) = cp"
         "nth l2 (intlen l2) = intlen \<sigma> -cp"
         "cp \<le> intlen \<sigma>"
         "l = fuse l1 (map (shift cp) l2)"
         "i \<le> intlen l2"
 shows   "nth l (intlen l1 + i) = cp + nth l2 i"
proof -
 have 1: "intlast l1 = intfirst (map (shift cp) l2)"
    using assms interval_idx_fuse_intfirst_intlast by blast
 have 2: "nth l (intlen l1 + i) = nth (map (shift cp) l2) i"
    using assms by (metis "1" interval_fuse_nth_a interval_intlen_map)
 have 3: "nth (map (shift cp) l2) i = nth l2 i +cp"
    by (simp add: Interval.shift_def interval_nth_map)
  show ?thesis using "2" "3" by auto 
qed


lemma interval_idx_fuse_idx:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "nth l1 (intlen l1) = cp"
         "nth l2 (intlen l2) = intlen \<sigma> -cp"
         "cp \<le> intlen \<sigma>"
         "l = fuse l1 (map (shift cp) l2)"
         "i \<le> intlen l2"
 shows   "index_sequence 0 l"
proof -
 have 1: "intlast l1 = intfirst (map (shift cp) l2)"
    using assms interval_idx_fuse_intfirst_intlast by blast
 have 2: " nth (fuse l1 (map (shift cp) l2)) 0 = nth l1 0"
     using "1" interval_fuse_nth by blast
 have 3: " intlen (fuse l1 (map (shift cp) l2)) = intlen l1 + intlen l2"
    by (simp add: interval_fuse_intlen_a)
 have 4: "\<forall>i. 0\<le>i  \<and> i\<le>intlen l1 \<longrightarrow> 
           nth (fuse l1 (map (shift cp) l2)) i  =
           nth l1 i"
    by (metis "1"  interval_nth_prefix interval_prefix_fuse)
 have 5: "\<forall>i. intlen l1 \<le>i  \<and> i\<le>intlen l1 + intlen l2 \<longrightarrow> 
           nth (fuse l1 (map (shift cp) l2)) i  =
           cp + nth l2 (i -intlen l1) " 
    by (metis (no_types, lifting) "1" "3" Interval.shift_def add.commute interval_fuse_nth 
        interval_nth_map)
 have 6: "\<forall>i. 0\<le>i  \<and> i<intlen l1+intlen l2 \<longrightarrow> 
            nth (fuse l1 (map (shift cp) l2)) i < nth (fuse l1 (map (shift cp) l2)) (Suc i)"    
    using assms by (metis "1" "3" "4" add.right_neutral index_sequence_def interval_idx_link 
        interval_idx_split interval_intlen_gr_zero interval_prefix_fuse interval_suffix_fuse 
        le_add1)
 have 7: "index_sequence 0 l =
          ((nth l1 0) = 0 \<and>
          (\<forall>i. 0\<le>i  \<and> i<intlen l1+intlen l2 \<longrightarrow> 
            nth (fuse l1 (map (shift cp) l2)) i < nth (fuse l1 (map (shift cp) l2)) (Suc i)))  "
    by (simp add: "2" "3" assms index_sequence_def)
 from 7 6 2 show ?thesis 
 using assms index_sequence_def by auto
qed


lemma interval_idx_fuse_intlen:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "nth l1 (intlen l1) = cp"
         "nth l2 (intlen l2) = intlen \<sigma> -cp"
         "cp \<le> intlen \<sigma>"
         "l = fuse l1 (map (shift cp) l2)"
         "i \<le> intlen l2"
 shows   "nth l (intlen l) = intlen \<sigma>"
using assms interval_idx_fuse_nth_cp[of "l1" "l2" "cp" "\<sigma>"] by (simp add: interval_fuse_intlen_a)

lemma interval_intfirst_lfuse_intfirst:
assumes "lastfirst (xs \<odot> xxs)"
shows  "intfirst(lfuse xxs) = intfirst(intfirst xxs)"
using assms 
proof 
 (induct xxs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xxs)
 then show ?case 
      proof (cases x1a)
      case (St x1)
      then show ?thesis using Cons.hyps Cons.prems by auto
      next
      case (Cons x21 x22)
      then show ?thesis by simp
      qed
qed 

lemma interval_intfirst_lfuse:
assumes "lastfirst (xs \<odot> xxs)"
shows "(intfirst (lfuse (xs \<odot> xxs)) ) = (intfirst xs)"
proof -
 have 1: "lastfirst (xs \<odot> xxs)"
  using assms by auto
 have 2: "( ((intlast xs) = (intfirst (intfirst xxs))) \<and> (lastfirst xxs))  "
   using "1" by simp
 have 3: "(intlast xs) = (intfirst (intfirst xxs))  "
   using "2" by auto
 have 4: "intfirst (lfuse (xs \<odot> xxs)) = intfirst(fuse xs (lfuse xxs))  "
   by simp
 have 5: "intfirst(lfuse xxs) = intfirst(intfirst xxs)"
   using assms interval_intfirst_lfuse_intfirst by blast
 have 6: "intfirst(fuse xs (lfuse xxs)) = intfirst xs "
   by (metis "3" "5" interval_intfirst_fuse)
 show ?thesis using "6" by auto 
qed 

lemma interval_lastfirst_lfuse_intlast:
 assumes "lastfirst xxs"
 shows   " intlast(lfuse xxs) = intlast(intlast xxs)"
using assms
 proof 
  (induct xxs)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xxs)
  then show ?case
  by (metis interval_fuse_intlen_a interval_fuse_nth_a interval_intfirst_lfuse_intfirst 
            interval_nth_intlen_intlast interval_nth_last lastfirst.simps(2) lfuse_Cons 
            order.order_iff_strict)
qed

lemma interval_lastfirst_lfuse:
assumes "lastfirst xxs"
shows "intfirst (lfuse xxs) = intfirst(intfirst(xxs))"
using assms 
proof
 (cases xxs)
 case (St x1)
 then show ?thesis by simp
 next
 case (Cons x21 x22)
 then show ?thesis 
 using assms interval_intfirst_lfuse by auto
 qed

lemma interval_lfuse_intlen :
 assumes " lastfirst xxs "
 shows   " intlen (lfuse xxs) = (\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k))"
using assms
proof
  (induct xxs)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xxs)
  then show ?case proof -
    have 1 : "intlen (lfuse (x1a \<odot> xxs)) = intlen (fuse x1a (lfuse xxs))" 
     by simp
    have 2 : "lastfirst (x1a \<odot> xxs)" using Cons.prems by auto
    have 3 : "intlast x1a = intfirst(lfuse xxs)" 
       using Cons.prems interval_intfirst_lfuse_intfirst by fastforce
    have 4 : "intlen (fuse x1a (lfuse xxs)) = (intlen x1a) + intlen(lfuse xxs)  "
     using "3" interval_fuse_intlen by blast
    have 5: "(intlen x1a) + intlen(lfuse xxs) =
             (intlen x1a) + (\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k))  "
     using Cons.hyps Cons.prems by auto
    have 6: "(\<Sum>k = 0..intlen (x1a \<odot> xxs). intlen (Interval.nth (x1a \<odot> xxs) k)) =
             (intlen(nth (x1a \<odot> xxs) 0)   ) + 
             (\<Sum>k = 1..1+intlen (xxs). intlen (Interval.nth (x1a \<odot> xxs) k)) "
        by (simp add: sum.atLeast_Suc_atMost)
    have 7: "(intlen(nth (x1a \<odot> xxs) 0)   ) = intlen(x1a) "
     by simp
    have 8: "(\<Sum>k = 1..1+intlen (xxs). intlen (Interval.nth (x1a \<odot> xxs) k)) =
             (\<Sum>k = 0..intlen (xxs). intlen (Interval.nth (x1a \<odot> xxs) (k+1))) "
     by (metis (mono_tags, lifting) Nat.add_0_right add.commute sum.cong sum.shift_bounds_cl_nat_ivl)
    have 9: "(\<Sum>k = 0..intlen (xxs). intlen (Interval.nth (x1a \<odot> xxs) (k+1))) =
              (\<Sum>k = 0..intlen (xxs). intlen (Interval.nth (xxs) (k))) "
     by auto
    have 10: "(intlen x1a) + (\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k)) =
             (\<Sum>k = 0..intlen (x1a \<odot> xxs). intlen (Interval.nth (x1a \<odot> xxs) k)) "
    using "6" "7" "8" "9" by linarith
    show ?thesis 
    by (simp add: "10" "4" "5")       
  qed
qed

lemma interval_idx_fuse:
 assumes "intlast l1 = intfirst l2" 
 shows  
 " (index_sequence (intfirst l1) (fuse l1 l2)) =
     ( index_sequence (intfirst l1) l1 \<and> index_sequence (intfirst l2) l2 ) "
using assms 
 proof
  (induct l1 arbitrary: l2)
  case (St x)
  then show ?case by (simp add: index_sequence_def)
  next
  case (Cons x1a l1)
  then show ?case 
   using interval_idx_expand1 interval_intfirst_fuse by force
qed

lemma interval_idx_lfuse_help1:
assumes " (\<forall>k. k<intlen (lfuse l) \<longrightarrow> 
       nth (fuse x1a (lfuse l)) (intlen x1a + k) < 
       nth (fuse x1a (lfuse l)) (intlen x1a + Suc k)) 
        "
        "intlen x1a \<le> n"
        "n< intlen x1a +intlen (lfuse l)"
shows " nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)"
using assms 
by (metis add_Suc_right add_less_imp_less_left le_Suc_ex)

lemma interval_idx_lfuse:
 assumes "lastfirst l"
 shows   "(index_sequence (intfirst (lfuse l)) (lfuse l)) = 
          (\<forall> i\<le>intlen l. index_sequence (intfirst (nth l i)) (nth l i))"
using assms 
 proof 
   (induct l)
   case (St x)
   then show ?case by (simp add: index_sequence_def)
   next
   case (Cons x1a l)
   then show ?case 
    proof -
     have 0: "lastfirst l" 
         using Cons.prems lastfirst.simps(2) by blast
     have 1: "index_sequence (intfirst (lfuse (x1a \<odot> l))) (lfuse (x1a \<odot> l)) =
              (nth (fuse x1a (lfuse l)) 0 = intfirst (fuse x1a (lfuse l)) \<and>
              (\<forall>n<intlen (fuse x1a (lfuse l)). 
                  nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n))) "
        by (simp add: index_sequence_def)
     have 2: "(nth (fuse x1a (lfuse l)) 0) = intfirst (fuse x1a (lfuse l)) "
        by simp
     have 3: "intlen (fuse x1a (lfuse l)) = intlen x1a + intlen(lfuse l) "
        by (simp add: interval_fuse_intlen_a)
     have 4: " (\<forall>n<intlen (fuse x1a (lfuse l)). 
                  nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)) =
               (\<forall>n<intlen x1a + intlen (lfuse l). 
                  nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n))"
        by (simp add: "3")
     have 5: "(\<forall>n<intlen x1a + intlen (lfuse l). 
                  nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)) =
              ((\<forall>n<intlen x1a. nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n))  \<and>
               (\<forall>n. 0 \<le> n-intlen x1a \<and> n - intlen x1a<intlen (lfuse l) \<longrightarrow> 
                       nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)))   "
         by auto
           (metis add.commute less_diff_conv2 not_less)
     have 6: "(\<forall>n<intlen x1a. nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)) =
              index_sequence (intfirst x1a) x1a"
          by (simp add: index_sequence_def)
             (metis Cons.prems Suc_leI interval_intfirst_lfuse_intfirst interval_nth_prefix 
              interval_prefix_fuse lastfirst.simps(2) le_simps(1))
     have 7: "(\<forall>n. intlen x1a \<le> n \<and> n < intlen x1a + intlen (lfuse l) \<longrightarrow> 
                       nth (fuse x1a (lfuse l)) n < nth (fuse x1a (lfuse l)) (Suc n)) =
              (\<forall>k.  k<intlen (lfuse l) \<longrightarrow> 
                       nth (fuse x1a (lfuse l)) (intlen x1a + k) < 
                       nth (fuse x1a (lfuse l)) (intlen x1a + Suc k)) "    
            using interval_idx_lfuse_help1 by auto
     have 8: "(\<forall>k.  k<intlen (lfuse l) \<longrightarrow> 
                       nth (fuse x1a (lfuse l)) (intlen x1a + k) < 
                       nth (fuse x1a (lfuse l)) (intlen x1a + Suc k)) =
              (\<forall>k.  k<intlen (lfuse l) \<longrightarrow> 
                       nth ( (lfuse l)) (k) < 
                       nth ( (lfuse l)) (Suc k)) "  (is "?L=?R")
            proof 
             show "?L \<Longrightarrow> ?R"
              by (metis Cons.prems Suc_leI add_Suc_right interval_fuse_nth_a
                  interval_intfirst_lfuse_intfirst lastfirst.simps(2) le_simps(1))
             show "?R \<Longrightarrow> ?L"
              by (metis Cons.prems Suc_leI add_Suc_right interval_fuse_nth_a 
                   interval_intfirst_lfuse_intfirst lastfirst.simps(2) less_imp_le_nat)      
           qed
     have 9: "(\<forall>k.  k<intlen (lfuse l) \<longrightarrow> 
                       nth ( (lfuse l)) (k) < 
                       nth ( (lfuse l)) (Suc k)) = index_sequence (intfirst (lfuse l)) (lfuse l)"
        by (simp add: index_sequence_def)   
     have 91: "index_sequence (intfirst (lfuse (x1a \<odot> l))) (lfuse (x1a \<odot> l)) =
               (index_sequence (intfirst x1a) x1a \<and> index_sequence (intfirst (lfuse l)) (lfuse l))"  
         using 1 2  4 5 6 7 8 9 
         by (metis Cons.prems interval_idx_fuse interval_intfirst_lfuse 
             interval_intfirst_lfuse_intfirst lastfirst.simps(2) lfuse_Cons)
     have 10: "index_sequence (intfirst (lfuse l)) (lfuse l) =
               (\<forall> i\<le>intlen l. index_sequence (intfirst (nth l i)) (nth l i)) "
          using "0" Cons.hyps by blast
     have 11: "(\<forall> i\<le> intlen(x1a\<odot>l). index_sequence (intfirst (nth (x1a\<odot>l) i)) (nth (x1a\<odot>l) i)) =
               (\<forall> i\<le> 1  +intlen l. index_sequence (intfirst (nth (x1a\<odot>l) i)) (nth (x1a\<odot>l) i)) "
        by auto
     have 12: "(\<forall> i\<le> 1  +intlen l. index_sequence (intfirst (nth (x1a\<odot>l) i)) (nth (x1a\<odot>l) i)) =
               ( index_sequence (intfirst (x1a)) (x1a) \<and> 
               (\<forall> i. 1 \<le> i \<and> i\<le> 1  +intlen l \<longrightarrow> 
                 index_sequence (intfirst (nth (x1a \<odot> l) i)) (nth (x1a \<odot> l) i))) "
         by (metis One_nat_def Suc_leI interval_intlen_gr_zero interval_nth_zero 
                interval_prefix_length_good intlen.simps(2) order.strict_iff_order)
     have 13: "(\<forall> i. 1 \<le> i \<and> i\<le> 1  +intlen l \<longrightarrow> 
                index_sequence (intfirst (nth (x1a \<odot> l) i)) (nth (x1a \<odot> l) i)) =
               (\<forall> j.  j\<le> intlen l \<longrightarrow> 
                index_sequence (intfirst (nth (x1a \<odot> l) (1+j))) (nth (x1a \<odot> l) (1+j))) "  
         by auto
           (simp add: Nitpick.case_nat_unfold)         
     have 14: "(\<forall> j.  j\<le> intlen l \<longrightarrow> 
                 index_sequence (intfirst (nth (x1a \<odot> l) (1+j))) (nth (x1a \<odot> l) (1+j))) =
               (\<forall> j.  j\<le> intlen l \<longrightarrow> 
                 index_sequence (intfirst (nth (l) (j))) (nth (l) (j))) "
         by simp
     show ?thesis 
     using "10" "12" "13" "91" by auto
   qed
qed



lemma interval_lfuse_intlen_a:
 assumes "lastfirst xxs"
  shows  " ( \<forall>i. i \<le>intlen (xxs) \<longrightarrow>
              (\<forall> j\<le>intlen (nth (xxs) (i)) . j\<le> intlen(lfuse xxs)) )"
using assms
 proof (induct xxs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xxs)
 then show ?case 
 proof -
  have 0: "intlast x1a = intfirst(intfirst xxs)"
     using Cons.prems lastfirst.simps(2) by blast
  have 1: "(\<forall>i. i \<le> intlen (x1a \<odot> xxs) \<longrightarrow> 
            (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i). j \<le> intlen (lfuse (x1a \<odot> xxs))))
           =
           ( (\<forall>j\<le>intlen (Interval.nth (x1a \<odot> xxs) 0). j \<le> intlen (lfuse (x1a \<odot> xxs))) \<and>
             (\<forall>i. 1 \<le> i \<and> i-1 \<le> intlen (xxs) \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i). j \<le> intlen (lfuse (x1a \<odot> xxs)))) )"
       by auto
          (metis One_nat_def add.commute le_diff_conv le_zero_eq not_less_eq_eq old.nat.simps(4) 
           plus_1_eq_Suc)
 have 2: "(\<forall>j\<le>intlen (Interval.nth (x1a \<odot> xxs) 0). j \<le> intlen (lfuse (x1a \<odot> xxs))) =
          (\<forall>j\<le>intlen (x1a). j \<le> intlen (fuse x1a (lfuse xxs)))"
        by simp   
 have 3: "(\<forall>j\<le>intlen (x1a). j \<le> intlen (fuse x1a (lfuse xxs))) =
          (\<forall>j\<le>intlen (x1a). j \<le> intlen (x1a) + intlen (lfuse xxs))  "
    by (simp add: interval_fuse_intlen_a)
 have 4: "(\<forall>j\<le>intlen (x1a). j \<le> intlen (x1a) + intlen (lfuse xxs)) "
    by linarith 
 have 5: "(\<forall>i. 1 \<le> i \<and> i-1 \<le> intlen (xxs) \<longrightarrow> 
           (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i). j \<le> intlen (lfuse (x1a \<odot> xxs))))
          = 
          (\<forall>i. 0 \<le> i \<and> i \<le> intlen (xxs) \<longrightarrow> 
            (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) (Suc i)). j \<le> intlen (lfuse (x1a \<odot> xxs))))"
       by (metis add_diff_cancel_left' interval_intlen_gr_zero interval_prefix_length_good le_add1 
          ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
 have 6: "(\<forall>i. 0 \<le> i \<and> i \<le> intlen (xxs) \<longrightarrow> 
           (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) (Suc i)). j \<le> intlen (lfuse (x1a \<odot> xxs)))) =
          (\<forall>i. 0 \<le> i \<and> i \<le> intlen (xxs) \<longrightarrow>
           (\<forall>j\<le>intlen (nth ( xxs) (i)). j \<le> intlen x1a + intlen (lfuse ( xxs))))" 
         by (simp add: interval_fuse_intlen_a)     
 have 7: "\<forall>i. 0 \<le> i \<and> i \<le> intlen xxs \<longrightarrow> (\<forall>j\<le>intlen (Interval.nth xxs i). j \<le> intlen (lfuse xxs)) "
     using Cons.hyps Cons.prems lastfirst.simps(2) by blast
 have 8: "(\<forall>i. 0 \<le> i \<and> i \<le> intlen (xxs) \<longrightarrow> 
           (\<forall>j\<le>intlen (nth ( xxs) (i)). j \<le> intlen x1a + intlen (lfuse ( xxs)))) " 
          by (simp add: "7" trans_le_add2)
 show ?thesis using "1" "3" "5" "6" "8" by auto
 qed
qed


lemma interval_lfuse_split:
 assumes "lastfirst xxs \<and> (\<forall> j \<le> intlen (xxs). intlen(nth (xxs) j) > 0)"
 shows   "(\<forall>i\<le> intlen(xxs).
               (\<forall>ia<intlen ((nth (xxs) i)).
                 f (sub (nth (nth (xxs) i) ia)
                        (nth (nth (xxs) i) (Suc ia))
                        \<sigma>))) =
                  (\<forall>j<intlen (lfuse (xxs)).
                 f (sub (nth (lfuse (xxs)) j) 
                        (nth (lfuse (xxs)) (Suc j)) 
                        \<sigma>))"
using assms
proof (induct xxs)
case (St x)
then show ?case by auto
next
case (Cons x1a xxs)
then show ?case 
 proof -
  have 1: "(\<forall>i\<le>intlen (x1a \<odot> xxs).
             \<forall>ia<intlen (nth (x1a \<odot> xxs) i).
              f (sub (nth (nth (x1a \<odot> xxs) i) ia) (nth (nth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>)) =
           (\<forall>i\<le>intlen (xxs)+1.
              \<forall>ia<intlen (nth (x1a \<odot> xxs) i).
               f (sub (nth (nth (x1a \<odot> xxs) i) ia) (nth (nth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))" 
     by simp
  have 2: "... =
           ( (\<forall>ia<intlen (nth (x1a \<odot> xxs) 0).
              f (sub (nth (nth (x1a \<odot> xxs) 0) ia) (nth (nth (x1a \<odot> xxs) 0) (Suc ia)) \<sigma>))
              \<and>
           (\<forall>i. 1 \<le> i \<and> i\<le>intlen (xxs)+1\<longrightarrow>
            (\<forall>ia<intlen (nth (x1a \<odot> xxs) i).
             f (sub (nth (nth (x1a \<odot> xxs) i) ia) (nth (Interval.nth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))))"
       by (metis One_nat_def Suc_leI add_nonneg_nonneg gr_zeroI interval_intlen_gr_zero zero_le_one)
  have 3: "(\<forall>ia<intlen (nth (x1a \<odot> xxs) 0).
              f (sub (nth (nth (x1a \<odot> xxs) 0) ia) (nth (nth (x1a \<odot> xxs) 0) (Suc ia)) \<sigma>)) =
           (\<forall>ia<intlen x1a.
              f (sub (nth (x1a) ia) (nth (x1a) (Suc ia)) \<sigma>))" 
     by simp
  have 4: "(\<forall>i. 1 \<le> i \<and> i\<le>intlen (xxs)+1\<longrightarrow>
             (\<forall>ia<intlen (nth (x1a \<odot> xxs) i).
              f (sub (nth (nth (x1a \<odot> xxs) i) ia) (nth (nth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))) =
           (\<forall>i. 0 \<le> i-1 \<and> i-1\<le>intlen (xxs)\<longrightarrow>
             (\<forall>ia<intlen (nth (x1a \<odot> xxs) ((i-1)+1)).
              f (sub (nth (nth (x1a \<odot> xxs) ((i-1)+1)) ia) 
                     (nth (nth (x1a \<odot> xxs) ((i-1)+1)) (Suc ia)) \<sigma>)))" 
       by (auto simp add: Nitpick.case_nat_unfold) 
  have 5: " ... =
            (\<forall>i. 0 \<le> i \<and> i\<le>intlen (xxs)\<longrightarrow>
             (\<forall>ia<intlen (nth (x1a \<odot> xxs) ((i)+1)).
              f (sub (nth (nth (x1a \<odot> xxs) ((i)+1)) ia) 
                     (nth (nth (x1a \<odot> xxs) ((i)+1)) (Suc ia)) \<sigma>)))" 
          using "4" by auto
  have 6: " ... =
            (\<forall>i. 0 \<le> i \<and> i\<le>intlen (xxs)\<longrightarrow>
             (\<forall>ia<intlen (nth (xxs) ((i))).
              f (sub (nth (nth (xxs) ((i))) ia) (nth (nth (xxs) ((i))) (Suc ia)) \<sigma>)))
          "
                by simp
  have 7: "lastfirst xxs"
    using Cons.prems lastfirst.simps(2) by blast 
  have 8: "intlast x1a = intfirst(intfirst xxs)"
    using Cons.prems lastfirst.simps(2) by blast
  have 9: "(\<forall> j \<le> intlen (x1a\<odot>xxs). intlen(nth (x1a\<odot>xxs) j) > 0)" 
    using Cons.prems by blast
  have 10: "intlen(nth (x1a\<odot>xxs) 0) > 0"
    using Cons.prems by blast 
  have 11: "(\<forall>j. 1 \<le> j \<and> j \<le> intlen (xxs)+1 \<longrightarrow> intlen(nth (x1a\<odot>xxs) j) > 0)" 
     using Cons.prems by auto
  have 12: "(\<forall>j. 0 \<le> j-1 \<and> j-1 \<le> intlen (xxs) \<longrightarrow> intlen(nth (x1a\<odot>xxs) ((j-1)+1)) > 0)" 
       using Cons.prems by auto
  have 13: "(\<forall>j. j \<le> intlen (xxs) \<longrightarrow> intlen(nth (x1a\<odot>xxs) ((j)+1)) > 0)"     
     using Cons.prems by auto
  have 14: "(\<forall>j.  j \<le> intlen (xxs) \<longrightarrow> intlen(nth (xxs) ((j))) > 0)" 
     using "13" by simp
  have 15: "(\<forall>i.  i\<le>intlen (xxs)\<longrightarrow>
             (\<forall>ia<intlen (nth (xxs) ((i))).
              f (sub (nth (nth (xxs) ((i))) ia) (nth (nth (xxs) ((i))) (Suc ia)) \<sigma>))) =
            (\<forall>j<intlen (lfuse (xxs)).
                 f (sub (nth (lfuse (xxs)) j) 
                        (nth (lfuse (xxs)) (Suc j)) 
                        \<sigma>))" 
        by (simp add: "14" "7" Cons.hyps)
  have 16: "(\<forall>j<intlen (lfuse (x1a \<odot> xxs)). 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j<intlen (fuse x1a (lfuse (xxs))). 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))" 
             by simp
  have 17: " ... =
             (\<forall>j<intlen x1a + intlen ((lfuse (xxs))). 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))"
     by (simp add: interval_fuse_intlen_a)
  have 18: " ... =
             ((\<forall>j<intlen x1a. 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) \<and>
             (\<forall>j. intlen x1a \<le> j \<and> j<intlen x1a + intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))) " 
      using le_add1 less_le_trans not_less by blast
  have 19: "(\<forall>j<intlen x1a. 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j<intlen x1a. 
              f (sub (nth (fuse x1a (lfuse (xxs))) j) (nth (fuse x1a (lfuse (xxs))) (Suc j)) \<sigma>))" 
          by simp
  have 20: " ... =
             (\<forall>j<intlen x1a. 
              f (sub (nth (x1a) j) (nth (x1a) (Suc j)) \<sigma>))" 
       by (metis Cons.prems Suc_leI  interval_intfirst_lfuse_intfirst 
           interval_nth_prefix interval_prefix_fuse lastfirst.simps(2) less_imp_le_nat)
  have 21: "(\<forall>j. intlen x1a \<le> j \<and> j<intlen x1a + intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth (lfuse (x1a \<odot> xxs)) j) (nth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j. intlen x1a \<le> j \<and> j<intlen x1a + intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth (fuse x1a (lfuse (xxs))) j) (nth (fuse x1a (lfuse (xxs))) (Suc j)) \<sigma>))"  
      by simp
  have 22: " ... =
             (\<forall>j. intlen x1a \<le> j \<and> j<intlen x1a + intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth ((lfuse (xxs))) (j -intlen x1a)) 
                     (nth (fuse x1a (lfuse (xxs))) ((j)+1)) \<sigma>)) " (is "?L=?R")
           proof rule
            show "?L\<Longrightarrow>?R"
              by auto 
             (metis "8" Cons.prems interval_fuse_intlen_a interval_fuse_nth 
              interval_intfirst_lfuse_intfirst less_imp_le_nat)
            show "?R \<Longrightarrow> ?L"
              by auto
             (metis "8" Cons.prems interval_fuse_intlen_a interval_fuse_nth 
              interval_intfirst_lfuse_intfirst less_imp_le_nat)
           qed
  have 23: " ... =
             (\<forall>j. intlen x1a \<le> j \<and> j<intlen x1a + intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth ((lfuse (xxs))) (j -intlen x1a)) 
                     (nth ( (lfuse (xxs))) (((Suc j)-intlen x1a))) \<sigma>))" (is "?L=?R")
           proof 
            show "?L\<Longrightarrow>?R"
             by auto
                (metis "8" Cons.prems Suc_leI interval_fuse_intlen_a interval_fuse_nth 
                 interval_intfirst_lfuse_intfirst le_SucI)
            show "?R \<Longrightarrow>?L"
             by auto
                (metis "8" Cons.prems Suc_leI interval_fuse_intlen_a interval_fuse_nth 
                 interval_intfirst_lfuse_intfirst le_SucI)         
           qed
  have 24: " ... =
             (\<forall>j. 0 \<le> j  \<and> j < intlen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (nth ((lfuse (xxs))) (j) ) (nth ( (lfuse (xxs))) (((Suc j) ))) \<sigma>))"
           by (rule interval_shift_index_to_zero_b)
  show ?thesis 
  using "15" "17" "18" "2" "20" "22" "23" "24" "4" "5" by auto
 qed
qed


end