(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Infinite ITL Semantics\<close>   

theory InfiniteSemantics
imports InfiniteInterval "HOL-TLA.Intensional" 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                     isar_proofs=false,
                    provers="
   cvc4 z3 e spass vampire 
"]
*)
(*
declare [[show_types]]

declare [[show_consts]]
*)

text \<open> 
  This theory mechanises a \emph{shallow} embedding of Infinite ITL using the
  \<open>InfiniteInterval\<close> and \<open>Intensional\<close> theories. 
  A similar embedding as finite ITL has been used with the difference that we use 
  a sum type which is either a finite or infinite interval. 
\<close>

subsection "Types of Formulas"

text \<open>
  To mechanise the Infinite ITL semantics, the following type abbreviations are used:
\<close>

type_synonym 'a intervals = "'a interval + 'a infinterval"

type_synonym ('a,'b) formfun    = "'a intervals \<Rightarrow> 'b"
type_synonym ('a,'b) finformfun = "'a interval \<Rightarrow> 'b"
type_synonym ('a,'b) infformfun = "'a infinterval \<Rightarrow> 'b"
type_synonym 'a formula         = "('a,bool) formfun"
type_synonym 'a finformula      = "('a,bool) finformfun"
type_synonym 'a infformula      = "('a,bool) infformfun"
type_synonym ('a,'b) stfun      = "'a \<Rightarrow> 'b"
type_synonym 'a stpred          = "('a,bool) stfun"

instance 
 "fun" :: (type,type) world ..

instance
 "prod" :: (type,type) world ..

instance
 "sum" :: (type,type) world ..

instance
 "interval" :: (type) world ..

text \<open>
  Pair, function, sum, and interval are instantiated to be of type class world.
  This allows use of the lifted Intensional logic for formulas, and 
  standard logical connectives can therefore be used.
\<close>

subsection \<open>Semantics of ITL\<close>

text \<open> The semantics of ITL is defined. \<close>

definition skip_d :: "('a ::world) formula "
where 
  " skip_d \<equiv> (\<lambda>s. (case s of (Inl s) \<Rightarrow> (intlen s =1) | (Inr s) \<Rightarrow> False))"


definition chop_d :: " ('a ::world) formula \<Rightarrow> ('a ::world) formula \<Rightarrow> ('a ::world) formula"
where 
 "chop_d F1 F2 \<equiv>
  (\<lambda>s. 
    (case s of (Inl s) \<Rightarrow> 
      (\<exists>n. n\<ge>0 \<and> n\<le>intlen s \<and> ((Inl (prefix n s)) \<Turnstile> F1)  \<and> ((Inl (suffix n s)) \<Turnstile> F2))
     | 
               (Inr s) \<Rightarrow> 
      ( (\<exists>n.  ((Inl (iprefix n s)) \<Turnstile> F1) \<and> ((Inr (isuffix n s)) \<Turnstile> F2) )
         \<or> ((Inr s) \<Turnstile> F1)
      )
    )
  )" 


definition current_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where 
 "current_val_d f = (\<lambda>s. (case s of (Inl s) \<Rightarrow> ((nth s 0) \<Turnstile> f) |  (Inr s) \<Rightarrow>  ((s 0) \<Turnstile> f)))"

definition next_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "next_val_d f \<equiv> 
   (\<lambda> s. (case s of (Inl s) \<Rightarrow> if intlen s >0 then ((nth s 1) \<Turnstile> f) else (\<some> (x::'b ). x=x)
                  | (Inr s) \<Rightarrow> (( s 1) \<Turnstile> f)
          )
   )" 

definition fin_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "fin_val_d f \<equiv> \<lambda> s. (case s of (Inl s) \<Rightarrow> (nth s (intlen s)) \<Turnstile> f 
                                   | (Inr s) \<Rightarrow> (\<some> (x::'b ). x=x))"

definition penult_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "penult_val_d f \<equiv> 
    (\<lambda> s. 
     (case s of (Inl s) \<Rightarrow> if intlen s >0 then (nth s ((intlen s)-1) \<Turnstile> f) else (\<some> (x::'b ). x=x) 
              | (Inr s) \<Rightarrow> (\<some> (x::'b ). x=x)
     )
    )" 


syntax
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83) 
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)
 "TEMP"            :: "lift \<Rightarrow> 'b"          ("(TEMP _)")

syntax (ASCII)
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83)
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)

translations
 "_skip_d"        \<rightleftharpoons> "CONST skip_d"
 "_chop_d"        \<rightleftharpoons> "CONST chop_d"
 "_current_val_d" \<rightleftharpoons> "CONST current_val_d"
 "_next_val_d"    \<rightleftharpoons> "CONST next_val_d"
 "_fin_val_d"     \<rightleftharpoons> "CONST fin_val_d"
 "_penult_val_d"  \<rightleftharpoons> "CONST penult_val_d" 
 "TEMP F"         \<rightharpoonup> "(F:: (_ intervals) \<Rightarrow> _)"


subsection "Abbreviations"

text \<open> Some standard temporal abbreviations, with their concrete syntax. \<close>

definition infinite_d :: "('a ::world) formula"
where
 "infinite_d \<equiv> LIFT(#True;#False) "

syntax
  "_infinite_d"  :: "lift"         ("inf")

syntax (ASCII)
  "_infinite_d"  :: "lift"         ("inf")

translations 
  "_infinite_d"  \<rightleftharpoons> "CONST infinite_d"

definition finite_d :: "('a ::world) formula"
where
 "finite_d \<equiv> LIFT(\<not>(inf)) "

syntax
 "_finite_d"    :: "lift"         ("finite")

syntax (ASCII)
  "_finite_d"    :: "lift"         ("finite")

translations 
  "_finite_d"    \<rightleftharpoons> "CONST finite_d"


definition schop_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula" 
 where "schop_d F1 F2 \<equiv> LIFT((F1 \<and> finite);F2)"

definition sometimes_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "sometimes_d F \<equiv> LIFT(finite;F)"

definition di_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "di_d F \<equiv> LIFT(F;#True)" 

definition da_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "da_d F \<equiv> LIFT(finite;(F;#True))" 

definition next_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "next_d F \<equiv> LIFT(skip;F)" 

definition prev_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "prev_d F \<equiv> LIFT(F;skip)"


syntax
  
  "_schop_d"     :: "[lift,lift] \<Rightarrow> lift" ("(_ \<frown> _)" [84,84] 83)
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(\<diamond>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87)
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(\<circle> _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


syntax (ASCII)
  "_schop_d"     :: "[lift,lift] \<Rightarrow> lift" ("(_ schop _)" [84,84] 83)
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(<>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87) 
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(next _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


translations 
  "_schop_d"     \<rightleftharpoons> "CONST schop_d"
  "_sometimes_d" \<rightleftharpoons> "CONST sometimes_d"
  "_di_d"        \<rightleftharpoons> "CONST di_d" 
  "_da_d"        \<rightleftharpoons> "CONST da_d"
  "_next_d"      \<rightleftharpoons> "CONST next_d" 
  "_prev_d"      \<rightleftharpoons> "CONST prev_d"


definition df_d :: "('a::world) formula \<Rightarrow> 'a formula "
 where "df_d F \<equiv> LIFT(F\<frown>#True)"

definition sda_d :: "('a::world) formula \<Rightarrow> 'a formula "
 where "sda_d F \<equiv> LIFT(#True\<frown>(F\<frown>#True))"

definition always_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "always_d F \<equiv> LIFT(\<not>(\<diamond>(\<not>F)))"

definition bi_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bi_d F \<equiv> LIFT(\<not>(di(\<not>F)))"

definition ba_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "ba_d F \<equiv> LIFT(\<not>(da(\<not>F)))"

definition wnext_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wnext_d F \<equiv> LIFT(\<not>(\<circle>(\<not>F)))"

definition wprev_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wprev_d F \<equiv> LIFT(\<not>(prev(\<not>F)))"

definition more_d :: "('a::world) formula "
where "more_d \<equiv> LIFT(\<circle>(#True))" 


syntax
 "_df_d"       :: "lift \<Rightarrow> lift" ("(df _)" [88] 87)
 "_sda_d"      :: "lift \<Rightarrow> lift" ("(sda _)" [88] 87)
 "_always_d"   :: "lift \<Rightarrow> lift" ("(\<box> _)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

syntax (ASCII)
 "_df_d"       :: "lift \<Rightarrow> lift" ("(df _)" [88] 87)
 "_sda_d"      :: "lift \<Rightarrow> lift" ("(sda _)" [88] 87)
 "_always_d"   :: "lift \<Rightarrow> lift" ("([] _)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

translations
 "_df_d"     \<rightleftharpoons> "CONST df_d"
 "_sda_d"    \<rightleftharpoons> "CONST sda_d"
 "_always_d" \<rightleftharpoons> "CONST always_d"
 "_bi_d"     \<rightleftharpoons> "CONST bi_d" 
 "_ba_d"     \<rightleftharpoons> "CONST ba_d"
 "_wnext_d"  \<rightleftharpoons> "CONST wnext_d"
 "_wprev_d"  \<rightleftharpoons> "CONST wprev_d"
 "_more_d"   \<rightleftharpoons> "CONST more_d" 


definition bf_d :: "('a::world) formula \<Rightarrow> 'a formula "
 where "bf_d F \<equiv> LIFT(\<not>(df(\<not>F)))"

definition sba_d :: "('a::world) formula \<Rightarrow> 'a formula "
 where "sba_d F \<equiv> LIFT(\<not>(sda(\<not>F)))"

definition empty_d :: "('a::world) formula "
where "empty_d \<equiv> LIFT(\<not>(more))" 

definition fmore_d :: "('a::world) formula "
 where "fmore_d \<equiv> LIFT(more \<and> finite)" 

definition dm_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "dm_d F \<equiv> LIFT(#True;(more \<and> F))" 

syntax
 "_bf_d"       :: "lift \<Rightarrow> lift" ("(bf _)" [88] 87)
 "_sba_d"      :: "lift \<Rightarrow> lift" ("(sba _)" [88] 87)
 "_empty_d"    :: "lift"         ("(empty)")
 "_fmore_d"    :: "lift"         ("(fmore)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

syntax (ASCII)
 "_bf_d"       :: "lift \<Rightarrow> lift" ("(bf _)" [88] 87)
 "_sba_d"      :: "lift \<Rightarrow> lift" ("(sba _)" [88] 87)
 "_empty_d"    :: "lift"         ("(empty)")
 "_fmore_d"    :: "lift"         ("(fmore)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

translations
 "_bf_d"    \<rightleftharpoons> "CONST bf_d"
 "_sba_d"   \<rightleftharpoons> "CONST sba_d"
 "_empty_d" \<rightleftharpoons> "CONST empty_d"
 "_fmore_d" \<rightleftharpoons> "CONST fmore_d"
 "_dm_d"    \<rightleftharpoons> "CONST dm_d"


definition bm_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bm_d F \<equiv> LIFT(\<not>(dm(\<not>F)))" 

definition init_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "init_d F \<equiv> LIFT((empty \<and> F);#True)" 

definition fin_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "fin_d F \<equiv> LIFT(\<box>(empty \<longrightarrow> F))" 

definition halt_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "halt_d F \<equiv> LIFT(\<box>(empty = F))" 

definition initonly_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "initonly_d F \<equiv> LIFT(bi(empty = F))" 

definition keep_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "keep_d F \<equiv> LIFT(ba(skip \<longrightarrow> F))" 

definition yields_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "yields_d F1 F2 \<equiv> LIFT(\<not>(F1;(\<not>F2)))" 

definition syields_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "syields_d F1 F2 \<equiv> LIFT(\<not>(F1\<frown>(\<not>F2)))" 


definition ifthenelse_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthenelse_d F G H \<equiv> LIFT((F \<and> G) \<or> (\<not>F \<and> H) ) "


primrec power_d :: "('a::world) formula \<Rightarrow> nat \<Rightarrow> 'a formula "
where  pow_0  : "(power_d F 0) = LIFT(empty)"
     | pow_Suc: "(power_d F (Suc n)) = LIFT((F \<and> finite);(power_d F n)) "

primrec spower_d :: "('a::world) formula \<Rightarrow> nat \<Rightarrow> 'a formula "
where  spow_0  : "(spower_d F 0) = LIFT(empty)"
     | spow_Suc: "(spower_d F (Suc n)) = LIFT(F\<frown>(spower_d F n)) "


syntax
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_syields_d"    :: "[lift,lift] \<Rightarrow> lift"      ("(_ syields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87) 
 "_spower_d"      :: "[lift,nat] \<Rightarrow> lift"      ("(spower _ _)" [88,88] 87) 


syntax (ASCII)
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_syields_d"    :: "[lift,lift] \<Rightarrow> lift"      ("(_ syields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87)
 "_spower_d"      :: "[lift,nat] \<Rightarrow> lift"      ("(spower _ _)" [88,88] 87) 

translations
 "_bm_d"         \<rightleftharpoons> "CONST bm_d"
 "_init_d"       \<rightleftharpoons> "CONST init_d"
 "_fin_d"        \<rightleftharpoons> "CONST fin_d"
 "_halt_d"       \<rightleftharpoons> "CONST halt_d"
 "_initonly_d"   \<rightleftharpoons> "CONST initonly_d" 
 "_keep_d"       \<rightleftharpoons> "CONST keep_d"
 "_yields_d"     \<rightleftharpoons> "CONST yields_d"
 "_syields_d"    \<rightleftharpoons> "CONST syields_d"
 "_ifthenelse_d" \<rightleftharpoons> "CONST ifthenelse_d" 
 "_power_d"      \<rightleftharpoons> "CONST power_d"
 "_spower_d"     \<rightleftharpoons> "CONST spower_d"



definition len_d :: "nat \<Rightarrow> ('a::world) formula "
where "len_d n \<equiv> LIFT(power skip n)"

definition fpowerstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "fpowerstar_d F \<equiv> LIFT(\<exists> k. power F k)"

definition spowerstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "spowerstar_d F \<equiv> LIFT(\<exists> k. spower F k)"


definition omega_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "omega_d F \<equiv> (\<lambda>s.
         (case s of (Inl s) \<Rightarrow> False 
                  | (Inr s) \<Rightarrow> 
                     (\<exists>(l::infiniteindex). infinite_index_sequence 0 l \<and> 
                         (\<forall>i.  
                             ((Inl (subinterval s (l i) (l (Suc i)))) \<Turnstile> F)
                           ) 
                  )
         ))"


syntax
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_fpowerstar_d"   :: "lift \<Rightarrow> lift"        ("(fpowerstar _)" [85] 85)
 "_spowerstar_d"   :: "lift \<Rightarrow> lift"        ("(spowerstar _)" [85] 85)
 "_omega_d"        :: "lift \<Rightarrow> lift"        ("(_\<^sup>\<omega>)" [85] 85)

syntax (ASCII)
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_fpowerstar_d"   :: "lift \<Rightarrow> lift"        ("(fpowerstar _)" [85] 85)
 "_spowerstar_d"   :: "lift \<Rightarrow> lift"        ("(spowerstar _)" [85] 85)
 "_omega_d"        :: "lift \<Rightarrow> lift"        ("(omega _)" [85] 85)

translations
 "_len_d"          \<rightleftharpoons> "CONST len_d" 
 "_fpowerstar_d"   \<rightleftharpoons> "CONST fpowerstar_d"
 "_spowerstar_d"   \<rightleftharpoons> "CONST spowerstar_d"
 "_omega_d"        \<rightleftharpoons> "CONST omega_d" 


definition powerstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "powerstar_d F \<equiv> LIFT((\<exists> k. power F k);(empty \<or> (F \<and> inf)))"

syntax
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

syntax (ASCII)
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

translations
 "_powerstar_d"    \<rightleftharpoons> "CONST powerstar_d"


definition chopstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "chopstar_d F \<equiv> LIFT(powerstar (F \<and> more))"

definition schopstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "schopstar_d F \<equiv> LIFT(spowerstar (F \<and> more))"


syntax
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(_\<^sup>\<star>)" [85] 85)
 "_schopstar_d"     ::"lift \<Rightarrow> lift"        ("(schopstar _)" [85] 85)


syntax (ASCII)
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(chopstar _)" [85] 85)
 "_schopstar_d"     ::"lift \<Rightarrow> lift"        ("(schopstar _)" [85] 85)


translations
 "_chopstar_d"     \<rightleftharpoons> "CONST chopstar_d"
 "_schopstar_d"    \<rightleftharpoons> "CONST schopstar_d"



definition sfin_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "sfin_d F \<equiv> LIFT(\<not> (fin (\<not> F)))" 

definition ifthen_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthen_d F G \<equiv> LIFT(if\<^sub>i F then G else #True ) "

definition while_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "while_d F G \<equiv> LIFT( ( F \<and> G)\<^sup>\<star> \<and> (fin ((\<not>F))) ) "

syntax
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)
 "_sfin_d"        :: "lift \<Rightarrow> lift"   ("(sfin _)" [88] 87)

syntax (ASCII)
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)
 "_sfin_d"        :: "lift \<Rightarrow> lift"   ("(sfin _)" [88] 87)

translations
 "_ifthen_d" \<rightleftharpoons> "CONST ifthen_d"
 "_while_d"  \<rightleftharpoons> "CONST while_d"
 "_sfin_d"   \<rightleftharpoons> "CONST sfin_d"

definition swhile_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "swhile_d F G \<equiv> LIFT( schopstar( F \<and> G) \<and> (sfin ((\<not>F))) ) "

definition repeat_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "repeat_d F G \<equiv> LIFT(F;while (\<not> G) do F ) "

syntax
 "_swhile_d"  :: "[lift,lift] \<Rightarrow> lift" ("(swhile _ do _  )"  [88,88] 87)
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

syntax (ASCII)
 "_swhile_d"  :: "[lift,lift] \<Rightarrow> lift" ("(swhile _ do _  )"  [88,88] 87)
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

translations
 "_swhile_d" \<rightleftharpoons> "CONST swhile_d"
 "_repeat_d" \<rightleftharpoons> "CONST repeat_d"

definition srepeat_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "srepeat_d F G \<equiv> LIFT(F\<frown>swhile (\<not> G) do F ) "

definition next_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "next_assign_d v e \<equiv> LIFT( v$ = e)" 

definition prev_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "prev_assign_d v e \<equiv> LIFT( finite \<longrightarrow> v! = e)"  

definition always_eq_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "always_eq_d v e \<equiv> \<lambda> s. s \<Turnstile> \<box>($v = e)"  

definition temporal_assign_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "temporal_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> finite \<longrightarrow> !v = e" 

definition gets_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "gets_d v e \<equiv> \<lambda> s. s \<Turnstile> keep( temporal_assign_d v e)" 

definition stable_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where "stable_d v \<equiv> \<lambda> s. s \<Turnstile> gets_d v (current_val_d v) "

definition padded_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where "padded_d v \<equiv> \<lambda> s. s \<Turnstile> (stable_d v);skip \<or> empty" 

definition padded_temp_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "padded_temp_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> (temporal_assign_d v e) \<and> (padded_d v)" 


syntax
 "_srepeat_d"            :: "[lift,lift] \<Rightarrow> lift" ("(srepeat _ until _  )"  [88,88] 87)
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ \<approx> _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ \<leftarrow> _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <\<sim> _)" [50,51] 50)

syntax (ASCII)
 "_srepeat_d"            :: "[lift,lift] \<Rightarrow> lift" ("(srepeat _ until _  )"  [88,88] 87)
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ alweqv _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ <-- _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <~ _)" [50,51] 50)

translations
 "_srepeat_d"            \<rightleftharpoons> "CONST srepeat_d"
 "_next_assign_d"        \<rightleftharpoons> "CONST next_assign_d"
 "_prev_assign_d"        \<rightleftharpoons> "CONST prev_assign_d"
 "_always_eq_d"          \<rightleftharpoons> "CONST always_eq_d"
 "_temporal_assign_d"    \<rightleftharpoons> "CONST temporal_assign_d"
 "_gets_d"               \<rightleftharpoons> "CONST gets_d"
 "_stable_d"             \<rightleftharpoons> "CONST stable_d"
 "_padded_d"             \<rightleftharpoons> "CONST padded_d"
 "_padded_temp_assign_d" \<rightleftharpoons> "CONST padded_temp_assign_d"


subsection \<open>Properties of Operators\<close>

text \<open> The following lemmas show that above operators have the expected semantics. \<close>

lemma skip_defs :
 "(w \<Turnstile> skip) = (case w of (Inl w) \<Rightarrow> (intlen w =1) | (Inr w) \<Rightarrow> False) "
by (simp add: skip_d_def)

lemma skip_defs_finite :
 "((Inl w) \<Turnstile> skip) = (intlen w =1)"
by (simp add: skip_d_def)

lemma skip_defs_infinite :
 "\<not> ((Inr w) \<Turnstile> skip)"
by (simp add: skip_d_def)

lemma chop_defs :
 "(w \<Turnstile> F1 ; F2) = 
  (case w of (Inl w) \<Rightarrow>
       (\<exists>n. n\<ge>0 \<and> n\<le>intlen w \<and> ((Inl (prefix n w)) \<Turnstile> F1 )  \<and> ((Inl (suffix n w)) \<Turnstile> F2 ) )
           | (Inr w) \<Rightarrow>
       ( (\<exists>n.  ((Inl (iprefix n w)) \<Turnstile> F1 ) \<and> ((Inr (isuffix n w)) \<Turnstile> F2))
         \<or> ((Inr w) \<Turnstile> F1)
       )
  )
 "
by (simp add: chop_d_def)

lemma chop_defs_finite:
 "( (Inl w) \<Turnstile> F1;F2) =
    (\<exists>n. n\<ge>0 \<and> n\<le>intlen w \<and> 
             (Inl (prefix n w) \<Turnstile> F1 )  \<and> (Inl (suffix n w) \<Turnstile> F2 )
       )"
by (simp add: chop_d_def)

lemma chop_defs_infinite:
 " ( (Inr w) \<Turnstile> F1;F2) =
       ( (\<exists>n.  (Inl (iprefix n w) \<Turnstile> F1 ) \<and> (Inr (isuffix n w) \<Turnstile> F2 ))
         \<or> (Inr w \<Turnstile> F1 )
       )"
by (simp add: chop_d_def) 


lemma infinite_defs:
 " (w \<Turnstile> inf) = (case w of (Inr w) \<Rightarrow> True | (Inl w) \<Rightarrow> False)"
by (simp add: infinite_d_def chop_d_def sum.case_eq_if)

lemma infinite_defs_1:
 " ((Inr w) \<Turnstile> inf) "
by (simp add: infinite_d_def chop_d_def sum.case_eq_if)

lemma finite_defs :
 " (w \<Turnstile> finite) = (case w of (Inl w) \<Rightarrow> True | (Inr w) \<Rightarrow> False)"
by (simp add: finite_d_def infinite_defs chop_d_def sum.case_eq_if)

lemma finite_defs_1 :
 " ((Inl w) \<Turnstile> finite)"
by (simp add: finite_defs sum.case_eq_if)

lemma schop_defs :
 "(w \<Turnstile> F1 \<frown> F2) = 
  (case w of (Inl w) \<Rightarrow>
       (\<exists>n. n\<ge>0 \<and> n\<le>intlen w \<and> ((Inl (prefix n w)) \<Turnstile> F1 )  \<and> ((Inl (suffix n w)) \<Turnstile> F2 ) )
           | (Inr w) \<Rightarrow>
       ( (\<exists>n.  ((Inl (iprefix n w)) \<Turnstile> F1 ) \<and> ((Inr (isuffix n w)) \<Turnstile> F2))
       )
  )
 "
by (simp add: schop_d_def chop_defs finite_defs sum.case_eq_if)

lemma schop_defs_finite :
 "((Inl w) \<Turnstile> F1 \<frown> F2) = 
       (\<exists>n. n\<ge>0 \<and> n\<le>intlen w \<and> ((Inl (prefix n w)) \<Turnstile> F1 )  \<and> ((Inl (suffix n w)) \<Turnstile> F2 ) )"
by (simp add: schop_defs)


lemma schop_defs_infinite :
 "((Inr w) \<Turnstile> F1 \<frown> F2) = 
       (\<exists>n.  ((Inl (iprefix n w)) \<Turnstile> F1 ) \<and> ((Inr (isuffix n w)) \<Turnstile> F2))
"       
by (simp add: schop_defs)

lemma sometimes_defs : 
  "(w \<Turnstile> \<diamond> F) = 
   (case w of (Inl w) \<Rightarrow> (\<exists> n. 0\<le>n \<and> n\<le>intlen w \<and> ((Inl (suffix n w)) \<Turnstile> F))
            | (Inr w) \<Rightarrow> (\<exists> n. ((Inr (isuffix n w)) \<Turnstile> F))
   )
"
by (simp add: sometimes_d_def finite_defs chop_defs sum.case_eq_if)

lemma always_defs : 
 " (w \<Turnstile> \<box> F) = 
    (case w of (Inl w) \<Rightarrow> (\<forall> n. 0\<le>n \<and> n\<le>intlen w \<longrightarrow> ((Inl (suffix n w)) \<Turnstile> F))
             | (Inr w) \<Rightarrow> (\<forall> n.  ((Inr (isuffix n w)) \<Turnstile> F ))
    ) "
by (simp add: always_d_def sometimes_defs sum.case_eq_if)

lemma di_defs : 
  "(w \<Turnstile> di F) = 
  (case w of (Inl w) \<Rightarrow> (\<exists> n. 0\<le>n \<and> n\<le>intlen w \<and> ((Inl (prefix n w)) \<Turnstile> F))
           | (Inr w) \<Rightarrow> (\<exists> n. ((Inl (iprefix n w)) \<Turnstile> F)) \<or> ((Inr w) \<Turnstile> F)
  )"
by (simp add: di_d_def chop_d_def sum.case_eq_if)


lemma df_defs : 
  "(w \<Turnstile> df F) = 
  (case w of (Inl w) \<Rightarrow> (\<exists> n. 0\<le>n \<and> n\<le>intlen w \<and> ((Inl (prefix n w)) \<Turnstile> F))
           | (Inr w) \<Rightarrow> (\<exists> n. ((Inl (iprefix n w)) \<Turnstile> F))
  )"
by (simp add: df_d_def schop_defs sum.case_eq_if)

lemma bi_defs : 
  "(w \<Turnstile> bi F) = 
   (case w of (Inl w) \<Rightarrow> (\<forall> n. 0\<le>n \<and> n\<le>intlen w \<longrightarrow> ((Inl (prefix n w)) \<Turnstile> F))
            | (Inr w) \<Rightarrow> (\<forall> n. ((Inl (iprefix n w)) \<Turnstile> F)) \<and> ((Inr w) \<Turnstile> F)
   )"
by (simp add: bi_d_def di_defs sum.case_eq_if)

lemma bf_defs : 
  "(w \<Turnstile> bf F) = 
   (case w of (Inl w) \<Rightarrow> (\<forall> n. 0\<le>n \<and> n\<le>intlen w \<longrightarrow> ((Inl (prefix n w)) \<Turnstile> F))
            | (Inr w) \<Rightarrow> (\<forall> n. ((Inl (iprefix n w)) \<Turnstile> F)) 
   )"
by (simp add: bf_d_def df_defs sum.case_eq_if)

lemma da_defs : 
  "(w \<Turnstile> da F) = 
   (case w of (Inl w) \<Rightarrow> (\<exists> n na.   n+na \<le> intlen w \<and> ((Inl (sub n (n+ na) w)) \<Turnstile> F))
            | (Inr w) \<Rightarrow> (\<exists> n na.  ((Inl (subinterval w n (n+na) )) \<Turnstile> F) 
                         \<or> ( (Inr (isuffix n w)) \<Turnstile> F))
   )"
proof 
 (auto simp add: da_d_def chop_defs finite_d_def infinite_d_def iprefix_isuffix sum.case_eq_if)
 show "\<And>n na.
       isl w \<Longrightarrow>
       n \<le> intlen (projl w) \<Longrightarrow>
       na \<le> intlen (projl w) - n \<Longrightarrow> F (Inl (prefix na (suffix n (projl w)))) \<Longrightarrow> 
      \<exists>n na. n + na \<le> intlen (projl w) \<and> F (Inl (sub n (n + na) (projl w))) " 
 by (metis Nat.le_diff_conv2 add.commute interval_sub_prefix_suffix_0 zero_le)
 show "\<And>n na.
       isl w \<Longrightarrow>
       n + na \<le> intlen (projl w) \<Longrightarrow>
       F (Inl (sub n (n + na) (projl w))) \<Longrightarrow> 
       \<exists>n\<le>intlen (projl w). \<exists>na\<le>intlen (projl w) - n. F (Inl (prefix na (suffix n (projl w))))" 
 by (metis Interval.sub_def add_leD1 interval_intlen_sub interval_pref_intlen_bound 
     interval_suffix_length le_add1) 
 show "\<And>n na.
       \<not> isl w \<Longrightarrow> F (Inl (subinterval (projr w) n (na + n))) \<Longrightarrow>
       \<exists>n. (\<exists>na. F (Inl (subinterval (projr w) n (n + na)))) \<or> F (Inr (isuffix n (projr w)))" 
 by (metis add.commute)
 show "\<And>n na.
       \<not> isl w \<Longrightarrow> F (Inl (subinterval (projr w) n (n + na))) \<Longrightarrow> 
       \<exists>n. (\<exists>na. F (Inl (subinterval (projr w) n (na + n)))) \<or> F (Inr (isuffix n (projr w))) " 
 by (metis add.commute)
qed

lemma ba_defs :
 "(w \<Turnstile> ba F) = 
  (case w of (Inl w) \<Rightarrow> (\<forall> n na. n+na \<le> intlen w \<longrightarrow> ((Inl (sub n (n+na) w)) \<Turnstile> F))
           | (Inr w) \<Rightarrow> (\<forall> n na. ((Inl (subinterval w n (n+na) )) \<Turnstile> F) 
                         \<and> ( (Inr (isuffix n w)) \<Turnstile> F))
  )"
by (simp add: ba_d_def da_defs sum.case_eq_if)

lemma sda_defs : 
  "(w \<Turnstile> sda F) = 
   (case w of (Inl w) \<Rightarrow> (\<exists> n na.  n+na \<le> intlen w \<and> ((Inl (sub n (n+na) w)) \<Turnstile> F))
            | (Inr w) \<Rightarrow> (\<exists> n na.  ((Inl (subinterval w n (n+na) )) \<Turnstile> F))
   )"
proof 
 (auto simp add: sda_d_def schop_defs iprefix_isuffix sum.case_eq_if)
 show "\<And>n na.
       isl w \<Longrightarrow>
       n \<le> intlen (projl w) \<Longrightarrow>
       na \<le> intlen (projl w) - n \<Longrightarrow> F (Inl (prefix na (suffix n (projl w)))) \<Longrightarrow> 
       \<exists>n na. n + na \<le> intlen (projl w) \<and> F (Inl (sub n (n + na) (projl w)))" 
 by (metis Nat.le_diff_conv2 interval_sub_prefix_suffix_0 le_add_diff_inverse 
     nat_add_left_cancel_le zero_le)
 show "\<And>n na.
       isl w \<Longrightarrow>
       n + na \<le> intlen (projl w) \<Longrightarrow>
       F (Inl (sub n (n + na) (projl w))) \<Longrightarrow> 
       \<exists>n\<le>intlen (projl w). \<exists>na\<le>intlen (projl w) - n. F (Inl (prefix na (suffix n (projl w))))" 
 by (metis Interval.sub_def add_leD1 interval_intlen_sub interval_pref_intlen_bound 
     interval_suffix_length le_add1)
 show "\<And>n na. \<not> isl w \<Longrightarrow> F (Inl (subinterval (projr w) n (na + n))) \<Longrightarrow> 
       \<exists>n na. F (Inl (subinterval (projr w) n (n + na)))" 
 by (metis add.commute)
 show "\<And>n na. \<not> isl w \<Longrightarrow> F (Inl (subinterval (projr w) n (n + na))) \<Longrightarrow> 
       \<exists>n na. F (Inl (subinterval (projr w) n (na + n))) " 
 by (metis add.commute)
qed


lemma sba_defs :
 "(w \<Turnstile> sba F) = 
  (case w of (Inl w) \<Rightarrow> (\<forall> n na.  n+na \<le> intlen w \<longrightarrow> ((Inl (sub n (n+na) w)) \<Turnstile> F))
           | (Inr w) \<Rightarrow> (\<forall> n na.  ((Inl (subinterval w n (n+na) )) \<Turnstile> F))
  )"
by (simp add: sba_d_def sda_defs sum.case_eq_if)

lemma next_defs :
 " (w \<Turnstile> \<circle> F) = 
 (case w of (Inl w) \<Rightarrow> (intlen w > 0 \<and> ((Inl (suffix 1 w))\<Turnstile> F) )
          | (Inr w) \<Rightarrow> ((Inr (isuffix 1 w)) \<Turnstile> F)
 )"
using Suc_le_eq  min.absorb1
by (simp add: next_d_def chop_defs skip_d_def iprefix_length  sum.case_eq_if)
    force

lemma wnext_defs :
 "(w \<Turnstile> wnext F) = 
  (case w of (Inl w) \<Rightarrow> (intlen w = 0 \<or> ((Inl (suffix 1 w))\<Turnstile> F) )
          |  (Inr w) \<Rightarrow> ((Inr (isuffix 1 w)) \<Turnstile> F)
  )"
by (simp add: wnext_d_def next_defs sum.case_eq_if)

lemma prev_defs :
 " (w \<Turnstile> prev F) = 
   (case w of (Inl w) \<Rightarrow> (intlen w > 0 \<and> ((Inl (prefix ((intlen w)-1) w)) \<Turnstile> F) )
           |  (Inr w) \<Rightarrow> (Inr w) \<Turnstile> F
   ) "
by (simp add: prev_d_def chop_defs skip_d_def sum.case_eq_if)
   (metis One_nat_def Suc_leI diff_diff_cancel diff_le_self interval_suffix_length_good 
             le_zero_eq neq0_conv zero_neq_one)

lemma wprev_defs :
 " (w \<Turnstile> wprev F) = 
  (case w of (Inl w) \<Rightarrow> (intlen w = 0 \<or> ((Inl (prefix ((intlen w)-1) w)) \<Turnstile> F) )
           | (Inr w) \<Rightarrow> (Inr w) \<Turnstile> F 
  )"
by (simp add: wprev_d_def prev_defs sum.case_eq_if)

lemma more_defs :
 " (w \<Turnstile> more) = 
   (case w of (Inl w) \<Rightarrow> (intlen w>0) 
            | (Inr w) \<Rightarrow> True
   )"
by (simp add: more_d_def next_defs sum.case_eq_if)

lemma fmore_defs :
 " (w \<Turnstile> fmore) = 
   (case w of (Inl w) \<Rightarrow> (intlen w>0) 
            | (Inr w) \<Rightarrow> False
   )"
by (simp add: fmore_d_def more_defs finite_defs sum.case_eq_if) 

lemma empty_defs :
 " (w \<Turnstile> empty) = 
   (case w of (Inl w) \<Rightarrow> (intlen w = 0)
            | (Inr w) \<Rightarrow> False
   )"
by (simp add: empty_d_def more_defs sum.case_eq_if)

lemma init_defs :
 " (w \<Turnstile> init F) = 
   (case w of (Inl w) \<Rightarrow> ( (Inl (prefix 0 w)) \<Turnstile> F )
            | (Inr w) \<Rightarrow> ( (Inl (iprefix 0 w)) \<Turnstile> F )
   )"
using min.absorb1 
by (simp add: init_d_def chop_defs empty_defs iprefix_length sum.case_eq_if) 
   force


lemma init_defs_finite:
 " ((Inl w) \<Turnstile> init F) = ( (Inl (prefix 0 w)) \<Turnstile> F )"
by (simp add: init_defs)

lemma init_defs_infinite:
 " ((Inr w) \<Turnstile> init F) = ( (Inl (iprefix 0 w)) \<Turnstile> F )"
by (simp add: init_defs)

lemma initalt_defs :
 " (w \<Turnstile> bi( empty \<longrightarrow> F)) = 
   (case w of (Inl w) \<Rightarrow> ( (Inl (prefix 0 w)) \<Turnstile> F )
            | (Inr w) \<Rightarrow> ( (Inl (iprefix 0 w)) \<Turnstile> F )
   )" 
using min.absorb1
by (simp add: bi_defs empty_defs iprefix_length sum.case_eq_if)
    force

lemma fin_defs :
 " (w \<Turnstile> fin F) = 
   (case w of (Inl w) \<Rightarrow> ( (Inl (suffix (intlen w) w)) \<Turnstile> F)
            | (Inr w) \<Rightarrow> True
   )"
by (simp add: fin_d_def empty_defs always_defs sum.case_eq_if) 

lemma finalt_defs :
 " (w \<Turnstile> #True;(F \<and> empty)) = 
   (case w of (Inl w) \<Rightarrow> ( (Inl (suffix (intlen w) w)) \<Turnstile> F)
            | (Inr w) \<Rightarrow> True
   )"
by (simp add: chop_defs empty_defs sum.case_eq_if)  fastforce 

lemma sfin_defs :
 " (w \<Turnstile> sfin F) =
  (case w of (Inl w) \<Rightarrow> ( (Inl (suffix (intlen w) w)) \<Turnstile> F)
            | (Inr w) \<Rightarrow> False
   )
"
by (simp add: sfin_d_def fin_defs sum.case_eq_if)

lemma halt_defs :
 " (w \<Turnstile> halt(F)) = 
    (case w of (Inl w) \<Rightarrow> (\<forall>n\<le>intlen w. (intlen w = n) = ( (Inl (suffix n w)) \<Turnstile> F))
             | (Inr w) \<Rightarrow> (\<forall>n. \<not>( (Inr (isuffix n w)) \<Turnstile> F))
    ) "
by (simp add: halt_d_def empty_defs always_defs sum.case_eq_if)

lemma initonly_defs :
 " (w \<Turnstile> initonly(F)) = 
   (case w of (Inl w) \<Rightarrow> (\<forall>n\<le>intlen w. (n = 0) = ( (Inl (prefix n w)) \<Turnstile> F ))
            | (Inr w) \<Rightarrow> (\<forall>n. (n = 0) = ( (Inl (iprefix n w)) \<Turnstile> F )) \<and> \<not>((Inr w) \<Turnstile> F)
   )"
by (simp add: min.absorb1 initonly_d_def bi_defs empty_defs iprefix_length sum.case_eq_if)

lemma ifthenelse_defs:
 " (w \<Turnstile> if\<^sub>i F then G else H) = 
   ( ((w \<Turnstile> F) \<and> (w \<Turnstile> G)) \<or> (( \<not>(w \<Turnstile> F) \<and> (w \<Turnstile> H))) )"
by (simp add: ifthenelse_d_def)

lemma len_defs :
 " (w \<Turnstile> len n) = 
    (case w of (Inl w) \<Rightarrow> (intlen w =n)
             | (Inr w) \<Rightarrow> False
    ) "
proof  
   (simp add: len_d_def  sum.case_eq_if) 
   show "(isl w \<longrightarrow> (w \<Turnstile> (power skip n)) = (intlen (projl w) = n)) \<and> 
         (\<not> isl w \<longrightarrow> \<not> (w \<Turnstile> (power skip n)))  " 
   proof (induct n arbitrary:w)
   case 0
   then show ?case by (simp add: empty_defs sum.case_eq_if)
   next
   case (Suc n)
   then show ?case 
      by (auto simp add: len_d_def chop_defs skip_defs finite_defs sum.case_eq_if)
          auto
   qed
qed

lemma currentval_defs :
 " (s \<Turnstile> $v) = 
   (case s of (Inl s) \<Rightarrow> (v (nth s 0))
            | (Inr s) \<Rightarrow> (v (s 0))
   )"
by (simp add: current_val_d_def)

lemma nextval_defs :
 "(s \<Turnstile> v$) = 
  (case s of (Inl s) \<Rightarrow> (if intlen s >0 then (v (nth s 1)) else (\<some> x. x=x))
           | (Inr s) \<Rightarrow> (v (s 1))
  )"
by (simp add: next_val_d_def)

lemma finval_defs :
 " (s \<Turnstile> !v) = 
   (case s of (Inl s) \<Rightarrow> (v (nth s (intlen s)))
            | (Inr s) \<Rightarrow> (\<some> x. x=x)
   )"
by (simp add: fin_val_d_def)

lemma penultval_defs :
 " (s \<Turnstile> v!) = 
   (case s of (Inl s) \<Rightarrow> (if intlen s >0 then (v (nth s ((intlen s)-1))) else (\<some> x. x=x))
            | (Inr s) \<Rightarrow> (\<some> x. x=x)
   )"
by (simp add: penult_val_d_def) 

lemma next_assign_defs :
 " (s \<Turnstile> v := e) = 
    (case s of (Inl s) \<Rightarrow> if 0 < intlen s then v (Interval.nth s 1) else (\<some> x. x=x)
     | Inr s \<Rightarrow> v (s 1)
    ) =
    e s
"
by (auto simp: next_assign_d_def next_val_d_def)

lemma prev_assign_defs :
 " (s \<Turnstile> v =: e) = 
  (case s of (Inl s) \<Rightarrow> 
        if 0 < intlen s then (v (Interval.nth s ((intlen s)-1)) = e (Inl s))  
                        else ((\<some> x. x=x) = e (Inl s))
           | (Inr s) \<Rightarrow>  True
   )  "
by (simp add: prev_assign_d_def penult_val_d_def finite_defs sum.case_eq_if)

lemma always_eqv_defs :
 " (s \<Turnstile> v \<approx> e) = 
   (case s of (Inl s) \<Rightarrow> (\<forall> i\<le> intlen s. v (Interval.nth s i) = e (Inl (suffix i s)))
            | (Inr s) \<Rightarrow> (\<forall> i. v (s i) = e (Inr (isuffix i s)))
   ) "
by (simp add: always_eq_d_def always_defs current_val_d_def isuffix_def sum.case_eq_if) 

lemma temporal_assign_defs :
 " (s \<Turnstile> v \<leftarrow> e) = 
   (case s of (Inl s) \<Rightarrow> (v (Interval.nth s (intlen s))) = e (Inl s)
            | (Inr s) \<Rightarrow> True
   ) "
by (simp add: temporal_assign_d_def fin_val_d_def finite_defs sum.case_eq_if)

lemma gets_defs :
 " (s \<Turnstile> v gets e) = 
  (case s of (Inl s) \<Rightarrow> (\<forall> i < intlen s. v (Interval.nth s (Suc i)) = e (Inl (sub i (i+1) s)) )
           | (Inr s) \<Rightarrow> (\<forall> i. v ( s (Suc i)) = e (Inl (subinterval s i (i+1))) )
  )"
using Suc_leI Suc_le_lessD 
by (simp add: min.absorb1 finite_defs gets_d_def keep_d_def ba_defs skip_defs sub_def 
              temporal_assign_d_def fin_val_d_def subinterval_length subinterval_nth sum.case_eq_if)
 auto

lemma stable_defs_helpa:
assumes "(\<forall>i<intlen s. v (Interval.nth s (Suc i)) = v (Interval.nth s i))"  
        " i \<le> intlen s" 
 shows  "(v (Interval.nth s i) = v (Interval.nth s 0))" 
using assms
proof (induct s arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a s)
then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by blast
  next
  case (Suc nat)
  then show ?thesis 
     by (metis Cons.hyps Cons.prems(1) Cons.prems(2) Suc_le_mono Suc_mono interval_nth_Suc 
         interval_nth_zero intlen.simps(2) plus_1_eq_Suc zero_less_Suc)
  qed
qed

lemma stable_defs_helpb:
assumes "(\<forall>i\<le>intlen s. v (Interval.nth s i) = v (Interval.nth s 0))"  
        " i < intlen s" 
 shows  "v (Interval.nth s (Suc i)) = v (Interval.nth s i)" 
using assms
proof (induct s arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a s)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis using Suc_leI Cons.prems(1) Cons.prems(2) by blast
   next
   case (Suc nat)
   then show ?thesis using Cons.prems(1) Cons.prems(2) Suc_leI less_imp_le_nat by presburger
   qed
qed


lemma stable_defs_help:
 "(\<forall>i<intlen s. v (Interval.nth s (Suc i)) = v (Interval.nth s i)) = 
  (\<forall>i\<le>intlen s. v (Interval.nth s i) = v (Interval.nth s 0))"
proof -
 have 1: "(\<forall>i<intlen s. v (Interval.nth s (Suc i)) = v (Interval.nth s i)) \<longrightarrow>
          (\<forall>i\<le>intlen s. v (Interval.nth s i) = v (Interval.nth s 0)) " 
  using stable_defs_helpa by auto
 have 2: "(\<forall>i\<le>intlen s. v (Interval.nth s i) = v (Interval.nth s 0)) \<longrightarrow>
          (\<forall>i<intlen s. v (Interval.nth s (Suc i)) = v (Interval.nth s i))" 
   using stable_defs_helpb by blast
 show ?thesis using "1" "2" by blast
qed

lemma stable_defs_infinite:
 "(\<forall>i. v (s (Suc i)) = v (s i)) = (\<forall>i. v (s i) = v (s 0))"
proof -
 have 1: "(\<forall>i. v (s (Suc i)) = v (s i)) \<Longrightarrow>(\<And>j. v (s j) = v (s 0))"
 proof -
  assume A1: "(\<forall>i. v (s (Suc i)) = v (s i))"
  show "(\<And>j. v (s j) = v (s 0))"
  proof -
   fix j
   show "v (s j) = v (s 0)"
   using A1 by (induct j) simp_all
  qed
 qed
 have 2: "(\<forall>i. v (s (Suc i)) = v (s i)) \<Longrightarrow> (\<forall>j. v (s j) = v (s 0))"
 using 1 by blast
 have 3: "(\<forall>j. v (s j) = v (s 0)) \<Longrightarrow> (\<And>j. v (s (Suc j)) = v (s j))"
 proof -
  assume A2: "(\<forall>j. v (s j) = v (s 0))"
  show "(\<And>j. v (s (Suc j)) = v (s j))"
  proof -
   fix j
   show "v (s (Suc j)) = v (s j)"
   using A2  
     proof (induct j)
     case 0
     then show ?case by auto
     next
     case (Suc j)
     then show ?case by metis
     qed
  qed
 qed
 have 4: "(\<forall>j. v (s j) = v (s 0)) \<Longrightarrow> (\<forall>j. v (s (Suc j)) = v (s j))"
 using 3 by blast
 from 2 4 show ?thesis by blast 
qed


lemma stable_defs:
 "(s \<Turnstile> stable v) =
  (case s of (Inl s) \<Rightarrow>  (\<forall> i\<le>intlen s. (v (nth s i)) = (v (nth s 0)))
           | (Inr s) \<Rightarrow>  (\<forall> i. (v (s i)) = (v (s 0)))
  )"
by (simp add: stable_d_def gets_defs current_val_d_def sub_def stable_defs_help 
       subinterval_def upt_same stable_defs_infinite sum.case_eq_if )

lemma padded_defs :
 " (s \<Turnstile> padded v) = 
  (case s of (Inl s) \<Rightarrow> ((\<forall> i< intlen s. (v (nth s i)) = (v (nth s 0))) \<or> intlen s =0)
           | (Inr s) \<Rightarrow> ((\<forall> i. (v ( s i)) = (v ( s 0))) )
  ) "
proof 
 (simp add: padded_d_def stable_defs chop_d_def skip_defs empty_defs sum.case_eq_if)
 show "isl s \<longrightarrow>
    ((\<exists>n\<le>intlen (projl s).
         (\<forall>i. i \<le> n \<and> i \<le> intlen (projl s) \<longrightarrow> 
            v (nth (projl s) i) = v (nth (projl s) 0)) \<and> intlen (projl s) - n = Suc 0) \<or>
     intlen (projl s) = 0) =
    ((\<forall>i<intlen (projl s). v (nth (projl s) i) = v (nth (projl s) 0)) \<or> intlen (projl s) = 0) "
 proof rule+
  show "\<And>i. isl s \<Longrightarrow>
         (\<exists>n\<le>intlen (projl s).
             (\<forall>i. i \<le> n \<and> i \<le> intlen (projl s) \<longrightarrow> v (nth (projl s) i) = v (nth (projl s) 0))
              \<and> intlen (projl s) - n = Suc 0) \<or>
         intlen (projl s) = 0 \<Longrightarrow>
         i < intlen (projl s) \<Longrightarrow> v (Interval.nth (projl s) i) = v (Interval.nth (projl s) 0)" 
  by (metis One_nat_def Suc_leI Suc_le_mono le_add_diff_inverse2 less_imp_le_nat not_less_zero 
      plus_1_eq_Suc)
  show "isl s \<Longrightarrow>
    (\<forall>i<intlen (projl s). 
       v (nth (projl s) i) = v (nth (projl s) 0)) \<or> intlen (projl s) = 0 \<Longrightarrow>
    (\<exists>n\<le>intlen (projl s).
        (\<forall>i. i \<le> n \<and> i \<le> intlen (projl s) \<longrightarrow> 
       v (nth (projl s) i) = v (nth (projl s) 0)) \<and> intlen (projl s) - n = Suc 0) \<or>
    intlen (projl s) = 0 " 
  by (metis Suc_leI Suc_pred diff_diff_cancel diff_le_self gr_zeroI le_imp_less_Suc)
 qed
qed

lemma padded_temporal_assign_defs :
 " (s \<Turnstile> v <\<sim> e) = 
  ((s \<Turnstile> padded v) \<and>
    (case s of (Inl s) \<Rightarrow> (v (Interval.nth s (intlen s)) ) = e (Inl s)
             | (Inr s) \<Rightarrow> True)
   )"
by (auto simp add: padded_temp_assign_d_def padded_defs temporal_assign_defs)



subsection \<open>Soundness Axioms\<close>

subsubsection \<open>ChopAssoc\<close>

lemma ChopAssocSemHelpa:
assumes "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (Inl (prefix i \<sigma>) \<Turnstile> f) \<and>
    (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h))" 
shows " (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and> 
    (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h))" 
proof -
 have 1: "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (Inl (prefix i \<sigma>) \<Turnstile> f) \<and>
    (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h))" 
 using assms by auto
 obtain i ia where 2: "i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (Inl (prefix i \<sigma>) \<Turnstile> f) \<and>
    (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h)" 
   using "1" by auto
 have 3: "(Inl (suffix (ia+i) \<sigma>) \<Turnstile> h) "
    using "2" by auto
 have 4: "ia +i \<le>intlen \<sigma> " 
    using "2" Nat.le_diff_conv2 by blast
 have 5: " i \<le> ia+i"
    by simp 
 have 6: "(Inl (suffix i (prefix (ia +i) \<sigma>)) \<Turnstile> g)" 
    using "2" "4" interval_suffix_prefix_swap by force
 have 7: "(Inl (prefix i (prefix (ia +i) \<sigma>)) \<Turnstile> f)" 
    by (simp add: "2" add.commute)
 show ?thesis using "2" "4" "5" "6" "7" by blast
qed

lemma ChopAssocSemHelpb:
assumes " (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and> 
    (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h))" 
shows "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (Inl (prefix i \<sigma>) \<Turnstile> f) \<and>
    (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h))" 
proof -
 have 1: "(\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and> 
    (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h))"
  using assms by auto
 obtain j ja where 2: "j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and> 
    (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h) "  
 using "1" by auto
 have 3: "ja\<le>intlen \<sigma>" 
   using "2" le_trans by blast
 have 4: "j -ja \<le> intlen \<sigma> -ja" 
   by (simp add: "2" diff_le_mono)
 have 5: "(Inl (prefix ja \<sigma>) \<Turnstile> f)" 
    by (metis "2" interval_pref_pref_3 le_add_diff_inverse)
 have 6: "(Inl (prefix (j -ja) (suffix ja \<sigma>)) \<Turnstile> g)" 
   by (simp add: "2" interval_suffix_prefix_swap)
 have 7: "(Inl (suffix ((j -ja) + ja) \<sigma>) \<Turnstile> h)" 
   by (simp add: "2")
 show ?thesis using "3" "4" "5" "6" "7" by blast
qed


lemma ChopAssocSemHelp:
 "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (Inl (prefix i \<sigma>) \<Turnstile> f) \<and>
    (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h)) =
  (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and> 
    (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h))"
 using ChopAssocSemHelpa[of "\<sigma>" "f" "g" "h"] 
      ChopAssocSemHelpb[of "\<sigma>" "f" "g" "h"] by auto 


lemma ChopAssocSemHelpFinite:
 " ((Inl \<sigma>) \<Turnstile>  f ; (g ; h)) = ((Inl \<sigma>) \<Turnstile> (f;g);h)"
proof -
 have " ((Inl \<sigma>) \<Turnstile>  f ; (g ; h)) =
        ((\<exists>i\<le>intlen \<sigma>. (Inl (prefix i \<sigma>) \<Turnstile> f) \<and> (\<exists>ia\<le>intlen (suffix i \<sigma>). 
          (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h))))" 
 by (simp add: chop_defs)
 also have " ... = 
             (\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> ((Inl (prefix i \<sigma>)) \<Turnstile> f) \<and>
         (Inl (prefix ia (suffix i \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix (ia + i) \<sigma>) \<Turnstile> h))" 
 by fastforce
 also have " ... =
            (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (Inl (prefix ja (prefix j \<sigma>)) \<Turnstile> f) \<and>
            (Inl (suffix ja (prefix j \<sigma>)) \<Turnstile> g) \<and> (Inl (suffix j \<sigma>) \<Turnstile> h))" 
 using ChopAssocSemHelp[of \<sigma> f g h]  by blast
 also have " ... =
            (\<exists>i\<le>intlen \<sigma>. (\<exists>ia\<le>intlen (prefix i \<sigma>).  (Inl (prefix ia (prefix i \<sigma>)) \<Turnstile> f) \<and>
             (Inl (suffix ia (prefix i \<sigma>)) \<Turnstile> g)) \<and> (Inl (suffix i \<sigma>) \<Turnstile> h)) " 
 by fastforce
 also have " ... =
             (Inl \<sigma> \<Turnstile> (f;g);h)" by (simp add: chop_defs)
 finally show "(Inl \<sigma> \<Turnstile>  f ; (g ; h)) = (Inl \<sigma> \<Turnstile> (f;g);h) " .
qed

lemma ChopAssocSemHelpInFinite:
 " ((Inr \<sigma>) \<Turnstile>  f ; (g ; h)) = ((Inr \<sigma>) \<Turnstile> (f;g);h)"
proof -
 have "((Inr \<sigma>) \<Turnstile>  f ; (g ; h)) =
         ((\<exists>n.
        f (Inl (iprefix n \<sigma>)) \<and>
        ((\<exists>na. g (Inl (iprefix na (isuffix n \<sigma>))) \<and> h (Inr (isuffix na (isuffix n \<sigma>)))) \<or>
         g (Inr (isuffix n \<sigma>)))) \<or>
    f (Inr \<sigma>))" 
    by (metis chop_defs_infinite)
 also have " ... =
             ((\<exists>n na.
        f (Inl (iprefix n \<sigma>)) \<and>
        (( g (Inl (iprefix na (isuffix n \<sigma>))) \<and> h (Inr (isuffix na (isuffix n \<sigma>)))) \<or>
         g (Inr (isuffix n \<sigma>)))) \<or>
    f (Inr \<sigma>))"  
 by fastforce
 also have " ... =
             ((\<exists>n na. na\<le>intlen (iprefix n \<sigma>) \<and>
        ( f (Inl (prefix na (iprefix n \<sigma>))) \<and> g (Inl (suffix na (iprefix n \<sigma>)))) \<and>
        h (Inr (isuffix n \<sigma>))) \<or>
    (\<exists>n::nat. f (Inl (iprefix n \<sigma>)) \<and> g (Inr (isuffix n \<sigma>))) \<or> f (Inr \<sigma>))
   " 
  by (auto simp add: add.commute interval_iprefix_isuffix_swap iprefix_length isuffix_isuffix)
     (metis interval_pref_ipref_3 ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 also have " ... = 
             ((Inr \<sigma>) \<Turnstile> (f;g);h)"
  by (metis chop_defs_finite chop_defs_infinite le_add2 le_add_same_cancel2)
 finally show "((Inr \<sigma>) \<Turnstile>  f ; (g ; h)) = ((Inr \<sigma>) \<Turnstile> (f;g);h)" .
qed

lemma ChopAssocSem:
  " (\<sigma> \<Turnstile>  f ; (g ; h) =  (f;g);h)"
by auto
   (metis ChopAssocSemHelpFinite ChopAssocSemHelpInFinite sum.collapse(1) sum.collapse(2))+


subsubsection \<open>OrChopImp\<close>

lemma OrChopImpSem:
  " (\<sigma> \<Turnstile>  ( f \<or> g);h   \<longrightarrow>  f;h \<or> g;h ) "
by (auto simp add: chop_defs sum.case_eq_if)
 

subsubsection \<open>ChopOrImp\<close>

lemma ChopOrImpSem:
  " (\<sigma> \<Turnstile> f;(g \<or> h) \<longrightarrow>   f;g \<or>  f;h ) "
by (auto simp add: chop_defs sum.case_eq_if)

subsubsection \<open>EmptyChop\<close>

lemma EmptyChopSemFinite:
  " ((Inl \<sigma>) \<Turnstile> empty ; f = f ) "
using min.absorb1 by (simp add: empty_defs chop_defs) force

lemma EmptyChopSemInfinite:
  " ((Inr \<sigma>) \<Turnstile> empty ; f = f ) "
by (simp add: chop_defs empty_defs iprefix_length isuffix_0)

lemma EmptyChopSem:
  " (\<sigma> \<Turnstile> empty ; f = f ) "
using EmptyChopSemFinite EmptyChopSemInfinite 
by (metis sum.collapse(1) sum.collapse(2))



subsubsection \<open>ChopEmpty\<close>

lemma ChopEmptySemFinite:
  " ((Inl \<sigma>) \<Turnstile> f;empty = f ) "
by (simp add: empty_defs chop_defs) auto

lemma ChopEmptySemInfinite:
  " ((Inr \<sigma>) \<Turnstile> f;empty = f ) "
by (simp add: chop_defs empty_defs)

lemma ChopEmptySem:
  " (\<sigma> \<Turnstile> f;empty = f ) "
using ChopEmptySemFinite ChopEmptySemInfinite
by (metis sum.collapse(1) sum.collapse(2))

subsubsection \<open>StateImpBi\<close>

lemma StateImpBiSem:
  " (\<sigma> \<Turnstile> init f \<longrightarrow>   bi (init f) ) "
by (simp add: init_defs bi_defs  sum.case_eq_if)
   (metis conc_def conc_iprefix_isuffix interval_intlen_gr_zero iprefix_0)

subsubsection \<open>NextImpNotNextNot\<close>

lemma NextImpNotNextNotSem:
  " (\<sigma> \<Turnstile> \<circle> f \<longrightarrow>  \<not> (\<circle> (\<not> f)) ) "
by (simp add: next_defs sum.case_eq_if) 

subsubsection \<open>BiBoxChopImpChop\<close>

lemma BiBoxChopImpChopSem:
  " (\<sigma> \<Turnstile> bi ( f \<longrightarrow> f1) \<and> \<box>(g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f1;g1 ) "
by (simp add: bi_defs always_defs chop_defs sum.case_eq_if)
   fastforce

subsubsection \<open>BoxInduct\<close>

lemma box_induct_help_1 :
  "((Inl \<sigma>) \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (Inl (suffix i \<sigma>) \<Turnstile> f) \<longrightarrow> (Inl (suffix (Suc i) \<sigma>) \<Turnstile> f)) 
     \<Longrightarrow> (\<forall> j. j \<le> intlen \<sigma> \<longrightarrow> (Inl (suffix j \<sigma>) \<Turnstile> f)) "
proof
    fix j
    show " ((Inl \<sigma>) \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (Inl (suffix i \<sigma>) \<Turnstile> f) \<longrightarrow> (Inl (suffix (Suc i) \<sigma>) \<Turnstile> f)) 
          \<Longrightarrow>  j \<le> intlen \<sigma> \<longrightarrow> (Inl (suffix j \<sigma>) \<Turnstile> f)" 
     proof (induct j arbitrary: \<sigma>)
     case 0
     then show ?case by simp
     next
     case (Suc j)
     then show ?case by (metis Nat.le_diff_conv2 One_nat_def Suc_eq_plus1_left Suc_leD)
     qed
qed

lemma box_induct_help_infinite :
  "((Inr \<sigma>) \<Turnstile> f) \<and> (\<forall>i. 
           ((Inr (isuffix i \<sigma>) \<Turnstile> f) \<longrightarrow> (Inr (isuffix (Suc i) \<sigma>) \<Turnstile> f))) 
     \<Longrightarrow> (\<forall> j.   (Inr (isuffix j \<sigma>) \<Turnstile> f)) "
proof
    fix j
    show " ((Inr \<sigma>) \<Turnstile> f) \<and> (\<forall>i.  
          (Inr (isuffix i \<sigma>) \<Turnstile> f) \<longrightarrow> (Inr (isuffix (Suc i) \<sigma>) \<Turnstile> f)) 
          \<Longrightarrow>  (Inr (isuffix j \<sigma>) \<Turnstile> f)" 
     proof (induct j arbitrary: \<sigma>)
     case 0
     then show ?case by (simp add: isuffix_0)
     next
     case (Suc j)
     then show ?case by blast
     qed
qed


lemma BoxInductSem:
  " (\<sigma> \<Turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f)"
proof 
 (auto simp add: always_defs wnext_defs sum.case_eq_if)
 show "\<And>n. isl \<sigma> \<Longrightarrow>
         \<forall>n\<le>intlen (projl \<sigma>). f (Inl (suffix n (projl \<sigma>))) \<longrightarrow> 
         intlen (projl \<sigma>) = n \<or> f (Inl (suffix (Suc n) (projl \<sigma>))) \<Longrightarrow>
         f \<sigma> \<Longrightarrow> n \<le> intlen (projl \<sigma>) \<Longrightarrow> f (Inl (suffix n (projl \<sigma>)))" 
 by (metis One_nat_def box_induct_help_1 cancel_comm_monoid_add_class.diff_cancel not_one_le_zero 
     sum.collapse(1))
 show "\<And>n. \<not> isl \<sigma> \<Longrightarrow> 
      \<forall>n. f (Inr (isuffix n (projr \<sigma>))) \<longrightarrow> f (Inr (isuffix (Suc 0) (isuffix n (projr \<sigma>)))) \<Longrightarrow> 
      f \<sigma> \<Longrightarrow> f (Inr (isuffix n (projr \<sigma>))) " 
 by (metis (mono_tags) add.right_neutral add_Suc_right box_induct_help_infinite isuffix_isuffix 
     sum.collapse(2))
qed

subsubsection \<open>ChopStarEqv\<close>

lemma ChopExist:
 "\<turnstile>  (\<exists> k. f;g k) = f;(\<exists> k. g k)"
by (auto simp add: chop_defs Valid_def sum.case_eq_if)

lemma SChopExist:
 "\<turnstile>  (\<exists> k. f\<frown>g k) = f\<frown>(\<exists> k. g k)"
by (auto simp add: schop_defs Valid_def sum.case_eq_if)

lemma ExistChop:
 "\<turnstile> (\<exists> k. (g k);f) =  (\<exists> k. g k);f"
by (auto simp add: chop_defs Valid_def sum.case_eq_if)

lemma ExistSChop:
 "\<turnstile> (\<exists> k. (g k)\<frown>f) =  (\<exists> k. g k)\<frown>f"
by (auto simp add: schop_defs Valid_def sum.case_eq_if)

lemma powersem1:
 "(\<sigma> \<Turnstile>  (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k)))) "
proof auto
 show "\<And>x. \<sigma> \<Turnstile> (power f x)  \<Longrightarrow> \<forall>k. \<not> (\<sigma> \<Turnstile>  (f \<and> finite);power f k)  \<Longrightarrow> \<sigma> \<Turnstile> empty " 
 by (metis not0_implies_Suc pow_0 pow_Suc)
 show " \<sigma> \<Turnstile> empty  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (power f x)  "
 by (metis pow_0)
 show " \<And>k. \<sigma> \<Turnstile> ((f \<and> finite);power f k)  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (power f x)   " 
 by (metis pow_Suc)
qed

lemma spowersem1:
 "(\<sigma> \<Turnstile>  (\<exists> k. spower f k) = ( empty \<or> (\<exists> k. spower f (Suc k)))) "
proof auto
 show "\<And>x. \<sigma> \<Turnstile> (spower f x)  \<Longrightarrow> \<forall>k. \<not> (\<sigma> \<Turnstile>  f \<frown>spower f k)  \<Longrightarrow> \<sigma> \<Turnstile> empty" 
 by (metis not0_implies_Suc spow_0 spow_Suc)
 show " \<sigma> \<Turnstile> empty  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (spower f x)  "
 by (metis spow_0)
 show "\<And>k. \<sigma> \<Turnstile> (f \<frown> spower f k)  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (spower f x) " 
 by (metis spow_Suc)
qed

lemma powersem:
 " \<turnstile> (\<exists> k. power f k) = (  empty \<or> (f \<and> finite);(\<exists> k. (power f k)))"
proof -
 have 1: "\<turnstile> (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k))) "
 using powersem1 by blast
 have 2: "\<turnstile> (\<exists> k. power f (Suc k)) = (\<exists> k. (f \<and> finite);power f k) "
 by simp
 have 3: "\<turnstile> (\<exists> k. (f \<and> finite);(power f k)) = (f \<and> finite);(\<exists> k. (power f k)) " 
 using ChopExist by blast
 from 1 2 3 show ?thesis by fastforce
qed

lemma spowersem:
 " \<turnstile> (\<exists> k. spower f k) = (  empty \<or> f\<frown>(\<exists> k. (spower f k)))"
proof -
 have 1: "\<turnstile> (\<exists> k. spower f k) = ( empty \<or> (\<exists> k. spower f (Suc k))) "
 using spowersem1 by blast
 have 2: "\<turnstile> (\<exists> k. spower f (Suc k)) = (\<exists> k. f\<frown>spower f k) "
 by simp
 have 3: "\<turnstile> (\<exists> k. f\<frown>(spower f k)) = f\<frown>(\<exists> k. (spower f k)) " 
 using SChopExist by blast
 from 1 2 3 show ?thesis by fastforce
qed

lemma PowerstarEqvSemhelp1:
 " \<turnstile> empty;(empty \<or> (f \<and> inf)) = (empty \<or> (f \<and> inf))"
using EmptyChopSem by blast

lemma PowerstarEqvSemhelp2:
 " \<turnstile> (f \<and> inf) = (f \<and> inf);g"
by (simp add: Valid_def infinite_defs chop_defs sum.case_eq_if)
  
lemma PowerstarEqvSemhelp3:
 " \<turnstile> ((f \<and> inf);g \<or> (f \<and> finite);g) = (f ;g)"
by (auto simp add: Valid_def finite_defs infinite_defs chop_defs sum.case_eq_if)

lemma PowerstarEqvSem:
  " (\<sigma> \<Turnstile> (powerstar f) =  (empty \<or>  f;(powerstar f) ))"
proof -
 have 1: "(\<sigma> \<Turnstile> (powerstar f)) = 
          ( \<sigma> \<Turnstile> (\<exists> k. power f k);(empty \<or> f \<and> inf)) "
 by (simp add: powerstar_d_def)
 have 2: "( \<sigma> \<Turnstile> (\<exists> k. power f k);(empty \<or> f \<and> inf)) =
          (\<sigma> \<Turnstile> (empty \<or> (f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))"
 using powersem by (metis inteq_reflection)
 have 3: " (\<sigma> \<Turnstile> (empty \<or> (f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf)) =
           (\<sigma> \<Turnstile> empty;(empty \<or> (f \<and> inf)) \<or> 
                ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))"
 by (auto simp add: chop_defs sum.case_eq_if)
 have 4: "(\<sigma> \<Turnstile> empty;(empty \<or> (f \<and> inf)) \<or> 
                ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf)) =
          (\<sigma> \<Turnstile> (empty \<or> (f \<and> inf) \<or> ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))) "
 using PowerstarEqvSemhelp1 
 by (metis (mono_tags, lifting) inteq_reflection unl_lift2)
 have 5: "(\<sigma> \<Turnstile> (empty \<or> (f \<and> inf) \<or> ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))) =
          (\<sigma> \<Turnstile> (empty \<or> (f \<and> inf);((\<exists> k. (power f k));(empty \<or> f \<and> inf)) \<or>
                     ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))) " 
 using PowerstarEqvSemhelp2 
 by (metis (mono_tags, lifting) inteq_reflection)
 have 6: "(\<sigma> \<Turnstile> (empty \<or> (f \<and> inf);((\<exists> k. (power f k));(empty \<or> f \<and> inf)) \<or>
                     ((f \<and> finite);(\<exists> k. (power f k)));(empty \<or> f \<and> inf))) =
          (\<sigma> \<Turnstile> (empty \<or> (f \<and> inf);((\<exists> k. (power f k));(empty \<or> f \<and> inf)) \<or>
                      (f \<and> finite);((\<exists> k. (power f k));(empty \<or> f \<and> inf))))
         "
  by auto
     (metis  ChopAssocSemHelpFinite ChopAssocSemHelpInFinite sum.split_sel)+
 have 7 : "(\<sigma> \<Turnstile> (empty \<or> (f \<and> inf);((\<exists> k. (power f k));(empty \<or> f \<and> inf)) \<or>
                      (f \<and> finite);((\<exists> k. (power f k));(empty \<or> f \<and> inf)))) =
           (\<sigma> \<Turnstile> (empty \<or> f;((\<exists> k. (power f k));(empty \<or> f \<and> inf))))"
 using PowerstarEqvSemhelp3 by fastforce
 have 8: "(\<sigma> \<Turnstile> (empty \<or> f;((\<exists> k. (power f k));(empty \<or> f \<and> inf)))) =
          (\<sigma> \<Turnstile> (empty \<or> f;(powerstar f))) "
 by (simp add:  powerstar_d_def)
 from 1 2 3 4 5 6 7 8 show ?thesis by fastforce
qed


lemma FPowerstarEqvSem:
  " (\<sigma> \<Turnstile> (fpowerstar f) =  (empty \<or>  (f \<and> finite);(fpowerstar f) ))"
proof -
 have 1: "(\<sigma> \<Turnstile> (fpowerstar f)) = 
          ( \<sigma> \<Turnstile> (\<exists> k. power f k)) "
 by (simp add: fpowerstar_d_def)
 have 2: "( \<sigma> \<Turnstile> (\<exists> k. power f k)) =
          (\<sigma> \<Turnstile> (empty \<or> (f \<and> finite);(\<exists> k. (power f k))))"
 using powersem by (metis inteq_reflection)
 from 1 2 show ?thesis by (simp add: fpowerstar_d_def)
qed

lemma SPowerstarEqvSem:
  " (\<sigma> \<Turnstile> (spowerstar f) =  (empty \<or>  f\<frown>(spowerstar f) ))"
proof -
 have 1: "(\<sigma> \<Turnstile> (spowerstar f)) = 
          ( \<sigma> \<Turnstile> (\<exists> k. spower f k)) "
 by (simp add: spowerstar_d_def)
 have 2: "( \<sigma> \<Turnstile> (\<exists> k. spower f k)) =
          (\<sigma> \<Turnstile> (empty \<or> f\<frown>(\<exists> k. (spower f k))))"
 using spowersem by (metis inteq_reflection)
 from 1 2 show ?thesis by (simp add: spowerstar_d_def)
qed

lemma powerchopsem:
 " \<turnstile> (\<exists> k. power (f \<and> more) k) = 
      ( empty \<or> ((f \<and> more) \<and> finite);(\<exists> k. (power (f \<and> more) k))
      )"
using powersem by auto

lemma spowerchopsem:
 " \<turnstile> (\<exists> k. spower (f \<and> more) k) = 
      ( empty \<or> (f \<and> more)\<frown>(\<exists> k. (spower (f \<and> more) k))
      )"
using spowersem by auto

lemma ChopstarEqvSem:
  " (\<sigma> \<Turnstile> f\<^sup>\<star> =  (empty \<or>  (f \<and> more); f\<^sup>\<star>) )"
by (metis PowerstarEqvSem chopstar_d_def)

lemma SChopstarEqvSem:
  " (\<sigma> \<Turnstile> (schopstar f) =  (empty \<or>  (f \<and> more)\<frown> (schopstar f) ))"
by (metis SPowerstarEqvSem schopstar_d_def)

subsubsection \<open>OmegaUnroll\<close>

lemma omega_unroll_chain:
 "(\<exists>l. infinite_index_sequence 0 l \<and> (\<forall>i. f (Inl (subinterval \<sigma> (l i) (l (Suc i)))))) 
  =
   (\<exists>n.
        f (Inl (iprefix n \<sigma>)) \<and>
        0 < n \<and>
        (\<exists>l.
            infinite_index_sequence 0 l \<and> 
            (\<forall>i. f (Inl (subinterval (isuffix n \<sigma>) (l i) (l (Suc i)))))
        )
   ) 
    "
proof -
 have  " (\<exists>l. infinite_index_sequence 0 l \<and> 
              (\<forall>i. f (Inl (subinterval \<sigma> (l i) (l (Suc i)))))) =
         (\<exists>l.  infinite_index_sequence 0 l \<and> l = conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) \<and>
              (\<forall>i. f (Inl (subinterval \<sigma> (l i) (l (Suc i))))))" 
         using iidx_1 by blast
 also have " ... =
             (\<exists>l.  infinite_index_sequence 0 (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1)))) \<and> 
                   l = conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) \<and>
                   (\<forall>i. f (Inl (subinterval \<sigma> (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) i) 
                                              (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) (Suc i))))
                   )
             )"
      by force
 also have " ... =
             (\<exists>l.  (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
                   infinite_index_sequence (l 1) (\<lambda>x. (l (x+1))) \<and> 
                   l = conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) \<and>
                   (\<forall>i. f (Inl (subinterval \<sigma> (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) i) 
                                              (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) (Suc i))))
                   )
             )
           " 
          using iidx_2 by force
 also have " ... =
             (\<exists>l.  (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
                   infinite_index_sequence (l 1) (\<lambda>x. (l (x+1))) \<and> 
                   l = conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1))) \<and>
                   f (Inl (subinterval \<sigma> (l 0) (l 1))) \<and> 
                   (\<forall>i. f (Inl (subinterval \<sigma>  ((\<lambda>x. (l (x+1))) i) ( (\<lambda>x. (l (x+1))) (Suc i)))))
             )
           " (is "?L=?R")
       proof rule
         show " ?L \<Longrightarrow> ?R" 
        by (metis One_nat_def Suc_eq_plus1)
        show "?R \<Longrightarrow> ?L" 
       by (metis (lifting) One_nat_def conc_empty_suc not0_implies_Suc)
     qed
 also have " ... =
             (\<exists>l ls.  ls = (\<lambda>x. (l (x+1))) \<and> (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
                      infinite_index_sequence (l 1) ls \<and> 
                      l = conc \<langle>(l 0)\<rangle> ls \<and>
                      f (Inl (subinterval \<sigma> (l 0) (l 1))) \<and> 
                      (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )
           "       
          by force
 also have " ... =
             (\<exists>l ls.  ls = (\<lambda>x. (l (x+1))) \<and> (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
                      infinite_index_sequence (l 1) ls \<and> 
                      l = conc \<langle>(l 0)\<rangle> ls \<and>
                      f (Inl (iprefix (l 1) \<sigma>)) \<and> 
                      (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )
           "
       by (metis iprefix_def)
 also have " ... =
             (\<exists>l ls n. n = (ls 0) \<and> ls = (\<lambda>x. (l (x+1))) \<and> 0 < n \<and> 
                       infinite_index_sequence n ls \<and> l = conc \<langle>0\<rangle> ls \<and>
                       f (Inl (iprefix n \<sigma>)) \<and> 
                       (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )
           "
        by (metis (no_types, lifting)  One_nat_def conc_empty_suc conc_empty_zero)
 also have " ... = 
             (\<exists> ls n. n = (ls 0) \<and> 0 < n \<and> 
                      infinite_index_sequence n ls \<and> 
                      f (Inl (iprefix n \<sigma>)) \<and> 
                      (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )
            "
           using iidx_0  by rule (metis (no_types, lifting) Suc_eq_plus1 conc_empty_suc)
 also have " ... =
             (\<exists> ls n lsk.  n = (ls 0) \<and> 0 < n \<and> lsk = (shiftm n) \<circ> ls \<and>
                           infinite_index_sequence n ls \<and> 
                           f (Inl (iprefix n \<sigma>)) \<and> 
                           (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )"
   by blast
 also have " ... =
             (\<exists> ls n lsk.  n = (ls 0) \<and> 0 < n \<and> lsk = (shiftm n) \<circ> ls \<and>
                           infinite_index_sequence n ls \<and> infinite_index_sequence 0 lsk \<and>
                           f (Inl (iprefix n \<sigma>)) \<and> 
                           (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
             )"
   using iidx_5 by auto
 also have "... =
            (\<exists> ls n lsk.  n = (ls 0) \<and> 0 < n \<and> ls = (shift n) \<circ> lsk \<and>
                          infinite_index_sequence n ls \<and> infinite_index_sequence 0 lsk \<and>
                          f (Inl (iprefix n \<sigma>)) \<and> 
                          (\<forall>i. f (Inl (subinterval \<sigma>  (ls i) (ls (Suc i)))))
            )"
   using iidx_6 by blast
  also have " ... =
             (\<exists> ls n lsk.  n = (((shift n) \<circ> lsk) 0) \<and> 0 < n \<and> ls = (shift n) \<circ> lsk \<and>
                           infinite_index_sequence n ls \<and> infinite_index_sequence 0 lsk \<and>
                           f (Inl (iprefix n \<sigma>)) \<and> 
                           (\<forall>i. f (Inl (subinterval \<sigma>  (((shift n) \<circ> lsk) i) 
                                                       (((shift n) \<circ> lsk) (Suc i)) ))
                           )
             )"
  by metis
  also have " ... =
              (\<exists> ls n lsk. 0 = (lsk 0) \<and> 0 < n \<and> ls = (shift n) \<circ> lsk \<and>
                           infinite_index_sequence n ls \<and> infinite_index_sequence 0 lsk \<and>
                           f (Inl (iprefix n \<sigma>)) \<and> 
                           (\<forall>i. f (Inl (subinterval \<sigma>  ((lsk i)+n) ((lsk (Suc i)+n)) )))
              )"
  using shift_def by auto
  also have " ... =
              (\<exists> ls n lsk. 0 = (lsk 0) \<and> 0 < n \<and> ls = (shift n) \<circ> lsk \<and>
                           infinite_index_sequence 0 lsk \<and>
                           f (Inl (iprefix n \<sigma>)) \<and> 
                           (\<forall>i. f (Inl (subinterval \<sigma>  ((lsk i)+n) ((lsk (Suc i)+n)) )))
               )"
  using iidx_8 by blast
  also have " ... =
              (\<exists> n lsk. 0 = (lsk 0) \<and> 0 < n \<and> 
                        infinite_index_sequence 0 lsk \<and>
                        f (Inl (iprefix n \<sigma>)) \<and> 
                        (\<forall>i. f (Inl (subinterval \<sigma>  ((lsk i)+n) ((lsk (Suc i)+n)) )))
              )"
   by blast

  also have " ... =
              (\<exists> n lsk. 0 = (lsk 0) \<and> 0 < n \<and> 
                        infinite_index_sequence 0 lsk \<and>
                        f (Inl (iprefix n \<sigma>)) \<and> 
                        (\<forall>i. f (Inl (subinterval (isuffix n \<sigma>)  ((lsk i)) ((lsk (Suc i))) )))
              )"
  using  subinterval_sub_isuffix_iidx by (metis  calculation calculation )
 also have " ... =
             (\<exists>n. f (Inl (iprefix n \<sigma>)) \<and> 0 < n \<and>
                  (\<exists>l. infinite_index_sequence 0 l \<and>
                       (\<forall>i. f (Inl (subinterval (isuffix n \<sigma>) (l i) (l (Suc i)))))
                  )
             )"
  using infinite_index_sequence_def by auto
  finally show "(\<exists>l. infinite_index_sequence 0 l \<and> 
                     (\<forall>i. f (Inl (subinterval \<sigma> (l i) (l (Suc i)))))
                ) 
                =
                (\<exists>n. f (Inl (iprefix n \<sigma>)) \<and> 0 < n \<and>
                     (\<exists>l. infinite_index_sequence 0 l \<and> 
                          (\<forall>i. f (Inl (subinterval (isuffix n \<sigma>) (l i) (l (Suc i)))))
                     )
                )" .
qed
        

lemma omega_unroll_sem:
 "((Inr \<sigma>) \<Turnstile> (f \<and> fmore);(omega f) = (omega f))"
proof 
 (simp add: fmore_defs chop_defs omega_d_def iprefix_length) 
 show "(\<exists>n. f (Inl (iprefix n \<sigma>)) \<and>
         0 < n \<and> (\<exists>l. infinite_index_sequence 0 l \<and> 
        (\<forall>i. f (Inl (subinterval (isuffix n \<sigma>) (l i) (l (Suc i))))))) =
    (\<exists>l. infinite_index_sequence 0 l \<and> (\<forall>i. f (Inl (subinterval \<sigma> (l i) (l (Suc i))))))" 
    using omega_unroll_chain by metis
qed

lemma OmegaUnrollSem:
 "\<sigma> \<Turnstile> (omega f) = (f \<and> fmore);(omega f)  "
proof (cases \<sigma>)
case (Inl a)
then show ?thesis by (simp add: omega_d_def chop_defs_finite)
next
case (Inr b)
then show ?thesis using omega_unroll_sem by fastforce
qed


subsubsection \<open>OmegaInduct\<close>

lemma OmegaInductSem_help:
 "(\<sigma> \<Turnstile>  inf \<and> g \<and> \<box>(g \<longrightarrow> (f \<and> fmore);g)) =
  ( case \<sigma> of (Inl \<sigma>) \<Rightarrow> False
             | (Inr \<sigma>) \<Rightarrow> 
               g (Inr \<sigma>) \<and>
               (\<forall>n::nat. g (Inr (isuffix n \<sigma>)) \<longrightarrow>
                       (\<exists>na::nat. f (Inl (subinterval \<sigma> n (na+n))) \<and> 
                            (0::nat) < na \<and> g (Inr (isuffix (n + na) \<sigma>))))
   ) 
  "
by (simp add: infinite_defs always_defs chop_defs fmore_defs isuffix_isuffix sum.case_eq_if)
   (metis iprefix_isuffix iprefix_length) 

lemma OmegaInductSem_help_infinite:
 "((Inr \<sigma>) \<Turnstile>  inf \<and> g \<and> \<box>(g \<longrightarrow> (f \<and> fmore);g)) =
               (g (Inr \<sigma>) \<and>
               (\<forall>n::nat. g (Inr (isuffix n \<sigma>)) \<longrightarrow>
                       (\<exists>na::nat. f (Inl (subinterval \<sigma> n (na+n))) \<and> 
                            (0::nat) < na \<and> g (Inr (isuffix (na+n) \<sigma>)))))
   
  "
by ( simp add: infinite_defs always_defs chop_defs fmore_defs isuffix_isuffix sum.case_eq_if)
 (metis iprefix_isuffix iprefix_length add.commute) 


primrec cpoint :: " ('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> nat \<Rightarrow> 'a infinterval \<Rightarrow> nat"
 where "cpoint f g 0 \<sigma> = 0"
       | "cpoint f g (Suc n) \<sigma> =
          (\<some> x. (\<exists> m. (Inl (subinterval \<sigma> (cpoint f g n \<sigma>) (m+(cpoint f g n \<sigma>))  ) \<Turnstile> f) 
                 \<and> m>0 \<and>(Inr(isuffix (m+(cpoint f g n \<sigma>)) \<sigma>) \<Turnstile> g) \<and>
                 x=m+(cpoint f g n \<sigma>)   
                )
          )"

lemma cpoint_order_0a:
 "na\<le>x \<Longrightarrow>  intlen (suffix na (iprefix x \<sigma>)) = x-na"
by (simp add: iprefix_length) 

lemma cpoint_expand_0:
 "(cpoint f g 0 \<sigma>) =0"
by simp 

lemma cpoint_expand_1:
 "(cpoint f g 1 \<sigma>) = 
   (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> 0 (m) ))
                  \<and> m>0 \<and> g (Inr (isuffix (m) \<sigma>))
                  \<and> x=m))"
by (simp add: fmore_defs subinterval_length)

lemma cpoint_expand_n:
 "(cpoint f g (Suc n) \<sigma>) = 
  (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g n \<sigma>) (m+(cpoint f g n \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g n \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g n \<sigma>)) 
  )"
by (simp add: fmore_defs subinterval_length)

lemma cpoint_0:
 assumes "g (Inr \<sigma>) \<and>
               (\<forall>k. g (Inr (isuffix k \<sigma>)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval \<sigma> k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) \<sigma>))))
         "
 shows " g (Inr(isuffix (cpoint f g i \<sigma>) \<sigma>)) " 
proof
 (induct i)
 case 0
 then show ?case by (simp add: assms isuffix_0)
 next
 case (Suc i)
 then show ?case 
 proof -
  have 1: "g (Inr(isuffix (cpoint f g i \<sigma>) \<sigma>)) "
  by (simp add: Suc.hyps)
  have 2: "g (Inr(isuffix (cpoint f g i \<sigma>) \<sigma>)) \<longrightarrow>
           (\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g i \<sigma>) (m+(cpoint f g i \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g i \<sigma>)) \<sigma>))) "
  using assms by blast
  have 3: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g i \<sigma>) (m+(cpoint f g i \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g i \<sigma>)) \<sigma>))) "
  using 1 2 by auto
  have 4: "(cpoint f g (Suc i) \<sigma>) =
           (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g i \<sigma>) (m+(cpoint f g i \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g i \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g i \<sigma>)))  "
  by simp
  have 5: "g (Inr(isuffix ((cpoint f g (Suc i) \<sigma>)) \<sigma>)) "
  using 3 4 someI_ex[of "\<lambda>x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g i \<sigma>) (m+(cpoint f g i \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g i \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g i \<sigma>))"] by auto
  from 5 show ?thesis by auto
  qed
 qed



lemma cpoint_1:
 assumes "g (Inr \<sigma>) \<and>
               (\<forall>k. g (Inr (isuffix k \<sigma>)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval \<sigma> k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) \<sigma>))))
         "
 shows   "( g (Inr(isuffix (cpoint f g i \<sigma>) \<sigma>)) 
              \<Longrightarrow> g (Inr(isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>)))
  "
proof -
 have 1: "g (Inr \<sigma>) \<and>
               (\<forall>k. g (Inr (isuffix k \<sigma>)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval \<sigma> k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) \<sigma>)))) "
      using assms by blast
 have 2: "( g (Inr(isuffix (cpoint f g i \<sigma>) \<sigma>)) 
              \<Longrightarrow> g (Inr(isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>))) "
 proof  
    (induct i)
    case 0
    then show ?case 
    proof -
     have 01: "g (Inr(isuffix (cpoint f g 0 \<sigma>) \<sigma>)) =
               g (Inr(isuffix 0 \<sigma>)) " 
         by auto
     have 02: "g (Inr(isuffix 0 \<sigma>)) = g (Inr \<sigma>)"
         by (simp add: isuffix_0) 
     have 03: "(\<exists>m. f (Inl (subinterval \<sigma> 0 (m))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m) \<sigma>))) "
     using 02 1 by fastforce
     have 04: "(cpoint f g 1 \<sigma>) = 
                 (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> 0 (m) ))
                          \<and> m>0 \<and> g (Inr (isuffix (m) \<sigma>))
                          \<and> x=m)
                 )"
      using cpoint_expand_1 by blast
     have 05: "g (Inr (isuffix (cpoint f g 1 \<sigma>) \<sigma>)) "
     using 03 04 someI_ex by (metis (mono_tags, lifting))
     from 01 05 show ?thesis by auto
    qed
    next
    case (Suc i)
    then show ?case 
    proof - 
    have n1: "g (Inr (isuffix (cpoint f g i \<sigma>) \<sigma>)) \<Longrightarrow> g (Inr (isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>))"
    using Suc.hyps by blast
    have n2: "g (Inr (isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>)) "
    using Suc.prems by blast
    have n3: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))) "
    using assms n2 by auto
    have n4: "(cpoint f g (Suc (Suc i)) \<sigma>) = 
              (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>)) 
              ) "
           using cpoint_expand_n by blast
    have n5: "g (Inr (isuffix (cpoint f g (Suc (Suc i)) \<sigma>) \<sigma>)) "
    using n3 n4 someI_ex[of "\<lambda>x.  (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) 
                                                              (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>))"] by auto
    from n5 show ?thesis by auto
    qed 
   qed
  from 2 show ?thesis using assms cpoint_0 by blast
qed


lemma cpoint_2:
 assumes "g (Inr \<sigma>) \<and>
               (\<forall>k. g (Inr (isuffix k \<sigma>)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval \<sigma> k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) \<sigma>))))
         "
 shows   "f (Inl (subinterval \<sigma> (cpoint f g i \<sigma>) (cpoint f g (Suc i) \<sigma>)))
  "
proof 
 (induct i)
 case 0
 then show ?case 
 proof -
  have 1: "g (Inr(isuffix 0 \<sigma>)) "
       using assms cpoint_0 cpoint_expand_0 by (simp add: isuffix_0)
  have 2: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) (m+(cpoint f g 0 \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))) "
       using assms 1 by auto
  have 3: "(cpoint f g 1 \<sigma>) =
           (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) (m+(cpoint f g 0 \<sigma>)) ))
                          \<and> m>0 \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))
                          \<and> x=m+(cpoint f g 0 \<sigma>))
                 )
          "
       by simp
  have 4: "f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) ((cpoint f g 1 \<sigma>)) )) "
       using 2 3 someI_ex[of "\<lambda>x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) 
                                                              (m+(cpoint f g 0 \<sigma>)) ))
                          \<and> m>0 \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))
                          \<and> x=m+(cpoint f g 0 \<sigma>))"] by auto
  from 4 show ?thesis by auto
  qed
 next
 case (Suc i)
 then show ?case 
 proof -
  have n1: "g (Inr (isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>)) " 
       using assms cpoint_0 by blast
  have n2: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))) "
       using assms n1 by auto
  have n3: "(cpoint f g (Suc (Suc i)) \<sigma>) = 
           (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>)) 
              ) "
       using cpoint_expand_n by blast
  have n4: "f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) ((cpoint f g (Suc (Suc i)) \<sigma>)))) " 
       using n2 n3 someI_ex[of "\<lambda>x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) 
                                                                (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>)) "] by auto 
  from n4 show ?thesis by auto
  qed
qed
   
lemma cpoint_3a:
 "m>0 \<and> x=m+(cpoint f g (Suc i) \<sigma>) \<Longrightarrow> (cpoint f g (Suc i) \<sigma>) < x"
by auto


lemma cpoint_3:
 assumes "g (Inr \<sigma>) \<and>
               (\<forall>k. g (Inr (isuffix k \<sigma>)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval \<sigma> k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) \<sigma>))))
         "
 shows   "(cpoint f g i \<sigma>) < (cpoint f g (Suc i) \<sigma>) 
  "
proof 
 (induct i)
 case 0
 then show ?case 
 proof -
  have 1: "g (Inr(isuffix 0 \<sigma>)) "
       using assms cpoint_0 cpoint_expand_0 by (simp add: isuffix_0)
  have 2: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) (m+(cpoint f g 0 \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))) "
       using assms 1 by auto
  have 3: "(cpoint f g 1 \<sigma>) =
           (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) (m+(cpoint f g 0 \<sigma>)) ))
                          \<and> m>0 \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))
                          \<and> x=m+(cpoint f g 0 \<sigma>))
                 )
          "
       by simp
  have 4: "(cpoint f g 0 \<sigma>) < (cpoint f g 1 \<sigma>) "
       using 2 3 someI_ex[of "\<lambda>x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g 0 \<sigma>) (m+(cpoint f g 0 \<sigma>)) ))
                          \<and> m>0 \<and> g (Inr (isuffix (m+(cpoint f g 0 \<sigma>)) \<sigma>))
                          \<and> x=m+(cpoint f g 0 \<sigma>))"] by auto 
  from 4 show ?thesis by auto
  qed
 next
 case (Suc i)
 then show ?case 
 proof -
  have n1: "g (Inr (isuffix (cpoint f g (Suc i) \<sigma>) \<sigma>)) " 
       using assms cpoint_0 by blast
  have n2: "(\<exists>m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))) "
       using assms n1 by auto
  have n3: "(cpoint f g (Suc (Suc i)) \<sigma>) = 
           (SOME x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>)) 
              ) "
       using cpoint_expand_n by blast
  have n4: "(\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> (cpoint f g (Suc (Suc i)) \<sigma>)=m+(cpoint f g (Suc i) \<sigma>))  " 
       using n2 n3 someI_ex[of "\<lambda>x. (\<exists> m. f (Inl (subinterval \<sigma> (cpoint f g (Suc i) \<sigma>) (m+(cpoint f g (Suc i) \<sigma>)))) 
                 \<and> m>0 \<and> g (Inr(isuffix (m+(cpoint f g (Suc i) \<sigma>)) \<sigma>))
                 \<and> x=m+(cpoint f g (Suc i) \<sigma>)) "] by auto
  have n5: "(cpoint f g (Suc i) \<sigma>) < (cpoint f g (Suc (Suc i)) \<sigma>)  "
  using n4 using cpoint_3a by blast
  from n5 show ?thesis by auto
  qed
qed 


lemma OmegaInductSem:
 "( w \<Turnstile> (inf \<and> g \<and> \<box>(g \<longrightarrow> (f \<and> fmore);g)) \<longrightarrow> omega f)"
proof  (cases w)
  case (Inl a)
  then show ?thesis 
  by (metis (no_types, lifting) finite_d_def finite_defs_1 int_simps(21) inteq_reflection unl_con unl_lift2)
  next
  case (Inr b)
  then show ?thesis 
  proof -
   have 1: "((Inr b) \<Turnstile> (inf \<and> g \<and> \<box>(g \<longrightarrow> (f \<and> fmore);g))) = 
               (g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) "
    by ( simp add: infinite_defs always_defs chop_defs fmore_defs isuffix_isuffix sum.case_eq_if)
       (metis iprefix_isuffix iprefix_length add.commute) 
   have 2: "(g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) \<longrightarrow>
                     infinite_index_sequence 0 (\<lambda>i.  (cpoint f g  i b  )) "
   using 1 infinite_index_sequence_def cpoint_3 by (metis cpoint_expand_0)
   have 3: "(g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) \<longrightarrow>
                              (\<forall>i. f (Inl (subinterval b ((\<lambda>i.  (cpoint f g  i b  )) i) 
                                                         ((\<lambda>i.  (cpoint f g  i b  )) (Suc i))))) "
   using 1 cpoint_2 by (metis )
   have 4: "(g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) \<longrightarrow>
                             infinite_index_sequence 0 (\<lambda>i.  (cpoint f g  i b  )) \<and>
                             (\<forall>i. f (Inl (subinterval b ((\<lambda>i.  (cpoint f g  i b  )) i) 
                                                        ((\<lambda>i.  (cpoint f g  i b  )) (Suc i)))))"
    using 2 3 by auto
   have 5: "(g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) \<longrightarrow>
                             (\<exists>(ls::infiniteindex). infinite_index_sequence 0 ls \<and> 
                                                    ls = (\<lambda>i.  (cpoint f g  i b  )))"
    using "4" by auto
   have 6: "(g (Inr b) \<and>
               (\<forall>k. g (Inr (isuffix k b)) \<longrightarrow>
                       (\<exists>m. f (Inl (subinterval b k (m+k))) \<and> 
                            0 < m \<and> g (Inr (isuffix (m+k) b))))) \<longrightarrow> ((Inr b) \<Turnstile> omega f) "
    using "3" "5" unfolding omega_d_def by auto
   have 7: "((Inr b) \<Turnstile> (inf \<and> g \<and> \<box>(g \<longrightarrow> (f \<and> fmore);g)) \<longrightarrow> (omega f)  ) "
   using "6" "1" by auto
   from 7 show ?thesis using Inr by blast
  qed
qed
 
  


subsection \<open>Quantification over State (Flexible) Variables\<close>

text \<open>
  Quantification in Infinite ITL is done similarly as in Finite ITL.
\<close>

typedecl state

instance state :: world ..

type_synonym 'a statefun = "(state,'a) stfun"
type_synonym statepred   = "bool statefun"
type_synonym 'a tempfun  = "(state,'a) formfun"
type_synonym temporal    = "state formula"


subsection \<open>Temporal Quantifiers\<close>

definition exist_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Eex " 10)
where "exist_state_d  F  \<equiv> (\<lambda>s.  (\<exists> x.  s \<Turnstile> F x ))"

syntax
 "_Eex" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<exists>\<exists> _./ _)" [0,10] 10)

translations
 "_Eex v A"  == "Eex v. A" 

definition forall_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Aall " 10)
where "forall_state_d F \<equiv> LIFT(\<not>(\<exists>\<exists> x. \<not>(F x)))"

syntax 
 "_Aall" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<forall>\<forall> _./ _)" [0,10] 10)

translations
 "_Aall v A" == "Aall v. A"



end
