(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Finite Intervals\<close>

theory Interval
 imports
   Main 
begin

(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]

*)
(*
declare [[smt_timeout=120]]
*)


text \<open>
   An interval is a finite sequence of elements of a particular type. Intervals are similar to list 
   in Isabelle/HOL, the difference is that intervals have instead of nil a single
   element at the end. So an interval of length zero is a single element whereas for Isabelle's list 
   we have that an empty list is nil (no element present). 

   The usual operations on intervals are defined: length (\<open>intlen\<close>), \<open>prefix\<close>, 
   \<open>suffix\<close>, \<open>sub\<close>,  \<open>nth\<close>, \<open>intfirst\<close>, \<open>intlast\<close>, \<open>intapp\<close> 
   and \<open>intrev\<close>. 

   We also introduce \<open>index_sequence\<close> which is a sequence of chop (fuse) points. This sequence
   is used in the old definition of chopstar which is an existential quantification over this 
   sequence. The type \<open>index_sequence\<close> is again of type interval but the elements are 
   natural numbers. Two functions \<open>shift\<close> and \<open>shiftm\<close> are introduced that are used to add (shift) 
   and subtract a natural number of each element in the sequence of chop (fuse) points. 
   The operation \<open>upt\<close> to produce a sequence of consecutive chop points between two natural numbers.
    
\<close>

subsection \<open>Definitions\<close>

datatype (set: 'a) interval =
    St  'a ("\<lceil>_\<rceil>")
  | Cons 'a "'a interval"  (infixr "\<odot>" 65)
for
  map: map
  rel: interval_all2
  pred: interval_all

(*print_theorems*)

type_synonym  index = "nat interval"  

syntax
  \<comment> \<open>interval Enumeration\<close>
  "_interval" :: "args => 'a interval"    ("\<langle>(_)\<rangle>")

translations
  "\<langle>x, xs\<rangle>" == "x\<odot>\<langle>xs\<rangle>"
  "\<langle>x\<rangle>" == "\<lceil>x\<rceil>"
  
primrec (nonexhaustive) intlen :: "'a interval \<Rightarrow> nat" where
    " intlen \<langle>x\<rangle> = 0"
  | " intlen (x\<odot>xs) = 1+ (intlen xs)"
 
primrec (nonexhaustive) nth :: "'a interval => nat => 'a" where
    "nth \<langle>x\<rangle> n      = x"
  | "nth (x\<odot>xs) n = (case n of 0 \<Rightarrow> x | Suc k \<Rightarrow> nth xs k)"
   
primrec prefix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
    "prefix n \<langle>x\<rangle> = \<langle>x\<rangle>"   
  | "prefix n (x \<odot>xs) = (case n of 0 \<Rightarrow> \<langle>x\<rangle> | Suc m \<Rightarrow> x \<odot> (prefix m xs))"

primrec suffix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
   "suffix n \<langle>x\<rangle> = \<langle>x\<rangle>" 
 | "suffix n (x\<odot>xs) = (case n of 0 \<Rightarrow> (x\<odot>xs)  | Suc m \<Rightarrow> suffix m xs)"

(*
definition sub:: "nat \<Rightarrow> nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
where
 "sub n k xs = (if k<n then prefix 0 (suffix n xs)
                       else prefix (k-n) (suffix n xs)
               )" 
*)

definition sub:: "nat \<Rightarrow> nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
where
 "sub n k xs = prefix (k-n) (suffix n xs)"



definition intfirst ::  "'a interval \<Rightarrow> 'a" where
 "intfirst xs = (nth xs 0)" 

(*
primrec intfirst :: "'a interval \<Rightarrow> 'a" where
    "intfirst \<langle>x\<rangle>     = x"
  | "intfirst (Cons x _) = x"
*)

definition intlast ::  "'a interval \<Rightarrow> 'a" where
 "intlast xs = (nth xs (intlen xs))" 

(*
primrec intlast :: "'a interval \<Rightarrow> 'a" where
    "intlast \<langle>x\<rangle>      = x"
  | "intlast (Cons _ xs) = intlast xs"  
*)

primrec intapp :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval" (infixr "\<ominus>" 65) where
intapp_St:   "\<langle>x\<rangle> \<ominus> ys = x \<odot> ys" |
intapp_Cons: "(x\<odot>xs) \<ominus> ys = x \<odot> (xs \<ominus> ys)"

primrec intrev :: "'a interval \<Rightarrow> 'a interval" where
  " intrev \<langle>x\<rangle> = \<langle>x\<rangle>"
| " intrev (x \<odot>xs) = (intrev xs) \<ominus> \<langle>x\<rangle>"

definition index_sequence :: " nat \<Rightarrow> index \<Rightarrow> bool" where
  "index_sequence x idx \<equiv> (nth idx 0 = x) \<and> (\<forall> n. n<intlen idx \<longrightarrow> nth idx n < nth idx (Suc n))"  

definition shift :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shift k = (\<lambda> x. x+k)"

definition shiftm ::  "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shiftm k = (\<lambda> x. x-k)" 

primrec upt :: "nat \<Rightarrow> nat \<Rightarrow> nat interval" ("(1[_..\<le>/_'])")
 where
    upt_0  : "[i..\<le>0]       = \<langle>0\<rangle>"
  | upt_Suc: "[i..\<le>(Suc j)] = (if i \<le> j then [i..\<le>j] \<ominus> \<langle>(Suc j)\<rangle> else \<langle>(Suc j)\<rangle> )"


subsection \<open>Lemmas\<close>

text \<open>
 Basic lemmas are introduced for each of the above operations on intervals.  
\<close>

subsubsection \<open>Shifting indexes to zero\<close>

lemma interval_shift_index_to_zero_a:
shows "(\<forall> (i::nat). a \<le> i \<and> i < a+b \<longrightarrow> f ( g1(i) ) ( g2(Suc i))) =
       (\<forall> i. 0 \<le> i \<and> i < b \<longrightarrow> f (g1(i+a)) (g2((Suc i)+a)))" 
by (metis add_diff_cancel_left' le0 le_add2 le_add_diff_inverse2 less_diff_conv plus_nat.simps(2))


lemma interval_shift_index_to_zero_b:
shows "(\<forall> (i::nat). a \<le> i \<and> i < a+b \<longrightarrow> f ( g(i-a) ) ( g2((Suc i)-a))) =
       (\<forall> i. 0 \<le> i \<and> i < b \<longrightarrow> f (g(i)) (g2(Suc i)))" (is "?L=?R")
proof -
 have 1: "?L \<Longrightarrow> ?R" 
   by (metis  add_Suc add_diff_cancel_left' add_diff_cancel_right' le_add2 less_diff_conv)
 have 2: "?R \<Longrightarrow> ?L"
   by (metis Nat.add_diff_assoc add_diff_cancel_left' le0 le_add_diff_inverse2 less_diff_conv 
       plus_1_eq_Suc)
 show ?thesis using "1" "2" by blast
qed


subsubsection \<open>Interval Length\<close>

lemma interval_intlen_gr_zero [simp]:
    "intlen xs \<ge> 0"
by auto

lemma interval_intlen_cons [simp]: 
    " (intlen (x\<odot>xs)) = (intlen xs) +1"
by simp
  
lemma interval_intlen_cons_1 :
    " intlen l > 0 = (\<exists> x ls. l = x\<odot>ls) "
by (induct l) simp_all

lemma interval_intlen_map [simp]:
    " intlen (map f xs) = intlen xs"
by (induct xs) simp_all

lemma interval_intlen_intapp [simp]:
 " intlen (xs \<ominus> ys) = (intlen xs) + (intlen ys) +1"
by (induct xs arbitrary: ys) simp_all

lemma interval_intrev_intlen [simp]:
 " intlen (intrev xs) = intlen xs"
by (induct xs) simp_all



subsubsection \<open>nth\<close>

lemma interval_nth_zero [simp]:
    " nth (x\<odot>xs) 0  = x"
by simp

lemma interval_nth_Suc [simp]:
    "nth (x\<odot>xs) (Suc n) = nth xs n"
by auto

lemma interval_nth_last:
    " nth (x\<odot>xs) (intlen (x\<odot>xs)) = nth xs (intlen xs) "
by simp


lemma interval_nth_last_stutter:
 " nth xs (intlen xs + i) = nth xs (intlen xs)"
proof (induction xs arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by simp
   next
   case (Suc nat)
   then show ?thesis 
   by (metis Cons.IH ab_semigroup_add_class.add_ac(1) add_Suc_right interval_intlen_cons 
       interval_nth_Suc interval_nth_last)
   qed
qed

lemma interval_nth_cons_a:
 assumes "0<i"
 shows   "nth(x\<odot>xs) i     = nth xs (i-1)"
using assms by (metis Suc_diff_1 interval_nth_Suc)

lemma interval_nth_cons_b:
 shows   "nth(x\<odot>xs) (i+1) = nth xs i"
by simp

lemma interval_nth_cons:
 assumes "0<i"
 shows   "nth(x\<odot>xs) i     = nth xs (i-1) \<and>
          nth(x\<odot>xs) (i+1) = nth xs i"
by (meson assms interval_nth_cons_a interval_nth_cons_b)

lemma interval_nth_zero_intfirst [simp]:
 " intfirst xs = nth xs 0   "
by (simp add: intfirst_def)

lemma interval_nth_intlen_intlast [simp]:
 " intlast xs = nth xs (intlen xs)  "
by (simp add: intlast_def)

lemma interval_st_intlen :
    " (xs = \<langle>x\<rangle>) \<longleftrightarrow> intlen xs = 0 \<and> nth xs 0 = x" 
by (cases xs) auto

lemma interval_eq_nth_eq :
    " (xs =ys) = (intlen xs =intlen ys \<and> (\<forall> i\<le> intlen xs. nth xs i = nth ys i))"
proof 
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case by (metis interval_st_intlen le_numeral_extra(3))
 next
 case (Cons x1a xs)
 then show ?case 
  proof (cases ys)
  case (St x1)
  then show ?thesis by simp
  next
  case (Cons x21 x22)
  then show ?thesis 
  using Cons.hyps by fastforce
  qed
 qed

lemma interval_nth_map :
    "nth (map f xs) i = f (nth xs i)"
proof 
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by simp
   next
   case (Suc nat)
   then show ?thesis using Cons.hyps by simp
   qed
 qed




subsubsection \<open>prefix, suffix and sub\<close>



lemma interval_prefix_state [simp]: 
    " prefix m \<langle>x\<rangle> = \<langle>x\<rangle> " 
by simp

lemma interval_prefix_suc [simp]:
    " prefix (Suc m) (x\<odot>xs) = x \<odot> (prefix m xs)" 
by auto

lemma interval_prefix_zero [simp]:
    " prefix 0 (x\<odot>xs) = \<langle>x\<rangle>"
by auto

lemma interval_prefix_zero_intfirst [simp]:
  " prefix 0 xs = \<langle>nth xs 0\<rangle>"
by (induct xs) simp_all

lemma interval_intfirst_prefix [simp]:
 shows " intfirst (prefix i xs) = intfirst xs"
proof 
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using Cons.hyps by auto
   qed
 qed

  
lemma interval_intlast_suffix [simp]:
 shows   "intlast (suffix i xs) = intlast xs" 
proof 
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using Cons.hyps by auto
   qed
 qed
 

lemma interval_prefix_intlen [simp]:
    " (prefix (intlen xs) xs) = xs"
by (induct xs) simp_all

lemma interval_prefix_intlen_gr_1 [simp]: 
    "(prefix ((intlen xs)+i) xs) = xs"
by (induct xs) simp_all

lemma interval_intlen_prefix_cons [simp]:
    " intlen( prefix (Suc i) (x\<odot>xs)) = 1 + intlen(prefix i xs)"
using interval_intlen_cons by auto

lemma interval_prefix_length_code [code]: 
    " intlen (prefix i xs) = (if i\<le> intlen xs then i else intlen xs)"
proof 
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using Cons.hyps by auto
   qed
 qed

lemma interval_prefix_length [simp]: 
    " intlen (prefix i xs) = min i (intlen xs)"
by (simp add: interval_prefix_length_code min_def)

lemma interval_prefix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " (intlen ( prefix i xs)) = i" 
using assms by simp 

lemma interval_prefix_length_bad : 
  assumes   " i > intlen xs"
  shows     "  intlen (prefix i xs) = intlen xs"
using assms by simp 
   
lemma interval_pref_intlen_bound :
 shows     " intlen (prefix i xs) \<le> intlen xs " 
by simp

lemma interval_suffix_length_code [code]:
    " intlen (suffix i xs) = (if i\<le> intlen xs then (intlen xs)-i else 0) "
proof
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis using Cons.hyps by auto
  qed
qed
  
lemma interval_suffix_length [simp]:
 " intlen (suffix i xs ) = (intlen xs)-i"
by (simp add: interval_suffix_length_code)
 
lemma interval_suffix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " intlen (suffix i xs ) = (intlen xs)-i"
using assms by simp
  
lemma interval_suffix_length_bad:
  assumes  " i> intlen xs"
  shows    "  intlen (suffix i xs ) = 0"
using assms by simp
 
lemma interval_suffix_intlen_bound:
 "intlen(suffix i xs) \<le> intlen xs"
by simp

lemma interval_nth_prefix [simp]:
 assumes "  k \<le> i"
 shows   "  nth (prefix i xs) k = nth xs k"
using assms
 proof 
  (induct i arbitrary: xs k)
  case 0
  then show ?case 
    proof (cases xs)
    case (St x1)
    then show ?thesis by auto
    next
    case (Cons x21 x22)
    then show ?thesis using "0.prems" by auto
    qed
  next
  case (Suc i)
  then show ?case 
    proof (cases xs)
    case (St x1)
    then show ?thesis by auto
    next
    case (Cons x21 x22)
    then show ?thesis 
       proof (cases k)
       case 0
       then show ?thesis by (simp add: local.Cons)
       next
       case (Suc nat)
       then show ?thesis 
       using Suc.hyps Suc.prems local.Cons by auto
       qed
   qed
qed
 
lemma interval_nth_suffix [simp]:
 assumes " k \<le> intlen xs -i"
 shows   " nth (suffix i xs) k = nth xs (i+k)"
using assms
proof (induct xs arbitrary: i k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis 
     proof auto
       assume a0: " i = Suc nat" 
       show "Interval.nth (suffix nat xs) k = Interval.nth xs (nat + k) "
       using a0 Cons.hyps Cons.prems by auto
     qed
   qed
qed

lemma interval_suffix_prefix_help_1:
 assumes " ia+i \<le> intlen xs"
         "  k \<le> ia"
 shows " nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k"
proof -
 have 1: " nth (prefix ia (suffix i xs)) k = nth (suffix i xs) k" 
 using interval_nth_prefix assms by metis 
 have 2: " nth (suffix i xs) k = nth xs (i+k) " 
 using interval_nth_suffix assms by (simp add: add_le_imp_le_diff)
 have 3: " nth xs (i+k) = nth (prefix (ia+i) xs) (i+k)" 
 using interval_nth_prefix assms by simp
 have 4: " nth (prefix (ia+i) xs) (i+k) = nth (suffix i (prefix (ia +i) xs)) k" 
 using interval_nth_suffix assms by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma interval_suffix_prefix_help_2:
 assumes "ia+i \<le> intlen xs"
 shows   " (\<forall> k \<le> ia . nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k)"
using interval_suffix_prefix_help_1 using assms by fastforce

lemma interval_suffix_prefix_help_3:
 assumes "ia+i \<le> intlen xs "
 shows   "intlen (prefix ia (suffix i xs)) = intlen (suffix i (prefix (ia +i) xs)) "
using assms interval_prefix_length_good interval_suffix_length_good by auto

lemma interval_suffix_prefix_swap:
 assumes "ia+i \<le> intlen xs"
 shows   " prefix ia (suffix i xs) = suffix i (prefix (ia +i) xs)"
using assms using interval_eq_nth_eq by fastforce


lemma interval_prefix_prefix_zero [simp]:
    " prefix 0 ( prefix 0 xs ) = prefix 0 xs" 
by (induct xs) simp_all
  
lemma interval_pref_pref [simp]:
    " (prefix i (prefix i xs)) = prefix i xs"
by (metis interval_prefix_intlen interval_prefix_intlen_gr_1 interval_prefix_length_good 
    less_imp_add_positive not_less)

lemma interval_pref_pref_3 [simp]:
    " (prefix i (prefix (i+k) xs)) = prefix i xs"
proof 
  (induct xs arbitrary: i k)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
    by (auto simp add: Nitpick.case_nat_unfold)
   next
   case (Suc nat)
   then show ?thesis 
    by (simp add: Cons.hyps)
   qed
qed
 
lemma interval_pref_help:
 assumes "i\<le>intlen (prefix (intlen xs - Suc 0) xs)"
 shows   " (prefix i (prefix (intlen xs - Suc 0) xs)) = (prefix i xs)  "
using assms 
by (metis diff_le_self interval_pref_pref_3 interval_prefix_length_good 
     ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma interval_pref_pref_help:
 assumes "intlen xs >0"
         "  ia<intlen (xs)"
 shows   "(prefix ia (prefix (intlen xs - Suc 0) xs)) = (prefix ia  xs) "
using assms 
by (metis Suc_leI Suc_le_mono Suc_pred diff_le_self interval_pref_help interval_prefix_length_good)

lemma interval_pref_pref_help_1:
 assumes "i>0"
         "i\<le> intlen xs"
 shows   "(prefix (intlen (prefix i xs) - Suc 0) (prefix i xs)) = 
          (prefix (intlen (prefix i xs) - Suc 0) xs) "
using assms interval_pref_pref_3 by (metis diff_le_self interval_prefix_length_good le_iff_add)

lemma interval_suffix_suc [simp]:
    " suffix (Suc m) (x \<odot> xs) = suffix m xs"
by auto
  
lemma interval_suffix_zero [simp]:
    " suffix 0 xs = xs "
by (induct xs) simp_all
    
lemma interval_hd_tail:
 assumes "intlen xs >0"
 shows   "xs = (intfirst xs) \<odot> (suffix 1 xs)"
by (metis One_nat_def assms interval_intlen_cons_1 interval_nth_zero interval_nth_zero_intfirst 
     interval_suffix_suc interval_suffix_zero)

lemma interval_suffix_intlen [simp]:
    " suffix (intlen xs) xs = \<langle>(nth xs (intlen xs))\<rangle> "
by (induct xs) simp_all

lemma interval_suffix_intlast [simp]:
   " suffix (intlen xs) xs = \<langle>intlast xs\<rangle>"
by (induct xs) simp_all

lemma interval_suffix_suffix [simp]:
    " suffix i (suffix j xs) = suffix (i+j) xs "
proof 
  (induct xs arbitrary: i j)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
     by auto
   next
   case (Suc nat)
   then show ?thesis 
     by (simp add: Nitpick.case_nat_unfold add.commute local.Cons)
   qed
qed

lemma interval_prefix_suffix_intlen_code [code]:
    "intlen (prefix ia (suffix i xs)) =
     (if i \<le> intlen xs then
      (if ia\<le> intlen xs -i  then ia else (intlen xs) -i )
      else 0)"
using interval_suffix_length_code by auto

lemma interval_prefix_suffix_intlen [simp]:
    "intlen (prefix ia (suffix i xs)) =
     min ia (intlen xs -i)
     "
by auto

lemma interval_prefix_suffix_intlen_good [simp]:
  assumes  "ia\<le> intlen xs -i"
           "i \<le> intlen xs"
  shows    " intlen (prefix ia (suffix i xs)) = ia "
using assms by auto

lemma interval_prefix_suffix_intlen_bad_0:
  assumes  "i> intlen xs"
  shows    "intlen (prefix ia (suffix i xs)) = 0"
using assms by simp 

lemma interval_prefix_suffix_intlen_bad_1 :
  assumes  "i \<le> intlen xs"
           " ia > intlen xs -i"
  shows    "intlen (prefix ia (suffix i xs)) = (intlen xs) -i  "
using assms by simp

lemma interval_suffix_suffix_3:
 assumes "i>0"
         "ia<i"
         "i\<le> intlen xs"
 shows   "(suffix (i-ia) (suffix ((intlen xs)-i) xs)) = (suffix (((intlen xs)-ia)) xs)"
using assms by simp

lemma interval_sub_zero_prefix :
    " sub 0 k xs = prefix k xs"
by (simp add: Interval.sub_def) 

lemma interval_sub_suffix : 
  assumes  "i < j"
           "j\<le> (intlen xs)-k"
  shows    "(sub (i+k) (j+k) xs) = (sub i j (suffix k xs))"
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix_0:
  assumes "0 \<le> i"
          "ia+i \<le> intlen xs"
  shows   "(sub i (i+ia) xs) = (prefix (ia) (suffix i xs))" 
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix:
 assumes "0 \<le> i"
         "i\<le>j"
         "j \<le> intlen xs"
 shows   "(sub i j xs) = (prefix (j-i) (suffix i xs))"
using assms by (simp add: Interval.sub_def) 

lemma interval_intlast_prefix:
 assumes "k \<le> intlen xs"
 shows   "intlast(prefix k xs) = (nth xs k)"
using assms interval_prefix_length_good by fastforce

lemma interval_intfirst_suffix:
 assumes "k \<le> intlen xs"
 shows   "intfirst(suffix k xs) = (nth xs k)"
by (simp add: assms)

lemma interval_suffix_gr:
 assumes "i>intlen xs"
 shows   "suffix i xs = \<langle>intlast(xs)\<rangle>"
 by (metis add.commute assms interval_suffix_intlast interval_suffix_suffix 
     less_imp_add_positive suffix.simps(1))

lemma interval_intlast_intfirst:
    " (intlast (prefix i xs)) = (intfirst (suffix i xs))"
proof -
  have 1: "(intlast (prefix i xs)) = (nth (prefix i xs) (intlen (prefix i xs)))" 
   by simp  
  have 2: "i \<le> intlen xs \<longrightarrow> intlen (prefix i xs) = i  "
   using interval_prefix_length_good by blast
  have 3: "i>intlen xs \<longrightarrow> intlen (prefix i xs) = intlen xs"
    using interval_prefix_length_bad by blast
  have 4: "i \<le> intlen xs \<longrightarrow> 
           (nth (prefix i xs) (intlen (prefix i xs))) = (nth xs i)"
    using interval_intlast_prefix by auto
  have 5: "i>intlen xs \<longrightarrow> 
           (nth (prefix i xs) (intlen (prefix i xs))) = (nth xs (intlen xs))"
    using "3" by auto
  have 6: "(intfirst (suffix i xs)) = (nth (suffix i xs) 0)"
      by simp 
  have 7: "i \<le> intlen xs \<longrightarrow> 
           (nth (suffix i xs) 0) = (nth xs i)"
    by simp 
  have 8: "i>intlen xs \<longrightarrow> 
           (nth (suffix i xs) 0) = (nth xs (intlen xs)) "
     by (simp add: interval_suffix_gr)
  show ?thesis using "4" "5" "8" by auto 
qed


lemma interval_intlen_sub [simp]:
 assumes "k\<le> n"
         "n \<le> intlen xs"
 shows   "intlen(sub k n xs ) = (n-k)"
using assms 
by (metis Interval.sub_def interval_prefix_length_good interval_suffix_length 
    interval_suffix_prefix_swap le_add_diff_inverse2)

lemma interval_nth_sub [simp]:
 assumes "k\<le> n"
         "n \<le> intlen xs"
         "j\<le>n-k"
 shows    "nth(sub k n xs) j = (nth xs (k+j))"
proof -
 have 1: "nth(sub k n xs) j = nth (prefix (n-k) (suffix k xs)) j "
   by (simp add: Interval.sub_def) 
 have 2: "n-k \<le> intlen (suffix k xs) "
    using Interval.sub_def assms by auto
 have 3: "j\<le> (n-k)"
    using assms by auto
 have 4: "nth (prefix (n-k) (suffix k xs)) j =
          nth (suffix k xs) j"
    using "2" assms interval_nth_prefix by blast  
 have 5: "nth (suffix k xs) j = nth xs (k+j)"
       using assms by auto 
 show ?thesis 
 by (simp add: "1" "4" "5")
qed       

lemma interval_intlast_sub:
 assumes "k \<le> n"
         "n \<le> intlen xs"
 shows   "intlast (sub k n xs) = (nth xs n)"
by (simp add: assms)

lemma interval_intfirst_sub:
 assumes "k \<le> n"
         "n \<le> intlen xs"
 shows   "intfirst (sub k n xs) = (nth xs k)"
by (simp add: assms)


lemma interval_sub_sub:
 assumes "n1 \<le> n2"
         "n0 \<le> n4"
         "n2 \<le> n4-n0"
         "n4 \<le> n3"
         "n3 \<le> intlen xs"
 shows   "  (sub n1 n2 (sub n0 n3 xs)) = (sub n1 n2 (sub n0 n4 xs))"
proof -
 have 1: " intlen(sub n0 n3 xs) = n3 -n0"
     by (meson assms interval_intlen_sub le_trans)
 have 2: " intlen (sub n1 n2 (sub n0 n3 xs)) = n2 -n1 "
    using interval_intlen_sub assms by auto
 have 3: " intlen(sub n0 n4 xs) = n4 -n0"
     using assms interval_intlen_sub le_trans by blast
 have 4: " intlen (sub n1 n2 (sub n0 n4 xs)) = n2 -n1"  
    by (simp add: "3" assms)
 have 5: " \<And>i. i\<le> (n3 -n0) \<longrightarrow> (nth (sub n0 n3 xs) i ) = (nth xs (n0+i))"
   using assms interval_nth_sub le_trans by blast
 have 6: " \<And>i . i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n3 xs)) i) = (nth (sub n0 n3 xs) (n1+i))"
    using interval_nth_sub assms
    by (metis "1" Nat.le_diff_conv2 le_trans)
 have 7: "\<And>i . i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n3 xs)) i) = (nth xs (n0+(n1+i))) "
    using "5" "6" assms by auto
 have 8: " n0\<le> n4 \<and> n4 \<le> intlen xs"
   using assms le_trans by blast
 have 9: " \<And>i. i\<le> (n4 -n0) \<longrightarrow> (nth (sub n0 n4 xs) i ) = (nth xs (n0+i))"
   using "8" interval_nth_sub by blast
 have 10: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n4 xs)) i) = (nth (sub n0 n4 xs) (n1+i))"
   by (simp add: "3" assms)
 have 11: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n4 xs)) i) = (nth xs (n0 +(n1+i))) "
    by (metis "3" Nat.le_diff_conv2 add.commute assms interval_nth_sub le_trans)
 have 12: "\<And>i. i\<le> (n2-n1) \<longrightarrow> 
           (nth (sub n1 n2 (sub n0 n3 xs)) i) = (nth (sub n1 n2 (sub n0 n4 xs)) i) "
   by (simp add: "11" "7")
 from 12 2 4 show ?thesis by (simp add: interval_eq_nth_eq)
qed

lemma interval_sub_sub_1:
 assumes "n1 \<le> n2"
         "n0 \<le> n3"
         "n2 \<le> n3-n0"
         "n3 \<le> intlen xs"
 shows   "(sub n1 n2 (sub n0 n3 xs)) = (sub (n0+n1) (n0+n2) xs)"
proof -
 have 1: "intlen(sub n1 n2 (sub n0 n3 xs)) = intlen(sub (n1+n0) (n2+n0) xs)"
   using assms by auto
 have 2: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n3 xs)) i) = (nth xs (n0+(n1+i))) "
   by (simp add: add.commute assms le_trans ordered_cancel_comm_monoid_diff_class.le_diff_conv2)
 have 3: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (nth (sub (n0+n1) (n0+n2) xs) i) = (nth xs (n0+(n1+i))) "
     by (metis (no_types, hide_lams) add.commute add_diff_cancel_left add_mono assms(1) assms(2) 
         assms(3) assms(4) interval_nth_sub le_refl le_trans 
         ordered_cancel_comm_monoid_diff_class.le_diff_conv2 semiring_normalization_rules(25))
 have 4: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (nth (sub n1 n2 (sub n0 n3 xs)) i) = (nth (sub (n0+n1) (n0+n2) xs) i)"
   by (simp add: "2" "3")
 show ?thesis 
 by (metis "1" "4" add.commute assms interval_eq_nth_eq interval_intlen_sub)
qed

lemma interval_suf_first_upto:
assumes "(\<exists>i<k. f (suffix i xs))"
        "k \<le> intlen xs +1"
shows   "(\<exists>i<k. f (suffix i xs) \<and>
           (\<forall>j<i . \<not> (f (suffix j xs))))"
using assms
proof (induct xs arbitrary:k)
case (St x)
then show ?case by auto
next
case (Cons x1a xs)
then show ?case 
  proof -
   have 0: "k=0 \<longrightarrow> ( \<exists>i<k. f (suffix i (x1a \<odot> xs)) \<and> (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) "
     using Cons.prems(1) by blast
   have 1: "\<And>n. k = (Suc n) \<longrightarrow> (\<exists>i< (Suc n). f (suffix i (x1a \<odot> xs)) \<and> 
            (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) =
            ( f (x1a \<odot> xs) \<or>
            (\<exists>i. 1\<le> i \<and> i<(Suc n) \<and> f (suffix i (x1a \<odot> xs)) \<and> 
                 (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))))
           " 
          by auto
             (metis One_nat_def less_one nat.split_sels(1) not_le_imp_less)
  have 2: "\<And>n. k = (Suc n) \<longrightarrow> (\<exists>i. 1\<le> i \<and> i<(Suc n) \<and> f (suffix i (x1a \<odot> xs)) \<and> 
                 (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) =
            (\<exists>i.  i<n  \<and> f (suffix (Suc i) (x1a \<odot> xs)) \<and> 
                 (\<forall>j<(Suc i). \<not> f (suffix j (x1a \<odot> xs))))
           "         
        by auto
          (metis Nitpick.case_nat_unfold Suc_le_D diff_Suc_1 less_Suc_eq_0_disj)
  have 3: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i.  i<n  \<and> f (suffix (Suc i) (x1a \<odot> xs)) \<and> 
                 (\<forall>j<(Suc i). \<not> f (suffix j (x1a \<odot> xs)))) =
           (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 \<not>(f (suffix 0 (x1a\<odot>xs))) \<and>
                 (\<forall>j. 1 \<le> j \<and> j<(Suc i) \<longrightarrow> \<not> f (suffix j (x1a \<odot> xs))))
          " 
         by (metis interval_suffix_suc le_add1 less_Suc_eq_0_disj plus_1_eq_Suc)
  have 4: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 \<not>(f (suffix 0 (x1a\<odot>xs))) \<and>
                 (\<forall>j. 1 \<le> j \<and> j<(Suc i) \<longrightarrow> \<not> f (suffix j (x1a \<odot> xs)))) =
           ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 (\<forall>j.  j<i \<longrightarrow> \<not> f (suffix (Suc j) (x1a \<odot> xs)))) )
            " using Cons.hyps Cons.prems by (auto, auto simp add: less_Suc_eq_0_disj)
  have 5: "\<And>n. k =(Suc n) \<longrightarrow> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 (\<forall>j.  j<i \<longrightarrow> \<not> f (suffix (Suc j) (x1a \<odot> xs)))) ) =
             ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) ) "
      by auto  
  have 6: "\<And>n. k =(Suc n) \<longrightarrow>  (\<exists>i< (Suc n). f (suffix i (x1a \<odot> xs)) \<and> 
            (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) = 
            ( f (x1a \<odot> xs) \<or> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) )  )"  
       using "1" "2" "3" "4" by auto
  have 7: "\<And>n. k =(Suc n) \<longrightarrow> ( f (x1a \<odot> xs) \<or> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) )  ) =
               (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))))
              " 
         by auto
  have 8: "\<And>n. k= (Suc n) \<longrightarrow> n \<le> intlen xs +1"
      using Cons.prems(2) by auto
  have 9: "\<And>n. k =(Suc n) \<longrightarrow> (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs)))) =
             (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs))) "
       using 7 8 Cons.hyps by fastforce
  have 10: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i<k. f (suffix i (x1a \<odot> xs))) =
            (f (x1a \<odot> xs)  \<or> (\<exists>i<n. f (suffix i (xs))))
          " 
       using Nitpick.case_nat_unfold less_Suc_eq_0_disj by auto 
  have 11: "\<And>n. k =(Suc n) \<longrightarrow>  
             ( \<exists>i<k. f (suffix i (x1a \<odot> xs)) \<and> (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs))))  "  
          using 9 10 6 Cons.prems(1) by force 
   show ?thesis 
  using 0 11 less_imp_Suc_add by blast
 qed
qed

subsubsection \<open>intapp\<close>

lemma interval_intlen_snoc_1:
 "intlen l > 0 = (\<exists> x ls. l = ls \<ominus> \<langle>x\<rangle>)"
proof (induct l)
case (St x)
then show ?case by fastforce
next
case (Cons x1a l)
then show ?case 
   by (metis Suc_eq_plus1 intapp_Cons intapp_St interval.exhaust interval_intlen_intapp nat.simps(3) 
       neq0_conv)
qed 


lemma interval_prefix_intapp [simp]:
 "prefix (intlen xs -k) (xs \<ominus> ys) = prefix (intlen xs -k) xs"
proof 
  (induct xs arbitrary:k)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis 
       by (metis Cons.hyps diff_zero intapp_Cons interval_prefix_suc intlen.simps(2) plus_1_eq_Suc)
    next
    case (Suc nat)
    then show ?thesis 
       by (auto simp add: Cons.hyps Nitpick.case_nat_unfold)
    qed   
qed


lemma interval_prefix_intapp2 [simp]:
 "prefix (intlen xs + k +1) (xs \<ominus> ys) = xs \<ominus> prefix k ys"
proof 
 (induct xs arbitrary: k)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases k)
   case 0
   then show ?thesis 
      by auto
         (metis Cons.hyps Suc_eq_plus1 add.right_neutral interval_prefix_zero_intfirst)
   next
   case (Suc nat)
   then show ?thesis 
      by auto
         (metis Cons.hyps Suc_eq_plus1 add_Suc_right)
   qed
qed


lemma interval_suffix_intapp [simp]:
 "suffix (intlen xs +m +1) (xs \<ominus> ys) = suffix (m) ys"
proof 
  (induct xs arbitrary:m)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
   proof (cases m)
   case 0
   then show ?thesis 
    by  auto
        (metis Cons.hyps One_nat_def Suc_eq_plus1 interval_suffix_zero plus_1_eq_Suc 
         semiring_normalization_rules(23))
   next
   case (Suc nat)
   then show ?thesis 
    by auto
       (metis Cons.hyps Suc_eq_plus1 interval_suffix_suffix)
   qed
qed

lemma interval_suffix_intapp2 [simp]: 
 "(suffix (intlen xs - k)  xs) \<ominus> ys = suffix (intlen xs - k) (xs \<ominus> ys)"
proof 
 (induct xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
      by (auto simp add: Nitpick.case_nat_unfold)
qed

lemma interval_intapp_assoc [simp]:
 " (xs \<ominus> ys) \<ominus> zs = xs \<ominus> ( ys \<ominus> zs ) "
by (induct xs) simp_all

lemma interval_intapp_nth:
 " nth (xs \<ominus> ys) k = (if k \<le> intlen xs 
                      then (nth xs k) 
                      else (nth ys (k - (intlen xs) -1)) )"
proof 
  (induct xs arbitrary: k)
  case (St x)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis by simp
    next
    case (Suc nat)
    then show ?thesis by simp
    qed
  next
  case (Cons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis by simp
    next
    case (Suc nat)
    then show ?thesis by (simp add: Cons.hyps)
    qed
qed


lemma interval_rev_intapp [simp]:
 " intrev (xs \<ominus> ys) = (intrev ys) \<ominus> (intrev xs)"
by (induct xs) simp_all

lemma interval_intlast_intapp [simp]:
 "intlast(xs \<ominus> \<langle>x\<rangle>) = x"
by (induct xs) simp_all

lemma interval_intlast_intapp2 [simp]:
 "intlast (xs \<ominus> ys) = intlast ys"
by (induct xs arbitrary: ys) simp_all

lemma interval_intfirst_intapp [simp]:
 "intfirst (\<langle>x\<rangle> \<ominus> xs) = x"
by (induct xs) simp_all

lemma interval_intfirst_intapp2 [simp]:
 "intfirst(xs \<ominus> ys) = intfirst xs"
by (induct xs arbitrary: ys) simp_all

lemma interval_intapp_not_state [simp]:
 "xs \<ominus> ys \<noteq> \<langle>x\<rangle>"
by (induct xs arbitrary: ys) simp_all

lemma interval_intapp_eq_intapp_conv [simp]:
 assumes "intlen xs = intlen ys \<or> intlen us = intlen vs"
 shows   "(xs \<ominus> us = ys \<ominus> vs) = (xs =ys \<and> us=vs)"
using assms 
proof 
  (induct xs arbitrary: ys)
  case (St x)
  then show ?case 
    proof (cases ys)
    case (St x1)
    then show ?thesis by simp
    next
    case (Cons x21 x22)
    then show ?thesis using St.prems by auto
    qed
  next
  case (Cons x1a xs)
  then show ?case 
    proof (cases ys)
    case (St x1)
    then show ?thesis using Cons.prems by auto
    next
    case (Cons x21 x22)
    then show ?thesis using Cons.hyps Cons.prems by auto
    qed
qed

lemma interval_intapp_eq_intapp_conv2: 
 "(xs \<ominus> ys = zs \<ominus> ts) =
  (\<exists> us. xs = zs \<ominus> us \<and> us \<ominus> ys = ts \<or>
         xs =zs \<and> ys=ts \<or>
         xs \<ominus> us = zs \<and> ys = us \<ominus> ts)"
proof 
 (induct xs arbitrary: ys zs ts)
case (St x)
then show ?case 
 proof (cases zs)
 case (St x1)
 then show ?thesis by simp
 next
 case (Cons x21 x22)
 then show ?thesis by simp
 qed
next
case (Cons x1a xs)
then show ?case  
 proof (cases zs)
 case (St x1)
 then show ?thesis by simp
 next
 case (Cons x21 x22)
 then show ?thesis by (auto simp add: Cons.hyps)
 qed
qed

lemma interval_same_intapp_eq[iff, induct_simp]:
 "(xs \<ominus> ys = xs \<ominus> zs) = (ys = zs)"
using interval_suffix_intapp by (metis interval_suffix_zero)

lemma interval_intapp_eq_conv[iff]:
 "(xs \<ominus> \<langle>x\<rangle> = ys \<ominus> \<langle>y\<rangle>) = (xs = ys \<and> x = y)"
by auto

lemma interval_intapp_same_eq[iff, induct_simp]:
 "(ys \<ominus> xs = zs \<ominus> xs) = (ys = zs)"
by auto

lemma interval_suffix1_intapp:
 "suffix 1 (xs \<ominus> ys) = (case xs of \<langle>x\<rangle> \<Rightarrow> ys | x \<odot> zs \<Rightarrow> zs \<ominus> ys)"
by (cases xs) simp_all


lemma interval_cons_eq_intapp_conv:
 "(x\<odot> xs = ys \<ominus> zs) =
  ((\<langle>x\<rangle> = ys \<and> xs =zs) \<or> (\<exists> ys'. x\<odot>ys' =  ys \<and>  xs = ys' \<ominus> zs))"
by (cases ys) simp_all

lemma interval_intapp_eq_cons_conv:
 "(ys \<ominus> zs = x \<odot> xs) =
  ((\<langle>x\<rangle> = ys \<and> zs =xs) \<or> (\<exists> ys'. ys = x\<odot>ys'  \<and>  ys' \<ominus> zs = xs))"
by (cases ys) auto

lemma interval_cons_eq_intappI:
assumes "x \<odot> xs1 = ys"
        "xs = xs1 \<ominus> zs"
shows   "x \<odot> xs = ys \<ominus> zs"
using assms by auto

lemma interval_intapp_eq_intappI:
assumes "xs \<ominus> xs1 = zs"
        "ys = xs1 \<ominus> us"
shows   "xs \<ominus> ys = zs \<ominus> us"
using assms by auto


lemma intlen_intapp_gr_zero:
 " intlen (xs \<ominus> ys) >0"
by auto


lemma interval_intapp_prefix_suffix:
assumes "i+1 \<le> intlen xs"
        "intlen xs >0 "
shows "xs = (prefix i xs) \<ominus> (suffix (i+1) xs)"
using assms 
proof (induct xs arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using Cons.hyps Cons.prems(1) by auto
   qed
qed 


subsubsection \<open>Reverse\<close>


lemma interval_rev_rev_ident [simp]: 
 " intrev (intrev xs) = xs"
by (induct xs) auto

lemma interval_rev_swap :
 " ((intrev xs) = ys) = (xs = intrev ys)"
by auto

lemma interval_rev_singleton_conv [simp]:
 "( intrev xs = \<langle>x\<rangle>) = (xs = \<langle>x\<rangle>)"
by (metis interval_rev_rev_ident intrev.simps(1)) 

lemma interval_single_rev_conv [simp]:
 " (\<langle>x\<rangle> = intrev xs) = (\<langle>x\<rangle> = xs)"
by (metis interval_rev_rev_ident intrev.simps(1))

lemma interval_rev_is_rev_conv [iff]:
 "(intrev xs = intrev ys) = (xs = ys)"
proof
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
 using interval_rev_swap by force
qed

lemma interval_rev_induct [case_names St snoc]:
assumes "\<And> y. P \<langle>y\<rangle>"
        "\<And>x xs. P xs \<Longrightarrow> P(xs \<ominus> \<langle>x\<rangle>)"
shows   "P xs"
using assms
using interval.induct[of "\<lambda> xs. P (intrev xs)" "intrev xs"]
by simp
  

lemma interval_rev_exhaust [case_names St snoc]:
assumes "\<And>x. xs = \<langle>x\<rangle> \<Longrightarrow> P"
        "\<And>ys y. xs = ys \<ominus> \<langle>y\<rangle> \<Longrightarrow> P"
shows   "  P"
using assms 
by (induct xs rule:interval_rev_induct) auto

lemmas interval_rev_cases = interval_rev_exhaust

lemma interval_rev_eq_cons_iff [iff]:
 "(intrev xs = y\<odot>ys) = (xs = (intrev ys) \<ominus> \<langle>y\<rangle>)"
by (metis interval_rev_rev_ident intrev.simps(2)) 


lemma interval_intrev_intapp_cons:
 " intrev (xs \<ominus> \<langle>x\<rangle>) = x\<odot>intrev xs"
by (cases xs) simp_all

lemma interval_intlast_intrev:
 "intlast (intrev xs) = intfirst xs"
proof (cases xs)
case (St x1)
then show ?thesis 
 by simp
next
case (Cons x21 x22)
then show ?thesis 
 by  (metis interval_intlast_intapp interval_nth_zero interval_nth_zero_intfirst intrev.simps(2))
qed

lemma interval_intfirst_intrev:
 "intfirst (intrev xs) = intlast xs"
proof 
  (induct xs)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
    by (metis interval_intfirst_intapp2 interval_nth_intlen_intlast interval_nth_last 
        intrev.simps(2))
qed


lemma interval_intrev_nth:
assumes "k\<le> intlen (intrev xs)"
shows   "  (nth  (intrev xs) k) = (nth  xs ((intlen xs) -k))"
using assms 
proof 
  (induct xs)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis 
     by auto
        (metis Cons.hyps diff_zero interval_intapp_nth interval_intlen_gr_zero)
    next
    case (Suc nat)
    then show ?thesis 
     using Cons.hyps Suc_diff_le by (auto simp add: interval_intapp_nth) fastforce
    qed
qed


lemma interval_intrev_prefix:
assumes "k\<le> intlen xs"
shows   "  intrev( prefix k xs) = suffix ((intlen xs) - k) (intrev xs)  "
proof 
 (induct xs arbitrary: k)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases k)
   case 0
   then show ?thesis 
    by auto
       (metis Suc_eq_plus1 add.right_neutral interval_intlast_intapp interval_intlen_intapp 
        interval_intrev_intlen interval_suffix_intlast intlen.simps(1))
   next
   case (Suc nat)
   then show ?thesis 
    by auto
       (metis Cons.hyps interval_intrev_intlen interval_suffix_intapp2)
   qed
qed


lemma interval_intrev_suffix:
assumes "k\<le> intlen xs"
shows   " intrev( suffix k xs) = prefix ((intlen xs) - k) (intrev xs)  "
using assms
proof 
 (induct xs arbitrary: k)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case by (simp add: interval_intrev_prefix interval_rev_swap)
qed

lemma interval_intrev_sub:
 assumes " 0 \<le> i "
         " i\<le>j "
         " j \<le> intlen xs"
 shows   " intrev (sub i j xs) = sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs)"
using assms 
proof -
 have 1: "intrev (sub i j xs) = intrev (prefix (j-i) (suffix i xs))" 
   using assms interval_sub_prefix_suffix by (simp add: interval_sub_prefix_suffix)
 have 2: "intrev (prefix (j-i) (suffix i xs)) = suffix ((intlen xs) - j) (intrev (suffix i xs))"
   using assms interval_intrev_prefix[of "j-i" "suffix i xs"]  by auto
 have 3: "suffix ((intlen xs) - j) (intrev (suffix i xs)) = 
           suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs)) "
    using assms interval_intrev_suffix[of "i" "xs"] by auto
 have 4: "suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs)) =
          sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs) "
    using assms by (simp add: diff_le_mono2 interval_sub_prefix_suffix interval_suffix_prefix_swap)
 from 1 2 3 4 show ?thesis by auto
qed


subsubsection \<open>Induction rule\<close>

lemma interval_length_induct:
assumes "(\<And>xs. \<forall>ys. intlen ys < intlen xs \<longrightarrow> P ys \<Longrightarrow> P xs)"
shows   "  P xs"
using assms by (fact measure_induct)

lemma interval_induct_12:
 assumes "\<And>x. P \<langle>x\<rangle>"
         "\<And>x y. P \<langle>x,y\<rangle>"
         "\<And>x y zs. P (y\<odot>zs) \<Longrightarrow> P (x\<odot>y\<odot>zs)"
 shows   "P xs" 
using assms 
proof (induction xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
by auto
   (metis add_cancel_right_right interval_intlen_cons_1 interval_st_intlen le_add1 le_eq_less_or_eq 
    less_add_same_cancel1)
qed

lemma interval_induct2 [consumes 1, case_names St Cons]:
assumes "intlen xs = intlen ys"
        " (\<And> x y. P \<langle>x\<rangle> \<langle>y\<rangle>) "  
        "(\<And> x xs y ys. intlen xs = intlen ys \<Longrightarrow> P xs ys \<Longrightarrow> P (x\<odot>xs) (y\<odot>ys))"
 shows " P xs ys"
using assms 
proof (induction xs arbitrary: ys)
case (St x)
then show ?case 
by (metis interval_st_intlen)
next
case (Cons x1a xs)
then show ?case 
by (metis interval_intlen_cons_1 intlen.simps(2) nat.simps(1) plus_1_eq_Suc)
qed



subsubsection \<open>Map\<close>

lemma map_ext:
 assumes "(\<forall>x. x \<in> set xs \<longrightarrow> f x = g x)"
 shows   " map f xs = map g xs"
using assms
by (induct xs) simp_all


lemma map_ident [simp]:
 "map (\<lambda>x. x) = (\<lambda>xs .xs)"
 proof (rule ext) 
  show " \<And>xs. interval.map (\<lambda>x. x) xs = xs " 
   by (simp add: interval.map_ident)
qed


lemma map_intapp [simp]:
 "map f (xs \<ominus> ys) = map f xs \<ominus> map f ys"
by (induct xs) auto

lemma map_map [simp]:
 "map f (map g xs) = map (f \<circ> g) xs"
by (simp add: interval.map_comp)

lemma map_comp_map [simp]:
 "((map f) \<circ> (map g)) = map(f \<circ> g)"
by (rule ext) simp

lemma intrev_map:
 "intrev (map f xs) = map f(intrev xs)"
by (induct xs) auto

lemma map_eq_conv [simp]:
 "(map f xs = map g xs) = (\<forall> x\<in> set xs. (f x) = (g x))"
by (induct xs) auto

lemma map_cong [fundef_cong]:
 assumes "xs = ys"
         "(\<forall>x. x \<in> set ys \<longrightarrow> f x = g x)"
 shows   " map f xs = map g ys"
using assms by simp

lemma map_injective:
 assumes "map f xs = map f ys"
         " inj f"
 shows  " xs = ys"
using assms by (meson injD interval.inj_map)

lemma inj_map_eq_map [simp]:
 assumes "inj f"
 shows   "  (map f xs = map f ys)  = (xs = ys)"
using assms by (blast dest: map_injective)

lemma inj_mapI:
 assumes "inj f"
 shows   " inj (map f)"
using assms interval.inj_map by blast

lemma inj_mapD:
 assumes "inj (map f)"
 shows   "  inj f"
using assms by (metis inj_def interval.map(1) interval.simps(1))

lemma inj_map[iff]:
 "inj (map f) = inj f"
by (blast dest: inj_mapD intro: inj_mapI)

lemma map_idI: 
 assumes "(\<forall>x. x \<in> set xs \<longrightarrow> f x = x)"
 shows   "  map f xs = xs"
using assms by (induct xs) auto

lemma map_is_state_conv[iff]:
 "(map f xs = \<langle>x\<rangle>) = (\<exists> y. xs = \<langle>y\<rangle> \<and> f y = x)"
proof (cases xs)
case (St x1)
then show ?thesis by simp
next
case (Cons x21 x22)
then show ?thesis by simp
qed

lemma state_is_map_conv [iff]:
 "(\<langle>x\<rangle> = map f xs) = (\<exists> y. \<langle>y\<rangle> = xs \<and> f y = x)"
proof (cases xs)
case (St x1)
then show ?thesis by auto
next
case (Cons x21 x22)
then show ?thesis by auto
qed

lemma map_eq_cons_conv:
 "(map f xs = y\<odot>ys) = (\<exists> z zs. xs = z\<odot>zs \<and> f z =y \<and> map f zs = ys)"
by (cases xs) auto

lemma cons_eq_map_conv:
 "(y\<odot>ys= map f xs ) = (\<exists> z zs. z\<odot>zs=xs \<and> f z =y \<and> ys = map f zs)"
by (cases xs) auto

lemma ex_map_conv:
  "(\<exists>xs. ys = map f xs) = (\<forall>y \<in> set ys. \<exists>x. y = f x)"
by (induct ys) (auto simp add: cons_eq_map_conv)


functor map: map
by (simp_all add: id_def)

declare map.id [simp]


lemma intfirst_map:
 "intfirst (map f xs) = f (intfirst xs)"
by (cases xs) simp_all


lemma intlast_map:
 "intlast (map f xs) = f (intlast xs)"
proof (cases xs rule: interval_rev_cases)
case (St x)
then show ?thesis by simp
next
case (snoc ys y)
then show ?thesis by (simp add: interval_intapp_nth)
qed  


lemma map_tail:
shows   " map f (suffix 1 xs) = (suffix 1 (map f xs))"
by (cases xs) simp_all

lemma map_eq_imp_intlen_eq:
 assumes "map f xs = map g ys"
 shows "intlen xs = intlen ys"
using assms
proof (induct ys arbitrary: xs)
case (St x)
then show ?case by auto
next
case (Cons x1a ys)
then show ?case by (metis interval_intlen_map)
qed

lemma interval_set_map [simp]: 
  "set (map f xs) = f`(set xs)"
by (induct xs) auto

lemma map_inj_on:
 assumes map: "map f xs = map f ys" and
         inj: "inj_on f (set xs \<union> set ys)"
  shows "xs =ys"
using map_eq_imp_intlen_eq[OF map] assms
proof (induction rule: interval_induct2)
case (St x y)
then show ?case by auto
next
case (Cons x xs y ys)
then show ?case
by (metis (mono_tags, hide_lams) UnI1 UnI2 inj_on_def interval.inj_map_strong) 
qed

lemma inj_on_map_eq_map:
assumes "inj_on f (set xs \<union>  set ys)"
shows   "(map f xs = map f ys) = (xs = ys)"
using assms by (blast dest:map_inj_on)

lemma inj_on_mapI:
 assumes "inj_on f (\<Union> (set ` A))"
 shows   " inj_on (map f) A"
using assms by (blast intro:inj_onI dest:inj_onD map_inj_on)

lemma map_prefix:
 assumes "k \<le> intlen xs"
 shows   " map f (prefix k xs) = prefix k (map f xs)"
using assms
proof (induction xs arbitrary: k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof (cases k)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis using Cons.IH Cons.prems by auto
  qed
qed


lemma map_suffix:
 assumes "k \<le> intlen xs"
 shows   " map f (suffix k xs) = suffix k (map f xs)"
using assms
proof (induction xs arbitrary: k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
 proof (cases k)
 case 0
 then show ?thesis by auto
 next
 case (Suc nat)
 then show ?thesis using Cons.IH Cons.prems by auto
 qed
qed


subsubsection \<open>index sequence\<close>

lemma interval_idx_less:
    assumes  " index_sequence x idx"
             " n+k<intlen idx" 
    shows "  nth idx n < nth idx (Suc(n+k)) "
using index_sequence_def assms by (induct k) auto

lemma interval_idx_less_eq:
 assumes "index_sequence x l"
         "k\<le>j"
         "j\<le>intlen l"
 shows    "nth l k \<le> nth l j"
using assms 
proof (cases "k=j")
 show "index_sequence x l \<Longrightarrow> k \<le> j \<Longrightarrow> j \<le> intlen l \<Longrightarrow> k = j \<Longrightarrow> nth l k \<le> nth l j" 
   by blast
 show "index_sequence x l \<Longrightarrow> k \<le> j \<Longrightarrow> j \<le> intlen l \<Longrightarrow> k \<noteq> j \<Longrightarrow> nth l k \<le> nth l j " 
   by (metis Suc_le_lessD interval_idx_less le_SucE le_eq_less_or_eq le_zero_eq 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse zero_induct)
qed


lemma interval_idx_mono:
 assumes "index_sequence x l"
 shows    "mono (\<lambda> x. nth l x)"
proof - 
 have 1: "\<forall>x y. x \<le> y \<longrightarrow> nth l x \<le> nth l y"
   proof 
     fix x 
     show "\<forall>y\<ge>x. nth l x \<le> nth l y " 
     proof
      fix y
      show "x \<le> y \<longrightarrow> nth l x \<le> nth l y" 
       proof -
        have 2: "x\<le> y \<and> y \<le> intlen l \<Longrightarrow> nth l x \<le> nth l y" 
         using assms interval_idx_less_eq by blast
        have 3: "x\<le> y  \<and> x>intlen l \<Longrightarrow> nth l x \<le> nth l y" 
           using assms interval_nth_last_stutter 
           by (metis "2" le_cases less_imp_add_positive 
               ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
        have 4: "x\<le> y  \<and> x\<le>intlen l \<and> y > intlen l \<Longrightarrow> nth l x \<le> nth l y"   
           by (metis assms interval_idx_less_eq interval_nth_last_stutter 
               less_imp_add_positive order_refl)
        show ?thesis 
         using "2" "3" "4" not_less by blast
       qed
      qed
     qed
    show ?thesis 
    by (simp add: "1" monoI)
qed


lemma interval_idx_less_last :
 assumes " index_sequence x idx"
         " i<intlen idx "
         " i+(intlen idx - (i+1))< intlen idx" 
 shows   " nth idx i < nth idx (Suc(i+(intlen idx -(i+1))))"
using assms interval_idx_less by blast

lemma interval_idx_less_last_1:
 assumes  " index_sequence x idx"
          " i<intlen idx" 
 shows    " nth idx i < nth idx (intlen idx) "
using assms interval_idx_less_last by auto

lemma interval_idx_greater_first:
 assumes  " index_sequence x idx"
          " 0 < i"
          " i\<le> intlen idx"
 shows    " x < nth idx i"
using assms 
proof -
 have 1: " nth idx 0 = x "
    using assms by (simp add: index_sequence_def)
 have 2: "\<forall> i. 0 < i \<and> i\<le> intlen idx \<longrightarrow> (nth idx 0) < (nth idx i) " 
     proof 
       fix i
       show "0 < i \<and> i \<le> intlen idx \<longrightarrow> nth idx 0 < nth idx i " 
        by (meson Suc_leI assms(1) dual_order.strict_trans1 index_sequence_def interval_idx_less_eq)
     qed
 show ?thesis 
 using "1" "2" assms(2) assms(3) by blast 
qed

lemma interval_idx_cons:
    " index_sequence y (x\<odot>ls) = 
       (x=y \<and> x<nth ls 0 \<and> index_sequence (nth ls 0) ls)"
using less_Suc_eq_0_disj by (simp add: index_sequence_def) auto

lemma interval_idx_shift_mono:
    "mono (shift k)" 
by (simp add: Interval.shift_def mono_def)     

lemma interval_idx_expand:
 assumes  "index_sequence 0 l"
          "(nth l (intlen l)) = (intlen xs)"
          " i< (intlen l)"
 shows    "  (nth l i) \<le>  (nth l (i+1)) \<and> (nth l (i+1)) \<le> (intlen xs)"
using assms
by (metis Suc_eq_plus1 Suc_lessI add.right_neutral interval_idx_less interval_idx_less_last_1 
    less_imp_le_nat order_refl)

lemma interval_idx_shift_idx [simp]:
    " ( index_sequence (x+k) (map (shift k)  idx)) = (index_sequence x idx) " 
by (simp add: Interval.shift_def index_sequence_def interval_nth_map)    

lemma interval_idx_shiftm :
  assumes "index_sequence k lsk"
 shows    "index_sequence 0 (map (shiftm k) lsk) \<and> k \<le> (nth lsk 0)"
using assms
proof (auto simp add:  index_sequence_def shiftm_def interval_nth_map )
  show "\<And>n. \<forall>n<intlen lsk. nth lsk n < nth lsk (Suc n) \<Longrightarrow>
         k = nth lsk 0 \<Longrightarrow>
         n < intlen lsk \<Longrightarrow> nth lsk n - nth lsk 0 < nth lsk (Suc n) - nth lsk 0" 
  by (metis assms diff_less_mono interval_idx_greater_first le_eq_less_or_eq neq0_conv)
qed  

lemma interval_idx_shiftm_a :
  assumes "index_sequence 0 (map (shiftm k) lsk)"
          "k \<le> (nth lsk 0)"
 shows    "index_sequence k lsk"
using assms 
by (auto simp add: index_sequence_def shiftm_def interval_nth_map) 

lemma interval_idx_shiftm_b :
 "index_sequence k lsk = (index_sequence 0 (map (shiftm k) lsk) \<and> k \<le> (nth lsk 0))"
using interval_idx_shiftm interval_idx_shiftm_a by blast

lemma interval_idx_shiftm_c :
 "index_sequence (nth lsk 0) lsk = index_sequence 0 (map (shiftm (nth lsk 0)) lsk)"
using interval_idx_shiftm_b by blast

lemma interval_lsk_ls :
    "(index_sequence k (lsk) \<and> lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
      (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> index_sequence 0 (ls) ) "
proof (simp add: interval_eq_nth_eq index_sequence_def shift_def shiftm_def interval_nth_map)
  show "(nth lsk 0 = k \<and>
     (\<forall>n<intlen lsk. nth lsk n < nth lsk (Suc n)) \<and>
     intlen lsk = intlen ls \<and>
     (\<forall>i\<le>intlen lsk. nth lsk i = nth ls i + k) \<and>
     nth ls 0 = 0 \<and> (\<forall>n<intlen ls. nth ls n < nth ls (Suc n))) =
    (nth lsk 0 = k \<and>
     (\<forall>n<intlen lsk. nth lsk n < nth lsk (Suc n)) \<and>
     intlen ls = intlen lsk \<and>
     (\<forall>i\<le>intlen ls. nth ls i = nth lsk i - k) \<and>
     nth ls 0 = 0 \<and> (\<forall>n<intlen ls. nth ls n < nth ls (Suc n)))" (is "?L=?R")
    proof rule
     show "?L \<Longrightarrow> ?R"
    by (metis (no_types, lifting) add_diff_cancel_right')
     show "?R \<Longrightarrow> ?L "
      by (metis add.commute add_diff_inverse_nat diff_is_0_eq less_eq_Suc_le less_irrefl_nat 
          less_nat_zero_code less_or_eq_imp_le old.nat.exhaust)
   qed
qed

lemma interval_idx_link_shiftm:
    "(index_sequence k (lsk) \<and> ls = map (shiftm k) lsk ) =
     (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> 
      index_sequence 0 (ls) \<and> (intlen ls) =(intlen lsk)) "
using interval_idx_shiftm using interval_intlen_map by blast
    
lemma interval_idx_link:
    "(lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
     (lsk = map (shift k) ls \<and> index_sequence k (lsk) \<and> index_sequence 0 (ls)\<and> 
      (intlen ls) =(intlen lsk))"
by (metis add.left_neutral interval_idx_shift_idx interval_intlen_map)

lemma interval_idx_bound_0 :
 assumes "index_sequence 0 ls"
         "nth ls (intlen ls) = intlen (suffix k xs)"
         " i\<le>intlen ls"
 shows   " nth ls i \<le>  intlen (suffix k xs) "
using assms
by (metis eq_iff interval_idx_less_last_1 le_neq_implies_less less_imp_le_nat) 
 
lemma interval_idx_bound_1:
  "(index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs))) =
     (index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs)) \<and>
     (\<forall>i. (i\<le>intlen ls) \<longrightarrow> ((nth ls (i)) \<le> (intlen (suffix k xs)))) )" 
using interval_idx_bound_0 by blast


lemma interval_idx_less_equal:
assumes "index_sequence 0 l"
        "(nth l (intlen l)) = intlen xs"
        "i \<le> intlen l"
        "n \<le> intlen l"
shows   "\<forall> j . j\<le>n \<longrightarrow> nth l j \<le> nth l n"
using assms 
using interval_idx_less_eq by blast


lemma interval_idx_less_than:
assumes "index_sequence 0 l"
        "(nth l (intlen l)) = intlen xs"
        "i \<le> intlen l"
        "n \<le> intlen l"
shows   "\<forall> j. j>n \<and> j \<le> intlen l\<longrightarrow> nth l j > nth l n"
by (meson Suc_leI assms interval_idx_less_equal index_sequence_def less_le_trans)


lemma interval_idx_sub:
 assumes "k\<le>n"
         "n \<le> intlen l"
         "index_sequence 0 l"
 shows   "index_sequence (nth l k) (sub k n l)"
proof -
 have 1: "index_sequence (nth l k) (sub k n l) = 
          ((nth (sub k n l) 0) = (nth l k) \<and> 
           (\<forall>i. i<intlen(sub k n l) \<longrightarrow> (nth (sub k n l) i) < (nth (sub k n l) (Suc i))))"
    using index_sequence_def by auto 
 have 2: "(nth (sub k n l) 0) = (nth l k) "
       using assms interval_intfirst_sub by auto
 have 3: " intlen(sub k n l) = (n -k)"
   by (simp add: assms)
 have 4: "(\<forall>i. i<intlen(sub k n l) \<longrightarrow> (nth (sub k n l) i) < (nth (sub k n l) (Suc i))) "
  proof 
   fix i
   show "i < intlen (sub k n l) \<longrightarrow> nth (sub k n l) i < nth (sub k n l) (Suc i) "
   proof -
    have 41: "i < (n-k) \<longrightarrow> nth (sub k n l) i = nth l (k+i) "
    by (simp add: assms)
    have 42: "i < (n-k) \<longrightarrow> nth (sub k n l) (Suc i) = nth l (k+(Suc i)) "
      by (simp add: assms)
    have 43: "i < (n-k) \<longrightarrow> nth l (k+i) < nth l (k+(Suc i))"
      using assms 
      using index_sequence_def by auto
    show ?thesis using "3" "41" "42" "43" by auto
   qed
  qed
 show ?thesis by (simp add: "1" "2" "4")
qed

lemma interval_idx_split:
 assumes " n\<le> intlen l"
 shows   "index_sequence 0 l = 
          (index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l))"
proof -
 have 1: "index_sequence 0 l \<longrightarrow> 
          index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l) "
   using interval_idx_sub using assms index_sequence_def by auto
 have 2: "index_sequence 0 (prefix n l) =
          ( (nth (prefix n l) 0) = 0 \<and>
            (\<forall>i. i< intlen(prefix n l) \<longrightarrow> (nth (prefix n l) i) < (nth (prefix n l) (Suc i))))"
   using index_sequence_def by blast
 have 3: "( (nth (prefix n l) 0) = 0 \<and>
            (\<forall>i. i< intlen(prefix n l) \<longrightarrow> (nth (prefix n l) i) < (nth (prefix n l) (Suc i)))) =
            ( (nth l 0) = 0 \<and> 
              (\<forall>i. i<n \<longrightarrow> (nth l i) < (nth l (Suc i)))) "
    using assms by auto      
 have 4: "index_sequence (nth l n) (suffix n l) =
          ( (nth (suffix n l) 0) = (nth l n) \<and>
            (\<forall> i. i<intlen(suffix n l) \<longrightarrow> (nth (suffix n l) i) < (nth (suffix n l) (Suc i)))) "
    using index_sequence_def by blast
 have 5: "( (nth (suffix n l) 0) = (nth l n) \<and>
            (\<forall> i. i<intlen(suffix n l) \<longrightarrow> (nth (suffix n l) i) < (nth (suffix n l) (Suc i)))) =
            ( (nth l n) = (nth l n) \<and>
            (\<forall> i. i<intlen l - n \<longrightarrow> (nth l (i+n) ) < (nth l ((Suc i)+n)))) "
       by (metis (no_types, lifting) Suc_leI add.commute assms interval_intlast_intfirst 
           interval_intlast_prefix interval_nth_suffix interval_nth_zero_intfirst 
           interval_suffix_length_good less_imp_le_nat)
 have 6: "(\<forall> i. i<intlen l - n \<longrightarrow> (nth l (i+n) ) < (nth l ((Suc i)+n))) =
          (\<forall> i. n \<le> (i+n) \<and> (i+n)<intlen l  \<longrightarrow> (nth l (i+n) ) < (nth l ((Suc i)+n))) "
       by auto
 have 7: " (index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l)) \<longrightarrow>
          (nth l 0) = 0 \<and> (\<forall>i. i<n \<longrightarrow> (nth l i) < (nth l (Suc i))) " 
       using "2" "3" by blast
 have 8: "(index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l)) \<longrightarrow>
           (\<forall> i. n \<le> i \<and> i<intlen l  \<longrightarrow> (nth l (i) ) < (nth l ((Suc i)))) "
       by (metis (no_types, lifting) "4" "5" "6" add_Suc diff_add)
 have 9: "(index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l)) \<longrightarrow>
          (nth l 0) = 0 \<and> (\<forall>i. i<intlen l \<longrightarrow> (nth l i) < (nth l (Suc i))) "
  using "7" "8" not_le by blast
 have 10: "(index_sequence 0 (prefix n l) \<and> index_sequence (nth l n) (suffix n l)) \<longrightarrow>
           index_sequence 0 l "
    using "9" index_sequence_def by blast
 from 10 1 show ?thesis by blast
qed

lemma interval_idx_suffixa:
 assumes "n\<le> intlen l"
         "index_sequence (nth l n) (suffix n l)"
 shows   "index_sequence 0 ((map (shiftm (nth l n))  (suffix n l)))"
using assms interval_idx_shiftm by blast

lemma interval_idx_greater:
 assumes "index_sequence k l"
 shows   "(\<forall>i. i\<le> intlen l \<longrightarrow> k \<le> (nth l i))"
by (metis assms eq_iff index_sequence_def interval_idx_greater_first less_imp_le neq0_conv)

lemma interval_idx_suffixb:
 assumes "n\<le> intlen l"
         "index_sequence 0 ((map (shiftm (nth l n))  (suffix n l)))"
 shows   "index_sequence (nth l n) (suffix n l)"
by (metis assms interval_idx_shiftm_c interval_intlast_intfirst interval_intlast_prefix 
    interval_nth_zero_intfirst)

lemma interval_idx_suffix:
 assumes "n\<le> intlen l"
 shows   "index_sequence (nth l n) (suffix n l) =
          index_sequence 0 ((map (shiftm (nth l n))  (suffix n l)))"
using assms interval_idx_shiftm interval_idx_suffixb by blast
 
lemma interval_idx_intfirst:
 assumes "index_sequence 0 (x1a\<odot> l)"
  shows  " x1a < intfirst(l)"
by (metis assms interval_idx_cons interval_nth_zero_intfirst) 

lemma interval_idx_expand1:
 " (index_sequence x1a (x1a \<odot> l)) = ( x1a < intfirst l \<and> index_sequence (intfirst l) l )"
using interval_nth_zero_intfirst less_Suc_eq_0_disj by (auto simp add: index_sequence_def)


lemma interval_idx_intlen_leq_intlast_intfirst:
 assumes "index_sequence (intfirst l) l"
 shows   "intlen (l) \<le> (intlast l - intfirst l)"
using assms
proof
 (induct l)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a l)
 then show ?case 
   proof -
    have 1: "intlen (x1a \<odot> l) = intlen l +1"
     by simp
    have 2: "index_sequence (intfirst l) l" 
     using Cons.prems interval_idx_expand1 by auto 
    have 3: "intlast (x1a \<odot> l) = intlast l" 
     by simp
    have 4: "intfirst (x1a \<odot> l) = x1a"
     by simp
    have 5: "x1a < intfirst l"
     using Cons.prems interval_idx_expand1 by auto
    have 6: "intlen l \<le> intlast l - intfirst l" 
     using "2" Cons.hyps by blast
    have 7: "intlen l +1 \<le> (intlast l - intfirst l)+1"
       using "6" add_le_cancel_right by blast
    have 8: " (intlast l - intfirst l) +1 \<le> intlast l - x1a "
      by (metis "5" "6" Suc_eq_plus1 diff_is_0_eq diff_less_mono diff_less_mono2 
          interval_nth_intlen_intlast interval_nth_zero_intfirst le_antisym less_eq_Suc_le 
          nat_le_linear not_le not_less0)
    show ?thesis using "7" "8" by auto
   qed      
 qed

lemma interval_idx_intlen_leq:
assumes "index_sequence (intfirst l) l"
        "intlast(l) \<le> intlen xs"
shows "intlen (l) \<le> intlen (sub (intfirst l) (intlast l) xs) "
proof -
 have 1: "intfirst l \<le> intlast l"
   using assms by (metis eq_iff gr0I index_sequence_def interval_idx_less_last_1 
      interval_nth_intlen_intlast less_imp_le_nat) 
 have 2: "intlen (sub (intfirst l) (intlast l) xs) = (intlast l - intfirst l)"
     using "1" assms interval_intlen_sub by blast
 have 3: "intlen (l) \<le> (intlast l - intfirst l)" 
     using assms interval_idx_intlen_leq_intlast_intfirst by blast
 show ?thesis using "2" "3" by linarith
qed


lemma interval_idx_shiftm_sub_nth:
assumes "index_sequence 0 l"
        " (nth l (intlen l)) = intlen xs"
        "  k \<le> n "
        " n \<le> intlen l"
 shows  "\<forall> j. j \<le> n-k \<longrightarrow> 
           nth (map (shiftm (nth l k)) (sub k n l)) j = nth l (k+j) - (nth l k)"
by (simp add: Nat.le_diff_conv2 assms interval_nth_map shiftm_def)


lemma interval_idx_shiftm_suffix_nth:
assumes "index_sequence 0 l"
        "  (nth l (intlen l)) = intlen xs"
        "  n \<le> intlen l"
 shows  "\<forall> j. j \<le> intlen l -n \<longrightarrow> 
           nth (map (shiftm (nth l n)) (suffix n l)) j = nth l (n+j) - (nth l n)"
using assms by (metis  interval_nth_map interval_nth_suffix shiftm_def) 
 


subsubsection \<open>upt\<close>

lemma upt_rec[code]:  
 "[i..\<le>j] = (if i<j then i \<odot> [Suc i..\<le>j] else \<langle>j\<rangle>)"
by (induct j) auto

lemma upt_conv_st [simp]:
assumes "j< i"
shows   "[i..\<le>j] = \<langle>j\<rangle>"
using assms
by (metis Interval.upt.simps(1) Interval.upt.simps(2) Suc_leD less_Suc_eq_0_disj not_le)

lemma upt_same:
 "[i..\<le>i] = \<langle>i\<rangle>"
by (metis Interval.upt.simps(1) Interval.upt.simps(2) less_Suc_eq less_Suc_eq_0_disj not_le)

lemma upt_eq_st_conv[simp]:
 "([i..\<le>j] = \<langle>j\<rangle>) = (j\<le>i)"
by (simp add: Interval.upt_rec)

lemma upt_eq_cons_conv:
 "([i..\<le>j] = x\<odot>xs) = (i<j \<and> i=x \<and> [Suc i..\<le>j] = xs)"
using Interval.upt_rec by (induct j arbitrary: x xs) auto

lemma upt_suc_append:
assumes "i \<le> j"
shows  " [i..\<le>(Suc j)] = [i..\<le>j]\<ominus>\<langle>(Suc j)\<rangle>"
using assms by simp

lemma upt_conv_cons:
assumes "i<j"
shows   "  [i..\<le>j] = i \<odot> [(Suc i)..\<le>j]"
using assms by (simp add: upt_rec)

lemma upt_conv_cons_cons:
 "( m \<odot> n \<odot> ns = [m..\<le>q]) = (n \<odot> ns = [(Suc m)..\<le>q])"
proof (cases "m \<le> q")
case True
then show ?thesis by (simp add: Interval.upt_rec)
next
case False
then show ?thesis by auto
qed

lemma upt_add_eq_append:
assumes "i \<le> j"
        " k>0"
shows   "[i..\<le>j+k] = [i..\<le>j]\<ominus>[Suc j..\<le>j+k]"
using assms 
proof
 (induct k)
 case 0
 then show ?case by blast
 next
 case (Suc k)
 then show ?case using Suc_less_eq le_simps(2) by auto
qed

lemma upt_length:
 "intlen [i..\<le>j] = j - i"
by (induct j) (auto simp add: Suc_diff_le)

lemma upt_nth_help:
 "Interval.nth [i..\<le>i + k] k = i + k"
proof 
   (induct k arbitrary: i)
   case 0
   then show ?case by (simp add: upt_same)
   next
   case (Suc k)
   then show ?case 
   by (metis Interval.upt_rec add_Suc_shift interval_nth_Suc less_add_same_cancel1 
           zero_less_Suc)
qed

lemma upt_nth:
assumes "i + k \<le> j"
shows   "(nth [i..\<le>j] k) = i + k"
using assms 
proof 
  (induct j arbitrary: k i)
  case 0
  then show ?case by simp
  next
  case (Suc j)
  then show ?case 
   by (metis Interval.upt.upt_Suc Nat.le_diff_conv2 add.commute add_leD1 
          interval_intapp_nth le_SucE upt_length upt_nth_help)
qed

lemma upt_intfirst:
assumes "i \<le> j"
shows   "intfirst [i..\<le>j] = i"
using assms by (simp add: Interval.upt_rec)

lemma upt_intlast:
 " intlast [i..\<le>j] = j"
by (metis add_diff_inverse_nat interval_nth_intlen_intlast interval_st_intlen order_refl 
          upt_conv_st upt_length upt_nth)

lemma prefix_upt:
assumes "i+m \<le> n"
shows   "prefix m [i..\<le>n] = [i..\<le>i+m]"
using assms 
proof
 (induct m arbitrary: i)
 case 0
 then show ?case by (simp add: upt_nth upt_same)
 next
 case (Suc m)
 then show ?case using Interval.upt_rec by auto
qed

lemma suffix_upt:
 " suffix m [i..\<le>j] = [i+m..\<le>j]"
proof
 (induct m arbitrary: i j)
 case 0
 then show ?case by simp
 next
 case (Suc j)
 then show ?case using Interval.upt_rec 
 by (metis add_Suc_shift interval_suffix_suc not_less_eq not_less_iff_gr_or_eq suffix.simps(1))
qed

lemma map_suc_upt:
 " map Suc [m..\<le>n] = [Suc m..\<le>Suc n]"
proof
 (induct n arbitrary: m)
 case 0
 then show ?case by simp
 next
 case (Suc n)
 then show ?case by simp
qed

lemma map_add_upt:
 "map (\<lambda>i. i + n) [0..\<le>m] = [n..\<le>m+n]"
proof
 (induct m)
 case 0
 then show ?case by (simp add: upt_same)
 next
 case (Suc m)
 then show ?case by simp
qed



subsubsection \<open>Set\<close>

lemma interval_set_intapp [simp]:
 "set (xs \<ominus> ys) = (set xs \<union> set ys)"
by (induct xs) auto

lemma interval_finite_set [iff]: 
  "finite (set (xs:: 'a interval) )"
by (induct xs) auto

lemma interval_hd_in_set [simp]: 
  "  x \<in> set (x\<odot>xs)"
by simp

lemma interval_set_subset_Cons: 
   "set xs \<subseteq> set (x \<odot> xs)"
by auto

lemma interval_set_ConsD: 
 assumes "y \<in> set (x \<odot> xs)"
 shows   " y=x \<or> y \<in> set xs"
using assms by auto

lemma interval_exists_cons:
 "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys) \<longleftrightarrow> 
  (P x \<and> (\<exists> ys \<in> set xs. P ys)) \<or> (\<not> P x \<and> (\<exists> ys \<in> set xs. P ys)) \<or>
  (P x \<and> (\<forall> ys \<in> set xs. \<not> P ys))"
by auto

lemma interval_set_nonempty:
 " set xs \<noteq> {} "
by (induct xs) auto

lemma interval_set_intrev [simp]: 
  "set (intrev xs) = set xs"
by (induct xs) auto



lemma interval_split_interval:
assumes "x \<in> set xs " 
shows   " \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs= ys \<ominus> \<langle>x\<rangle>"
using assms 
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
by (metis intapp_St interval.inject(2) interval.set_cases interval_intapp_assoc)
qed

lemma interval_in_set_conv_decomp:
 "x \<in> set xs =
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs= ys \<ominus> \<langle>x\<rangle>)"
by (auto elim: interval_split_interval)

lemma interval_split_interval_first:
assumes "x \<in> set xs"
shows   "  \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> set ys \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle> \<and> x\<notin> set ys"
using assms 
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof (cases "x = x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
   using Cons interval_cons_eq_intappI 
   proof auto  
    show "\<And>ys zs.
       x \<noteq> x1a \<Longrightarrow>
       (\<And>x xs1 ys xs zs. x \<odot> xs1 = ys \<Longrightarrow> xs = xs1 \<ominus> zs \<Longrightarrow> x \<odot> xs = ys \<ominus> zs) \<Longrightarrow>
       x \<notin> interval.set ys \<Longrightarrow>
       xs = ys \<ominus> x \<odot> zs \<Longrightarrow>
       \<exists>ysa. (\<exists>zsa. x1a \<odot> ys \<ominus> x \<odot> zs = ysa \<ominus> x \<odot> zsa) \<and> x \<notin> interval.set ysa \<or>
             x1a \<odot> ys \<ominus> x \<odot> zs = ysa \<ominus> \<langle>x\<rangle> \<and> x \<notin> interval.set ysa"
      by (metis Un_iff empty_iff insert_iff intapp_St interval.simps(15) interval_intapp_assoc 
          interval_set_intapp)  
     show " x \<noteq> x1a \<Longrightarrow>
           (\<And>x xs1 ys xs zs. x \<odot> xs1 = ys \<Longrightarrow> xs = xs1 \<ominus> zs \<Longrightarrow> x \<odot> xs = ys \<ominus> zs) \<Longrightarrow>
           xs = \<langle>x\<rangle> \<Longrightarrow>
           \<exists>ys. (\<exists>zs. \<langle>x1a, x\<rangle> = ys \<ominus> x \<odot> zs) \<and> x \<notin> interval.set ys \<or> \<langle>x1a, x\<rangle> = ys \<ominus> \<langle>x\<rangle> \<and> 
               x \<notin> interval.set ys"
      by (metis insert_absorb insert_iff interval.simps(15) interval_intrev_intapp_cons 
          interval_rev_intapp interval_set_nonempty intrev.simps(1))
     show "\<And>zs. x \<noteq> x1a \<Longrightarrow>
          (\<And>x xs1 ys xs zs. x \<odot> xs1 = ys \<Longrightarrow> xs = xs1 \<ominus> zs \<Longrightarrow> x \<odot> xs = ys \<ominus> zs) \<Longrightarrow>
          xs = x \<odot> zs \<Longrightarrow>
          \<exists>ys. (\<exists>zsa. x1a \<odot> x \<odot> zs = ys \<ominus> x \<odot> zsa) \<and> x \<notin> interval.set ys \<or>
               x1a \<odot> x \<odot> zs = ys \<ominus> \<langle>x\<rangle> \<and> x \<notin> interval.set ys"
       by (metis Interval.nth.simps(1) intapp_St interval.set_cases interval_intapp_not_state)
     show "\<And>ys. x \<noteq> x1a \<Longrightarrow>
          (\<And>x xs1 ys xs zs. x \<odot> xs1 = ys \<Longrightarrow> xs = xs1 \<ominus> zs \<Longrightarrow> x \<odot> xs = ys \<ominus> zs) \<Longrightarrow>
          x \<notin> interval.set ys \<Longrightarrow>
          xs = ys \<ominus> \<langle>x\<rangle> \<Longrightarrow>
          \<exists>ysa. (\<exists>zs. x1a \<odot> ys \<ominus> \<langle>x\<rangle> = ysa \<ominus> x \<odot> zs) \<and> x \<notin> interval.set ysa \<or>
                x1a \<odot> ys \<ominus> \<langle>x\<rangle> = ysa \<ominus> \<langle>x\<rangle> \<and> x \<notin> interval.set ysa "
      by (metis insert_iff intapp_Cons interval.simps(16)) 
    qed
  qed
qed

lemma in_set_conv_decomp_first:
 "x \<in> set xs = 
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> set ys \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle> \<and> x\<notin> set ys)"
by (auto dest!: interval_split_interval_first)

lemma interval_split_interval_last:
 assumes "x \<in> set xs"
  shows " \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> set zs \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> x \<notin> set zs\<or> 
           xs = ys \<ominus> \<langle>x\<rangle> "
using assms
proof (induct xs rule: interval_rev_induct)
case (St y)
then show ?case by simp
next
case (snoc x1a xs)
  then show ?case proof (cases "x = x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis using snoc by fastforce
qed 
qed

lemma interval_in_set_conv_decomp_last:
 "x \<in> set xs = 
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> set zs \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> x \<notin> set zs\<or> 
           xs = ys \<ominus> \<langle>x\<rangle>)" 
by (auto dest!: interval_split_interval_last)

lemma interval_list_prop:
assumes "\<exists> x \<in> set xs. P x"
shows   "(\<exists> ys x zs. (xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle>) \<and> P x)"
using assms
proof (induct xs)
case (St x)
then show ?case by auto
next
  case (Cons x1a xs)
  then show ?case 
  by (meson interval_split_interval)
qed

lemma interval_split_interval_propE:
 assumes "\<exists> x\<in> set xs. P x"
 obtains ys x zs where "xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle>" and "P x" 
using interval_list_prop [OF assms] by blast

lemma interval_split_interval_first_prop:
assumes "\<exists> x \<in> set xs. P x" 
 shows  "(\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> set ys. \<not> P y))) \<and> P x 
  )"
using assms
proof (induct xs)
case (St x)
then show ?case by auto
next
  case (Cons x1a xs)
  then show ?case proof (cases "P x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
   proof -
    have 1: " \<exists> x \<in> set xs. P x" 
     using Cons.prems False by auto
    have 2: "\<not> P x1a "
      by (simp add: False)
    have 3: " \<exists>ys x zs.
       (xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set ys. \<not> P y) \<or>
        xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set ys. \<not> P y)) \<and>
       P x"
     using "1" Cons.hyps by blast
    obtain ys x zs where 4: " (xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set ys. \<not> P y) \<or>
        xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set ys. \<not> P y)) \<and>
       P x"
        using "3" by auto
    have 5: "(xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set ys. \<not> P y))  \<longrightarrow> 
              (x1a \<odot> xs = (x1a \<odot> ys) \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y)) 
         \<or> (x1a\<odot> xs) = \<langle>x\<rangle> \<or> (x1a\<odot> xs) = x \<odot> zs \<or>
           x1a \<odot> xs = (x1a\<odot>ys) \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y)" 
       using False by auto
    have 6: "xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set ys. \<not> P y) \<longrightarrow>
             (x1a \<odot> xs = (x1a \<odot> ys) \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y)) \<or>
             x1a \<odot> xs = (x1a\<odot>ys) \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y) \<or>
          (x1a\<odot> xs) = \<langle>x\<rangle> \<or> (x1a\<odot> xs) = x \<odot> zs"
        using False by auto 
    have 7: "P x "
     using "4" by auto
    have 8: "xs = \<langle>x\<rangle> \<longrightarrow> 
             (x1a \<odot> xs = (x1a \<odot> ys) \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y)) \<or>
             x1a \<odot> xs = \<langle>x1a\<rangle> \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set (\<langle>x1a\<rangle>). \<not> P y) \<or>
          (x1a\<odot> xs) = \<langle>x\<rangle> \<or> (x1a\<odot> xs) = x \<odot> zs
            " 
           by (simp add: False)
    have 9: " xs = x \<odot> zs \<longrightarrow>
             (x1a \<odot> xs = (\<langle>x1a\<rangle>) \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set (\<langle>x1a\<rangle>). \<not> P y)) \<or>
             x1a \<odot> xs = (x1a\<odot>ys) \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>interval.set (x1a\<odot>ys). \<not> P y) \<or>
          (x1a\<odot> xs) = \<langle>x\<rangle> \<or> (x1a\<odot> xs) = x \<odot> zs " 
        by (simp add: False)
    show ?thesis 
    using "4" "5" "6" "8" "9" by blast
   qed
  qed 
qed

lemma interval_split_interval_first_propE:
 assumes "\<exists> x \<in> set xs. P x"
 obtains ys x zs where "((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> set ys. \<not> P y)))" and "P x"
 using interval_split_interval_first_prop [OF assms] by blast

lemma interval_split_first_prop_iff:
 "(\<exists> x \<in> set xs. P x) \<longleftrightarrow>
  (\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> set ys. \<not> P y))) \<and> P x 
  )"
by (rule, erule interval_split_interval_first_prop) auto

lemma interval_split_interval_last_prop:
assumes "\<exists> x \<in> set xs. P x "
shows   "(\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> set zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>)) \<and> P x 
  )"
using assms
proof (induct xs rule: interval_rev_induct)
case (St y)
then show ?case by auto
next
case (snoc x1a xs)
  then show ?case proof (cases "P x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
    proof - 
     have 1: "\<exists> x \<in> set xs. P x" 
       using False snoc.prems by auto
     have 2: "\<not> P x1a"
       by (simp add: False) 
     have 3: " \<exists>ys x zs.
       (xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y) \<or>
        xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y) \<or> xs = ys \<ominus> \<langle>x\<rangle>) \<and>
       P x" 
       using "1" snoc.hyps by blast
     obtain ys x zs where 4: "(xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y) \<or>
        xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y) \<or> xs = ys \<ominus> \<langle>x\<rangle>) \<and>
       P x" 
      using "3" by auto
     have 5: " P x"
      using "4" by auto
     have 6: "(xs = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y)) \<longrightarrow>
              (xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> x \<odot> (zs \<ominus> \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set (zs \<ominus> \<langle>x1a\<rangle>). \<not> P y) \<or>
               xs \<ominus> \<langle>x1a\<rangle> = \<langle>x\<rangle> \<or>
               xs \<ominus> \<langle>x1a\<rangle> = x \<odot> (zs \<ominus> \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set (zs\<ominus> \<langle>x1a\<rangle>). \<not> P y) \<or> 
               xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> \<langle>x\<rangle>) 
             " 
        by (simp add: False)
     have 7: "xs = \<langle>x\<rangle> \<longrightarrow>
              (xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> x \<odot> ( \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set ( \<langle>x1a\<rangle>). \<not> P y) \<or>
               xs \<ominus> \<langle>x1a\<rangle> = \<langle>x\<rangle> \<or>
               xs \<ominus> \<langle>x1a\<rangle> = x \<odot> ( \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set ( \<langle>x1a\<rangle>). \<not> P y) \<or> 
               xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> \<langle>x\<rangle>) 
             " 
       using False by auto
     have 8: "xs = x \<odot> zs \<and> (\<forall>y\<in>interval.set zs. \<not> P y) \<longrightarrow>
              (xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> x \<odot> (zs \<ominus> \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set (zs \<ominus> \<langle>x1a\<rangle>). \<not> P y) \<or>
               xs \<ominus> \<langle>x1a\<rangle> = \<langle>x\<rangle> \<or>
               xs \<ominus> \<langle>x1a\<rangle> = x \<odot> (zs \<ominus> \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set (zs\<ominus> \<langle>x1a\<rangle>). \<not> P y) \<or> 
               xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> \<langle>x\<rangle>) 
             " 
          by (simp add: False)
     have 9: " xs = ys \<ominus> \<langle>x\<rangle> \<longrightarrow>
               (xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> x \<odot> ( \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set ( \<langle>x1a\<rangle>). \<not> P y) \<or>
               xs \<ominus> \<langle>x1a\<rangle> = \<langle>x\<rangle> \<or>
               xs \<ominus> \<langle>x1a\<rangle> = x \<odot> ( \<langle>x1a\<rangle>) \<and> (\<forall>y\<in>interval.set ( \<langle>x1a\<rangle>). \<not> P y) \<or> 
               xs \<ominus> \<langle>x1a\<rangle> = ys \<ominus> \<langle>x\<rangle>) 
             " 
        by (simp add: False)
     show ?thesis 
     using "4" "6" "7" "8" "9" by (metis (full_types)) 
    qed
  qed
qed

lemma interval_split_interval_last_propE:
 assumes "\<exists> x \<in> set xs. P x" 
 obtains ys x zs where "((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> set zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>))" and "P x"
 using interval_split_interval_last_prop [OF assms] by blast

lemma interval_split_interval_last_prop_iff:
 "(\<exists> x \<in> set xs. P x) \<longleftrightarrow>
  (\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> set zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> set zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>)) \<and> P x 
  )"
by (rule, erule interval_split_interval_last_prop, auto)

lemma interval_nth_and_set:
 "x \<in> set xs = (\<exists> i\<le>intlen xs. (nth xs i) =x)"
proof (induct xs)
case (St x)
then show ?case by auto
next
case (Cons x1a xs)
then show ?case 
by (metis Suc_le_mono insert_iff interval.simps(16) interval_intlen_cons interval_intlen_gr_zero 
    interval_nth_Suc interval_nth_cons interval_nth_zero intlen.simps(2) le_diff_conv 
    neq0_conv plus_1_eq_Suc)
qed

lemma interval_card_intlen:
 "card (set xs) \<le> intlen xs+1"
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by (simp add: card_insert_le_m1)
qed

lemma set_nth:
 "set xs = { (nth xs k) | k. k\<le> intlen xs}"
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof auto
   show "set xs = {nth xs k |k. k \<le> intlen xs} \<Longrightarrow>
    \<exists>k. x1a = (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> nth xs x) \<and> k \<le> Suc (intlen xs)"
   by force
   show "\<And>k. set xs = {nth xs k |k. k \<le> intlen xs} \<Longrightarrow>
         k \<le> intlen xs \<Longrightarrow>
         \<exists>ka. nth xs k = (case ka of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> nth xs x) \<and> ka \<le> Suc (intlen xs)"
   by force
   show "\<And>k. set xs = {nth xs k |k. k \<le> intlen xs} \<Longrightarrow>
         (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> nth xs x) \<noteq> x1a \<Longrightarrow>
         k \<le> Suc (intlen xs) \<Longrightarrow>
         \<exists>ka. (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> nth xs x) = nth xs ka \<and> ka \<le> intlen xs "
   by (metis (mono_tags, lifting) Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
  qed
qed

lemma prefix_set:
assumes "k\<le> intlen xs"
 shows  "  set (prefix k xs) = {(nth xs i) |i. i\<le>k}"
proof -
 have 1: " set (prefix k xs) = { (nth (prefix k xs) i) | i. i\<le> intlen (prefix k xs)}"
   by (simp add: set_nth)
 have 2: " k\<le> intlen xs"
   using assms by auto
 have 3: "{ (nth (prefix k xs) i) | i. i\<le> intlen (prefix k xs)} =
          { (nth xs i) | i. i\<le> k}" 
   using "2" by (metis interval_nth_prefix interval_prefix_length_good)
 show ?thesis using "1" "3" by auto
qed

 lemma suffix_set:
assumes "k\<le> intlen xs"
 shows  "  set (suffix k xs) = {(nth xs (k+i)) |i. i\<le>intlen xs - k}"
proof -
 have 1: "set (suffix k xs) = { (nth (suffix k xs) i) | i. i\<le> intlen (suffix k xs)}" 
   by (simp add: set_nth)
 have 2: "k \<le> intlen xs"
   using assms by auto 
 have 3: "{ (nth (suffix k xs) i) | i. i\<le> intlen (suffix k xs)} =
          { (nth xs (k+i)) | i. i\<le> intlen xs - k}"
   using "2" by (metis interval_nth_suffix interval_suffix_length_good) 
 show ?thesis  using "1" "3" by auto
qed

lemma suffix_set_a:
assumes "k\<le> intlen xs"
 shows  "  set (suffix k xs) = {(nth xs i) |i. k\<le> i \<and> i\<le>intlen xs }"
proof -
 have 1: "set (suffix k xs) = { (nth xs (k+i)) | i. i\<le> intlen xs - k}" 
   using assms suffix_set by blast
 have 2: "\<forall> x \<in> { (nth xs (k+i)) | i. i\<le> intlen xs - k} . 
            x \<in> {(nth xs i) |i. k\<le> i \<and> i\<le>intlen xs }"
    using assms by auto
 have 3: "\<forall> x \<in> {(nth xs i) |i. k\<le> i \<and> i\<le>intlen xs } . 
            x \<in> { (nth xs (k+i)) | i. i\<le> intlen xs - k} "
     using nat_le_iff_add by force
 have 4: "{ (nth xs (k+i)) | i. i\<le> intlen xs - k} =
          {(nth xs i) |i. k\<le> i \<and> i\<le>intlen xs }" 
  using "2" "3" by blast  
 show ?thesis by (simp add: "1" "4")
qed

lemma sub_interval_set:
 assumes "k \<le> n"
         "n \<le> intlen xs"
 shows   " set (sub k n xs) = { (nth xs (k+i)) | i. i \<le> n-k }"
proof -
 have 1: " set (sub k n xs) = { (nth (sub k n xs) i) | i. i \<le> intlen (sub k n xs) }"
    by (simp add: set_nth)
 have 2: " k\<le> n "
   using assms by auto
 have 3: " n \<le> intlen xs"
   using assms by auto
 have 4: " { (nth (sub k n xs) i) | i. i \<le> intlen (sub k n xs) } =
           { (nth xs (k+i) ) | i. i \<le> n-k } "
   using "2" "3" by force
 show ?thesis by (simp add: "1" "4")
qed


lemma sub_interval_set_a:
 assumes "k \<le> n"
         "n \<le> intlen xs"
 shows   " set (sub k n xs) = { (nth xs i) | i. k \<le> i \<and> i \<le> n }"
proof -
 have 1: "set (sub k n xs) = { (nth xs (k+i)) | i. i \<le> n-k }" 
   using assms sub_interval_set by blast
 have 2: "\<forall> x \<in> { (nth xs (k+i)) | i. i \<le> n-k } .
            x \<in> { (nth xs i) | i. k \<le> i \<and> i \<le> n }"
   using assms by auto 
 have 3: "\<forall> x \<in> { (nth xs i) | i. k \<le> i \<and> i \<le> n }.
            x \<in> { (nth xs (k+i)) | i. i \<le> n-k }" 
    using assms using le_Suc_ex by auto fastforce
 have 4: "{ (nth xs (k+i)) | i. i \<le> n-k } = { (nth xs i) | i. k \<le> i \<and> i \<le> n }" 
     using "2" "3" by blast
  show ?thesis by (simp add: "1" "4")
qed

lemma nth_set:
 assumes "k\<le> intlen xs"
 shows   " (nth xs k) \<in> set xs"
using assms 
by (meson interval_nth_and_set)



end
