(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Until and Since operator\<close>   

theory UntilSince
imports Semantics Fuse Theorems TimeReversal
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)

text \<open> 
This theory introduces the weak and strong versions of the until and since operators. 
The theorems from \cite{vegapaper2019} are proven in a mostly deductive style. 
\<close>

subsection \<open>Definitions\<close>

definition until_d :: " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
 where "until_d F G \<equiv> \<lambda>s. ( (\<exists> k \<le> intlen s. ((suffix k s ) \<Turnstile> G) \<and> 
                               (\<forall>j. j < k  \<longrightarrow>  ((suffix j s ) \<Turnstile> F) )) )"

definition suntil_d :: " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
 where "suntil_d F G \<equiv> \<lambda>s. ( intlen s > 0 \<and> (\<exists> k. 0<k \<and> k \<le> intlen s \<and> ((suffix k s ) \<Turnstile> G) \<and> 
                               (\<forall>j. 0<j \<and> j < k \<longrightarrow>  ((suffix j s ) \<Turnstile> F) )) )"

definition since_d :: " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
 where "since_d F G \<equiv> \<lambda>s. ( (\<exists> k \<le> intlen s. ( (prefix k s) \<Turnstile> G) \<and> 
                               (\<forall> j. k< j  \<and> j \<le> intlen s\<longrightarrow> ((prefix j s) \<Turnstile> F) )) )"

definition ssince_d :: " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
 where "ssince_d F G \<equiv> \<lambda>s. ( intlen s > 0 \<and> (\<exists> k .  k < intlen s \<and> ( (prefix k s) \<Turnstile> G) \<and> 
                               (\<forall> j. k< j \<and> j < intlen s \<longrightarrow> ((prefix j s) \<Turnstile> F) )) )"


syntax 
 "_until_d"  :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<U> _)" [84,84] 83)
 "_since_d"  :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<S> _)" [84,84] 83)
 "_suntil_d" :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<U>\<^sup>s _)" [84,84] 83)
 "_ssince_d" :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<S>\<^sup>s _)" [84,84] 83)
 


syntax (ASCII)
 "_until_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ until _)" [84,84] 83)
 "_since_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ since _)" [84,84] 83)
 "_suntil_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ suntil _)" [84,84] 83)
 "_ssince_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ ssince _)" [84,84] 83)



translations
 "_until_d"    \<rightleftharpoons> "CONST until_d"
 "_since_d"    \<rightleftharpoons> "CONST since_d"
 "_suntil_d"   \<rightleftharpoons> "CONST suntil_d"
 "_ssince_d"   \<rightleftharpoons> "CONST ssince_d"
 
definition wait_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "wait_d F G \<equiv> LIFT(\<box> F \<or> F \<U> G)"

definition pwait_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "pwait_d F G \<equiv> LIFT(bi F \<or> F \<S> G)"

definition release_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "release_d F G \<equiv> LIFT(\<not>((\<not> F) \<U> (\<not> G)))"

definition prelease_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "prelease_d F G \<equiv> LIFT(\<not>((\<not> F) \<S> (\<not> G)))"


syntax 
 "_wait_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<W> _)" [84,84] 83)
 "_pwait_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<P>\<W> _)" [84,84] 83)
 "_release_d"  :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<R> _)" [84,84] 83)
 "_prelease_d" :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<P>\<R> _)" [84,84] 83)


syntax (ASCII)
 "_wait_d"        :: "[lift,lift] \<Rightarrow> lift"        ("(_ wait _)" [84,84] 83)
 "_pwait_d"       :: "[lift,lift] \<Rightarrow> lift"        ("(_ pwait _)" [84,84] 83)
 "_release_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ release _)" [84,84] 83)
 "_prelease_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ prelease _)" [84,84] 83)


translations
 "_wait_d"       \<rightleftharpoons> "CONST wait_d" 
 "_pwait_d"      \<rightleftharpoons> "CONST pwait_d" 
 "_release_d"    \<rightleftharpoons> "CONST release_d" 
 "_prelease_d"   \<rightleftharpoons> "CONST prelease_d" 


definition srelease_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "srelease_d F G \<equiv> LIFT(\<not>((\<not> F) \<W> (\<not> G)))"

definition psrelease_d ::  " ('a :: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
  where "psrelease_d F G \<equiv> LIFT(\<not>((\<not> F) \<P>\<W> (\<not> G)))"

syntax 
 "_srelease_d"  :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<M> _)" [84,84] 83)
 "_psrelease_d" :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<P>\<M> _)" [84,84] 83)


syntax (ASCII)
 "_srelease_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ srelease _)" [84,84] 83)
 "_psrelease_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ psrelease _)" [84,84] 83)


translations
 "_srelease_d"    \<rightleftharpoons> "CONST srelease_d" 
 "_psrelease_d"   \<rightleftharpoons> "CONST psrelease_d" 


subsection \<open>Semantic Lemmas\<close>

lemma SUntilNextUntilsema:
 assumes "\<sigma> \<Turnstile>  f \<U>\<^sup>s g"
 shows   "\<sigma> \<Turnstile> \<circle> ( f \<U> g)"
proof -
 have 1: " \<sigma> \<Turnstile>  f \<U>\<^sup>s g "
    using assms by auto
 have 2: "0 < intlen \<sigma> \<and> 
          (\<exists>k>0. k \<le> intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> (\<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>)))"
   using "1" by (simp add: suntil_d_def)
 have 3: "(\<exists>k>0. k \<le> intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> (\<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>)))"
   using "2" by auto
 obtain k where 4: "(Suc k)>0 \<and> (Suc k) \<le> intlen \<sigma> \<and> g (suffix (Suc k) \<sigma>) \<and> 
      (\<forall>j. 0 < j \<and> j < (Suc k) \<longrightarrow> f (suffix j \<sigma>))"
  using "3" by (metis Suc_pred)
 have 5: "k\<le>intlen (suffix (Suc 0) \<sigma>)" 
   using "4" by auto
 have 6: "g (suffix (Suc k) \<sigma>)"
   using "4" by auto
 have 7: "(\<forall>j<k. f (suffix (Suc j) \<sigma>))"
   using "4" by blast  
 have 8: "(\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>))) " 
   using "4" "5" by blast
 have 9: "0 < intlen \<sigma> \<and>
          (\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))"
    using "2" "8" by blast 
 from 9 show ?thesis by (auto simp add: next_defs until_d_def) 
qed

lemma SUntilNextUntilsemb:
 assumes "\<sigma> \<Turnstile> \<circle> ( f \<U> g)"
 shows   "\<sigma> \<Turnstile>  f \<U>\<^sup>s g"
proof -
 have 1: "0 < intlen \<sigma> \<and> 
          (\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))"
   using assms by (auto simp add: next_defs until_d_def) 
 have 2: "(\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))" 
   using "1" by auto
 obtain k where 3: "k\<le>intlen (suffix (Suc 0) \<sigma>) \<and> g (suffix (Suc k) \<sigma>) \<and> 
         (\<forall>j<k. f (suffix (Suc j) \<sigma>)) "
   using "2" by auto
 have 4: "(Suc k)>0"
  by simp
 have 5: "g (suffix (Suc k) \<sigma>) " 
     using "3" by auto
 have 6: "(Suc k) \<le> intlen \<sigma>" 
   using "1" "3" by auto
 have 7: "(\<forall>j. 0 < j \<and> j < (Suc k) \<longrightarrow> f (suffix j \<sigma>))" 
   using "3" less_Suc_eq_0_disj by auto
 have 8: "(\<exists>k>0. k \<le> intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> (\<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>)))"   
   using "3" "6" "7" by blast
 have 9: "0 < intlen \<sigma> \<and> 
          (\<exists>k>0. k \<le> intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> (\<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>)))" 
   using "1" "8" by blast
 from 9 show ?thesis by (simp add: suntil_d_def)
qed

lemma SUntilNextUntilsem:
 "\<sigma> \<Turnstile>  f \<U>\<^sup>s g = \<circle> ( f \<U> g)"
using SUntilNextUntilsema SUntilNextUntilsemb unl_lift2 by blast


lemma SSincePrevSincesema:
 assumes " \<sigma> \<Turnstile> f \<S>\<^sup>s g "
 shows   " \<sigma> \<Turnstile> prev (f \<S> g)"
proof -
 have 1: "0 < intlen \<sigma> \<and> 
          (\<exists>k<intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>)))" 
    using assms by (simp add: ssince_d_def)
 have 2: "(\<exists>k<intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>)))"
   using "1" by auto
 obtain k where 3: "k<intlen \<sigma> \<and> g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>))" 
   using "2" by auto
  have 4: "k\<le>intlen \<sigma> - Suc 0"
    using "3" by linarith 
  have 5: "g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>))"
    by (simp add: "1" "3" interval_pref_pref_help) 
  have 6: "(\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>)))"
    using "3" interval_pref_pref_help by fastforce 
  have 7: "(\<exists>k\<le>intlen \<sigma> - Suc 0.
        g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>)) \<and> 
    (\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>))))" 
   using "4" "5" "6" by blast
  have 8: "0 < intlen \<sigma> \<and>
    (\<exists>k\<le>intlen \<sigma> - Suc 0.
        g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>)) \<and> 
    (\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>))))" 
   using "1" "7" by blast
  from 8 show ?thesis by (simp add: min.absorb1 prev_defs since_d_def)
qed


lemma SSincePrevSincesemb:
 assumes " \<sigma> \<Turnstile> prev (f \<S> g)  "
 shows   " \<sigma> \<Turnstile> f \<S>\<^sup>s g "
proof -
 have 1: "0 < intlen \<sigma> \<and>
    (\<exists>k\<le>intlen \<sigma> - Suc 0.
        g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>)) \<and> 
    (\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>))))" 
  using assms by (simp add: min.absorb1 prev_defs since_d_def)
 have 2: "(\<exists>k\<le>intlen \<sigma> - Suc 0.
        g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>)) \<and> 
    (\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>))))"
   using "1" by auto
 obtain k where 3: "k\<le>intlen \<sigma> - Suc 0 \<and> g (prefix k (prefix (intlen \<sigma> - Suc 0) \<sigma>)) \<and> 
    (\<forall>j. k < j \<and> j \<le> intlen \<sigma> - Suc 0 \<longrightarrow> f (prefix j (prefix (intlen \<sigma> - Suc 0) \<sigma>)))"  
   using "2" by auto
 have 4: "k<intlen \<sigma>"
    using "1" "3" by linarith 
 have 5: "g (prefix k \<sigma>)"
   using "3" "4" interval_pref_pref_help by force 
 have 6: "(\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>))" 
   using "3" interval_pref_pref_help by fastforce
 have 7: "(\<exists>k<intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>)))" 
   using "4" "5" "6" by blast
 have 8: "0 < intlen \<sigma> \<and> 
          (\<exists>k<intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>)))" 
  using "1" "7" by blast
 from 8 show ?thesis by (simp add: ssince_d_def)
qed

lemma SSincePrevSincesem:
 "\<sigma> \<Turnstile>   f \<S>\<^sup>s g = prev ( f \<S> g)"
using SSincePrevSincesema SSincePrevSincesemb unl_lift2 by blast

lemma UntilSUntilsem:
 "\<sigma> \<Turnstile> f \<U> g = (g \<or> (f \<and> f \<U>\<^sup>s g))"
proof (auto simp add: suntil_d_def until_d_def)
  show " \<And>k. k \<le> intlen \<sigma> \<Longrightarrow> g (suffix k \<sigma>) \<Longrightarrow> \<forall>j<k. f (suffix j \<sigma>) \<Longrightarrow> \<not> g \<sigma> \<Longrightarrow> f \<sigma>" 
   by fastforce
  show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow> g (suffix k \<sigma>) \<Longrightarrow> \<forall>j<k. f (suffix j \<sigma>) \<Longrightarrow> \<not> g \<sigma> \<Longrightarrow> 0 < intlen \<sigma>" 
   using gr0I by force
  show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow>
         g (suffix k \<sigma>) \<Longrightarrow>
         \<forall>j<k. f (suffix j \<sigma>) \<Longrightarrow> 
         \<not> g \<sigma> \<Longrightarrow>
          \<exists>k>0. k \<le> intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> (\<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>))" 
   by (metis interval_suffix_zero neq0_conv)
  show "\<And>k. f \<sigma> \<Longrightarrow>
         0 < k \<Longrightarrow>
         k \<le> intlen \<sigma> \<Longrightarrow>
         g (suffix k \<sigma>) \<Longrightarrow> 
         \<forall>j. 0 < j \<and> j < k \<longrightarrow> f (suffix j \<sigma>) \<Longrightarrow> 
          \<exists>k\<le>intlen \<sigma>. g (suffix k \<sigma>) \<and> (\<forall>j<k. f (suffix j \<sigma>)) " 
   by (metis Suc_pred interval_suffix_zero less_Suc_eq_0_disj)
qed

lemma SinceSSincesem:
 "\<sigma> \<Turnstile> f \<S> g = ( g \<or> (f \<and> f \<S>\<^sup>s g))"
proof (auto simp add: ssince_d_def since_d_def)
 show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow>
           g (prefix k \<sigma>) \<Longrightarrow> 
           \<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>) \<Longrightarrow> 
           \<not> g \<sigma> \<Longrightarrow> 
           f \<sigma>" 
  using le_eq_less_or_eq by auto
 show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow> 
           g (prefix k \<sigma>) \<Longrightarrow> 
           \<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>) \<Longrightarrow> 
           \<not> g \<sigma> \<Longrightarrow> 
           0 < intlen \<sigma>" 
  by (metis gr0I interval_prefix_intlen le_zero_eq)
 show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow>
          g (prefix k \<sigma>) \<Longrightarrow>
          \<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>) \<Longrightarrow>
          \<not> g \<sigma> \<Longrightarrow> 
          \<exists>k<intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>))" 
  using le_eq_less_or_eq by auto
 show "\<And>k. f \<sigma> \<Longrightarrow>
          k < intlen \<sigma> \<Longrightarrow>
          g (prefix k \<sigma>) \<Longrightarrow>
          \<forall>j. k < j \<and> j < intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>) \<Longrightarrow> 
           \<exists>k\<le>intlen \<sigma>. g (prefix k \<sigma>) \<and> (\<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (prefix j \<sigma>))" 
  by (metis antisym_conv2 interval_prefix_intlen less_imp_le_nat)
qed

lemma UntilAndDistsem:
 "\<sigma> \<Turnstile> (f \<and> g) \<U> h = ((f \<U> h) \<and> (g \<U> h))"
by (auto simp add: until_d_def) 
   (metis dual_order.strict_trans linorder_cases)

lemma UntilOrDistsem:
 " \<sigma> \<Turnstile> f \<U> (g \<or> h) = ( f \<U> g \<or> f \<U> h) " 
by (auto simp add: until_d_def) 


lemma NextUntilsema:
 assumes "(\<sigma> \<Turnstile> \<circle>(f \<U> g))"
 shows   "(\<sigma> \<Turnstile> (\<circle> f) \<U> (\<circle> g))"
proof -
 have 0: "0 < intlen \<sigma> \<and> 
         (\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))"
   using assms by  (auto simp add: next_defs until_d_def) 
 have 1: "0 < intlen \<sigma> "
    using "0" by auto
 have 2: "(\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))"
    using "0" by auto
 obtain k where 3: "k\<le>intlen (suffix (Suc 0) \<sigma>) \<and> g (suffix (Suc k) \<sigma>) \<and> 
      (\<forall>j<k. f (suffix (Suc j) \<sigma>))"
 using "2" by auto   
 have 4: "g (suffix (Suc k) \<sigma>)" 
   using "3" by auto
 have 5: " k\<le> intlen \<sigma>"
    using "3" "0" by auto 
 have 6: "0 < intlen (suffix k \<sigma>) " 
     using "1" "3" by auto
 have 7: "(\<forall>j<k. 0 < intlen (suffix j \<sigma>) \<and> f (suffix (Suc j) \<sigma>))" 
   using "3" "5" by force
 have 8: "\<exists>k\<le>intlen \<sigma>.
            0 < intlen (suffix k \<sigma>) \<and>
            g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. 0 < intlen (suffix j \<sigma>) \<and> f (suffix (Suc j) \<sigma>))"
   using "3" "5" "6" "7" by blast
 from 8 show ?thesis by (simp add: next_defs until_d_def)
qed  

lemma NextUntilsemb:
 assumes "(\<sigma> \<Turnstile> (\<circle> f) \<U> (\<circle> g))"
 shows   "(\<sigma> \<Turnstile> \<circle>(f \<U> g))"
proof -
 have 1: "\<exists>k\<le>intlen \<sigma>.
            0 < intlen (suffix k \<sigma>) \<and>
            g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. 0 < intlen (suffix j \<sigma>) \<and> f (suffix (Suc j) \<sigma>))" 
   using assms by (auto simp add: next_defs until_d_def)
 obtain k where 2: "k\<le>intlen \<sigma> \<and> 0 < intlen (suffix k \<sigma>) \<and>
            g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. 0 < intlen (suffix j \<sigma>) \<and> f (suffix (Suc j) \<sigma>))"
   using 1 by auto
 have 3: "0 < intlen \<sigma>  "  
   using "2" by auto
 have 4: "k\<le>intlen (suffix (Suc 0) \<sigma>)" 
    using "2" interval_suffix_length_good by auto
 have 5: "g (suffix (Suc k) \<sigma>)" 
    using "2" by auto
 have 6: "(\<forall>j<k. f (suffix (Suc j) \<sigma>))" 
    using "2" by blast
 have 7: "0 < intlen \<sigma> \<and> 
          (\<exists>k\<le>intlen (suffix (Suc 0) \<sigma>). g (suffix (Suc k) \<sigma>) \<and> (\<forall>j<k. f (suffix (Suc j) \<sigma>)))" 
  using "2" "3" "4" by blast
 from 7 show ?thesis by (auto simp: next_defs until_d_def)
qed

lemma NextUntilsem:
 "\<sigma> \<Turnstile> \<circle>(f \<U> g) = (\<circle> f) \<U> (\<circle> g)"
using NextUntilsema NextUntilsemb using unl_lift2 by blast

lemma UntilUntilsem:
 "\<sigma> \<Turnstile> f \<U> g = f \<U> (f \<U> g) "
proof (auto simp add: until_d_def)
 show "\<And>k. k \<le> intlen \<sigma> \<Longrightarrow>
         g (suffix k \<sigma>) \<Longrightarrow>
         \<forall>j<k. f (suffix j \<sigma>) \<Longrightarrow>
         \<exists>k\<le>intlen \<sigma>. 
          (\<exists>ka\<le>intlen \<sigma> - k. 
            g (suffix (ka + k) \<sigma>) \<and> (\<forall>j<ka. f (suffix (j + k) \<sigma>))) \<and> (\<forall>j<k. f (suffix j \<sigma>))" 
  by force
 show "\<And>k ka.
        k \<le> intlen \<sigma> \<Longrightarrow>
        \<forall>j<k. f (suffix j \<sigma>) \<Longrightarrow>
        ka \<le> intlen \<sigma> - k \<Longrightarrow>
        g (suffix (ka + k) \<sigma>) \<Longrightarrow> 
        \<forall>j<ka. f (suffix (j + k) \<sigma>) \<Longrightarrow> \<exists>k\<le>intlen \<sigma>. g (suffix k \<sigma>) \<and> (\<forall>j<k. f (suffix j \<sigma>)) " 
  by (metis leI le_add_diff_inverse2 less_diff_conv2 
      ordered_cancel_comm_monoid_diff_class.le_diff_conv2)
qed


lemma LFPUntilsem1:
assumes "\<forall>n\<le>intlen \<sigma>.
            (g (suffix n \<sigma>) \<longrightarrow> h (suffix n \<sigma>)) \<and>
            (f (suffix n \<sigma>) \<and> n < intlen \<sigma> \<and> h (suffix (Suc n) \<sigma>) \<longrightarrow>
             h (suffix n \<sigma>))"
         "k \<le> intlen \<sigma>"
         "g (suffix k \<sigma>) " 
         "\<forall>j<k. f (suffix j \<sigma>)"
shows     " h \<sigma>"
using assms
proof (induct k arbitrary: \<sigma>)
case 0
then show ?case by auto
next
case (Suc k)
then show ?case 
   proof (cases \<sigma>)
   case (St x1)
   then show ?thesis using Suc.prems Suc.hyps by (metis suffix.simps(1)) 
   next
   case (Cons x21 x22)
   then show ?thesis 
         using Suc.prems Suc.hyps 
         by (metis  Suc_less_eq interval_suffix_suc interval_suffix_zero 
             intlen.simps(2) le_eq_less_or_eq not_less_eq_eq plus_1_eq_Suc zero_less_Suc) 
   qed
qed

lemma LFPUntilsem:
  "\<sigma> \<Turnstile> \<box>((g \<or> (f \<and> \<circle>h)) \<longrightarrow> h) \<longrightarrow> (f \<U> g \<longrightarrow> h)"
using LFPUntilsem1 by (simp add: always_defs next_defs until_d_def, blast)


lemma RevUntilsema:
assumes "\<sigma> \<Turnstile> (f \<U> g)\<^sup>r" 
shows   "\<sigma> \<Turnstile> ( (f\<^sup>r) \<S> (g\<^sup>r) )"
proof -
 have 1: "\<sigma> \<Turnstile> (f \<U> g)\<^sup>r" 
  using assms by auto
 have 2: "\<exists>k\<le>intlen \<sigma>.
       g (suffix k (intrev \<sigma>)) \<and> (\<forall>j<k. f (suffix j (intrev \<sigma>)))"
   using "1" by (simp add: until_d_def reverse_d_def)
 obtain k where 3: "k\<le>intlen \<sigma> \<and> g (suffix k (intrev \<sigma>)) \<and> 
     (\<forall>j<k. f (suffix j (intrev \<sigma>)))" 
  using "2" by auto
 have 4: "g (intrev (prefix (intlen \<sigma> -k) \<sigma>)) "
    by (simp add: "3" interval_intrev_prefix) 
 have 5: "(\<forall>j<k. f (intrev (prefix (intlen \<sigma> -j) \<sigma>)))"
    by (simp add: "3" interval_intrev_prefix)
 have 6: " (intlen \<sigma> -k) \<le> intlen \<sigma>" 
   using diff_le_self by blast
 have 7: "(\<forall>j. (intlen \<sigma> -k) < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (intrev (prefix j \<sigma>)))"
   by (simp add: "3" interval_intrev_prefix less_diff_conv2) 
 have 8: "\<exists>k\<le>intlen \<sigma>. g (intrev (prefix k \<sigma>)) \<and> 
          (\<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (intrev (prefix j \<sigma>)))"
  using "4" "6" "7" by blast
 have 10: "\<sigma> \<Turnstile> ( (f\<^sup>r) \<S> (g\<^sup>r) ) "
   using "8" by (simp add: since_d_def reverse_d_def) 
 show ?thesis by (simp add: "10")
qed

lemma RevUntilsemb:
assumes "\<sigma> \<Turnstile> ( (f\<^sup>r) \<S> (g\<^sup>r) )" 
shows   "\<sigma> \<Turnstile> (f \<U> g)\<^sup>r"
proof -
 have 1: "\<sigma> \<Turnstile> ( (f\<^sup>r) \<S> (g\<^sup>r) )" 
  using assms by auto
 have 2: "\<exists>k\<le>intlen \<sigma>. g (intrev (prefix k \<sigma>)) \<and> 
          (\<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (intrev (prefix j \<sigma>)))"
  using "1" by (simp add: since_d_def reverse_d_def)
 obtain k where 3: "k\<le>intlen \<sigma> \<and> g (intrev (prefix k \<sigma>)) \<and> 
          (\<forall>j. k < j \<and> j \<le> intlen \<sigma> \<longrightarrow> f (intrev (prefix j \<sigma>)))"
   using "2" by auto
 have 4: "g (suffix (intlen \<sigma> -k) (intrev \<sigma>))"
    using "3" interval_intrev_prefix by fastforce
 have 5: "(\<forall>j. j< intlen \<sigma> -k \<longrightarrow> f (suffix j (intrev \<sigma>)))"
    by (metis "3" add.commute diff_le_self dual_order.trans interval_intrev_intlen
         interval_intrev_suffix interval_rev_rev_ident less_diff_conv less_imp_le_nat)
 have 6: "\<exists>k\<le>intlen \<sigma>. g (suffix k (intrev \<sigma>)) \<and> 
          (\<forall>j<k. f (suffix j (intrev \<sigma>)))" 
  using "4" "5" diff_le_self by blast
 have 7: "\<sigma> \<Turnstile> (f \<U> g)\<^sup>r"   
   using "6" by (simp add: until_d_def reverse_d_def)
 show ?thesis 
 using "7" by blast 
qed
 
lemma RevUntilsem:
 " \<sigma> \<Turnstile> (f \<U> g)\<^sup>r = ( (f\<^sup>r) \<S> (g\<^sup>r) )"
using RevUntilsema RevUntilsemb using unl_lift2 by blast

lemma UntilRightAndsem:
assumes "(\<sigma> \<Turnstile> f \<U> (g \<and> h))"
shows   "(\<sigma> \<Turnstile> (f \<U> g) \<U> h)"
proof -
 have 1: "\<exists>k\<le>intlen \<sigma>. g (suffix k \<sigma>) \<and> h (suffix k \<sigma>) \<and> (\<forall>j<k. f (suffix j \<sigma>))"
   using assms by (simp add: until_d_def) 
 obtain k where 2: "k\<le>intlen \<sigma> \<and> g (suffix k \<sigma>) \<and> h (suffix k \<sigma>) \<and> (\<forall>j<k. f (suffix j \<sigma>))"
   using "1" by auto
 have 3: "h (suffix k \<sigma>)"
  using "2" by auto
 have 4: "k\<le>intlen \<sigma>"  
  using "2" by auto
 have 5: "(\<forall>j<k. 
           \<exists>ka\<le>intlen (suffix j \<sigma>). g (suffix (ka + j) \<sigma>) \<and> (\<forall>ja<ka. f (suffix (ja + j) \<sigma>)))"
    proof 
      fix j
      show "j < k \<longrightarrow> 
            (\<exists>ka\<le>intlen (suffix j \<sigma>). g (suffix (ka + j) \<sigma>) \<and> (\<forall>ja<ka. f (suffix (ja + j) \<sigma>))) "
       proof 
         assume a0: "j < k"
         show "(\<exists>ka\<le>intlen (suffix j \<sigma>). g (suffix (ka + j) \<sigma>) \<and> (\<forall>ja<ka. f (suffix (ja + j) \<sigma>)))"  
          proof -
           have 51: "k-j \<le> intlen (suffix j \<sigma>)"
                using "4" a0 by auto
           have 52: " g (suffix ((k-j) + j) \<sigma>)" 
             by (simp add: "2" a0 less_imp_le_nat)
           have 53: "(\<forall>ja<(k-j). f (suffix (ja + j) \<sigma>))" 
             using "2" less_diff_conv by blast
          show ?thesis 
          using "51" "52" "53" by blast
        qed
      qed
     qed
 have 6: "\<exists>k\<le>intlen \<sigma>.
       h (suffix k \<sigma>) \<and>
       (\<forall>j<k. \<exists>k\<le>intlen (suffix j \<sigma>). g (suffix (k + j) \<sigma>) \<and> (\<forall>ja<k. f (suffix (ja + j) \<sigma>)))" 
  using "2" "5" by blast
 from 6 show ?thesis by (simp add: until_d_def)
qed

lemma interval_suf_first:
assumes "(\<exists>i\<le>intlen xs. f (suffix i xs))"
shows   "(\<exists>i\<le>intlen xs. f (suffix i xs) \<and>
           (\<forall>j. j<i \<longrightarrow> \<not> (f (suffix j xs))))"
using assms interval_suf_first_upto[of "intlen xs +1" "f" "xs" ] by (simp add: discrete) 

lemma NotSuffixFirst:
assumes "(\<exists>n\<le>intlen xs. \<not> f (suffix n xs))"
shows   "(\<exists>n \<le> intlen xs. \<not> f (suffix n xs) \<and> (\<forall>k. k<n \<longrightarrow> f (suffix k xs)))"
using assms interval_suf_first[of "xs" "LIFT(\<not> f)"] by auto  

lemma NotSuffixFirst_upto:
assumes "(\<exists>i<k. \<not> f (suffix i xs))"
        "k \<le> intlen xs +1"
shows   "(\<exists>i<k. \<not>f (suffix i xs) \<and>
           (\<forall>j<i .  (f (suffix j xs))))"
using assms interval_suf_first_upto[of "k" "LIFT(\<not> f)" "xs" ]  by auto


lemma WaitNotDistUntilsem1:
assumes "(\<sigma> \<Turnstile> \<not>(f \<W> g))"
shows   "(\<sigma> \<Turnstile> ((\<not> g) \<U> ((\<not> f) \<and> (\<not>g))))"
proof -
 have 1: "(\<forall>k. g (suffix k \<sigma>) \<longrightarrow> k \<le> intlen \<sigma> \<longrightarrow> (\<exists>j<k. \<not> f (suffix j \<sigma>))) \<and>
    (\<exists>n\<le>intlen \<sigma>. \<not> f (suffix n \<sigma>))"
    using assms by (simp add: wait_d_def until_d_def always_defs)
 have 2: "(\<forall>k. k \<le> intlen \<sigma> \<longrightarrow> \<not> g (suffix k \<sigma>) \<or> (\<exists>j<k. \<not> f (suffix j \<sigma>)))"
   using "1" by auto
 have 3: "(\<exists>n\<le>intlen \<sigma>. \<not> f (suffix n \<sigma>))"  
    using "1" by auto
 obtain n where 4: "n\<le> intlen \<sigma> \<and> \<not> f (suffix n \<sigma>) \<and>
                    (\<forall> k <n . f (suffix k \<sigma>))"
   using "3" using NotSuffixFirst by blast
 have 16: "n \<le> intlen \<sigma>"
   by (simp add: "4") 
 have 17: "\<not> g (suffix n \<sigma>)" 
   using "1" "4" by blast
 have 18: "(\<forall>j<n. \<not> g (suffix j \<sigma>))" 
  by (meson "2" "4" le_eq_less_or_eq less_le_trans)
 have 19: "\<exists>k\<le>intlen \<sigma>. \<not> f (suffix k \<sigma>) \<and> \<not> g (suffix k \<sigma>) \<and> (\<forall>j<k. \<not> g (suffix j \<sigma>))"
  using 16 17 18 4 by blast
 have 20: "(\<sigma> \<Turnstile> ((\<not> g) \<U> (\<not> f \<and> \<not>g)))" 
   using "19" by (simp add: until_d_def)
 show ?thesis using "20" by auto 
qed

lemma WaitNotDistUntilsem2:
assumes "(\<sigma> \<Turnstile> ((\<not> g) \<U> ((\<not> f) \<and> (\<not>g))))"
shows   "(\<sigma> \<Turnstile> \<not>(f \<W> g))"
using assms not_less_iff_gr_or_eq by (auto simp add: always_defs wait_d_def until_d_def)

lemma WaitNotDistUntilsem:
 "(\<sigma> \<Turnstile> (\<not>(f \<W> g))  =  ((\<not> g) \<U> ((\<not> f) \<and> (\<not>g)))) " 
using WaitNotDistUntilsem1 WaitNotDistUntilsem2 unl_lift2 by blast




lemma SUntilNextUntil:
 "\<turnstile> f \<U>\<^sup>s g = \<circle> ( f \<U> g)"
using SUntilNextUntilsem Valid_def by blast

lemma SSincePrevSince:
 "\<turnstile> f \<S>\<^sup>s g = prev ( f \<S> g)"
 using SSincePrevSincesem Valid_def by blast

lemma UntilSUntil:
 "\<turnstile> f \<U> g = (g \<or> (f \<and> f \<U>\<^sup>s g))"
using UntilSUntilsem Valid_def by blast

lemma SinceSSince:
 "\<turnstile> f \<S> g = (g \<or> (f \<and> f \<S>\<^sup>s g))"
 using SinceSSincesem Valid_def by blast

lemma UntilAndDist:
 "\<turnstile> (f \<and> g) \<U> h = ((f \<U> h) \<and> (g \<U> h))"
using UntilAndDistsem Valid_def by blast

lemma UntilOrDist:
 "\<turnstile> f \<U> (g \<or> h) = ( f \<U> g \<or> f \<U> h)"
using UntilOrDistsem Valid_def by blast

lemma NextUntil:
 "\<turnstile> \<circle>(f \<U> g) = (\<circle> f) \<U> (\<circle> g)"
using NextUntilsem Valid_def by blast

lemma UntilUntil: 
"\<turnstile> f \<U> g = f \<U> (f \<U> g)"
using UntilUntilsem Valid_def by blast

lemma LFPUntil:
  "\<turnstile> \<box>((g \<or> (f \<and> \<circle>h)) \<longrightarrow> h) \<longrightarrow> (f \<U> g \<longrightarrow> h)"
using LFPUntilsem Valid_def by blast

lemma RevUntil:
 " \<turnstile> (f \<U> g)\<^sup>r = ( (f\<^sup>r) \<S> (g\<^sup>r) )"
using RevUntilsem Valid_def by blast

lemma UntilEqvUntil:
 assumes "\<turnstile> f0 = f1"
         "\<turnstile> g0 = g1"
 shows   "\<turnstile> f0 \<U> g0 = f1 \<U> g1"
using assms by (simp add: until_d_def Valid_def)

lemma UntilImpUntil:
 assumes "\<turnstile> f0 \<longrightarrow> f1"
         "\<turnstile> g0 \<longrightarrow> g1"
 shows   "\<turnstile> f0 \<U> g0 \<longrightarrow> f1 \<U> g1"
using assms by (auto simp add: until_d_def Valid_def)


lemma SinceEqvSince:
 assumes "\<turnstile> f0 = f1"
         "\<turnstile> g0 = g1"
 shows   "\<turnstile> f0 \<S> g0 = f1 \<S> g1"
using assms by (simp add: since_d_def Valid_def)

lemma SinceImpSince:
 assumes "\<turnstile> f0 \<longrightarrow> f1"
         "\<turnstile> g0 \<longrightarrow> g1"
 shows   "\<turnstile> f0 \<S> g0 \<longrightarrow> f1 \<S> g1"
using assms by (auto simp add: since_d_def Valid_def)

lemma UntilLeftDistAnd:
 "\<turnstile> f \<U> (g \<and> h) \<longrightarrow> f \<U> g \<and> f \<U> h"
by (auto simp add: Valid_def until_d_def)

lemma UntilRightDistOr:
 "\<turnstile> f \<U> h \<or> g \<U> h \<longrightarrow> (f \<or> g) \<U> h"
by (auto simp add: Valid_def until_d_def)

lemma UntilNotImp:
 "\<turnstile> f \<U> g \<and> (\<not> g) \<U> h \<longrightarrow> f \<U> h"
by (auto simp add: Valid_def until_d_def)
   (metis less_trans linorder_cases)

lemma UntilRightOr:
 "\<turnstile> f \<U> ( g \<U> h ) \<longrightarrow> (f \<or> g) \<U> h"
by (auto simp add: Valid_def until_d_def)
   (metis Nat.le_diff_conv2 leI le_add_diff_inverse2 less_diff_conv2) 

lemma DiamondEqvTrueUntil:
 "\<turnstile> \<diamond> f = #True \<U> f" 
by (simp add: Valid_def sometimes_defs until_d_def)

lemma UntilRightAnd:
 "\<turnstile> f \<U> (g \<and> h) \<longrightarrow> (f \<U> g) \<U> h"
using UntilRightAndsem Valid_def by auto

lemma WaitNotDistUntil:
 "\<turnstile> (\<not>(f \<W> g)) = ((\<not> g) \<U> (\<not> f \<and> \<not>g))"
using WaitNotDistUntilsem Valid_def by blast

lemma UntilAlwaysAndDist:
 "\<turnstile> \<box> f \<and> g \<U> h \<longrightarrow> (f \<and> g) \<U> (f \<and> h)"
by (auto simp add: Valid_def always_defs until_d_def)

lemma UntilRightMono:
 "\<turnstile> \<box>(f \<longrightarrow> g) \<longrightarrow> (h \<U> f \<longrightarrow> h \<U> g)"
by (auto simp add: Valid_def always_defs until_d_def)

lemma UntilLeftMono:
 "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (f \<U> h \<longrightarrow> g \<U> h)"
by (auto simp add: Valid_def always_defs until_d_def)


lemma UntilInduction_help:
 "\<turnstile> \<box> (f \<longrightarrow> \<not> g \<and> \<circle> f) \<longrightarrow> \<box>(g \<or> #True \<and> \<circle> (\<not> f) \<longrightarrow> \<not> f)" 
 by (auto simp add: Valid_def always_defs next_defs)

lemma UntilInduction_a_help:
 "\<turnstile> (f \<longrightarrow> ((\<circle> f) \<and> g) \<or> h) \<longrightarrow> ( ((\<not>g \<and> \<not>h) \<or> (\<not>h \<and> \<circle>(\<not> f)))\<longrightarrow> \<not>f)" 
by (auto simp add: Valid_def next_defs)

lemma UntilInduction_b_help:
  "\<turnstile> \<box>(f \<longrightarrow> (\<circle>f) \<or> g) \<longrightarrow>   \<box>( ((\<not> f \<and> \<not> g) \<or> (\<not>g \<and> \<circle> (\<not> f))) \<longrightarrow> \<not> f)" 
by (auto simp add: Valid_def always_defs next_defs)

lemma UntilImpNot:
 "\<turnstile> f \<U> g \<longrightarrow> (f \<and> \<not>g) \<U> g"
by (auto simp add: until_d_def Valid_def)
   (metis (full_types) interval_suf_first leI less_le_trans)

lemma WaitAndRule:
 "\<turnstile> f \<W> g = (f \<and> \<not> g) \<W> g"
proof (auto simp add: Valid_def wait_d_def until_d_def always_defs )
 show "\<And>w n.
         \<forall>k. g (suffix k w) \<longrightarrow> k \<le> intlen w \<longrightarrow> (\<exists>j<k. f (suffix j w) \<longrightarrow> g (suffix j w)) \<Longrightarrow>
           n \<le> intlen w \<Longrightarrow> 
           \<forall>n\<le>intlen w. f (suffix n w) \<Longrightarrow> 
           g (suffix n w) \<Longrightarrow> 
           False" 
  by (metis (full_types) dual_order.order_iff_strict interval_suf_first less_le_trans)
 show "\<And>w n k.
       \<forall>k. g (suffix k w) \<longrightarrow> k \<le> intlen w \<longrightarrow> (\<exists>j<k. f (suffix j w) \<longrightarrow> g (suffix j w)) \<Longrightarrow>
       n \<le> intlen w \<Longrightarrow> 
       k \<le> intlen w \<Longrightarrow> 
       g (suffix k w) \<Longrightarrow>
        \<forall>j<k. f (suffix j w) \<Longrightarrow>
        f (suffix n w)"
  by (metis (full_types) interval_suf_first leI less_le_trans) 
 show "\<And>w n k.
       \<forall>k. g (suffix k w) \<longrightarrow> k \<le> intlen w \<longrightarrow> (\<exists>j<k. f (suffix j w) \<longrightarrow> g (suffix j w)) \<Longrightarrow>
       n \<le> intlen w \<Longrightarrow> 
       k \<le> intlen w \<Longrightarrow> 
       g (suffix k w) \<Longrightarrow> 
       \<forall>j<k. f (suffix j w) \<Longrightarrow>  
       g (suffix n w) \<Longrightarrow> 
       False "
    by (metis (full_types) interval_suf_first less_le_trans not_le_imp_less) 
qed

lemma WaitLeftDistAnd:
 "\<turnstile> f \<W> (g \<and> h) \<longrightarrow> f \<W> g \<and> f \<W> h"
by (auto simp add: Valid_def wait_d_def until_d_def always_defs)

lemma WaitRightDistAnd:
 "\<turnstile> (f \<and> g) \<W> h = (f \<W> h \<and> g \<W> h)"
proof (auto simp add: Valid_def wait_d_def until_d_def always_defs)
 show " \<And>w n k ka.
       \<forall>k. h (suffix k w) \<longrightarrow> k \<le> intlen w \<longrightarrow> (\<exists>j<k. f (suffix j w) \<longrightarrow> \<not> g (suffix j w)) \<Longrightarrow>
       n \<le> intlen w \<Longrightarrow>
       k \<le> intlen w \<Longrightarrow>
       h (suffix k w) \<Longrightarrow> 
      \<forall>j<k. f (suffix j w) \<Longrightarrow> 
        ka \<le> intlen w \<Longrightarrow> 
        h (suffix ka w) \<Longrightarrow> 
        \<forall>j<ka. g (suffix j w) \<Longrightarrow> 
        f (suffix n w)" 
  by (metis less_trans linorder_cases)
 show "\<And>w n k ka.
       \<forall>k. h (suffix k w) \<longrightarrow> k \<le> intlen w \<longrightarrow> (\<exists>j<k. f (suffix j w) \<longrightarrow> \<not> g (suffix j w)) \<Longrightarrow>
       n \<le> intlen w \<Longrightarrow>
       k \<le> intlen w \<Longrightarrow>
       h (suffix k w) \<Longrightarrow> 
       \<forall>j<k. f (suffix j w) \<Longrightarrow>
          ka \<le> intlen w \<Longrightarrow> 
          h (suffix ka w) \<Longrightarrow>
          \<forall>j<ka. g (suffix j w) \<Longrightarrow> 
       g (suffix n w) " 
  by (metis less_trans linorder_cases)
qed

lemma WaitRightDistImp:
 "\<turnstile> (f \<longrightarrow> g) \<W> h \<longrightarrow> (f \<W> h \<longrightarrow> g \<W> h)"
by (auto simp add: Valid_def wait_d_def until_d_def always_defs)
   (metis less_trans linorder_cases)

lemma WaitImpRule:
 "\<turnstile> (f \<longrightarrow> g) \<W> f"
by (auto simp add: Valid_def wait_d_def until_d_def always_defs)
   (metis interval_suf_first)

lemma WaitInductionc_help:
  "\<turnstile> \<box>(f \<longrightarrow> \<circle> f) \<longrightarrow> \<box> (f \<longrightarrow> wnext f)"
 by (simp add: ImpBoxRule intI next_defs wnext_defs)

lemma WaitInductiond_help:
 "\<turnstile> \<box>(f \<longrightarrow> g \<and> \<circle> f) \<longrightarrow> \<box> (f \<longrightarrow>  wnext f)"
  by (simp add: Valid_def always_defs next_defs wnext_defs)

lemma WaitOrder_help: 
  "\<turnstile> (\<box> (\<not> f) \<or> \<box> (\<not> g)) \<or> \<diamond>(f \<or> g) "
by (auto simp add: always_defs sometimes_defs Valid_def) 

subsection \<open>Lemmas\<close>

lemma NextFalseSUntil:
 "\<turnstile> \<circle> g = #False \<U>\<^sup>s g"
 by (metis SUntilNextUntil UntilSUntil int_simps(19) int_simps(25) inteq_reflection)

lemma PrevFalseSSince:
 "\<turnstile> prev g = #False \<S>\<^sup>s g"
 by (metis SSincePrevSince SinceSSince int_simps(19) int_simps(25) inteq_reflection)

lemma UntilUnrol:
 "\<turnstile> f \<U> g = (g \<or> (f \<and> \<circle>(f \<U> g)))"
by (metis SUntilNextUntil UntilSUntil inteq_reflection)

lemma WNextUntil:
 "\<turnstile> wnext(f \<U> g) = (empty \<or> (\<circle> f) \<U> (\<circle> g))"
by (meson NextUntil Prop06 WnextEqvEmptyOrNext)

lemma UntilRelease:
 "\<turnstile> f \<R> g = (\<not> ( (\<not> f) \<U> (\<not> g)))"
by (simp add: release_d_def)

lemma SReleaseWait:
 "\<turnstile> f \<M> g = (\<not> (\<not> f) \<W> (\<not> g))" 
by (simp add: srelease_d_def)

lemma ReleaseUntil:
 "\<turnstile> f \<U> g = (\<not> ( (\<not> f) \<R> (\<not> g)))"
by (simp add: release_d_def)

lemma WaitSRelease:
 "\<turnstile> f \<W> g = (\<not> (\<not> f) \<M> (\<not> g))"
by (simp add: srelease_d_def)

lemma NotUntilRelease:
 "\<turnstile> \<not>(f \<U> g) = (\<not> f) \<R> (\<not> g)"
by (simp add: ReleaseUntil)

lemma NotWaitSRelease:
 "\<turnstile> \<not>(f \<W> g) = (\<not> f) \<M> (\<not> g)"
by (simp add: WaitSRelease)

lemma NotReleaseUntil:
 "\<turnstile> \<not>(f \<R> g) = (\<not> f) \<U> (\<not> g)"
by (simp add: UntilRelease)

lemma NotSReleaseWait:
 "\<turnstile> \<not>(f \<M> g) = (\<not> f) \<W> (\<not> g)"
by (simp add: SReleaseWait)

lemma BoxEqvFalseRelease:
 "\<turnstile> \<box> f = #False \<R> f"
 by (metis DiamondEqvTrueUntil EqvReverseReverse always_d_def int_simps(3) inteq_reflection 
     release_d_def)

lemma RevSince:
 "\<turnstile> (f \<S> g)\<^sup>r = ( (f\<^sup>r) \<U> (g\<^sup>r) )"
proof -
 have 1: "\<turnstile> (f\<^sup>r \<U> g\<^sup>r)\<^sup>r = ( (f\<^sup>r)\<^sup>r \<S> (g\<^sup>r)\<^sup>r )" 
   by (simp add: RevUntil)
 show ?thesis 
 by (metis "1" EqvReverseReverse inteq_reflection)
qed    

lemma LFPSince:
 "\<turnstile> bi ((g \<or> (f \<and> prev h)) \<longrightarrow> h) \<longrightarrow> (f \<S> g \<longrightarrow> h)"
by (metis (no_types, lifting) LFPUntil RBiEqvBox RPrevEqvNext RevSince ReverseEqv 
    all_rev_eq(3) inteq_reflection)

lemma UntilTrue:
 " \<turnstile> f \<U> #True"
using UntilSUntil by fastforce

lemma NotUntilFalse:
 "\<turnstile> \<not> (f \<U> #False)"
by (simp add: intI until_d_def)

lemma UntilIdempotent:
 "\<turnstile> f \<U> f = f"
using UntilSUntil by fastforce

lemma UntilRightDistImp:
 "\<turnstile> (f \<longrightarrow> g) \<U> h \<longrightarrow> (f \<U> h \<longrightarrow> g \<U> h)"
proof -
 have 1: "\<turnstile> (f \<longrightarrow> g) \<U> h \<longrightarrow> (f \<U> h \<longrightarrow> g \<U> h) =
            ((f \<longrightarrow> g) \<U> h \<and> f \<U> h \<longrightarrow> g \<U> h)" 
      by auto
 have 2: "\<turnstile> ((f \<longrightarrow> g) \<U> h \<and> f \<U> h) = ((f \<longrightarrow> g) \<and> f) \<U> h "
   by (simp add: UntilAndDist int_iffD1 int_iffD2 int_iffI)
 have 3: "\<turnstile> ((f \<longrightarrow> g) \<and> f) = (f \<and> g)"
   by auto
 have 4: "\<turnstile> h = h"
   by auto 
 have 5: "\<turnstile> ((f \<longrightarrow> g) \<and> f) \<U> h = (f \<and> g) \<U> h"
  using "3" "4" using UntilEqvUntil by blast
 have 6: "\<turnstile> (f \<and> g) \<U> h = (f \<U> h \<and> g \<U> h)"
   by (simp add: UntilAndDist) 
 show ?thesis 
 using "2" "5" "6" by fastforce
qed    

lemma FalseUntil:
 "\<turnstile> #False \<U> g = g"
 by (metis  Prop10 Prop12 TrueW UntilUnrol int_simps(14) int_simps(21) int_simps(25) int_simps(3)
      inteq_reflection)

lemma UntilExclMid:
 "\<turnstile> f \<U> g \<or> f \<U> (\<not> g)"
 using UntilOrDist UntilTrue by fastforce

lemma NotUntilImp:
 "\<turnstile> (\<not> f) \<U> (g \<U> h) \<and> f \<U> h \<longrightarrow> g \<U> h"
proof -
 have 1: "\<turnstile> (\<not> f) \<U> (g \<U> h) \<longrightarrow> (\<not> f \<or> g) \<U> h"
   by (simp add: UntilRightOr)
 have 2: "\<turnstile> (\<not> f \<or> g) = (f \<longrightarrow> g)" 
  by auto
 have 3: "\<turnstile> h = h"
  by auto
 have 4: "\<turnstile>  (\<not> f \<or> g) \<U> h = (f \<longrightarrow> g) \<U> h"
   by (simp add: "2" UntilEqvUntil)
 have 5: "\<turnstile> (f \<longrightarrow> g) \<U> h \<longrightarrow> (f \<U> h \<longrightarrow> g \<U> h)"
   by (simp add: UntilRightDistImp) 
 have 6: "\<turnstile> (\<not> f) \<U> (g \<U> h) \<longrightarrow> (f \<U> h \<longrightarrow> g \<U> h)"
   using "1" "4" "5" by fastforce
 from 6 show ?thesis by auto
qed

lemma UntilNotImpa:
 "\<turnstile> f \<U> ((\<not> g) \<U> h) \<and> g \<U> h \<longrightarrow> f \<U> h"
proof -
 have 1: "\<turnstile> f \<U> ((\<not> g) \<U> h) \<longrightarrow> (f \<or> (\<not> g)) \<U> h "
   by (simp add: UntilRightOr) 
 have 2: "\<turnstile> (f \<or> (\<not> g)) = (g \<longrightarrow> f)"
   by auto
 have 3: "\<turnstile> h = h"
   by auto
 have 4: "\<turnstile> (f \<or> (\<not> g)) \<U> h = (g \<longrightarrow> f) \<U> h"
   by (simp add: "2" UntilEqvUntil) 
 have 5: "\<turnstile> (g \<longrightarrow> f) \<U> h \<longrightarrow> ( g \<U> h \<longrightarrow> f \<U> h)"
   by (simp add: UntilRightDistImp) 
 have 6: "\<turnstile>  f \<U> ((\<not> g) \<U> h) \<longrightarrow> ( g \<U> h \<longrightarrow> f \<U> h)"
  using "1" "4" "5" by fastforce
 from 6 show ?thesis by auto
qed

lemma UntilNotUntilImp:
 "\<turnstile> f \<U> g \<and> (\<not> g) \<U> f \<longrightarrow> f"
proof -
 have 1: "\<turnstile> f \<U> g \<and> (\<not> g) \<U> f \<longrightarrow> f \<U> f"
   using UntilNotImp by auto
 have 2: "\<turnstile>  f \<U> f = f "
  using UntilIdempotent by auto
 from 1 2 show ?thesis by fastforce
qed

lemma AndNotUntilImp:
 "\<turnstile> f \<and> (\<not> f) \<U> g \<longrightarrow> g"
proof -
 have 1: "\<turnstile> f = f \<U> f"
    by (simp add: UntilIdempotent int_iffD1 int_iffD2 int_iffI)
 have 2: "\<turnstile> g =  #False \<U> g"
   by (meson FalseUntil Prop11)
 have 3: "\<turnstile> f \<U> f \<and> (\<not> f) \<U> g \<longrightarrow> #False \<U> g"
  by (metis "1" FalseUntil UntilNotImp inteq_reflection)
 from 1 2 3 show ?thesis by fastforce
qed

lemma UntilImpOr:
 "\<turnstile> f \<U> g \<longrightarrow> f \<or> g"
 proof -
 have "\<turnstile> f \<and> f \<U>\<^sup>s g \<longrightarrow> f \<or> g"
 by force
 then show ?thesis
 by (metis (no_types) FalseUntil Prop02 Prop03 UntilSUntil inteq_reflection)
 qed

lemma UntilIntro:
 "\<turnstile> g \<longrightarrow> f \<U> g"
proof -
 have 1: "\<turnstile> g = #False \<U> g"
  by (meson FalseUntil Prop11)
 have 2: "\<turnstile> #False \<longrightarrow> f"
   by auto
 have 3: "\<turnstile> g \<longrightarrow> g"
   by auto
 have 4: "\<turnstile> #False \<U> g \<longrightarrow> f \<U> g"
   by (simp add: UntilImpUntil)
 from 1 4 show ?thesis by fastforce
qed

lemma OrImpUntil:
 "\<turnstile> f \<and> g \<longrightarrow> f \<U> g"
by (simp add: Prop01 Prop05 UntilIntro)

lemma UntilAbsorp_a:
 "\<turnstile> (f \<or> f \<U> g) = (f \<or> g)"
proof -
 have 1: "\<turnstile> (f \<or> f \<U> g) \<longrightarrow> f \<or> g"
   using UntilImpOr by fastforce
 have 2: "\<turnstile> f \<or> g \<longrightarrow> (f \<or> f \<U> g)"
   using UntilIntro by fastforce
 from 1 2 show ?thesis by fastforce
qed

lemma UntilAbsorp_b:
 "\<turnstile> (f \<U> g \<or> g) = f \<U> g"
using UntilSUntil by fastforce

lemma UntilAbsorp_c:
 "\<turnstile> (f \<U> g \<and> g) = g"
using UntilIntro by fastforce

lemma UntilAbsorp_d:
 "\<turnstile> (f \<U> g \<or> (f \<and> g)) = f \<U> g"
using UntilSUntil by fastforce 

lemma UntilAbsorp_e:
 "\<turnstile> (f \<U> g \<and> (f \<or> g)) = f \<U> g"
by (meson Prop10 Prop11 UntilImpOr)

lemma LeftUntilAbsorp:
 "\<turnstile> f \<U> (f \<U> g) = f \<U> g"
by (meson Prop11 UntilUntil)

lemma RightUntilAbsorp:
 "\<turnstile> (f \<U> g) \<U> g = f \<U> g"
by (metis Prop11 UntilAbsorp_b UntilAbsorp_c UntilImpOr UntilRightAnd UntilUntil inteq_reflection)

lemma UntilAbsorpAndDiamond:
 "\<turnstile> (f \<U> g \<and> \<diamond> g) = f \<U> g"
by (metis DiamondEqvTrueUntil Prop11 Prop12 UntilAbsorp_c UntilRightAnd int_simps(17) 
      inteq_reflection)

lemma UntilAbsorpOrDiamond:
 "\<turnstile> (f \<U> g \<or> \<diamond> g) = \<diamond> g"
using UntilAbsorpAndDiamond by fastforce

lemma UntilAbsorpDiamond:
 "\<turnstile> f \<U> (\<diamond> g) = \<diamond> g"
using DiamondDiamondEqvDiamond UntilAbsorpOrDiamond UntilAbsorp_b by fastforce

lemma UntilImpDiamond:
 "\<turnstile> f \<U> g \<longrightarrow> \<diamond> g"
using UntilAbsorpAndDiamond by fastforce


lemma UntilInduction_a:
 "\<turnstile> \<box>(f \<longrightarrow> ((\<circle> f) \<and> g) \<or> h) \<longrightarrow> (f \<longrightarrow> \<box> g \<or> g \<U> h)"
proof -
 have 1: "\<turnstile> (\<box> g \<or> g \<U> h) = g \<W> h "
  by (auto simp add: wait_d_def)
 have 2: "\<turnstile> (f \<longrightarrow> \<box> g \<or> g \<U> h) = ( (\<not> h) \<U> (\<not> g \<and> \<not>h) \<longrightarrow> \<not> f)"
   using "1" WaitNotDistUntil by fastforce 
 have 3: "\<turnstile> \<box>( ((\<not>g \<and> \<not>h) \<or> (\<not>h \<and> \<circle>(\<not> f)))\<longrightarrow> \<not>f) \<longrightarrow> ( (\<not> h) \<U> (\<not> g \<and> \<not>h) \<longrightarrow> \<not> f)  "
   using LFPUntil by blast
 have 4: "\<turnstile> (f \<longrightarrow> ((\<circle> f) \<and> g) \<or> h) \<longrightarrow> ( ((\<not>g \<and> \<not>h) \<or> (\<not>h \<and> \<circle>(\<not> f)))\<longrightarrow> \<not>f)"
   using UntilInduction_a_help by simp
 have 5: "\<turnstile> \<box>(f \<longrightarrow> ((\<circle> f) \<and> g) \<or> h) \<longrightarrow> \<box>( ((\<not>g \<and> \<not>h) \<or> (\<not>h \<and> \<circle>(\<not> f)))\<longrightarrow> \<not>f)" 
   using "4" by (rule ImpBoxRule)
 show ?thesis 
 using "2" "3" "5" by fastforce
qed

lemma UntilInduction_b:
 "\<turnstile> \<box>(f \<longrightarrow> (\<circle>f) \<or> g) \<longrightarrow> (f \<longrightarrow> \<box> f \<or> f \<U> g)"
proof -
 have 1: "\<turnstile> (\<box> f \<or> f \<U> g) = f \<W> g"
   by (auto simp add: wait_d_def)
 have 2: "\<turnstile> (f \<longrightarrow> \<box> f \<or> f \<U> g) = ( (\<not> g) \<U> (\<not> f \<and> \<not> g) \<longrightarrow> \<not> f)"
     using "1" WaitNotDistUntil by fastforce 
 have 3: "\<turnstile> \<box> ( ((\<not> f \<and> \<not> g) \<or> (\<not>g \<and> \<circle> (\<not> f))) \<longrightarrow> \<not> f) \<longrightarrow> ( (\<not> g) \<U> (\<not> f \<and> \<not> g) \<longrightarrow> \<not> f)"
   using LFPUntil by blast
 have 4: "\<turnstile> \<box>(f \<longrightarrow> (\<circle>f) \<or> g) \<longrightarrow>   \<box>( ((\<not> f \<and> \<not> g) \<or> (\<not>g \<and> \<circle> (\<not> f))) \<longrightarrow> \<not> f)"
   using UntilInduction_b_help by simp
 show ?thesis 
 using "2" "3" "4" by fastforce
qed    

lemma AlwaysImpNotUntilNot:
 "\<turnstile> \<box> f \<longrightarrow> \<not>(g \<U> (\<not> f))"
by (simp add: UntilImpDiamond always_d_def)

lemma UntilAndImp:
 "\<turnstile> \<box> f \<and> \<diamond> g \<longrightarrow> f \<U> g"
proof -
have 1: "\<turnstile> \<diamond> g = #True \<U> g"
  by (simp add: DiamondEqvTrueUntil)
have 2: "\<turnstile> \<box> f \<and> #True \<U> g \<longrightarrow> (f \<and> #True) \<U> (f \<and> g)"
  using UntilAlwaysAndDist by blast 
have 3: "\<turnstile> (f \<and> #True) \<U> (f \<and> g) = f  \<U> (f \<and> g)"
   by simp
have 4: "\<turnstile> f  \<U> (f \<and> g) \<longrightarrow> (f \<U> f) \<U> g  " 
  by (simp add: UntilRightAnd)
have 5: "\<turnstile> (f \<U> f) = f "
  by (simp add: UntilIdempotent)
have 6: "\<turnstile> (f \<U> f) \<U> g  = f \<U> g"
  by (simp add: "5" UntilEqvUntil)
show ?thesis 
by (metis "1" "2" "3" "4" "5" inteq_reflection lift_imp_trans)
qed


lemma UntilCatRule:
" \<turnstile> \<box> ( (f \<longrightarrow> g \<U> h) \<and> (h \<longrightarrow> g \<U> i)) \<longrightarrow> (f \<longrightarrow> (g \<U> i)) "
proof -
 have 1: "\<turnstile> \<box> ( (f \<longrightarrow> g \<U> h) \<and> (h \<longrightarrow> g \<U> i)) \<longrightarrow> \<box> (f \<longrightarrow> g \<U> h)"
   by (metis BoxElim BoxEqvBoxBox BoxImpBoxRule Prop12 inteq_reflection)
 have 2: "\<turnstile> \<box> ( (f \<longrightarrow> g \<U> h) \<and> (h \<longrightarrow> g \<U> i)) \<longrightarrow> \<box> (h \<longrightarrow> g \<U> i) "
    by (metis BoxElim BoxEqvBoxBox BoxImpBoxRule Prop12 inteq_reflection)
 have 3: "\<turnstile> \<box> (h \<longrightarrow> g \<U> i) \<longrightarrow> \<box> ( g \<U> h \<longrightarrow> g \<U> (g \<U> i))"
    by (metis BoxEqvBoxBox BoxImpBoxRule UntilRightMono inteq_reflection) 
 have 4: "\<turnstile> \<box> ( g \<U> h \<longrightarrow> g \<U> (g \<U> i)) \<longrightarrow> \<box>( g \<U> h \<longrightarrow> g \<U> i)"
   by (metis BoxEqvBoxBox UntilUntil int_iffD1 inteq_reflection)
 have 5: "\<turnstile>  \<box> (f \<longrightarrow> g \<U> h) \<longrightarrow> (f \<longrightarrow> g \<U> h)"
    by (simp add: BoxElim)
 have 6: "\<turnstile> \<box>( g \<U> h \<longrightarrow> g \<U> i) \<longrightarrow> (g \<U> h \<longrightarrow> g \<U> i)"
    by (simp add: BoxElim)
 have 7: "\<turnstile> (f \<longrightarrow> g \<U> h) \<and> (g \<U> h \<longrightarrow> g \<U> i) \<longrightarrow> (f \<longrightarrow> g \<U> i)"
   by auto
 have 8: "\<turnstile> \<box> ( (f \<longrightarrow> g \<U> h) \<and> (h \<longrightarrow> g \<U> i)) \<longrightarrow> 
            (f \<longrightarrow> g \<U> h) \<and> (g \<U> h \<longrightarrow> g \<U> i) "
   using "1" "2" "3" "4" "5" "6" by fastforce
 from 7 8 show ?thesis by auto
qed
 

lemma UntilStrengthen:
 "\<turnstile> \<box> ( (f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<U> g \<longrightarrow> h \<U> i)"
proof -
 have 1: "\<turnstile> \<box> ( (f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<U> g \<longrightarrow> h \<U> g)"
   by (metis BoxImpBoxRule Prop01 UntilLeftMono  int_simps(12) int_simps(29) int_simps(9) 
       inteq_reflection lift_imp_trans)
 have 2: "\<turnstile> \<box> ( (f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> ( h \<U> g \<longrightarrow> h \<U> i) "
   by (metis BoxImpBoxRule Prop01 Prop05 UntilIdempotent UntilIntro UntilRightMono inteq_reflection 
       lift_imp_trans) 
 have 3: "\<turnstile> \<box> ( (f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<U> g \<longrightarrow> h \<U> g) \<and> ( h \<U> g \<longrightarrow> h \<U> i)"
   using "1" "2" by fastforce
 have 4: "\<turnstile> (f \<U> g \<longrightarrow> h \<U> g) \<and> ( h \<U> g \<longrightarrow> h \<U> i) \<longrightarrow> (f \<U> g \<longrightarrow> h \<U> i)"
   by auto
 from 3 4 show ?thesis by auto
qed

lemma UntilInduction:
 "\<turnstile> \<box> (f \<longrightarrow> \<not> g \<and> \<circle> f) \<longrightarrow> (f \<longrightarrow> \<not>(h \<U> g))"
proof -
 have 1: "\<turnstile> \<box> (\<not> g) \<longrightarrow> \<not>(h \<U> g)"
   by (simp add: UntilImpDiamond always_d_def)
 have 2: "\<turnstile> \<box> (f \<longrightarrow> \<not> g \<and> \<circle> f) \<longrightarrow> \<box>(g \<or> #True \<and> \<circle> (\<not> f) \<longrightarrow> \<not> f)"
   using UntilInduction_help by simp
 have 3: "\<turnstile> \<box> (f \<longrightarrow> \<not> g \<and> \<circle> f) \<longrightarrow> (#True \<U> g \<longrightarrow> \<not> f)"
   using "2" LFPUntil[of "g" "LIFT(#True)" "LIFT(\<not> f)"]  
  by fastforce
 have 4: "\<turnstile> (#True \<U> g \<longrightarrow> \<not> f) \<longrightarrow> (f \<longrightarrow> \<not>(#True \<U> g))"
      by auto
 have 5: "\<turnstile>  \<not>(#True \<U> g) = \<box> (\<not> g)"
    using BoxEqvFalseRelease NotUntilRelease inteq_reflection by fastforce
  from 5 4 3 1 show ?thesis by fastforce
qed


lemma UntilBoxImphelp1:
 " \<turnstile> f \<U> \<box>g \<longrightarrow> f \<U> g"
by (meson BoxElim BoxGen MP UntilRightMono)

lemma UntilBoxImphelp2:
 " \<turnstile> more \<and> f \<U> \<box>g \<longrightarrow> \<circle> (f \<U> \<box>g)"
proof -
  have f4: "\<turnstile> wnext (f \<U> \<box>g) = (empty \<or> \<circle> (f \<U> \<box>g))"
     by (meson WnextEqvEmptyOrNext) 
  have f5: "\<turnstile> \<box> g = (g \<and> wnext (\<box>g))"
     by (metis (no_types) BoxEqvAndWnextBox) 
  have f6: "\<turnstile>f \<U> \<box>g = (\<box>g \<or> f \<and> \<circle> (f \<U> \<box>g))"
     by (meson UntilUnrol) 
  have "\<turnstile> g \<and> wnext (\<box>g) \<longrightarrow> empty \<or> \<circle> (f \<U> \<box>g) "
     by (metis NextUntil Prop01 Prop05 Prop08 UntilIntro WnextEqvEmptyOrNext int_iffD1 
         inteq_reflection)
  then have "\<turnstile> more \<and> f \<U> \<box>g \<longrightarrow> empty \<or> \<circle> (f \<U> \<box>g)"
    using f6 f5 by fastforce
  then show ?thesis
    using f4 using WnextAndMoreEqvNext by fastforce
qed

lemma UntilBoxImp:
 "\<turnstile> f \<U> (\<box> g) \<longrightarrow> \<box>(f \<U> g)"
using BoxIntro[of "LIFT(f \<U> (\<box> g))" "LIFT(f \<U> g)"]  
by (simp add: UntilBoxImphelp1 UntilBoxImphelp2)

lemma UntilBoxEqvBox:
 "\<turnstile> f \<U> (\<box> f) = \<box> f"
proof -
 have 1: "\<turnstile>  f \<U> (\<box> f) \<longrightarrow> \<box>(f \<U> f)"
   using UntilBoxImp[of "f" "f"] by auto
 have 2: "\<turnstile> \<box>(f \<U> f) = \<box> f" 
    by (simp add: BoxEqvBox UntilIdempotent)
 have 3: "\<turnstile> \<box> f \<longrightarrow> f \<U> (\<box> f)" 
   by (simp add: UntilIntro)
 from 1 2 3 show ?thesis by fastforce
qed

lemma UntilRightStrengthen:
 "\<turnstile> f \<U> (g \<and> h) \<longrightarrow> f \<U> (g \<U> h)"
by (meson BoxGen MP OrImpUntil UntilRightMono) 

lemma UntilLeftStrengthen:
 "\<turnstile> (f \<and> g) \<U> h \<longrightarrow> (f \<U> g) \<U>  h"
  by (simp add: OrImpUntil UntilImpUntil)

lemma UntilLeftAndOrder:
 "\<turnstile> (f \<and> g) \<U> h \<longrightarrow> f \<U> (g \<U> h)"
by (metis Prop05 Prop07 ReleaseUntil RightUntilAbsorp UntilAbsorp_c UntilAndDist 
    UntilRightStrengthen inteq_reflection)

lemma UntilFrameNext:
 "\<turnstile> \<box> f \<longrightarrow> (\<circle> g \<longrightarrow> \<circle> (f \<U> g))"
by (simp add: NextImpNext Prop01 Prop05 Prop09 UntilIntro)

lemma UntilFrameDiamond:
 "\<turnstile> \<box> f \<longrightarrow> (\<diamond> g \<longrightarrow> \<diamond> (f \<U> g))"
by (meson NowImpDiamond Prop09 UntilAndImp lift_imp_trans)

lemma UntilFrameBox:
 "\<turnstile> \<box> f \<longrightarrow> (\<box> g \<longrightarrow> \<box> (f \<U> g))"
by (simp add: BoxAndBoxImpBoxRule OrImpUntil Prop09)

lemma UntilAndRule:
 "\<turnstile> f \<U> g = (f \<and> \<not>g) \<U> g"
proof -
 have 1: "\<turnstile> (f \<and> \<not>g) \<U> g \<longrightarrow> f \<U> g" 
    using UntilAndDist by fastforce
 show ?thesis by (simp add: "1" UntilImpNot int_iffI)
qed

lemma UntilWait:
 "\<turnstile> f \<U> g = (f \<W> g \<and> \<diamond> g)"
proof -
 have 1: "\<turnstile> f \<U> g \<longrightarrow> f \<W> g \<and> \<diamond> g" 
     by (simp add: Prop05 Prop12 UntilImpDiamond wait_d_def)
 have 2: "\<turnstile> (f \<W> g \<and> \<diamond> g)  = ((\<box> f \<or> f \<U> g) \<and> \<diamond> g) "
   by (auto simp add: wait_d_def)
 have 3: "\<turnstile> ((\<box> f \<or> f \<U> g) \<and> \<diamond> g) = ( (\<box> f \<and> \<diamond>g) \<or> (f \<U> g \<and> \<diamond> g))"
   by auto
 have 4: "\<turnstile> (\<box> f \<and> \<diamond>g) \<longrightarrow> f \<U> g"
   by (simp add: UntilAndImp)
 have 5: "\<turnstile> (f \<U> g \<and> \<diamond> g) \<longrightarrow> f \<U> g"
   by auto
 show ?thesis 
 using "1" "2" "4" by fastforce 
qed

lemma WaitUntilb:
 "\<turnstile> f \<W>  g = (\<box> ( f \<and> \<not> g) \<or> f \<U> g)"
proof -
 have 1: "\<turnstile> f \<W> g = (f \<and> \<not> g) \<W> g" 
   by (simp add: WaitAndRule)
 have 2: "\<turnstile> (f \<and> \<not> g) \<W> g = (\<box> (f \<and> \<not> g) \<or> (f \<and> \<not> g) \<U> g)"
   by (auto simp add: wait_d_def)
 have 3: "\<turnstile> (f \<and> \<not> g) \<U> g = f \<U> g"
   by (meson Prop11 UntilAndRule)
 show ?thesis 
 using "1" "2" "3" by fastforce
qed

lemma UntilNotDistWait:
 "\<turnstile> (\<not> (f \<U> g)) = (\<not> g) \<W> (\<not> f \<and> \<not> g)"
proof -
 have 1: "\<turnstile> (\<not> ((\<not> g) \<W> (\<not> f \<and> \<not> g))) = (\<not>(\<not> f \<and> \<not> g)) \<U> ( \<not> (\<not> g) \<and> \<not> (\<not> f \<and> \<not> g))"
  using WaitNotDistUntil by blast  
 have 2: "\<turnstile> (\<not>(\<not> f \<and> \<not> g)) = (f \<or> g)"
    by auto
 have 3: "\<turnstile> ( \<not> (\<not> g) \<and> \<not> (\<not> f \<and> \<not> g)) = g"
   by auto
 have 4: "\<turnstile> (\<not>(\<not> f \<and> \<not> g)) \<U> ( \<not> (\<not> g) \<and> \<not> (\<not> f \<and> \<not> g)) =
            (f \<or> g) \<U> g "
   using "2" "3" UntilEqvUntil by blast
 have 5: "\<turnstile> (f \<or> g) \<U> g = ((f \<or> g) \<and> \<not>g) \<U> g"
   by (simp add: UntilAndRule)
 have 6: "\<turnstile> ((f \<or> g) \<and> \<not>g) = (f \<and> \<not> g)"
  by auto
 have 7: "\<turnstile> ((f \<or> g) \<and> \<not>g) \<U> g = (f \<and> \<not>g) \<U> g"
   using "6" inteq_reflection by fastforce 
 have 8: "\<turnstile> (f \<and> \<not>g) \<U> g = f \<U> g"
    by (meson Prop11 UntilAndRule)
 have 9: "\<turnstile> f \<U> g = (\<not> ((\<not> g) \<W> (\<not> f \<and> \<not> g)))"
    using "1" "4" "5" "7" "8" by fastforce
 show ?thesis using "9" by auto
qed   
   
lemma UntilImpWait:
 "\<turnstile> f \<U> g \<longrightarrow> f \<W> g"
by (meson Prop03 WaitUntilb)

lemma WaitAndDist:
 "\<turnstile> (\<box> f \<and> g \<W> h) \<longrightarrow> (f \<and> g) \<W> (f \<and> h)"
proof -
 have 1: "\<turnstile> (\<box> f \<and> g \<W> h) = (\<box> f \<and> (\<box> g \<or> g \<U> h))"
   by (auto simp add: wait_d_def)
 have 2: "\<turnstile> (\<box> f \<and> (\<box> g \<or> g \<U> h)) = ((\<box>f \<and> \<box>g) \<or> (\<box> f \<and> g \<U> h))"
   by auto
 have 3: "\<turnstile> (\<box>f \<and> \<box>g) = \<box>(f \<and> g)"
   by (simp add: BoxAndBoxEqvBoxRule)
 have 4: "\<turnstile> (\<box> f \<and> g \<U> h) \<longrightarrow> (f \<and> g) \<U> (f \<and> h)"
   by (simp add: UntilAlwaysAndDist)
 have 5: "\<turnstile> ((\<box>f \<and> \<box>g) \<or> (\<box> f \<and> g \<U> h)) \<longrightarrow> \<box>(f \<and> g) \<or> (f \<and> g) \<U> (f \<and> h)"
   using "3" "4" by fastforce
 have 6: "\<turnstile> (\<box>(f \<and> g) \<or> (f \<and> g) \<U> (f \<and> h)) = (f \<and> g) \<W> (f \<and> h)"
   by (auto simp add: wait_d_def)
 show ?thesis 
 using "1" "5" "6" by fastforce
qed

lemma WaitDiamondOr:
 "\<turnstile> f \<W> (\<diamond> g) = (\<box> f \<or> \<diamond> g)"
proof -
 have 1: "\<turnstile> f \<W> (\<diamond> g) = (\<box> f \<or> f \<U> (\<diamond> g))"
  by (auto simp add: wait_d_def)
 have 2: "\<turnstile> f \<U> (\<diamond> g) = \<diamond> g"
  by (simp add: UntilAbsorpDiamond)
 show ?thesis using "1" "2" Prop06 by blast
qed

lemma WaitBoxImp:
 "\<turnstile> f \<W> (\<box> g) \<longrightarrow> \<box> ( f \<W> g)"
proof -
 have 1: "\<turnstile> f \<W> (\<box> g) = (\<box> f \<or> f \<U> (\<box> g))"
 by (auto simp add: wait_d_def)
 have 2: "\<turnstile> \<box> f = \<box> (\<box> f)"
   by (simp add: BoxEqvBoxBox)
 have 3: "\<turnstile> f \<U> (\<box> g) \<longrightarrow> \<box>(f \<U> g)"
  by (simp add: UntilBoxImp)
 have 4: "\<turnstile> (\<box> f \<or> f \<U> (\<box> g)) \<longrightarrow> (\<box> (\<box> f) \<or> \<box>(f \<U> g))"
   using "2" "3" by fastforce
 have 5: "\<turnstile> \<box> (\<box> f) \<longrightarrow> \<box>( \<box> f \<or> f \<U> g) "
   by (metis BoxImpBoxRule Prop08 UntilIdempotent UntilIntro int_simps(11) int_simps(25) 
       inteq_reflection)
 have 6: "\<turnstile> \<box>(f \<U> g) \<longrightarrow> \<box>( \<box> f \<or> f \<U> g)"
   by (metis BoxImpBoxRule UntilImpWait wait_d_def) 
 have 7: "\<turnstile> (\<box> (\<box> f) \<or> \<box>(f \<U> g)) \<longrightarrow> \<box>( \<box> f \<or> f \<U> g)"
    using "5" "6" by fastforce
 have 6: "\<turnstile> \<box>( \<box> f \<or> f \<U> g) = \<box> ( f \<W> g)"
    by (simp add: wait_d_def) 
 show ?thesis 
 by (metis "4" "7" lift_imp_trans wait_d_def)
qed

lemma WaitAbsorptionBox:
 "\<turnstile> f \<W> (\<box> f) = \<box> f"
by (metis Prop02 Prop11 UntilBoxEqvBox UntilImpWait inteq_reflection wait_d_def)


lemma BoxImpWait:
 "\<turnstile> \<box> f \<longrightarrow> f \<W> g"
by (auto simp add: wait_d_def)

lemma WaitDistNext:
 "\<turnstile> \<circle> ( f \<W> g) = (\<circle> f) \<W> (\<circle> g) "
\<comment> \<open>nitpick finds counterexample\<close>
oops

lemma WnextAlwaysEqvAlwaysWnext:
  "\<turnstile> wnext (\<box> f) = \<box>(wnext f)"
   by (metis (no_types, lifting) NextDiamondEqvDiamondNext always_d_def int_eq int_simps(4) 
       wnext_d_def)

lemma WaitExpand: 
 "\<turnstile> f \<W> g = (g \<or> (f \<and> \<circle>(f \<W> g)))"
\<comment> \<open> nitpick finds counterexample\<close>
oops

lemma WaitExpand: 
 "\<turnstile> f \<W> g = (g \<or> (f \<and> wnext(f \<W> g)))"
proof -
 have 1: "\<turnstile>  f \<W> g = (\<box> f \<or> f \<U> g)"
  by (simp add: wait_d_def)
 have 2: "\<turnstile> \<box> f = (f \<and> wnext(\<box> f))"
   by (simp add: BoxEqvAndWnextBox)
 have 3: "\<turnstile> f \<U> g = (g \<or> (f \<and> \<circle> (f \<U> g)))"
    using UntilUnrol by blast 
 have 4: "\<turnstile> (f \<and> wnext(\<box> f)) = (f \<and> (empty \<or> \<circle> (\<box> f)))"
   using "2" BoxEqvAndEmptyOrNextBox by fastforce
 have 5: "\<turnstile>  wnext(f \<W> g) = (empty \<or> \<circle> (f \<W> g))"
    using WnextEqvEmptyOrNext by blast
 have 6: "\<turnstile> (f \<and> (empty \<or> \<circle> (\<box> f))) = ( (f \<and> empty) \<or> (f \<and> \<circle> (\<box> f)))"
   by auto
 have 7: "\<turnstile> (( (f \<and> empty) \<or> (f \<and> \<circle> (\<box> f))) \<or>
             (g \<or> (f \<and> \<circle> (f \<U> g))) ) =
             (g \<or> (f \<and> (empty \<or> \<circle> (\<box>f)  \<or> \<circle> (f \<U> g) ) ))"
      by auto
 have 8: "\<turnstile> (\<circle> (\<box>f)  \<or> \<circle> (f \<U> g)) = \<circle>(\<box>f \<or> f \<U> g)"
   by (metis ChopOrEqv Prop11 next_d_def)
 show ?thesis 
 by (metis "1" "2" "3" "4" "5" "6" "7" "8" inteq_reflection)
qed

lemma WaitExclMid:
 "\<turnstile> f \<W> g \<or> f \<W> (\<not> g)"
using WaitExpand
proof -
 have 1: "\<turnstile> f \<W> g = (g \<or> f \<and> wnext (f \<W> g)) " 
   by (simp add: WaitExpand)
 have 2: "\<turnstile> f \<W> (\<not> g) = ((\<not> g) \<or> f \<and> wnext ( f \<W> (\<not> g)))"
   by (simp add: WaitExpand) 
 have 3: "\<turnstile> (f \<W> g \<or> f \<W> (\<not> g)) = 
            ( (g \<or> f \<and> wnext (f \<W> g)) \<or> ((\<not> g) \<or> f \<and> wnext ( f \<W> (\<not> g)))) "
  using "1" "2" by fastforce
 from 3 show ?thesis by fastforce
qed
 

lemma WaitleftZero:
 "\<turnstile> #True \<W> g"
 by (meson BoxGen BoxImpWait MP TrueW)

lemma WaitLeftDistOr:
 "\<turnstile> f \<W> (g \<or> h) = (f \<W> g \<or> f \<W> h)"
proof -
 have 1: "\<turnstile> f \<W> (g \<or> h) = ( \<box> f \<or> f \<U> (g \<or> h))"
   by (simp add: wait_d_def)
 have 2: "\<turnstile> (f \<W> g \<or> f \<W> h) = ((\<box> f \<or> f \<U> g) \<or> (\<box> f \<or> f \<U> h))  "
  by (simp add: wait_d_def)
 have 3: "\<turnstile> f \<U> (g \<or> h) = (f \<U> g \<or> f \<U> h)"
   by (simp add: UntilOrDist) 
 from 1 2 3 show ?thesis by fastforce
qed


lemma WaitRightDistOr:
 "\<turnstile> f \<W> h \<or> g \<W> h \<longrightarrow> (f \<or> g) \<W> h"
proof -
 have 0:"\<turnstile> \<box> g \<longrightarrow> \<box> (f \<or> g)"
   by (simp add: BoxImpBoxRule intI)
 have 1: "\<turnstile> \<box>f \<longrightarrow> \<box> (f \<or> g)"
   by (simp add: BoxImpBoxRule intI)
 have 11: "\<turnstile> \<box> f \<or> \<box> g \<longrightarrow> \<box> (f \<or> g)" 
   using "0" "1" Prop02 by blast
 have 2: "\<turnstile>  f \<W> h = (\<box> f \<or> f \<U> h)"
   by (simp add: wait_d_def)
 have 3:"\<turnstile> (f \<or> g) \<W> h = (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
   by (simp add: wait_d_def)
 have 4: "\<turnstile> g \<W> h = (\<box>g  \<or> g \<U> h)" 
   by (simp add: wait_d_def)
 have 5: "\<turnstile> f \<U> h \<or> g \<U> h \<longrightarrow> (f \<or> g) \<U> h" 
   using UntilRightDistOr by simp
 have 6:"\<turnstile> (f \<W> h \<or> g \<W> h) = ((\<box> f \<or> \<box> g) \<or> (f \<U> h \<or> g \<U> h))" 
   using "2" "4" by fastforce
 from 11 5 6 3 show ?thesis 
 using BoxImpWait by fastforce
qed 




lemma WaitOrRule:
 "\<turnstile> f \<W> g = (f \<or> g) \<W> g"
proof -
 have 1: "\<turnstile> f \<W> g \<longrightarrow> (f \<or> g) \<W> g"
   by (metis (no_types, lifting) Prop03 Prop10 UntilAbsorp_a WaitNotDistUntil int_iffD1 int_simps(14) 
       int_simps(32) int_simps(33) inteq_reflection)
 have 2: "\<turnstile> (f \<or> g) \<W> g \<longrightarrow> f \<W> g"
   by (metis (no_types, lifting) Prop03 Prop10 WaitNotDistUntil int_iffD2 int_simps(14) 
       int_simps(32) int_simps(33) inteq_reflection) 
 from 1 2 show ?thesis by fastforce
qed

lemma UntilOrRule:
 "\<turnstile> f \<U> g = (f \<or> g) \<U> g"
by (metis UntilWait WaitOrRule inteq_reflection)
 
lemma WaitRule:
 "\<turnstile> (\<not> f) \<W> f"
by (metis BoxGen BoxImpWait MP WaitOrRule int_eq_true int_simps(29) inteq_reflection)

lemma UntilRule:
 "\<turnstile> (\<not> f) \<U> f = \<diamond> f"
using DiamondEqvTrueUntil UntilOrRule inteq_reflection by fastforce

lemma DiamondUntilImpRule:
 "\<turnstile> \<diamond> f \<longrightarrow> (f \<longrightarrow> g) \<U> f"
using UntilWait WaitImpRule by fastforce

lemma WaitNotDist:
 "\<turnstile> (\<not> (f \<W> g)) = (f \<and> \<not> g) \<U> (\<not> f \<and> \<not> g)"
proof -
 have 1: "\<turnstile> (\<not> (f \<W> g)) = (\<not> g ) \<U> (\<not> f \<and> \<not> g)"
  using WaitNotDistUntil by blast 
 have 2: "\<turnstile> (\<not> g ) \<U> (\<not> f \<and> \<not> g) = (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) \<U> (\<not> f \<and> \<not> g)"
   using UntilAndRule by blast
 have 3: "\<turnstile> (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) = (f \<and> \<not> g)"
   by auto
 have 4: "\<turnstile> (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) \<U> (\<not> f \<and> \<not> g) = (f \<and> \<not> g) \<U> (\<not> f \<and> \<not> g)"
   using "3" inteq_reflection by force
 show ?thesis using "1" "2" "4" by fastforce
qed

lemma UntilNotDist:
 "\<turnstile> (\<not> (f \<U> g)) = (f \<and> \<not> g) \<W> (\<not> f \<and> \<not> g)"
proof -
 have 1: "\<turnstile> (\<not> (f \<U> g)) = (\<not> g ) \<W> (\<not> f \<and> \<not> g)"
   using UntilNotDistWait by blast
 have 2: "\<turnstile>  (\<not> g ) \<W> (\<not> f \<and> \<not> g) = (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) \<W> (\<not> f \<and> \<not> g) "
   by (simp add: WaitAndRule)
 have 3: "\<turnstile> (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) = (f \<and> \<not> g)"
   by auto
 have 4: "\<turnstile> (\<not> g \<and> \<not>(\<not> f \<and> \<not> g)) \<W> (\<not> f \<and> \<not> g) = (f \<and> \<not> g) \<W> (\<not> f \<and> \<not> g)"
   using "3" inteq_reflection by force
 show ?thesis using "1" "2" "4" by fastforce
qed

lemma UntilDuala:
 "\<turnstile> (\<not> ( (\<not> f) \<U> (\<not> g))) = g \<W> (f \<and> g)"
proof -
 have 1: "\<turnstile> (\<not> ( (\<not> f) \<U> (\<not> g))) = (\<not> f \<and> \<not> (\<not> g)) \<W> ( \<not>(\<not> f) \<and> \<not> (\<not> g))"
   using UntilNotDist by blast 
 have 2: "\<turnstile> (\<not> f \<and> g) \<W> ( f \<and> g) = g \<W> ( f \<and> g) "
   using "1" UntilNotDistWait int_eq by fastforce
 show ?thesis 
 using "1" "2" by fastforce
qed

lemma UntilDualb:
 "\<turnstile> (\<not> ( (\<not> f) \<U> (\<not> g))) = (\<not> f \<and> g) \<W> (f \<and> g)"
proof -
 have 1: "\<turnstile> (\<not> ( (\<not> f) \<U> (\<not> g))) = (\<not> f \<and> \<not> (\<not> g)) \<W> ( \<not>(\<not> f) \<and> \<not> (\<not> g))"
   using UntilNotDist by blast 
 show ?thesis 
 using "1" by auto
qed

lemma WaitDuala:
 "\<turnstile> (\<not> ( (\<not> f) \<W> (\<not> g))) = g \<U> (f \<and> g)"
proof -
 have 1: "\<turnstile> (\<not> ( (\<not> f) \<W> (\<not> g))) = (\<not> f \<and> \<not> (\<not> g)) \<U> ( \<not>(\<not> f) \<and> \<not> (\<not> g))"
   using WaitNotDist by blast 
 have 2: "\<turnstile> (\<not> f \<and> g) \<U> ( f \<and> g) = g \<U> ( f \<and> g) "
   using "1" WaitNotDistUntil int_eq by fastforce
 show ?thesis 
 using "1" "2" by fastforce
qed

lemma WaitDualb:
 "\<turnstile> (\<not> ( (\<not> f) \<W> (\<not> g))) = (\<not> f \<and> g) \<U> (f \<and> g)"
proof -
 have 1: "\<turnstile> (\<not> ( (\<not> f) \<W> (\<not> g))) = (\<not> f \<and> \<not> (\<not> g)) \<U> ( \<not>(\<not> f) \<and> \<not> (\<not> g))"
   using WaitNotDist by blast 
 show ?thesis using "1" by auto
qed

lemma WaitIdempotent:
 "\<turnstile> f \<W> f = f"
by (meson BoxElim Prop02 Prop12 UntilIdempotent UntilImpWait UntilIntro WaitUntilb int_iffD1 
    int_iffI lift_imp_trans)


lemma WaitRightZero:
 "\<turnstile> f \<W> #True "
by (meson MP TrueW UntilImpWait UntilIntro)

lemma WaitLeftIdentity:
 "\<turnstile> #False \<W> g = g"
by (metis (no_types, lifting) UntilAbsorp_c UntilNotDistWait WaitDuala WaitIdempotent WaitSRelease 
    int_eq int_simps(17) int_simps(3) srelease_d_def)

lemma WaitImpOr:
 "\<turnstile> f \<W> g \<longrightarrow> f \<or> g"
 by (metis Prop03 WaitIdempotent WaitLeftDistOr WaitOrRule inteq_reflection)

lemma BoxOrImpWait:
 "\<turnstile> \<box>(f \<or> g) \<longrightarrow> f \<W> g"
using BoxImpWait WaitOrRule by fastforce

lemma BoxImpImpWait:
 "\<turnstile> \<box>(\<not> g \<longrightarrow> f) \<longrightarrow> f \<W> g"
proof -
 have 1: "\<turnstile> (\<not> g \<longrightarrow> f) = (f \<or> g)"
   by auto
 have 2: "\<turnstile> \<box> (\<not> g \<longrightarrow> f) = \<box>(f \<or> g)" 
   using "1" BoxEqvBox by blast
 show ?thesis using "2" BoxOrImpWait by fastforce
qed

lemma WaitInsertion:
 "\<turnstile> g \<longrightarrow> f \<W> g"
by (simp add: Prop05 UntilIntro wait_d_def)

lemma WaitFrameNext:
 "\<turnstile> \<box> f \<longrightarrow> (\<circle> g \<longrightarrow> \<circle> (f \<W> g))"
by (simp add: NextImpNext Prop01 Prop05 Prop09 WaitInsertion)

lemma WaitFrameDiamond:
 "\<turnstile> \<box> f \<longrightarrow> (\<diamond> g \<longrightarrow> \<diamond> (f \<W> g))"
by (simp add: DiamondImpDiamond Prop01 Prop05 Prop09 WaitInsertion)

lemma WaitFrameBox:
 "\<turnstile> \<box> f \<longrightarrow> (\<box> g \<longrightarrow> \<box> (f \<W> g))"
by (meson BoxAndBoxImpBoxRule OrImpUntil Prop09 UntilImpWait lift_imp_trans)

lemma WaitInductiona:
 "\<turnstile> \<box> (f \<longrightarrow> (\<circle> f \<and> g) \<or> h) \<longrightarrow> (f \<longrightarrow> g \<W> h)"
by (simp add: UntilInduction_a wait_d_def)

lemma WaitInductionb:
 "\<turnstile> \<box> (f \<longrightarrow> \<circle> f \<or> g) \<longrightarrow> (f \<longrightarrow> f \<W> g)"
by (simp add: UntilInduction_b wait_d_def)

lemma WaitInductionc:
 "\<turnstile> \<box>(f \<longrightarrow> \<circle> f) \<longrightarrow> (f \<longrightarrow> f \<W> g)"
by (meson WaitInductionc_help BoxImpWait BoxInduct Prop09 lift_imp_trans) 

lemma WaitInductiond:
 "\<turnstile> \<box> (f \<longrightarrow> g \<and> \<circle> f) \<longrightarrow> (f \<longrightarrow> f \<W> g)"
 by (meson WaitInductiond_help BoxImpWait BoxInduct Prop09 lift_imp_trans) 

lemma WaitAbsorptiona:
 "\<turnstile> (f \<or> f \<W> g) = (f \<or> g)"
proof -
 have 1: "\<turnstile> (f \<or> f \<W> g) \<longrightarrow> (f \<or> g) "
   using WaitImpOr by fastforce
 have 2: "\<turnstile> f \<or> g \<longrightarrow> f \<or> f \<W> g"
   using WaitInsertion by fastforce
 show ?thesis using "1" "2" int_iffI by blast
qed

lemma WaitAbsorptionb:
 "\<turnstile> (f \<W> g \<or> g) = f \<W> g"
by (metis (no_types, lifting) BoxEqvBoxBox UntilAbsorp_a UntilAbsorp_b WaitAbsorptiona 
    WaitLeftDistOr WaitOrRule inteq_reflection wait_d_def)

lemma WaitAbsorptionc:
 "\<turnstile> (f \<W> g \<and> g) = g"
using WaitInsertion by fastforce

lemma WaitAbsorptiond:
 "\<turnstile> (f \<W> g \<and> (f \<or> g)) = f \<W> g"
by (meson Prop10 Prop11 WaitImpOr)

lemma WaitAbsorptione:
 "\<turnstile> (f \<W> g \<or> (f \<and> g)) = f \<W> g"
by (metis (no_types, lifting) BoxEqvBoxBox UntilAbsorp_a UntilAbsorp_d WaitAbsorptiona 
    WaitLeftDistOr WaitOrRule inteq_reflection wait_d_def)

lemma WaitLeftAbsorption:
 "\<turnstile> f \<W> (f \<W> g) = f \<W> g"
by (metis (no_types, lifting) BoxEqvBoxBox UntilUntil WaitAbsorptionBox WaitAbsorptiona 
    WaitLeftDistOr inteq_reflection wait_d_def)

lemma WaitRightAbsorption:
 "\<turnstile> (f \<W> g) \<W> g = f \<W> g"
 by (metis (no_types, lifting) LeftUntilAbsorp Prop10 WaitInsertion WaitNotDistUntil int_iffD1
      int_iffI int_simps(32) inteq_reflection)

lemma WaitBox:
 "\<turnstile> \<box> f = f \<W> #False"
 by (metis (no_types, lifting) BoxGen DiamondNotEqvNotBox UntilAbsorpAndDiamond UntilAbsorp_c 
     int_eq_true int_simps(2) int_simps(25) inteq_reflection wait_d_def)

lemma WaitDiamond:
 "\<turnstile> \<diamond> f = (\<not>( (\<not> f) \<W> #False))"
using DiamondNotEqvNotBox WaitBox by fastforce

lemma WaitImp:
 "\<turnstile> f \<W> g \<longrightarrow> \<box> f \<or> \<diamond> g"
by (metis Prop08 UntilImpDiamond WaitAbsorptionb WaitImpOr WaitRightAbsorption int_eq wait_d_def)

lemma WaitRightUntilAbsorption:
 "\<turnstile> f \<W> (f \<U> g) = f \<W> g"
by (metis UntilUntil WaitOrRule inteq_reflection wait_d_def)

lemma WaitLeftUntilAbsorption:
 "\<turnstile> (f \<U> g) \<W> g = f \<U> g"
by (metis Prop11 RightUntilAbsorp UntilAbsorp_b UntilImpWait WaitImpOr inteq_reflection)

lemma UntilRightWaitAbsorption:
 "\<turnstile> f \<U> (f \<W> g) = f \<W> g"
using UntilImpWait UntilIntro WaitLeftAbsorption by fastforce

lemma UntilLeftWaitAbsorption:
 "\<turnstile> (f \<W> g) \<U> g = f \<U> g"
by (metis UntilWait WaitRightAbsorption inteq_reflection)

lemma WaitDiamondAbsorption:
 "\<turnstile> (\<diamond> g) \<W> g = \<diamond> g"
by (metis DiamondEqvTrueUntil WaitLeftUntilAbsorption inteq_reflection)

lemma WaitAndBoxAbsorption:
 "\<turnstile> (\<box> f \<and> f \<W> g) = \<box> f"
by (meson BoxImpWait NotDiamondNotEqvBox Prop04 Prop10)

lemma WaitOrBoxAbsorption:
 "\<turnstile> (\<box> f \<or> f \<W> g) = f \<W> g"
by (metis UntilRightWaitAbsorption WaitLeftAbsorption inteq_reflection wait_d_def)

lemma WaitAndBoxImpBox:
 "\<turnstile> f \<W> g \<and> \<box> (\<not> g) \<longrightarrow> \<box> f"
by (metis (no_types, hide_lams) Prop02 Prop05 Prop07 Prop08 UntilIdempotent UntilImpDiamond 
    UntilIntro always_d_def int_simps(25) int_simps(4) inteq_reflection wait_d_def)

lemma BoxImpUntilOrBox:
 "\<turnstile> \<box> f \<longrightarrow> f \<U> g \<or> \<box> (\<not> g)"
proof -
 have 1: "\<turnstile> (\<box> f \<longrightarrow> f \<U> g \<or> \<box> (\<not> g)) = 
            ((\<box> f \<and> \<diamond> g)  \<longrightarrow> f \<U> g)"
  by (auto simp add: always_d_def)
 have 2: "\<turnstile> (\<box> f \<and> \<diamond> g)  \<longrightarrow> f \<U> g"
    using UntilAndImp by blast
  show ?thesis 
  using "1" "2" by fastforce
qed

lemma NotBoxAndWaitImpDiamond:
 "\<turnstile> \<not>( \<box> f) \<and> f \<W> g \<longrightarrow> \<diamond> g"
using WaitImp by fastforce

lemma DiamondImpNotBoxOrUntil:
 "\<turnstile> \<diamond> g \<longrightarrow> \<not> (\<box> f) \<or> f \<U> g"
proof -
 have 1: "\<turnstile> \<diamond> g \<and> \<box> f \<longrightarrow> f \<U> g"
   using UntilAndImp by fastforce
 show ?thesis using "1" by auto
qed

lemma WaitLeftMono:
 "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (f \<W> h \<longrightarrow> g \<W> h)"
by (meson BoxImpWait WaitRightDistImp lift_imp_trans)

lemma WaitRightMono:
 "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (h \<W> f \<longrightarrow> h \<W> g)"
proof -
 have 1: "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (h \<U> f \<longrightarrow> h \<U> g)"
   by (simp add: UntilRightMono)
 have 2: "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (\<box> h  \<longrightarrow> \<box> h \<or> h \<U> g)"
   by auto
 have 3: "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (h \<U> f  \<longrightarrow> \<box> h \<or> h \<U> g)"
   using "1" by auto
 have 4: "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (\<box> h \<or> h \<U> f \<longrightarrow> \<box> h \<or> h \<U> g)"
   using "2" "3" by fastforce
 from 4 show ?thesis by (simp add: wait_d_def)
qed

lemma WaitStrengthen:
 "\<turnstile> \<box>((f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<W> g \<longrightarrow> h \<W> i)"
proof -
 have 1: "\<turnstile> \<box>((f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<W> g \<longrightarrow> h \<W> g)"
 by (meson BoxAndBoxEqvBoxRule Prop01 Prop05 Prop11 WaitLeftMono lift_and_com lift_imp_trans)
 have 2: "\<turnstile>  \<box>((f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (h \<W> g \<longrightarrow> h \<W> i) "
 by (metis BoxImpBoxRule Prop01 Prop05 TrueW WaitRightMono int_simps(13) inteq_reflection 
     lift_imp_trans)
 have 3: "\<turnstile> \<box> ( (f \<longrightarrow> h) \<and> (g \<longrightarrow> i)) \<longrightarrow> (f \<W> g \<longrightarrow> h \<W> g) \<and> ( h \<W> g \<longrightarrow> h \<W> i)"
   using "1" "2" by fastforce
 have 4: "\<turnstile> (f \<W> g \<longrightarrow> h \<W> g) \<and> ( h \<W> g \<longrightarrow> h \<W> i) \<longrightarrow> (f \<W> g \<longrightarrow> h \<W> i)"
   by auto
 from 3 4 show ?thesis by auto
qed

lemma WaitCatRule:
 "\<turnstile> \<box>( (f \<longrightarrow> g \<W> h) \<and> (h \<longrightarrow> g \<W> i)) \<longrightarrow> (f \<longrightarrow> g \<W> i)"
proof -
 have 1: "\<turnstile> \<box> ( (f \<longrightarrow> g \<W>  h) \<and> (h \<longrightarrow> g \<W>  i)) \<longrightarrow> \<box> (f \<longrightarrow> g \<W> h)"
   by (metis BoxElim BoxEqvBoxBox BoxImpBoxRule Prop12 inteq_reflection)
 have 2: "\<turnstile> \<box> ( (f \<longrightarrow> g \<W> h) \<and> (h \<longrightarrow> g \<W> i)) \<longrightarrow> \<box> (h \<longrightarrow> g \<W> i) "
    by (metis BoxElim BoxEqvBoxBox BoxImpBoxRule Prop12 inteq_reflection)
 have 3: "\<turnstile> \<box> (h \<longrightarrow> g \<W> i) \<longrightarrow> \<box> ( g \<W> h \<longrightarrow> g \<W> (g \<W> i))"
    by (metis BoxEqvBoxBox BoxImpBoxRule WaitRightMono inteq_reflection) 
 have 4: "\<turnstile> \<box> ( g \<W> h \<longrightarrow> g \<U> (g \<W> i)) \<longrightarrow> \<box>( g \<W> h \<longrightarrow> g \<W> i)"
   by (metis BoxBoxImpBox BoxEqvBoxBox UntilRightWaitAbsorption inteq_reflection)
 have 5: "\<turnstile>  \<box> (f \<longrightarrow> g \<W> h) \<longrightarrow> (f \<longrightarrow> g \<W> h)"
    by (simp add: BoxElim)
 have 6: "\<turnstile> \<box>( g \<W> h \<longrightarrow> g \<W> i) \<longrightarrow> (g \<W> h \<longrightarrow> g \<W> i)"
    by (simp add: BoxElim)
 have 7: "\<turnstile> (f \<longrightarrow> g \<W> h) \<and> (g \<W> h \<longrightarrow> g \<W> i) \<longrightarrow> (f \<longrightarrow> g \<W> i)"
   by auto
 have 8: "\<turnstile> \<box> ( (f \<longrightarrow> g \<W> h) \<and> (h \<longrightarrow> g \<W> i)) \<longrightarrow> 
            (f \<longrightarrow> g \<W> h) \<and> (g \<W> h \<longrightarrow> g \<W> i) "
   using "1" "2" "3" "4" "5" "6" 
   by (metis Prop12 WaitLeftAbsorption inteq_reflection lift_imp_trans)
 from 7 8 show ?thesis by auto
qed

lemma LeftUntilWaitImp:
 "\<turnstile> (f \<U> g) \<W> h \<longrightarrow> (f \<W> g) \<W> h"
 by (meson BoxGen MP UntilImpWait WaitLeftMono)

lemma RightWaitUntilImp:
 "\<turnstile> f \<W> (g \<U> h) \<longrightarrow> f \<W> (g \<W> h)"
by (meson BoxGen MP UntilImpWait WaitRightMono)


lemma RightUntilUntilImp:
 "\<turnstile> f \<U> (g \<U> h) \<longrightarrow> f \<U> (g \<W> h)"
by (meson BoxGen MP UntilImpWait UntilRightMono)

lemma LeftUntilUntilImp:
 "\<turnstile> (f \<U> g) \<U> h \<longrightarrow> (f \<W> g) \<U> h"
by (simp add: UntilImpUntil UntilImpWait)

lemma LeftUntilOrStrengthen:
 "\<turnstile> (f \<U> g) \<U> h \<longrightarrow> (f \<or> g) \<U> h"
by (simp add: UntilImpOr UntilImpUntil)

lemma LeftWaitOrStrengthen:
 "\<turnstile> (f \<W> g) \<W> h \<longrightarrow> (f \<or> g) \<W> h"
by (meson BoxGen MP WaitImpOr WaitLeftMono)

lemma RightWaitOrStrengthen:
 "\<turnstile> f \<W> (g \<W> h) \<longrightarrow> f \<W> (g \<or> h)"
by (meson BoxGen MP WaitImpOr WaitRightMono)


lemma BoxImpBoxOr:
 "\<turnstile> \<box> f \<longrightarrow> \<box>(f \<or> g)"
by (metis BoxImpWait BoxIntro UntilBoxEqvBox UntilBoxImphelp2 WaitImpOr inteq_reflection 
     lift_imp_trans)

lemma RightWaitOrOrder:
 "\<turnstile> f \<W> (g \<W> h) \<longrightarrow> (f \<or> g) \<W> h"
proof -
 have 1: "\<turnstile> f \<W> (g \<W> h) = (\<box> f \<or> f \<U> (\<box> g \<or> g \<U> h))"
    by (simp add: wait_d_def)
 have 2: "\<turnstile> (f \<or> g) \<W> h = (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
    by (simp add: wait_d_def)
 have 3: "\<turnstile> \<box> f \<longrightarrow> (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
    using BoxImpBoxOr by fastforce
 have 4: "\<turnstile> f \<U> (\<box> g \<or> g \<U> h) = (f \<U> (\<box> g) \<or> f \<U> (g \<U> h))"
  using UntilOrDist by blast
 have 5: "\<turnstile> f \<U> (g \<U> h) \<longrightarrow> (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
    by (simp add: Prop05 UntilRightOr)
 have 6: "\<turnstile> f \<U> (\<box> g) \<longrightarrow> (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
    by (metis BoxImpBoxRule BoxImpWait UntilBoxImp UntilImpOr lift_imp_trans wait_d_def) 
 have 7: "\<turnstile> (f \<U> (\<box> g) \<or> f \<U> (g \<U> h)) \<longrightarrow> (\<box> (f \<or> g) \<or> (f \<or> g) \<U> h)"
    using "5" "6" by fastforce
 show ?thesis 
 using "1" "2" "3" "4" "7" by fastforce
qed

lemma RightWaitAndOrder:
 "\<turnstile> f \<W> (g \<and> h) \<longrightarrow> f \<W> (g \<W> h)"
by (metis Prop03 WaitAbsorptione WaitLeftDistOr inteq_reflection)

lemma UntilOrder:
 "\<turnstile> ((\<not> f) \<U> g \<or> (\<not> g) \<U> f) = \<diamond>(f \<or> g)"
proof -
 have 1: "\<turnstile> ((\<not> f) \<U> g \<or> (\<not> g) \<U> f) \<longrightarrow> \<diamond>(f \<or> g)"
    using UntilAbsorpAndDiamond UntilOrDist by fastforce 
 have 2: "\<turnstile> \<diamond>(f \<or> g) = #True \<U> (f \<or>  g) "
    by (metis DiamondEqvTrueUntil )
 have 3: "\<turnstile> #True \<U> (f \<or>  g) = (\<not> (f \<or> g) ) \<U> (f \<or> g)" 
   using "2" UntilRule by fastforce
 have 4: "\<turnstile> (\<not> (f \<or> g) ) \<U> (f \<or> g) = (\<not>f \<and> \<not>g) \<U> (f \<or> g)"
   by (metis UntilAbsorp_c int_eq int_simps(14) int_simps(33))
 have 5: "\<turnstile>  \<diamond>(f \<or> g) \<longrightarrow> (\<not> g) \<U> f \<or> (\<not> f) \<U> g"
  by (metis "2" "3" "4" UntilAndRule UntilOrDist int_iffD1 inteq_reflection lift_and_com)
 have 6: "\<turnstile> ( (\<not> g) \<U> f \<or> (\<not> f) \<U> g) = ( (\<not> f) \<U> g \<or> (\<not> g) \<U> f)"
     by fastforce
 show ?thesis  using  1 5 6 by fastforce
qed

lemma WaitOrder: 
 "\<turnstile> (\<not> f) \<W> g \<or> (\<not> g) \<W> f"
proof -
 have 1: "\<turnstile> (\<not> f) \<W> g = (\<box> (\<not> f) \<or> (\<not> f) \<U> g)"
  by (simp add: wait_d_def)
 have 2:"\<turnstile> (\<not> g) \<W> f = (\<box> (\<not> g) \<or> (\<not> g) \<U> f)" 
   by (simp add: wait_d_def)
 have 3: "\<turnstile> ((\<box> (\<not> f) \<or> (\<not> f) \<U> g) \<or> (\<box> (\<not> g) \<or> (\<not> g) \<U> f)) =
           ( (\<box> (\<not> f) \<or> \<box> (\<not> g)) \<or> ((\<not> f) \<U> g \<or> (\<not> g) \<U> f))"
     by auto
 have 4: "\<turnstile> ((\<not> f) \<U> g \<or> (\<not> g) \<U> f) = \<diamond>(f \<or> g)"
   using UntilOrder by blast
 show ?thesis 
 using "1" "2" "4" WaitOrder_help by fastforce
qed

lemma WaitImpOrder:
 "\<turnstile> f \<W> g \<and> (\<not> g) \<W> h \<longrightarrow> f \<W> h"
proof -
 have 1: "\<turnstile> f \<W> g = (\<box>f \<or> f \<U> g) "
    by (simp add: wait_d_def)
 have 2: "\<turnstile> f \<W> h = (\<box> f \<or> f \<U> h)"
    by (simp add: wait_d_def)
 have 3: "\<turnstile> (\<not> g) \<W> h = (\<box> (\<not> g) \<or> (\<not> g) \<U> h)"
   by (simp add: wait_d_def)
 have 4: "\<turnstile> \<box> f \<and> (\<box> (\<not> g) \<or> (\<not> g) \<U> h) \<longrightarrow> (\<box> f \<or> f \<U> h)"
   by auto
 have 5: "\<turnstile> (\<box>f \<or> f \<U> g) \<and> \<box> (\<not> g) \<longrightarrow> (\<box> f \<or> f \<U> h)"
    using "1" WaitAndBoxImpBox by fastforce
 have 6: "\<turnstile> f \<U> g \<and> (\<not> g) \<U> h \<longrightarrow> (\<box> f \<or> f \<U> h)"
   by (simp add: Prop05 UntilNotImp)
 have 7: "\<turnstile> \<box> f \<and> (\<not> g) \<U> h \<longrightarrow> (\<box> f \<or> f \<U> h)"
    by auto
 have 8: "\<turnstile> (\<box>f \<or> f \<U> g) \<and>  (\<not> g) \<U> h \<longrightarrow> (\<box> f \<or> f \<U> h)"
   using "6" "7" by fastforce
 have 9: "\<turnstile> (\<box>f \<or> f \<U> g) \<and> (\<box> (\<not> g)\<or> (\<not> g) \<U> h) \<longrightarrow> (\<box> f \<or> f \<U> h)"
   using "5" "8" by fastforce
 show ?thesis by (simp add: "9" wait_d_def) 
qed

end 