(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Filter on Intervals\<close>

theory IntervalFilter
 imports
   Interval
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]

*)

text \<open>
The filter operator on intervals is defined. The definition of filter is slightly more complicated than 
the one for lists as an interval has at least one state and one needs to ensure that the filter
operator always returns an interval. The lemmas involving the filter on intervals are similar to those for the
filter operator on lists only a bit more complicated.  
\<close>

subsection \<open>Definitions\<close>

definition opfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
 where "opfx xs ys = (\<exists>zs. ys = xs \<ominus> zs \<or> ys = xs)"

definition sopfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "sopfx xs ys \<longleftrightarrow> opfx xs ys \<and> xs \<noteq> ys"

interpretation opfx_order: order opfx sopfx
proof  standard
 show "\<And>x y. sopfx x y = (opfx x y \<and> \<not> opfx y x)" 
 by (auto simp add: opfx_def sopfx_def)
    (metis One_nat_def add_diff_cancel_left' diff_is_0_eq interval_intlen_intapp le_add1 nat.simps(3))
 show "\<And>x. opfx x x"
   by (auto simp add: opfx_def sopfx_def)
 show "\<And>x y z. opfx x y \<Longrightarrow> opfx y z \<Longrightarrow> opfx x z"  
  by (auto simp add: opfx_def sopfx_def)
 show "\<And>x y. opfx x y \<Longrightarrow> opfx y x \<Longrightarrow> x = y " 
  by (auto simp add: opfx_def sopfx_def)
     (metis One_nat_def add_diff_cancel_left' diff_is_0_eq interval_intlen_intapp le_add1 nat.simps(3))
qed

definition osfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "osfx xs ys = (\<exists>zs. ys = zs \<ominus> xs \<or> ys =xs)"


definition sosfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "sosfx xs ys \<longleftrightarrow> osfx xs ys \<and> xs \<noteq> ys "

interpretation osfx_order: order osfx sosfx
proof standard
 show "\<And>x y. sosfx x y = (osfx x y \<and> \<not> osfx y x)" 
  by (auto simp add: osfx_def sosfx_def)
     (metis add_diff_cancel_right' add_eq_0_iff_both_eq_0 diff_is_0_eq' interval_intapp_assoc 
       interval_intlen_intapp le_add1 not_one_le_zero)
 show "\<And>x. osfx x x" 
  by (auto simp add: osfx_def sosfx_def)
 show "\<And>x y z. osfx x y \<Longrightarrow> osfx y z \<Longrightarrow> osfx x z"
   by (auto simp add: osfx_def sosfx_def)
      (metis interval_intapp_assoc)
 show "\<And>x y. osfx x y \<Longrightarrow> osfx y x \<Longrightarrow> x = y "
   by (auto simp add: osfx_def sosfx_def)
      (metis add_diff_cancel_right' add_eq_0_iff_both_eq_0 diff_is_0_eq' interval_intapp_assoc 
      interval_intlen_intapp le_add1 not_one_le_zero)
qed

primrec distinct :: "'a interval \<Rightarrow> bool"
 where "distinct \<langle>x\<rangle> \<longleftrightarrow> True"
      | "distinct (x\<odot>xs) \<longleftrightarrow> x \<notin> set xs \<and> distinct xs"

primrec remdups :: "'a interval \<Rightarrow> 'a interval"
 where "remdups \<langle>x\<rangle> = \<langle>x\<rangle>"
      | "remdups (x\<odot>xs) = (if x \<in> set xs then remdups xs else x \<odot> remdups xs)"

primrec filter:: "('a \<Rightarrow> bool) \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
 where  "filter P \<langle>x\<rangle> = \<langle>x\<rangle>"
      | "filter P (x\<odot>xs) = (if (\<exists>y\<in>set xs. P y) then
                              (if P x then x \<odot> filter P xs else filter P xs)
                             else \<langle>x\<rangle>)"

primrec nfilter:: "('a \<Rightarrow> bool) \<Rightarrow> 'a interval \<Rightarrow> nat \<Rightarrow> nat interval" 
 where  "nfilter P \<langle>x\<rangle> n     = \<langle>n\<rangle>"
      | "nfilter P (x\<odot>xs) n = (if (\<exists>y\<in>set xs. P y) then
                                (if P x then n \<odot> (nfilter P xs (Suc n)) 
                                        else nfilter P xs (Suc n))
                               else \<langle>n\<rangle>)"


primrec prefixes :: "'a interval \<Rightarrow> 'a interval interval"
 where
    "prefixes \<langle>x\<rangle> = \<langle>\<langle>x\<rangle>\<rangle>"
  | "prefixes (x\<odot>xs) = \<langle>x\<rangle> \<odot> (map ((\<odot>) x) (prefixes xs))"


primrec suffixes :: "'a interval \<Rightarrow> 'a interval interval" 
where
  "suffixes \<langle>x\<rangle> = \<langle>\<langle>x\<rangle>\<rangle>"
| "suffixes (x\<odot>xs) = (x\<odot>xs) \<odot> (suffixes xs)"


subsection \<open>Lemmas\<close>

subsubsection \<open>opfx and sopfx\<close>

lemma opfxI [intro?]:
 assumes "ys = xs \<ominus> zs \<or> ys = xs"
 shows   "  opfx xs ys"
using assms unfolding opfx_def by blast

lemma opfxE [elim?]:
 assumes "opfx xs ys"
 obtains zs where "ys = xs \<ominus> zs \<or> ys = xs"
 using assms unfolding opfx_def by blast

lemma sopfxI' [intro?]:
 " ys = xs \<ominus> (zs) \<Longrightarrow> sopfx xs ys"
unfolding sopfx_def opfx_def 
by (metis add_le_same_cancel1 interval_intlen_intapp le_add1 not_one_le_zero)

lemma sopfxE' [elim?]:
 assumes "sopfx xs ys"
 obtains zs where "ys = xs \<ominus> zs"
 using assms  unfolding sopfx_def opfx_def by blast

lemma opfx_state [simp]:
 "opfx \<langle>intfirst xs\<rangle> xs"
unfolding opfx_def intfirst_def 
by (metis intapp_St interval_hd_tail interval_intfirst_suffix interval_suffix_intlen 
    interval_suffix_zero le0 order.order_iff_strict)

lemma opfx_snoc [simp]:
 "opfx xs (ys \<ominus> \<langle>y\<rangle>) \<longleftrightarrow> xs = ys\<ominus>\<langle>y\<rangle> \<or> opfx xs ys"
unfolding opfx_def 
by (metis interval_intapp_eq_intapp_conv2 interval_intapp_not_state) 

lemma cons_pfx_cons [simp]:
 " opfx (x\<odot>xs) (y\<odot>ys) = (x =y \<and> opfx xs ys)"
by (auto simp add: opfx_def)

lemma opfx_code [code]:
 "opfx \<langle>intfirst xs\<rangle> xs \<longleftrightarrow> True"
 "opfx (x\<odot>xs) \<langle>y\<rangle> \<longleftrightarrow> False"
 "opfx (x\<odot>xs) (y\<odot>ys) \<longleftrightarrow> (x=y \<and> opfx xs ys)"
proof simp_all
 show "opfx \<langle>nth xs 0\<rangle> xs"  
 using opfx_state by auto
 show "\<not> opfx (x \<odot> xs) \<langle>y\<rangle> " 
 by (simp add: opfx_def)
qed

lemma same_opfx_opfx [simp]:
 "opfx (xs \<ominus> ys) (xs\<ominus>zs) = opfx ys zs"
by (induct xs) simp_all

lemma same_opfx_state [simp]:
 " opfx (xs \<ominus> ys) (xs \<ominus> \<langle>y\<rangle>) =  (ys = \<langle>y\<rangle>)"
by (meson interval_same_intapp_eq opfx_order.less_le_not_le opfx_snoc sopfxI')

lemma opfx_opfx [simp]:
 assumes "opfx xs ys"
 shows   "  opfx xs (ys\<ominus>zs)"
using assms unfolding opfx_def by fastforce

lemma intapp_opfxD:
 assumes "opfx (xs \<ominus> ys) zs"
 shows   "   opfx xs zs"
using assms by (auto simp add: opfx_def)

lemma opfx_cons: 
 "opfx xs (y\<odot>ys) = (xs = \<langle>y\<rangle> \<or> (\<exists> zs.  xs = y\<odot>zs \<and> opfx zs ys))"
by (case_tac xs) (auto simp add: opfx_def)

lemma opfx_intapp:
 "opfx xs (ys\<ominus>zs) = (opfx xs ys \<or> (\<exists> us. xs = ys \<ominus> us \<and> opfx us zs))"
proof (induct zs rule: interval_rev_induct)
case (St y)
then show ?case by (meson opfx_snoc same_opfx_opfx)
next
case (snoc x xs)
then show ?case by (metis interval_intapp_assoc opfx_snoc)
qed

lemma intapp_one_opfx:
assumes "opfx xs ys"
        "intlen xs < intlen ys"
  shows "opfx (xs\<ominus> \<langle>nth ys ((intlen xs)+1)\<rangle>) ys"
using assms
proof (unfold opfx_def)
 assume 1: "\<exists>zs. ys = xs \<ominus> zs \<or> ys = xs" 
   then obtain sk where 2: " ys = xs \<ominus> sk \<or> ys = xs"
   by fastforce
 assume 3: "intlen xs < intlen ys"
 have  4: "intlen sk \<ge> 0"
   by auto
 have 5: "\<exists> v. ys = xs \<ominus> ((intfirst sk) \<odot> v) \<or> ys = xs \<ominus> \<langle>intfirst sk\<rangle> "
    by (metis "2" "3" "4" interval_hd_tail interval_nth_zero_intfirst interval_suffix_intlen 
        interval_suffix_zero le_eq_less_or_eq nat_neq_iff)
 have 6: "ys \<noteq> xs"
     using "3" by blast
 have 7: " (intfirst sk) = nth (xs \<ominus> sk) ((intlen xs) + 1) " 
    by (simp add: interval_intapp_nth)
 have 8: "  ys = xs \<ominus> sk"
   using "2" "6" by auto
 have 9: "nth (xs \<ominus> sk) ((intlen xs) + 1) = nth (ys) ((intlen xs) + 1)" 
   using "8" by blast  
 thus "\<exists>zs. ys = (xs \<ominus> \<langle>nth ys ((intlen xs) + 1)\<rangle>) \<ominus> zs \<or> ys = xs \<ominus> \<langle>nth ys ((intlen xs) + 1)\<rangle>"
     using "5" "7" "9"  
   by simp
qed   

lemma opfx_intlen_le:
 assumes "opfx xs ys"
 shows   "intlen xs \<le> intlen ys"
 using assms by (auto simp add: opfx_def)

lemma opfx_same_cases:
 assumes "opfx xs1 ys"
         "opfx xs2 ys"
 shows   "opfx xs1 xs2 \<or> opfx xs2 xs1"
using assms unfolding opfx_def using interval_intapp_eq_intapp_conv2 
by metis

lemma opfx_intlen_opfx:
 assumes "opfx ps xs"
         "opfx qs xs"
         "intlen ps \<le> intlen qs"
 shows   "opfx ps qs"
using assms 
by (auto simp: opfx_def) 
   (metis assms(2) dual_order.antisym interval_intapp_eq_intapp_conv opfx_intapp opfx_intlen_le)

lemma set_mono_opfx:
 assumes "opfx xs ys"
 shows   "set xs \<subseteq> set ys"
using assms by (auto simp: opfx_def)

lemma prefix_is_opfx:
 "opfx (prefix n xs) xs"
by (auto simp: opfx_def) 
 (metis Suc_eq_plus1 add_diff_inverse_nat interval_intapp_prefix_suffix interval_prefix_intlen_gr_1 
        less_Suc_eq_0_disj less_Suc_eq_le not_less)

lemma map_mono_opfx:
 assumes "opfx xs ys"
 shows   "opfx (map f xs) (map f ys)"
using assms by (auto simp: opfx_def)

lemma opfx_intlen_less:
 assumes "sopfx xs ys"
 shows   "intlen xs < intlen ys"
using assms by (auto simp: sopfx_def opfx_def)

lemma opfx_snocD:
 assumes "opfx (xs\<ominus>\<langle>x\<rangle>) ys"
 shows   "sopfx xs ys"
using assms intapp_opfxD opfx_order.antisym sopfxI' sopfx_def by blast

lemma sopfx_simps [simp, code]:
 "sopfx xs \<langle>y\<rangle> \<longleftrightarrow> False"
 "sopfx \<langle>x\<rangle> (x\<odot>xs) \<longleftrightarrow> True"
 "sopfx (x\<odot>xs) (y\<odot>ys) \<longleftrightarrow> x=y \<and> sopfx xs ys"
proof (simp_all add: sopfx_def)
 show "opfx xs \<langle>y\<rangle> \<longrightarrow> xs = \<langle>y\<rangle>" 
   by (metis interval_intapp_not_state opfxE)
 show "opfx \<langle>x\<rangle> (x \<odot> xs)" 
   by (simp add: opfx_cons)
 show "(x = y \<and> opfx xs ys \<and> (x = y \<longrightarrow> xs \<noteq> ys)) = (x = y \<and> opfx xs ys \<and> xs \<noteq> ys)" 
   by auto
qed

lemma prefix_sopfx:
 assumes "sopfx xs ys"
 shows   "sopfx (prefix n xs) ys"
using assms 
proof (induct n arbitrary: xs ys)
case 0
then show ?case 
  proof (cases ys)
  case (St x1)
  then show ?thesis 
   using "0.prems" by auto
  next
  case (Cons x21 x22)
  then show ?thesis 
   using "0.prems" opfx_order.le_less_trans prefix_is_opfx by blast
  qed
next
case (Suc n)
then show ?case 
 using opfx_order.order.strict_trans1 prefix_is_opfx by blast
qed

lemma osfxI [intro?]:
 assumes "ys = zs \<ominus> xs \<or> ys =xs"
 shows   "  osfx xs ys"
using assms unfolding osfx_def by blast

lemma osfxE [elim?]:
 assumes "osfx xs ys"
 obtains zs where "ys= zs \<ominus> xs \<or> ys = xs"   
  using assms unfolding osfx_def by blast

lemma osfx_tl [simp]:
 assumes "intlen xs >0"
 shows  "osfx (suffix 1 xs) xs"
using assms 
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by (auto simp: osfx_def) (metis intapp.simps(1))
qed


lemma osfx_suffix [simp]:
assumes "i\<le> intlen xs"
 shows  "osfx (suffix i xs) xs"
using assms 
proof (cases i)
case 0
then show ?thesis by simp
next
case (Suc nat)
then show ?thesis 
  proof (cases xs)
  case (St x1)
  then show ?thesis unfolding osfx_def by simp
  next
  case (Cons x21 x22)
  then show ?thesis unfolding osfx_def 
  by (metis Suc Suc_eq_plus1 assms interval_intapp_prefix_suffix less_le_trans zero_less_Suc)
  qed
qed

lemma osfx_prefix [simp]:
assumes  "osfx xs ys"
 shows   "ys = (prefix (intlen ys - intlen xs -1) ys) \<ominus> xs   \<or> ys =xs" 
using assms 
by (auto simp add: osfx_def)
   (metis diff_zero interval_prefix_intapp interval_prefix_intlen)


lemma osfx_snoc [simp]:
 "osfx xs (ys\<ominus> \<langle>y\<rangle>)  \<longleftrightarrow> 
  xs = \<langle>y\<rangle> \<or> (\<exists>zs. xs = zs \<ominus> \<langle>y\<rangle> \<and> osfx zs ys)  "
proof (cases xs)
case (St x1)
then show ?thesis by (metis interval_intlast_intapp interval_intlast_intapp2 osfxI osfx_prefix)
next
case (Cons x21 x22)
then show ?thesis 
  proof auto
   show "xs = x21 \<odot> x22 \<Longrightarrow> osfx (x21 \<odot> x22) (ys \<ominus> \<langle>y\<rangle>) \<Longrightarrow> \<exists>zs. x21 \<odot> x22 = zs \<ominus> \<langle>y\<rangle> \<and> osfx zs ys" 
    using interval_intapp_eq_intapp_conv2 interval_intapp_not_state unfolding osfx_def 
       by (metis interval.distinct(1) )
   show "\<And>zs. xs = zs \<ominus> \<langle>y\<rangle> \<Longrightarrow> x21 \<odot> x22 = zs \<ominus> \<langle>y\<rangle> \<Longrightarrow> osfx zs ys \<Longrightarrow> osfx (zs \<ominus> \<langle>y\<rangle>) (ys \<ominus> \<langle>y\<rangle>)  "
    unfolding osfx_def by (metis interval_intapp_assoc )
    qed
qed

lemma snoc_osfx_snoc [simp]: 
"osfx (xs \<ominus> \<langle>x\<rangle>) (ys \<ominus> \<langle>y\<rangle>) = (x = y \<and> osfx xs ys)"
by (simp add: osfx_def)
   (metis interval_intapp_assoc interval_intapp_eq_conv)

lemma same_osfx_osfx [simp]: 
"osfx (ys \<ominus> xs) (zs \<ominus> xs) = osfx ys zs"
unfolding osfx_def 
by (metis interval_intapp_assoc interval_intapp_same_eq)

lemma same_suffix_state [simp]: 
"osfx (ys \<ominus> xs) (x\<odot>xs) = (ys = \<langle>x\<rangle>)"
unfolding osfx_def 
by (metis intapp_St interval_intapp_assoc interval_intapp_not_state interval_intapp_same_eq)

lemma osfx_cons:
 "osfx xs (y\<odot>ys) \<longleftrightarrow> xs = y\<odot>ys \<or> osfx xs ys"
unfolding osfx_def 
by (auto simp: interval_cons_eq_intapp_conv)

lemma osfx_intapp:
 "osfx xs (ys \<ominus> zs) \<longleftrightarrow> 
   osfx xs zs  \<or> (\<exists>xs'. xs = xs' \<ominus> zs \<and> osfx xs' ys)"
by (auto simp: osfx_def interval_intapp_eq_intapp_conv2)

lemma osfx_intlen:
assumes "osfx xs ys"
 shows   "  intlen xs \<le> intlen ys"
using assms by (auto simp add: osfx_def)

lemma osfx_same_cases:
assumes "osfx xs\<^sub>1 ys"
        " osfx xs\<^sub>2 ys"
shows   "  osfx xs\<^sub>1 xs\<^sub>2 \<or> osfx xs\<^sub>2 xs\<^sub>1"
using assms unfolding osfx_def by (metis interval_intapp_eq_intapp_conv2)

lemma osfx_intlen_osfx:
assumes "osfx ps xs"
        "osfx qs xs"
        "intlen ps \<le> intlen qs"
shows   "osfx ps qs"
using assms 
by (auto simp: osfx_def interval_intapp_eq_intapp_conv2)

lemma osfx_intlen_less:
assumes "sosfx xs ys"
shows   "intlen xs < intlen ys"
using assms by (auto simp: sosfx_def osfx_def)

lemma osfx_ConsD':
assumes "osfx (x\<odot>xs) ys"
shows   "sosfx xs ys"
using assms 
by (simp add: sosfx_def osfx_def )
   (metis interval_intapp_assoc interval_rev_eq_cons_iff interval_rev_intapp 
    interval_rev_rev_ident opfx_order.dual_order.strict_iff_order sopfxI')


lemma suffix_sosfx:
assumes "sosfx xs ys"
shows   "sosfx (suffix n xs) ys"
using assms 
proof (induct n arbitrary: xs ys)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
    proof (cases xs)
    case (St x1)
    then show ?thesis using Suc.prems by auto
    next
    case (Cons x21 x22)
    then show ?thesis 
    using Suc.hyps Suc.prems osfx_ConsD' osfx_order.less_imp_le by fastforce
    qed
qed

lemma sosfx_tl [simp]:
 assumes "intlen xs >0"
 shows   "sosfx (suffix 1 xs) xs"
using assms 
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case using osfx_ConsD' by force
qed

lemma state_osfx [simp]:
 " osfx \<langle>intlast xs\<rangle> xs "
using osfx_suffix by fastforce

lemma osfx_state [simp]:
 " (osfx xs \<langle>intlast xs\<rangle>) = (xs = \<langle>intlast xs\<rangle>)"
using osfx_order.antisym state_osfx by auto

lemma osfx_ConsI:
assumes "osfx xs ys"
shows   "osfx xs (x\<odot> ys)" 
using assms 
by (metis interval.inject(2) interval_hd_tail intlen.simps(2) le_add1 
      osfx_suffix plus_1_eq_Suc osfx_order.order.trans zero_less_Suc)

lemma osfx_ConsD:
assumes "osfx (x\<odot>xs) ys"
shows   "osfx xs ys"
using assms 
by (meson osfx_ConsD' osfx_order.less_imp_le)

lemma osfx_intappI:
assumes "osfx xs ys"
shows   "osfx xs (zs \<ominus> ys)"
using assms by (metis interval_intapp_assoc osfx_def)

lemma osfx_intappD:
assumes "osfx (zs \<ominus> xs) ys"
shows   "osfx xs ys"
using assms osfxI osfx_order.dual_order.trans by blast

lemma sosfx_set_subset:
assumes "sosfx xs ys"
shows   "set xs \<subseteq> set ys"
using assms by (auto simp: sosfx_def osfx_def)

lemma set_mono_osfx:
assumes "osfx xs ys"
shows   "set xs \<subseteq> set ys"
using assms by (auto simp: osfx_def)

lemma osfx_ConsD2:
assumes "osfx (x\<odot>xs) (y\<odot>ys)"
shows   "osfx xs ys"
using assms 
proof -
 assume "osfx (x\<odot>xs) (y\<odot>ys)"
 then obtain zs where "y\<odot>ys = zs \<ominus> (x\<odot>xs) \<or> y\<odot>ys = x\<odot>xs" 
   using osfxE by blast
 then show ?thesis  
 by (metis assms interval.inject(2) osfx_ConsD osfx_cons)
qed

lemma osfx_to_opfx [code]:
 "osfx xs ys \<longleftrightarrow> opfx (intrev xs) (intrev ys)"
unfolding opfx_def 
by (metis interval_rev_intapp interval_rev_rev_ident osfx_def)

lemma sosfx_to_sopfx [code]:
 "sosfx xs ys \<longleftrightarrow> sopfx (intrev xs) (intrev ys)"
by (auto simp: osfx_to_opfx sosfx_def sopfx_def)


lemma map_mono_osfx:
assumes "osfx xs ys"
shows   "osfx (map f xs) (map f ys)"
using assms by (auto elim!: osfxE intro: osfxI)

lemma prefix_subset:
assumes "k\<le>intlen xs"
 shows  " set (prefix k xs) \<le> set xs"
using assms by (simp add: prefix_is_opfx set_mono_opfx)

lemma suffix_subset:
assumes "k\<le>intlen xs"
shows   " set (suffix k xs) \<le> set xs"
using assms by (simp add: set_mono_osfx)

subsubsection \<open>distinct and remdups\<close>

lemma distinct_intapp [simp]:
 "distinct (xs\<ominus>ys) = (distinct xs \<and> distinct ys \<and> set xs \<inter> set ys = {})"
by (induct xs)  auto

lemma distinct_osfx:
assumes "distinct ys"
        "osfx xs ys"
shows   "distinct xs"
using assms 
proof (clarsimp elim!: osfxE)
 show "\<And>zs. distinct ys \<Longrightarrow> ys = zs \<ominus> xs \<or> ys = xs \<Longrightarrow> distinct xs" 
 using distinct_intapp by blast
qed

lemma distinct_tl:
assumes "distinct xs"
        "intlen xs > 0"
shows   "distinct (suffix 1 xs)"
using assms 
by (cases xs) auto

lemma distinct_intrev [simp]:
 "distinct (intrev xs) = distinct xs"
by (induct xs)  auto

lemma set_remdups [simp]:
 "set (remdups xs) = set xs"
by (induct xs) auto

lemma distinct_remdups [iff]:
 "distinct (remdups xs)"
by (induct xs) auto

lemma distinct_remdups_id: 
assumes  "distinct xs"
shows    "remdups xs = xs"
using assms 
by (induct xs) auto
 
lemma remdups_id_iff_distinct [simp]: 
 "remdups xs = xs \<longleftrightarrow> distinct xs"
by (metis distinct_remdups distinct_remdups_id)

lemma distinct_map:
 "distinct (map f xs) = (distinct xs \<and> inj_on f (set xs))"
by (induct xs) auto

lemma distinct_prefix [simp]:
assumes "distinct xs"
shows   "distinct (prefix i xs)"
using assms 
proof (induct xs arbitrary: i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis 
  using Cons.hyps Cons.prems prefix_is_opfx set_mono_opfx by fastforce
  qed
qed

lemma distinct_suffix [simp]:
assumes "distinct xs"
shows   "distinct (suffix i xs)"
using assms 
proof 
 (induct xs arbitrary: i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis using Cons.prems by auto
   next
   case (Suc nat)
   then show ?thesis using interval_intapp_prefix_suffix  Cons.hyps Cons.prems by auto
   qed
qed


lemma distinct_conv_nth:
 "distinct xs = 
  (\<forall> i \<le> intlen xs. (\<forall> j \<le> intlen xs. i \<noteq> j \<longrightarrow> nth xs i \<noteq> nth xs j))"
proof 
 (induction xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   by (auto simp add: Cons nth_Cons interval_nth_and_set split: nat.split_asm) 
      blast+
qed

lemma distinct_Ex1:
assumes "distinct xs"
        "x \<in> set xs"
shows   "(\<exists>!i. i\<le> intlen xs \<and> (nth xs i) = x)"
using assms
by (metis distinct_conv_nth interval_nth_and_set)


lemma inj_on_nth:
assumes "distinct xs"
shows   "(\<forall> i \<in> I. i \<le> intlen xs \<Longrightarrow> inj_on (nth xs) I)"
using assms 
by (meson distinct_conv_nth inj_onI)

lemma bij_betw_nth:
 assumes "distinct xs"
         "A = {..< intlen xs+1}"
         "B = set xs"
 shows   "bij_betw ((nth) xs) A B"
 using assms unfolding bij_betw_def 
proof (auto intro!: inj_on_nth simp: set_conv_nth)
   show "\<And>xa. distinct xs \<Longrightarrow>
          A = {..<Suc (intlen xs)} \<Longrightarrow>
          B = set xs \<Longrightarrow> xa < Suc (intlen xs) \<Longrightarrow> nth xs xa \<in> set xs" 
     by (meson interval_nth_and_set less_Suc_eq_le)
   show "\<And>x. distinct xs \<Longrightarrow>
         A = {..<Suc (intlen xs)} \<Longrightarrow>
         B = set xs \<Longrightarrow> x \<in> set xs \<Longrightarrow> x \<in> nth xs ` {..<Suc (intlen xs)} " 
     using interval_nth_and_set by (metis image_iff lessThan_iff less_Suc_eq_le)
qed

lemma card_distinct:
 assumes "card (set xs) = intlen xs +1"
 shows   "distinct xs"
using assms 
proof 
 (induct xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  proof (cases "x1a \<in> set xs")
  case True
  then show ?thesis 
  by (metis Cons.prems remdups.simps(2) set_remdups add_Suc interval_card_intlen intlen.simps(2) 
      not_less_eq_eq plus_1_eq_Suc)
  next
  case False
  then show ?thesis using Cons.hyps Cons.prems by auto
  qed
qed


lemma finite_interval:
 assumes "finite A"
 shows   "(A\<noteq> {} \<longrightarrow> (\<exists>xs. set xs = A)) "
using assms 
proof (induct rule:finite_induct)
case empty
then show ?case by simp
next
case (insert x F)
then show ?case by (metis insert_is_Un interval.set(1) interval_set_intapp)
qed

lemma finite_distinct_interval:
assumes "finite A"
        "A\<noteq> {}"
shows   " (\<exists> xs. set xs = A \<and> distinct xs)"
using assms by (metis distinct_remdups finite_interval set_remdups)

lemma remdups_eq_state_iff [simp]: 
 "(remdups xs = \<langle>x\<rangle>) = (\<forall> i \<le> intlen xs. (nth xs i) = x )"
proof 
 (induct xs)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
 proof -
  have 1: "(\<forall>i\<le>intlen (x1a \<odot> xs). nth (x1a \<odot> xs) i = x) =
           ( x1a = x \<and> (\<forall>i\<le>intlen (xs). nth (xs) i = x)) "
      by auto
         (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
  have 2: "(remdups (x1a \<odot> xs) = \<langle>x\<rangle>) = 
           ((if x1a \<in> set xs then remdups xs else x1a \<odot> remdups xs) = \<langle>x\<rangle>)" 
       by simp
  have 3: "((if x1a \<in> set xs then remdups xs else x1a \<odot> remdups xs) = \<langle>x\<rangle>) = 
           (remdups xs  = \<langle>x\<rangle> \<and> x1a \<in> set xs)"
    by auto 
  have 4: "(remdups xs  = \<langle>x\<rangle> \<and> x1a \<in> set xs) =
           ((\<forall>i\<le>intlen (xs). nth (xs) i = x) \<and> x1a \<in> set xs)  "
    using Cons.hyps by blast
  have 5: "(\<forall>i\<le>intlen (xs). nth (xs) i = x) \<and> x1a \<in> set xs \<longrightarrow> x1a = x " 
           by (metis interval_nth_and_set)
  show ?thesis 
  by (metis "1" "2" "3" "5" Cons.hyps set_remdups interval.set_intros(1))
  qed 
qed

lemma remdups_eq_state_right_iff [simp]: 
 "(\<langle>x\<rangle> = remdups xs) = (\<forall> i \<le> intlen xs. (nth xs i) = x )"
by (metis remdups_eq_state_iff)

lemma length_remdups_leq [iff]: 
 "intlen(remdups xs) \<le> intlen xs"
by (induct xs) auto

lemma length_remdups_eq[iff]:
  "(intlen (remdups xs) = intlen xs) = (remdups xs = xs)"
proof 
 (induct xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
 by (metis length_remdups_leq remdups.simps(2) add_left_cancel intlen.simps(2) not_less_eq_eq 
     plus_1_eq_Suc)
qed

lemma distinct_card:
assumes "distinct xs"
shows   "card(set xs) = Suc(intlen xs)"
using assms
by (induct xs) simp_all

subsubsection \<open>prefixes and suffixes\<close>

lemma in_set_prefixes [simp]:
 "xs \<in> set (prefixes ys) \<longleftrightarrow> opfx xs ys"
proof 
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case 
   proof (cases ys)
   case (St x1)
   then show ?thesis using opfxE by force
   next
   case (Cons x21 x22)
   then show ?thesis unfolding opfx_def by auto
   qed
 next
 case (Cons x1a xs)
 then show ?case 
    proof (cases ys)
    case (St x1)
    then show ?thesis by (simp add: opfx_code(2))
    next
    case (Cons x21 x22)
    then show ?thesis 
      proof (auto simp add: opfx_def)
        show "ys = x21 \<odot> x22 \<Longrightarrow> 
              xs \<in> interval.set (prefixes x22) \<Longrightarrow> 
              x1a = x21 \<Longrightarrow> 
              x22 \<noteq> xs \<Longrightarrow> 
              \<exists>zs. x22 = xs \<ominus> zs" 
          using Cons.hyps opfxE by blast
        show "\<And>zs. ys = x1a \<odot> xs \<ominus> zs \<Longrightarrow> 
                   x21 = x1a \<Longrightarrow> x22 = xs \<ominus> zs \<Longrightarrow> 
                   x1a \<odot> xs \<in> (\<odot>) x1a ` interval.set (prefixes (xs \<ominus> zs))" 
          by (simp add: Cons.hyps)
        show "ys = x1a \<odot> xs \<Longrightarrow> 
              x22 = xs \<Longrightarrow> 
              x21 = x1a \<Longrightarrow> 
              x1a \<odot> xs \<in> (\<odot>) x1a ` interval.set (prefixes xs) " 
          using Cons.hyps by blast
       qed
    qed
qed


lemma intlen_prefixes [simp]:
 "intlen (prefixes xs) = intlen xs"
by (induction xs)  auto

lemma distinct_prefixes [intro]:
 "distinct (prefixes xs)"
proof (induction xs)
case (St x)
then show ?case by (auto simp: distinct_map)
next
case (Cons x1a xs)
then show ?case by (auto simp: distinct_map)
                   (meson inj_onI interval.inject(2))
qed 


lemma prefixes_snoc [simp]:
 "prefixes (xs\<ominus>\<langle>x\<rangle>) = (prefixes xs) \<ominus> \<langle>xs\<ominus>\<langle>x\<rangle>\<rangle>"
by (induction xs) auto

lemma intfirst_prefixes [simp]:
 "intfirst (prefixes xs) = \<langle>intfirst xs\<rangle>"
by (cases xs)  auto

lemma intlast_prefixes [simp]:
 "intlast (prefixes xs) = xs"
by (induction xs) 
   (simp_all add: intlast_map interval_nth_map)

lemma prefixes_intapp:
 "prefixes (xs\<ominus>ys) = 
  prefixes xs \<ominus> map (\<lambda>ys'. xs \<ominus> ys') (prefixes ys)"
proof 
 (induction xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof -
    have 1: "prefixes ((x1a \<odot> xs) \<ominus> ys) = \<langle>x1a\<rangle> \<odot> (map ((\<odot>) x1a) (prefixes (xs \<ominus> ys))) "
       by simp
    have 2: "prefixes (xs \<ominus> ys) = prefixes xs \<ominus> map ((\<ominus>) xs) (prefixes ys)"
       by (simp add: Cons.IH) 
    have 3: "\<langle>x1a\<rangle> \<odot> (map ((\<odot>) x1a) (prefixes (xs \<ominus> ys))) =
             \<langle>x1a\<rangle> \<odot> (map ((\<odot>) x1a) (prefixes xs \<ominus> map ((\<ominus>) xs) (prefixes ys)))
             " 
        using "2" by auto
    show ?thesis by (simp add: "3")
   qed
qed

lemma prefixes_eq_snoc:
 "prefixes ys = xs \<ominus> \<langle>x\<rangle> \<longleftrightarrow>
  (\<exists> z zs. ys = zs\<ominus>\<langle>z\<rangle> \<and> xs = prefixes zs) \<and> x = ys"
proof (cases ys rule: interval_rev_cases)
case (St x)
then show ?thesis using interval_intapp_not_state by (metis prefixes.simps(1))
next
case (snoc ys y)
then show ?thesis by auto
qed 

lemma set_prefixes_eq:
 "set (prefixes xs) = {ys. opfx ys xs}"
by auto

lemma card_set_prefixes [simp]:
 "card (set (prefixes xs)) = Suc (intlen xs)"
by (simp add: distinct_card distinct_prefixes)

lemma set_prefixes_append:
 "set (prefixes (xs\<ominus>ys)) = set (prefixes xs) \<union> {xs \<ominus> ys' | ys'. ys' \<in> set (prefixes ys)}"
by (subst prefixes_intapp) auto 
 

lemma in_set_suffixes [simp]:
 "xs \<in> set(suffixes ys) \<longleftrightarrow> osfx xs ys"
proof (induct ys)
case (St x)
then show ?case using osfx_prefix by fastforce
next
case (Cons x1a ys)
then show ?case by (simp add: osfx_cons)
qed

lemma interval_sfx_state:
 "set(suffixes \<langle>x\<rangle>) = {\<langle>x\<rangle>}"
by simp

lemma interval_sfx_cons:
 "set(suffixes (x\<odot>xs)) = {x\<odot>xs} \<union> set(suffixes xs)"
by auto

lemma set_suffixes_sfx:
 "set (suffixes xs) = {suffix i xs| i. i\<le>intlen xs}"
proof 
 (induction xs)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
     proof auto
      show "set (suffixes xs) = {suffix i xs |i. i \<le> intlen xs} \<Longrightarrow>
            \<exists>i. x1a \<odot> xs = (case i of 0 \<Rightarrow> x1a \<odot> xs | Suc m \<Rightarrow> suffix m xs) \<and> i \<le> Suc (intlen xs)" 
      by force
      show "\<And>i. set (suffixes xs) = {suffix i xs |i. i \<le> intlen xs} \<Longrightarrow>
         i \<le> intlen xs \<Longrightarrow> \<exists>ia. suffix i xs = (case ia of 0 \<Rightarrow> x1a \<odot> xs | Suc m \<Rightarrow> suffix m xs) \<and> 
                            ia \<le> Suc (intlen xs)" 
      by force 
      show "\<And>i. set (suffixes xs) = {suffix i xs |i. i \<le> intlen xs} \<Longrightarrow>
         (case i of 0 \<Rightarrow> x1a \<odot> xs | Suc m \<Rightarrow> suffix m xs) \<noteq> x1a \<odot> xs \<Longrightarrow>
         i \<le> Suc (intlen xs) \<Longrightarrow> \<exists>ia. (case i of 0 \<Rightarrow> x1a \<odot> xs | Suc m \<Rightarrow>
           suffix m xs) = suffix ia xs \<and> ia \<le> intlen xs" 
         by (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
      qed
qed

lemma set_sfx_exists:
 " (\<exists>i\<le>intlen xs. f (suffix i xs)) = (set (suffixes xs) \<inter> {ys. f(ys)} \<noteq> {})"
proof rule+
 show "\<exists>i\<le>intlen xs. f (suffix i xs) \<Longrightarrow> set (suffixes xs) \<inter> {ys. f ys} = {} \<Longrightarrow> False" 
 by (meson disjoint_iff_not_equal in_set_suffixes mem_Collect_eq osfx_suffix)
 show "set (suffixes xs) \<inter> {ys. f ys} \<noteq> {} \<Longrightarrow> \<exists>i\<le>intlen xs. f (suffix i xs) " 
  using in_set_suffixes mem_Collect_eq set_suffixes_sfx  by auto force
qed  

lemma set_suffixes_osfx:
 "set(suffixes xs) = {ys. osfx ys xs}"
by auto


lemma distinct_suffixes [intro]:
 " distinct(suffixes xs)"
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
by auto (metis Suc_n_not_n intlen.simps(2) osfx_ConsD osfx_cons osfx_order.antisym plus_1_eq_Suc)
qed

lemma intlen_suffixes [simp]:
 "intlen (suffixes xs) = intlen xs"
by (induct xs) simp_all

lemma suffixes_snoc [simp]:
"suffixes (xs \<ominus> \<langle>x\<rangle>) = (map (\<lambda> ys. ys \<ominus> \<langle>x\<rangle>) (suffixes xs)) \<ominus> \<langle>\<langle>x\<rangle>\<rangle>" 
by (induct xs) simp_all

lemma intfirst_suffixes [simp]:
 "intfirst (suffixes xs) = xs"
by (induct xs) simp_all

lemma intlast_suffixes [simp]:
"intlast (suffixes xs) = \<langle>intlast xs\<rangle>"
by (induct xs) simp_all


lemma suffixes_intapp:
 "suffixes (xs \<ominus> ys) = map (\<lambda> xs'. xs' \<ominus> ys) (suffixes xs) \<ominus> (suffixes ys)"
proof 
 (induction xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a ys)
 then show ?case 
   proof (cases ys)
   case (St x1)
   then show ?thesis by simp
   next
   case (Cons x21 x22)
   then show ?thesis by (auto simp add: Cons.IH)
   qed
qed

lemma card_set_suffixes [simp]:
" card (set (suffixes xs)) = Suc (intlen xs)"
by (simp add: distinct_card distinct_suffixes)

lemma set_suffixes_intapp:
"set (suffixes (xs\<ominus>ys)) = {xs' \<ominus> ys | xs'. xs' \<in> set (suffixes xs)} \<union> set (suffixes ys)"
proof (subst suffixes_intapp)
 show "set (map (\<lambda>xs'. xs' \<ominus> ys) (suffixes xs) \<ominus> suffixes ys) =
      {xs' \<ominus> ys |xs'. xs' \<in> set (suffixes xs)} \<union> set (suffixes ys) " 
 proof (cases xs)
 case (St x1)
 then show ?thesis by simp
 next
 case (Cons x21 x22)
 then show ?thesis by force
 qed
qed

lemma map_first_suffixes [simp]:
 "map (\<lambda> xs. nth xs 0) (suffixes xs) = xs"
by (induct xs) auto

lemma suffixes_conv_prefixes:
 " (suffixes xs) = intrev (map intrev (prefixes (intrev xs)))"
by (induction xs) auto

lemma prefixes_conv_suffixes:
 " (prefixes xs) = intrev (map intrev (suffixes (intrev xs)))"
by (induction xs) (auto simp add: intrev_map)

lemma prefixes_intrev:
 " prefixes (intrev xs) = intrev (map intrev (suffixes xs))"
by (induction xs) auto

lemma suffixes_intrev:
 " suffixes (intrev xs) = intrev (map intrev (prefixes xs))" 
by (induction xs) (auto simp add: intrev_map)

lemma nth_suffixes:
 assumes " i\<le> intlen(suffixes xs)"
 shows   " nth (suffixes xs) i = (suffix i xs)"
using assms 
proof (induct xs arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by (simp add: Nitpick.case_nat_unfold)
qed


lemma suffix_suffixes:
 assumes "i \<le> intlen (suffixes xs)"
 shows   "suffix i (suffixes xs) = suffixes (suffix i xs)"
using assms 
proof (induct xs arbitrary:i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by (simp add: Nitpick.case_nat_unfold) 
qed


subsubsection \<open>filter and nfilter\<close>

lemma sfxfilter_intlen_a:
 assumes "\<exists> ys \<in> set (suffixes \<langle>x\<rangle>). P ys"
 shows   "intlen (filter P (suffixes \<langle>x\<rangle>)) = 0"
using assms by simp

lemma sfxfilter_intlen_b:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
          "(\<exists> ys \<in> set (suffixes xs). P ys)"
          "P (x\<odot>xs)"
 shows   " intlen (filter P (suffixes (x\<odot>xs))) = intlen(filter P (suffixes xs))+1"
using assms by simp 

lemma sfxfilter_intlen_c:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set (suffixes xs). P ys)"
         "\<not> P (x\<odot>xs)"
 shows   " intlen (filter P (suffixes (x\<odot>xs))) = intlen(filter P (suffixes xs))"
using assms by simp

lemma sfxfilter_intlen_d:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set (suffixes xs). P ys)"
 shows   "intlen (filter P (suffixes (x\<odot>xs))) = 0"
using assms by simp

lemma filter_intlen_a:
 assumes "\<exists> ys \<in> set \<langle>x\<rangle>. P ys"
 shows   "intlen (filter P \<langle>x\<rangle>) = 0"
using assms by simp 

lemma nfilter_intlen_a:
 assumes "\<exists> ys \<in> set \<langle>x\<rangle>. P ys"
 shows   "intlen (nfilter P \<langle>x\<rangle> n) = 0"
using assms by simp 

lemma filter_intlen_b:
 assumes "(\<exists> ys \<in> set (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> set (xs). P ys)"
         "P (x)"
 shows   " intlen (filter P (x\<odot>xs)) = intlen(filter P (xs))+1"
using assms by simp

lemma nfilter_intlen_b:
 assumes "(\<exists> ys \<in> set (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> set (xs). P ys)"
         "P (x)"
 shows   " intlen (nfilter P (x\<odot>xs) n) = intlen(nfilter P (xs) (Suc n))+1"
using assms by simp

lemma filter_intlen_c:
 assumes "(\<exists> ys \<in> set (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> set (xs). P ys)"
         "\<not> P (x)"
 shows   " intlen (filter P (x\<odot>xs)) = intlen(filter P (xs))"
using assms by simp

lemma nfilter_intlen_c:
 assumes "(\<exists> ys \<in> set (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> set (xs). P ys)"
         "\<not> P (x)"
 shows   " intlen (nfilter P (x\<odot>xs) n) = intlen(nfilter P (xs) (Suc n))"
using assms by simp

lemma filter_intlen_d:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set ( xs). P ys)"
 shows   " intlen (filter P ( (x\<odot>xs))) = 0"
using assms by simp

lemma nfilter_intlen_d:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set ( xs). P ys)"
 shows   " intlen (nfilter P (x\<odot>xs) n) = 0"
using assms by simp

lemma sfxfilter_nth_a:
 assumes "(\<exists> ys \<in> set (suffixes \<langle>x\<rangle>). P ys)"
         "j\<le> intlen(filter P (suffixes \<langle>x\<rangle>))"
 shows   "nth (filter P (suffixes \<langle>x\<rangle>)) j = \<langle>x\<rangle>"
using assms by simp

lemma sfxfilter_nth_b1:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set (suffixes xs). P ys)"
         "P (x\<odot>xs)"
 shows   " nth (filter P (suffixes (x\<odot>xs))) 0 = x\<odot>xs"
using assms by simp

lemma sfxfilter_nth_b2:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set (suffixes xs). P ys)"
         "P (x\<odot>xs)"
         "(Suc j) \<le> intlen (filter P (suffixes (x\<odot>xs)))"
 shows   " nth (filter P (suffixes (x\<odot>xs))) (Suc j) = nth (filter P (suffixes (xs))) j"
using assms by auto

lemma sfxfilter_nth_c:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set (suffixes xs). P ys)"
          "\<not> P (x\<odot>xs)"
          "j \<le> intlen (filter P (suffixes (x\<odot>xs)))"
 shows   " nth (filter P (suffixes (x\<odot>xs))) j =  nth (filter P (suffixes (xs))) j"
using assms by auto

lemma sfxfilter_nth_d:
 assumes "(\<exists> ys \<in> set (suffixes (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set (suffixes xs). P ys)"
         "j\<le> intlen (filter P (suffixes (x\<odot>xs)))"
 shows   " nth (filter P (suffixes (x\<odot>xs))) j  = x\<odot>xs"
using assms by auto

lemma nfilter_nth_a:
 assumes "(\<exists> ys \<in> set ( \<langle>x\<rangle>). P ys)"
         "j\<le> intlen(nfilter P ( \<langle>x\<rangle>) n)"
 shows   "nth (nfilter P ( \<langle>x\<rangle>) n) j = n"
using assms by auto

lemma filter_nth_a:
 assumes "(\<exists> ys \<in> set ( \<langle>x\<rangle>). P ys)"
         "j\<le> intlen(filter P ( \<langle>x\<rangle>))"
 shows   "nth (filter P ( \<langle>x\<rangle>)) j = x"
using assms by simp

lemma nfilter_nth_b1:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "P (x)"
 shows   " nth (nfilter P ( (x\<odot>xs)) n) 0 = n"
using assms by simp

lemma filter_nth_b1:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "P (x)"
 shows   " nth (filter P ( (x\<odot>xs))) 0 = x"
using assms by simp

lemma nfilter_nth_b2:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "P (x)"
         "(Suc j) \<le> intlen (nfilter P ( (x\<odot>xs)) n)  "
 shows   " nth (nfilter P ( (x\<odot>xs)) n) (Suc j) = nth (nfilter P ( (xs)) (Suc n)) j"
using assms by auto

lemma filter_nth_b2:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "P (x)"
         "(Suc j) \<le> intlen (filter P ( (x\<odot>xs)))  "
 shows   " nth (filter P ( (x\<odot>xs))) (Suc j) = nth (filter P ( (xs))) j"
using assms by auto

lemma nfilter_nth_c:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "\<not> P (x)"
         "j \<le> intlen (nfilter P ( (x\<odot>xs)) n)"
 shows   " nth (nfilter P ( (x\<odot>xs)) n) j =  nth (nfilter P ( (xs)) (Suc n)) j"
using assms by auto

lemma filter_nth_c:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> set ( xs). P ys)"
         "\<not> P (x)"
         "j \<le> intlen (filter P ( (x\<odot>xs)))"
 shows   " nth (filter P ( (x\<odot>xs))) j =  nth (filter P ( (xs))) j"
using assms by auto

lemma nfilter_nth_d:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set ( xs). P ys)"
         "j\<le> intlen (nfilter P ( (x\<odot>xs)) n)"
 shows   " nth (nfilter P ( (x\<odot>xs)) n) j  = n"
using assms by auto

lemma filter_nth_d:
 assumes "(\<exists> ys \<in> set ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> set ( xs). P ys)"
         "j\<le> intlen (filter P ( (x\<odot>xs)))"
 shows   " nth (filter P ( (x\<odot>xs))) j  = x"
using assms by auto



lemma sfxfilter_nth_cons:
 "nth (filter P (suffixes (x\<odot>xs))) j =
  (if (\<exists> ys \<in> set (suffixes xs). P ys) then 
      (if P (x\<odot>xs) then 
        (if j=0 then (x\<odot>xs) else nth (filter P (suffixes xs)) (j-1))
       else nth (filter P (suffixes xs)) j)
   else (x\<odot>xs))"
by (simp add: Nitpick.case_nat_unfold)

lemma sfxfilter_nth_cons_a:
 "nth (filter P (suffixes (x\<odot>xs))) j =
  (if (\<exists> ys \<in> set (suffixes xs). P ys) then 
      (if P (x\<odot>xs) then 
        (case j of 0 \<Rightarrow> x\<odot>xs | Suc m \<Rightarrow> nth (filter P (suffixes xs)) m)
       else nth (filter P (suffixes xs)) j)
   else (x\<odot>xs))"
by (simp add: Nitpick.case_nat_unfold)

lemma nfilter_nth_cons:
 "nth (nfilter P ( (x\<odot>xs)) n) j =
  (if (\<exists> ys \<in> set ( xs). P ys) then 
      (if P (x) then 
        (if j=0 then (n) else nth (nfilter P ( xs) (Suc n)) (j-1))
       else nth (nfilter P  (xs) (Suc n)) j)
   else (n))"
by (simp add: Nitpick.case_nat_unfold)

lemma nfilter_nth_cons_a:
 "nth (nfilter P ( (x\<odot>xs)) n) j =
  (if (\<exists> ys \<in> set ( xs). P ys) then 
      (if P (x) then 
        (case j of 0 \<Rightarrow> n | Suc m \<Rightarrow> nth (nfilter P ( xs) (Suc n)) m)
       else nth (nfilter P  (xs) (Suc n)) j)
   else (n))"
by (simp add: Nitpick.case_nat_unfold)

lemma filter_nth_cons:
 "nth (filter P ( (x\<odot>xs))) j =
  (if (\<exists> ys \<in> set ( xs). P ys) then 
      (if P (x) then 
        (if j=0 then (x) else nth (filter P ( xs)) (j-1))
       else nth (filter P  (xs)) j)
   else (x))"
by (simp add: Nitpick.case_nat_unfold)

lemma filter_nth_cons_a:
 "nth (filter P ( (x\<odot>xs))) j =
  (if (\<exists> ys \<in> set ( xs). P ys) then 
      (if P (x) then 
        (case j of 0 \<Rightarrow> x | (Suc m) \<Rightarrow> nth (filter P ( xs)) m)
       else nth (filter P  (xs)) j)
   else (x))"
by (simp add: Nitpick.case_nat_unfold)

lemma sfxfilter_nth:
 assumes "(\<exists> ys \<in> set (suffixes xs). P ys)"
         "i \<le> intlen (filter P (suffixes xs))"
 shows   " P (nth (filter P (suffixes xs)) i)"
using assms
proof 
 (induction xs arbitrary: i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases "\<exists>a\<in>set (suffixes xs). P a")
    show "(\<And>i. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
            i \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
            P (nth (filter P (suffixes xs)) i)) \<Longrightarrow>
          \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
          i \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
          \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow> 
          P (nth (filter P (suffixes (x1a \<odot> xs))) i)" 
       by (auto simp add: nat.split_sels(2)) 
    show "(\<And>i. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
            i \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
          P (nth (filter P (suffixes xs)) i)) \<Longrightarrow>
          \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
          i \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
          \<not> (\<exists>a\<in>set (suffixes xs). P a) \<Longrightarrow> 
          P (nth (filter P (suffixes (x1a \<odot> xs))) i) " 
    by simp
   qed
qed


lemma nfilter_intlen:
 assumes "(\<exists> x \<in> set xs. P x)"
 shows   "intlen(nfilter P xs n) = intlen (filter P xs)"
using assms
proof 
 (induction xs arbitrary: n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case by simp
qed


lemma nfilter_upper_bound:
 assumes "(\<exists> x \<in> set xs. P x)"
         "i\<le>intlen (nfilter P xs n)"
 shows   "(nth (nfilter P xs n) i) \<le> n+intlen xs"
using assms
proof 
 (induct xs arbitrary: i n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
 proof (cases "\<exists> x \<in> set xs. P x") 
  show "(\<And>i n. \<exists>a\<in>set xs. P a \<Longrightarrow>
            i \<le> intlen (nfilter P xs n) \<Longrightarrow> nth (nfilter P xs n) i \<le> n + intlen xs) \<Longrightarrow>
        \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
        i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
        \<exists>x\<in>set xs. P x \<Longrightarrow> 
        nth (nfilter P (x1a \<odot> xs) n) i \<le> n + intlen (x1a \<odot> xs)"
    proof auto
     show "\<And>x. (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow> nth (nfilter P xs n) i \<le> n + intlen xs) \<Longrightarrow>
             i \<le> Suc (intlen (nfilter P xs (Suc n))) \<Longrightarrow>
          x \<in> set xs \<Longrightarrow>
          P x \<Longrightarrow> 
          P x1a \<Longrightarrow> 
          (case i of 0 \<Rightarrow> n | Suc k \<Rightarrow> nth (nfilter P xs (Suc n)) k) \<le> Suc (n + intlen xs)" 
         by (cases i, simp, fastforce) 
     show "\<And>x. (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow> nth (nfilter P xs n) i \<le> n + intlen xs) \<Longrightarrow>
           i \<le> intlen (nfilter P xs (Suc n)) \<Longrightarrow>
           x \<in> set xs \<Longrightarrow> P x \<Longrightarrow> \<not> P x1a \<Longrightarrow> nth (nfilter P xs (Suc n)) i \<le> Suc (n + intlen xs)"
       by force
   qed
  show "(\<And>i n. \<exists>a\<in>set xs. P a \<Longrightarrow>
            i \<le> intlen (nfilter P xs n) \<Longrightarrow> nth (nfilter P xs n) i \<le> n + intlen xs) \<Longrightarrow>
        \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
        i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
        \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow> 
        nth (nfilter P (x1a \<odot> xs) n) i \<le> n + intlen (x1a \<odot> xs) " 
     by simp
  qed
qed

lemma nfilter_lower_bound:
 assumes "(\<exists> x \<in> set xs. P x)"
         "i\<le>intlen (nfilter P xs n)"
 shows   " n \<le> (nth (nfilter P xs n) i) "
using assms
proof 
 (induction xs arbitrary: i n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case
  proof   (cases "\<exists> x \<in> set xs. P x")
    show "(\<And>i n. \<exists>a\<in>set xs. P a \<Longrightarrow> i \<le> intlen (nfilter P xs n) \<Longrightarrow> n \<le> nth (nfilter P xs n) i) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow> 
          \<exists>x\<in>set xs. P x \<Longrightarrow> 
          n \<le> nth (nfilter P (x1a \<odot> xs) n) i"
    proof auto
     show " \<And>x. (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow> n \<le> nth (nfilter P xs n) i) \<Longrightarrow>
                 i \<le> Suc (intlen (nfilter P xs (Suc n))) \<Longrightarrow>
                 x \<in> set xs \<Longrightarrow> 
                 P x \<Longrightarrow> 
                 P x1a \<Longrightarrow> 
                 n \<le> (case i of 0 \<Rightarrow> n | Suc k \<Rightarrow> nth (nfilter P xs (Suc n)) k)" 
      by (metis Suc_leD add_le_imp_le_left nat.split_sels(1) order_refl plus_1_eq_Suc)
     show "\<And>x. (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow> n \<le> nth (nfilter P xs n) i) \<Longrightarrow>
                i \<le> intlen (nfilter P xs (Suc n)) \<Longrightarrow>
                x \<in> set xs \<Longrightarrow> 
                P x \<Longrightarrow> 
                \<not> P x1a \<Longrightarrow> 
                n \<le> nth (nfilter P xs (Suc n)) i " 
      using Suc_leD by blast
    qed
    show " (\<And>i n. \<exists>a\<in>set xs. P a \<Longrightarrow> i \<le> intlen (nfilter P xs n) \<Longrightarrow> n \<le> nth (nfilter P xs n) i) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow> 
             \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow> 
             n \<le> nth (nfilter P (x1a \<odot> xs) n) i "
     by auto 
   qed
qed


lemma nfilter_filter:
 assumes "(\<exists> x \<in> set xs. P x)"
         "i\<le>intlen (nfilter P xs n)"
 shows   " (nth xs ((nth (nfilter P xs n) i)-n)  ) = (nth (filter P xs) i)"
using assms
proof 
 (induct xs arbitrary: i n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases "\<exists> x \<in> set xs. P x")
     show "(\<And>i n. \<exists>a\<in>set xs. 
            P a \<Longrightarrow>
            i \<le> intlen (nfilter P xs n) \<Longrightarrow>
            nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
            \<exists>a\<in>set (x1a \<odot> xs). 
                P a \<Longrightarrow>
                i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
                 \<exists>x\<in>set xs.  
                   P x \<Longrightarrow>
                   nth (x1a \<odot> xs) (nth (nfilter P (x1a \<odot> xs) n) i - n) =
                   nth (filter P (x1a \<odot> xs)) i "         
      proof auto
        show "\<And>x. (\<And>i n. 
               i \<le> intlen (nfilter P xs n) \<Longrightarrow>
                 nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
         i \<le> Suc (intlen (nfilter P xs (Suc n))) \<Longrightarrow>
         x \<in> set xs \<Longrightarrow>
         P x \<Longrightarrow>
         P x1a \<Longrightarrow>
         (case (case i of 0 \<Rightarrow> n | Suc k \<Rightarrow> nth (nfilter P xs (Suc n)) k) - n of 0 \<Rightarrow> x1a
          | Suc x \<Rightarrow> nth xs x) =
         (case i of 0 \<Rightarrow> x1a | Suc k \<Rightarrow> nth (filter P xs) k)" 
           proof (cases i)
            show "\<And>x. (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow>
                 nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
                 i \<le> Suc (intlen (nfilter P xs (Suc n))) \<Longrightarrow>
                 x \<in> set xs \<Longrightarrow>
                 P x \<Longrightarrow>
                 P x1a \<Longrightarrow>
                 i = 0 \<Longrightarrow>
                 (case (case i of 0 \<Rightarrow> n | Suc k \<Rightarrow> nth (nfilter P xs (Suc n)) k) - n of 0 \<Rightarrow> x1a
          | Suc x \<Rightarrow> nth xs x) =
         (case i of 0 \<Rightarrow> x1a | Suc k \<Rightarrow> nth (filter P xs) k)" 
           by simp
         show " \<And>x nat.
                (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow>
               nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
               i \<le> Suc (intlen (nfilter P xs (Suc n))) \<Longrightarrow>
               x \<in> set xs \<Longrightarrow>
               P x \<Longrightarrow>
               P x1a \<Longrightarrow>
               i = Suc nat \<Longrightarrow>
               (case (case i of 0 \<Rightarrow> n | Suc k \<Rightarrow> nth (nfilter P xs (Suc n)) k) - n of 0 \<Rightarrow> x1a
                     | Suc x \<Rightarrow> nth xs x) =
               (case i of 0 \<Rightarrow> x1a | Suc k \<Rightarrow> nth (filter P xs) k) "
         by simp 
         (metis (full_types) Suc_diff_le diff_Suc_Suc nfilter_lower_bound old.nat.simps(5))
      qed
        show "\<And>x. 
               (\<And>i n. i \<le> intlen (nfilter P xs n) \<Longrightarrow>
                 nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
               i \<le> intlen (nfilter P xs (Suc n)) \<Longrightarrow>
               x \<in> set xs \<Longrightarrow>
               P x \<Longrightarrow>
              \<not> P x1a \<Longrightarrow>
              (case nth (nfilter P xs (Suc n)) i - n of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> nth xs x) =
              nth (filter P xs) i " 
        by (metis Nitpick.case_nat_unfold Suc_eq_plus1 Suc_le_lessD diff_diff_left neq0_conv 
            nfilter_lower_bound zero_less_diff)
      qed
    show "(\<And>i n. 
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            i \<le> intlen (nfilter P xs n) \<Longrightarrow>
            nth xs (nth (nfilter P xs n) i - n) = nth (filter P xs) i) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). 
            P a \<Longrightarrow>
            i \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
           \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow>
            nth (x1a \<odot> xs) (nth (nfilter P (x1a \<odot> xs) n) i - n) =
            nth (filter P (x1a \<odot> xs)) i"
     by auto
   qed
qed 


lemma set_filter [simp]:
assumes "(\<exists> x \<in> set xs. P x)"
shows "set (filter P xs) = {x. x \<in> set xs \<and> P x}"
using assms 
proof 
 (induct xs)
 case (St x)
 then show ?case by (simp add: Collect_conv_if)
 next
 case (Cons x1a xs)
 then show ?case by auto
qed

lemma set_nfilter [simp]:
assumes "(\<exists> x \<in> set xs. P x)"
shows "set (nfilter P xs n) = {n+k|k .  k\<le> intlen xs \<and> P (nth xs k)}"
using assms
proof 
 (induction xs arbitrary: n)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
  proof (auto simp add: nat.split)
    show "\<And>y k. (\<And>n. \<exists>x\<in>set xs. P x \<Longrightarrow>
                 set (nfilter P xs n) = {n + k |k. k \<le> intlen xs \<and> P (nth xs k)}) \<Longrightarrow>
           P x1a \<Longrightarrow>
           y \<in> set xs \<Longrightarrow>
           P y \<Longrightarrow>
           0 < k \<Longrightarrow>
           k \<le> Suc (intlen xs) \<Longrightarrow>
           \<forall>x2. k = Suc x2 \<longrightarrow> P (nth xs x2) \<Longrightarrow> 
           \<exists>ka. k = Suc ka \<and> ka \<le> intlen xs \<and> P (nth xs ka)" 
     by (metis Suc_le_mono gr0_implies_Suc)
   show "\<And>k. P x1a \<Longrightarrow>
            \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
            k \<le> Suc (intlen xs) \<Longrightarrow> 
            \<forall>x2. k = Suc x2 \<longrightarrow> P (nth xs x2) \<Longrightarrow> 
            k = 0" 
     by (metis Suc_le_mono le_SucE le_zero_eq nth_set zero_induct)
   show "\<And>x y k.
       (\<And>n. \<exists>x\<in>set xs. P x \<Longrightarrow>
             set (nfilter P xs n) = {n + k |k. k \<le> intlen xs \<and> P (nth xs k)}) \<Longrightarrow>
       x \<in> set xs \<Longrightarrow>
       P x \<Longrightarrow>
       P x1a \<Longrightarrow>
       y \<in> set xs \<Longrightarrow>
       P y \<Longrightarrow>
       0 < k \<Longrightarrow>
       k \<le> Suc (intlen xs) \<Longrightarrow>
       \<forall>x2. k = Suc x2 \<longrightarrow> P (nth xs x2) \<Longrightarrow> 
       \<exists>ka. k = Suc ka \<and> ka \<le> intlen xs \<and> P (nth xs ka)" 
     by (metis Suc_le_mono gr0_implies_Suc)
   show "\<And>x y k.
       (\<And>n. \<exists>x\<in>set xs. P x \<Longrightarrow>
             set (nfilter P xs n) = {n + k |k. k \<le> intlen xs \<and> P (nth xs k)}) \<Longrightarrow>
       x \<in> set xs \<Longrightarrow>
       P x \<Longrightarrow>
       \<not> P x1a \<Longrightarrow>
       y \<in> set xs \<Longrightarrow>
       P y \<Longrightarrow>
       k \<le> Suc (intlen xs) \<Longrightarrow>
       0 < k \<Longrightarrow> 
       \<forall>x2. k = Suc x2 \<longrightarrow> P (nth xs x2) \<Longrightarrow> 
       \<exists>ka. k = Suc ka \<and> ka \<le> intlen xs \<and> P (nth xs ka)" 
      by (metis Suc_less_eq gr0_implies_Suc not_less)
   qed
qed

lemma set_minus_filter_out:
assumes "(\<exists> z \<in> set xs. (\<lambda> x. \<not>(x =y)) z)"
 shows  "set xs - {y} = set(filter (\<lambda> x. \<not>(x =y)) xs)"
using assms
by (induct xs) auto

lemma filter_filter [simp]:
assumes "\<exists> x \<in> set (filter Q xs). P x"
        "\<exists> x \<in> set xs. Q x"
        "\<exists> x \<in> set xs. P x \<and> Q x" 
shows "filter P (filter Q xs) = filter (\<lambda>x. P x \<and> Q x) xs"
using assms
by (induct xs) auto

lemma length_nfilter_le [simp]: 
 "intlen (nfilter P xs n) \<le> intlen xs"
by (induct xs arbitrary: n) (auto simp add: le_SucI)

lemma length_filter_le [simp]: 
 "intlen (filter P xs) \<le> intlen xs"
by (induct xs) (auto simp add: le_SucI)

lemma sfxfilter_bound:
 assumes "(\<exists> ys \<in> set (suffixes xs). P ys)"
 shows  " intlen (filter P (suffixes xs)) \<le> intlen xs"
using assms by (metis length_filter_le intlen_suffixes)

lemma filter_bound:
 assumes "(\<exists> ys \<in> set ( xs). P ys)"
 shows  " intlen (filter P (xs)) \<le> intlen xs"
using assms by auto

lemma sfxfilter_nth_bound:
 assumes "(\<exists> ys \<in> set (suffixes xs). P ys) "
         " j \<le> intlen (filter P (suffixes xs))"
 shows  " intlen ((nth (filter P (suffixes xs)) j)) \<le> intlen xs"
using assms 
by (metis (mono_tags, lifting) set_filter in_set_suffixes mem_Collect_eq nth_set osfx_intlen)

lemma sfxfilter_nth_suffix:
 assumes "(\<exists> ys \<in> set (suffixes xs). P ys)"
         "j\<le> intlen (filter P (suffixes xs))"
 shows " nth (filter P (suffixes xs)) j = 
         suffix (intlen xs - intlen(nth(filter P (suffixes xs)) j)) xs"
using assms 
proof 
 (induction xs arbitrary: j)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  proof (cases j) 
   show "(\<And>j. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
          j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
          nth (filter P (suffixes xs)) j =
          suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
         \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
         j \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
         j = 0 \<Longrightarrow>
         nth (filter P (suffixes (x1a \<odot> xs))) j =
         suffix (intlen (x1a \<odot> xs) - intlen (nth (filter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)"
     by  (simp_all add: Suc_diff_le sfxfilter_nth_bound)
   show "\<And>nat. (\<And>j. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
                 j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
                 nth (filter P (suffixes xs)) j =
                 suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
           \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
           j \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
           j = Suc nat \<Longrightarrow>
           nth (filter P (suffixes (x1a \<odot> xs))) j =
           suffix (intlen (x1a \<odot> xs) - intlen (nth (filter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)"
     proof (cases "(\<exists> ys \<in> set (suffixes xs). P ys)")
      show "\<And>nat. (\<And>j. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
                 j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
                 nth (filter P (suffixes xs)) j =
                 suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
           \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
           j \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
           j = Suc nat \<Longrightarrow>
           \<exists>ys\<in>set (suffixes xs). P ys \<Longrightarrow>
           nth (filter P (suffixes (x1a \<odot> xs))) j =
           suffix (intlen (x1a \<odot> xs) - intlen (nth (filter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)"
        proof auto
         show "\<And>nat x.
       (\<And>j. j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
             nth (filter P (suffixes xs)) j =
             suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
       nat \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
       j = Suc nat \<Longrightarrow>
       osfx x xs \<Longrightarrow>
       P x \<Longrightarrow>
       P (x1a \<odot> xs) \<Longrightarrow>
       nth (filter P (suffixes xs)) nat =
       (case Suc (intlen xs) - intlen (nth (filter P (suffixes xs)) nat) of 0 \<Rightarrow> x1a \<odot> xs
        | Suc m \<Rightarrow> suffix m xs)" 
         by (metis (full_types) Suc_diff_le diff_le_self interval_suffix_suc osfx_intlen osfx_suffix 
             suffix.simps(2))
        show "\<And>nat x.
               (\<And>j. j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
                 nth (filter P (suffixes xs)) j =
                 suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
              Suc nat \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
              j = Suc nat \<Longrightarrow>
              osfx x xs \<Longrightarrow>
              P x \<Longrightarrow>
              \<not> P (x1a \<odot> xs) \<Longrightarrow>
             nth (filter P (suffixes xs)) (Suc nat) =
            (case Suc (intlen xs) - intlen (nth (filter P (suffixes xs)) (Suc nat)) of 0 \<Rightarrow> x1a \<odot> xs
                | Suc m \<Rightarrow> suffix m xs) "        
          by (metis (full_types) Suc_diff_le diff_le_self interval_suffix_suc osfx_intlen osfx_suffix 
              suffix.simps(2))
     qed
    show "\<And>nat. 
            (\<And>j. \<exists>a\<in>set (suffixes xs). P a \<Longrightarrow>
                 j \<le> intlen (filter P (suffixes xs)) \<Longrightarrow>
                 nth (filter P (suffixes xs)) j =
                 suffix (intlen xs - intlen (nth (filter P (suffixes xs)) j)) xs) \<Longrightarrow>
           \<exists>a\<in>set (suffixes (x1a \<odot> xs)). P a \<Longrightarrow>
           j \<le> intlen (filter P (suffixes (x1a \<odot> xs))) \<Longrightarrow>
           j = Suc nat \<Longrightarrow>
           \<not> (\<exists>ys\<in>set (suffixes xs). P ys) \<Longrightarrow>
           nth (filter P (suffixes (x1a \<odot> xs))) j =
           suffix (intlen (x1a \<odot> xs) - intlen (nth (filter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)"
      by simp
    qed
   qed
qed

lemma initfilter_sfxfilter_exists:
 " (\<exists> ys \<in> set (suffixes xs). P ( (nth ys 0 ))) = (\<exists> x \<in> set xs. P x)"
by (metis interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)

lemma initfilter_sfxfilter:
 assumes "\<exists> ys \<in> set (suffixes xs). P ( (nth ys 0 ))"
 shows   "filter P xs =  map (\<lambda>s. (nth s 0)) (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) 
          "
using assms 
proof 
 (induction xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  proof auto
    show "\<And>y. (\<exists>ys\<in>set (suffixes xs). P (nth ys 0) \<Longrightarrow>
          filter P xs =
          map (\<lambda>s. nth s 0) (filter (\<lambda>ys. P (nth ys 0)) (suffixes xs))) \<Longrightarrow>
         osfx y xs \<Longrightarrow> 
         P (nth y 0) \<Longrightarrow> 
         P x1a \<Longrightarrow> 
         \<exists>x\<in>set xs. P x" 
    by (metis in_set_suffixes interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)
    show "\<And>x. \<forall>y\<in>set (suffixes xs). \<not> P (nth y 0) \<Longrightarrow> P x1a \<Longrightarrow> x \<in> set xs \<Longrightarrow> P x \<Longrightarrow> False"  
    by (metis interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)
    show "\<And>ys y.
           (\<exists>ys\<in>set (suffixes xs). P (nth ys 0) \<Longrightarrow>
           filter P xs =
           map (\<lambda>s. nth s 0) (filter (\<lambda>ys. P (nth ys 0)) (suffixes xs))) \<Longrightarrow>
           osfx ys xs \<Longrightarrow> 
           P (nth ys 0) \<Longrightarrow> 
           osfx y xs \<Longrightarrow> 
           P (nth y 0) \<Longrightarrow> 
           P x1a \<Longrightarrow> 
           \<exists>x\<in>set xs. P x" 
    by (metis in_set_suffixes interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)
    show "\<And>ys y.
           (\<exists>ys\<in>set (suffixes xs). P (nth ys 0) \<Longrightarrow>
           filter P xs =
           map (\<lambda>s. nth s 0) (filter (\<lambda>ys. P (nth ys 0)) (suffixes xs))) \<Longrightarrow>
           osfx ys xs \<Longrightarrow>
           P (nth ys 0) \<Longrightarrow>
           osfx y xs \<Longrightarrow>
           P (nth y 0) \<Longrightarrow>
           \<not> P x1a \<Longrightarrow>
           \<forall>x\<in>set xs. \<not> P x \<Longrightarrow>
           \<exists>y. \<langle>y\<rangle> = filter (\<lambda>ys. P (nth ys 0)) (suffixes xs) \<and> nth y 0 = x1a "  
    by (metis in_set_suffixes interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)
  qed
qed

lemma filter_nth:
 assumes "(\<exists> x \<in> set xs. P x)"
         "i\<le> intlen (filter P xs)"
 shows   "(\<exists> k \<le> intlen xs. nth(filter P xs) i = nth xs k)"
proof -
 have 1: "\<exists> ys \<in> set (suffixes xs). P ( (nth ys 0 ))"
    using assms
    by (simp add: initfilter_sfxfilter_exists)
 have 2: "\<And>i.  i\<le> intlen (filter P xs) \<longrightarrow> 
          nth(filter P xs) i = nth (map (\<lambda>s. (nth s 0)) (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs))) i" 
  using "1" initfilter_sfxfilter by force
 have 3: "\<And>i. i \<le>  intlen (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) \<longrightarrow>
           nth (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) i =
            suffix (intlen xs - intlen(nth(filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) i)) xs
          " 
     by (meson "1" sfxfilter_nth_suffix)
 have 4: "\<And>i. i \<le>  intlen (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) \<longrightarrow>
            nth (map (\<lambda>s. (nth s 0)) (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs))) i =
            (\<lambda>s. (nth s 0)) (nth (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) i)   
          "
        using interval_nth_map by blast
 have 5: "\<And>i. i \<le>  intlen (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) \<longrightarrow>
         (\<lambda>s. (nth s 0)) (nth (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) i) =
   (\<lambda>s. (nth s 0)) (suffix (intlen xs - intlen(nth(filter (\<lambda>ys. P(nth ys 0)) (suffixes xs)) i)) xs)
         " 
    using "3" by auto
 have 6: "(\<exists> k \<le> intlen xs. 
           nth (map (\<lambda>s. (nth s 0)) (filter (\<lambda>ys. P(nth ys 0)) (suffixes xs))) i = nth xs k)" 
   by (metis (no_types, lifting) "2" set_filter assms interval_nth_and_set 
       mem_Collect_eq nth_set)
 show ?thesis
 by (simp add: "2" "6" assms) 
qed

lemma interval_sfx_nth_zero:
 " set xs = {(nth ys 0)| ys. ys \<in> set(suffixes xs) }"
proof 
 (induct xs)
 case (St x)
 then show ?case 
 by auto
 next
 case (Cons x1a xs)
 then show ?case 
   by auto (metis   interval_nth_zero)
qed

lemma interval_sfx_1:
assumes "ys \<in> set(suffixes xs)"
shows   "(nth ys 0) \<in> set xs"
using assms interval_sfx_nth_zero by fastforce

lemma sum_length_filter_compl_help:
 assumes "\<exists> x \<in> set xs. P x"
         "\<exists> x \<in> set xs. \<not> P x"
 shows   "intlen xs > 0"
using assms 
proof 
 (induct xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  using interval_intlen_cons_1 by blast
qed

lemma filter_id_conv: 
 assumes "\<exists> x \<in> set xs. P x"
 shows "(filter P xs = xs) = (\<forall>x\<in>set xs. P x)"
using assms
proof 
 (induct xs)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
  by (auto simp add: interval_set_nonempty)
     (metis length_filter_le Suc_n_not_le_n intlen.simps(2) plus_1_eq_Suc)
qed

lemma sum_length_filter_compl:
assumes "\<exists> x \<in> set xs. P x"
        "\<exists> x \<in> set xs. \<not> P x"
 shows "intlen(filter P xs) + intlen(filter (\<lambda>x. \<not>P x) xs) +1 = intlen xs "
using assms
proof 
 (induct xs)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case 
  proof (cases "(\<exists>y\<in>set xs. P y)")
  case True
  then show ?thesis 
   proof (cases "(\<exists>y\<in>set xs. \<not> P y)")
   case True
   then show ?thesis
     proof auto
      show "\<And>y x. y \<in> set xs \<Longrightarrow>
           \<not> P y \<Longrightarrow>
           x \<in> set xs \<Longrightarrow>
           P x \<Longrightarrow>
           \<not> P x1a \<Longrightarrow>
           Suc (intlen (filter P xs) + intlen (filter (\<lambda>x. \<not> P x) xs)) = intlen xs" 
       using Cons.hyps by fastforce
      show "\<And>y x. y \<in> set xs \<Longrightarrow>
           \<not> P y \<Longrightarrow>
           x \<in> set xs \<Longrightarrow>
           P x \<Longrightarrow>
           P x1a \<Longrightarrow> Suc (intlen (filter P xs) + intlen (filter (\<lambda>x. \<not> P x) xs)) = intlen xs" 
       using Cons.hyps by fastforce
      show "\<And>y. y \<in> set xs \<Longrightarrow>
         \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> \<not> P x1a \<Longrightarrow> Suc (intlen (filter (\<lambda>x. \<not> P x) xs)) = intlen xs" 
       using Cons.prems(1) by auto
      show "\<And>y. y \<in> set xs \<Longrightarrow>
         \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> P x1a \<Longrightarrow> intlen (filter (\<lambda>x. \<not> P x) xs) = intlen xs " 
      by (metis filter_id_conv)
     qed
   next
   case False
   then show ?thesis 
     proof auto
      show "\<And>y. \<forall>y\<in>set xs. P y \<Longrightarrow> P x1a \<Longrightarrow> y \<in> set xs \<Longrightarrow> 
            Suc (intlen (filter P xs)) = intlen xs" 
       using Cons.prems by auto
      show "P x1a \<Longrightarrow> set xs = {} \<Longrightarrow> intlen xs = 0" 
       using True by blast
      show "\<And>y. \<forall>y\<in>set xs. P y \<Longrightarrow> \<not> P x1a \<Longrightarrow> y \<in> set xs \<Longrightarrow> 
            intlen (filter P xs) = intlen xs" 
       by (metis filter_id_conv)
      show "\<not> P x1a \<Longrightarrow> set xs = {} \<Longrightarrow> intlen xs = 0 " 
       using interval_set_nonempty by blast
      qed
   qed
 next
 case False
 then show ?thesis 
     proof auto
      show "\<And>y. \<forall>x\<in>set xs. \<not> P x \<Longrightarrow>
         \<not> P x1a \<Longrightarrow> y \<in> set xs \<Longrightarrow> 
            Suc (intlen (filter (\<lambda>x. \<not> P x) xs)) = intlen xs" 
       using Cons.prems(1) by auto
      show "\<not> P x1a \<Longrightarrow> set xs = {} \<Longrightarrow> intlen xs = 0" 
       using interval_set_nonempty by blast
      show "\<And>y. \<forall>x\<in>set xs. \<not> P x \<Longrightarrow>
           P x1a \<Longrightarrow> y \<in> set xs \<Longrightarrow> 
            intlen (filter (\<lambda>x. \<not> P x) xs) = intlen xs" 
       by (metis filter_id_conv)
      show "P x1a \<Longrightarrow> set xs = {} \<Longrightarrow> intlen xs = 0 " 
      using interval_set_nonempty by blast
    qed
 qed
qed
 
lemma filter_intlen_imp:
assumes "\<exists> x \<in> set xs. P x \<and> Q x"
shows   "intlen (filter (\<lambda>x. P x \<and> Q x) xs) \<le> intlen (filter P xs)"
using assms 
proof (induct xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by force
qed 

lemma subset_filter:
assumes "(\<exists> x \<in> set xs. P x)"
shows   " set (filter P xs) \<le> set (filter (\<lambda> x. P x \<or> Q x) xs)"
proof -
 have 1: "(\<exists> x \<in> set xs. P x \<or> Q x)"
   using assms by blast 
 have 2: " set (filter (\<lambda> x. P x \<or> Q x) xs) = {x \<in> set xs. P x \<or> Q x}" 
   using assms set_filter[of "xs" "(\<lambda> x. P x \<or> Q x)"] by blast
 have 3: "set (filter P xs) = {x \<in> set xs. P x}"
   using assms set_filter by auto
 have 4: "{x \<in> set xs. P x} \<le> {x \<in> set xs. P x \<or> Q x}"
    by auto
 show ?thesis by (simp add: "2" "3" "4")  
qed

lemma set_filter_not:
 assumes " \<exists> x \<in> set xs. P x" 
         " \<exists> x \<in> set xs. \<not> (P x)"
 shows   "set (filter (\<lambda>x. \<not> (P x)) xs) = set xs - set (filter P xs)"
using assms
 proof (induct xs)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a xs)
 then show ?case
  proof -
   have 1: "set (filter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = 
            set (if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then x1a \<odot> filter (\<lambda> x. \<not> P x) xs else filter (\<lambda> x. \<not> P x) xs)
                 else \<langle>x1a\<rangle>)
            " 
    by simp
   have 2: "set (if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then x1a \<odot> filter (\<lambda> x. \<not> P x) xs else filter (\<lambda> x. \<not> P x) xs)
                 else \<langle>x1a\<rangle>) =
            (if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then set (x1a \<odot> filter (\<lambda> x. \<not> P x) xs) 
                              else set (filter (\<lambda> x. \<not> P x) xs))
                 else set (\<langle>x1a\<rangle>)) "
       by simp
   have 3: "(if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then set (x1a \<odot> filter (\<lambda> x. \<not> P x) xs) 
                               else set (filter (\<lambda> x. \<not> P x) xs))
                 else set (\<langle>x1a\<rangle>)) =
            (if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then set (\<langle>x1a\<rangle>) \<union> set( filter (\<lambda> x. \<not> P x) xs) 
                              else set (filter (\<lambda> x. \<not> P x) xs))
                 else set (\<langle>x1a\<rangle>))
           "  
        by simp
   have 4: "set (filter P (x1a \<odot> xs))  =
            set (if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then x1a \<odot> filter P xs else filter P xs)
                 else \<langle>x1a\<rangle>)
           " 
      by simp
   have 5: "set (if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then x1a \<odot> filter P xs else filter P xs)
                 else \<langle>x1a\<rangle>) =
             (if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then set (x1a \<odot> filter P xs) else set (filter P xs))
                 else set \<langle>x1a\<rangle>)
           " 
       by simp
   have 6: "(if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then set (x1a \<odot> filter P xs) else set (filter P xs))
                 else set \<langle>x1a\<rangle>) =
             (if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then set (\<langle>x1a\<rangle>) \<union> set( filter P xs) else set (filter P xs))
                 else set \<langle>x1a\<rangle>)
           " 
        by simp
   have 7: "(if (\<exists>y \<in> set xs. \<not> P y) then 
                  (if \<not> P x1a then set (\<langle>x1a\<rangle>) \<union> set( filter (\<lambda> x. \<not> P x) xs) 
                              else set (filter (\<lambda> x. \<not> P x) xs))
                 else set (\<langle>x1a\<rangle>)) =
               set (x1a \<odot> xs) -
                (if (\<exists>y \<in> set xs.  P y) then 
                  (if  P x1a then set (\<langle>x1a\<rangle>) \<union> set( filter P xs) else set (filter P xs))
                 else set \<langle>x1a\<rangle>)
              " 
         using Cons.prems by auto
   show ?thesis using "1" "3" "4" "6" "7" by presburger
  qed
qed

lemma set_filter_cap:
 assumes "\<exists> x \<in> set xs. f x"
         "\<exists> x \<in> set xs. \<not> f x"
 shows   " set (filter f xs) \<inter> set (filter (\<lambda> x. \<not> f x) xs) = {}"
using assms by auto

lemma filter_nth_or:
 assumes "\<exists> x \<in> set xs. P x"
 shows   "\<exists> x \<in> set (filter (\<lambda>x. P x \<or> Q x) xs). P x"
proof -
 have 1: "\<exists> i \<le> intlen(filter (\<lambda>x. P x \<or> Q x) xs). P (nth (filter (\<lambda>x. P x) xs) i)"
   using assms by (metis (mono_tags, lifting) set_filter interval_intlen_gr_zero 
   mem_Collect_eq nth_set)  
 obtain i where 2: "i \<le> intlen(filter (\<lambda>x. P x \<or> Q x) xs) \<and> P (nth (filter (\<lambda>x. P x) xs) i) \<and>
                    (nth (filter (\<lambda>x. P x) xs) i) \<in> set (filter P xs)" 
    using "1" 
    by (metis (mono_tags, lifting) set_filter assms interval_intlen_gr_zero 
        mem_Collect_eq nth_set)
 have 3: "set (filter P xs) \<le> set (filter (\<lambda>x. P x \<or> Q x) xs)"
    using assms subset_filter[of "xs" "P"] by auto
 have 4: "(nth (filter (\<lambda>x. P x) xs) i) \<in> set (filter P xs)"  
   using "2" by auto
 have 5: "(nth (filter (\<lambda>x. P x) xs) i) \<in> set (filter (\<lambda>x. P x \<or> Q x) xs)" 
   using "3" "4" by blast
  from "2" "5" show ?thesis by blast
qed

lemma filter_intapp1:
assumes "\<forall>x\<in>set xs. \<not> P x"
        "P x1a"
        "P y"
        "y \<in> set ys"
shows   "filter P (xs \<ominus> ys) = filter P ys"
using assms
by (induct xs) auto

lemma filter_intapp [simp]:
assumes "(\<exists> x \<in> set (xs \<ominus> ys). P x)"
        "(\<exists> x \<in> set xs. P x)"
        "(\<exists> x \<in> set ys. P x)"
shows "filter P (xs\<ominus>ys) = (filter P xs) \<ominus> (filter P ys)"
using assms 
proof 
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
  proof 
   (cases "(\<exists> x \<in> set xs. P x)")
   case True
   then show ?thesis using Cons.hyps Cons.prems(3) by auto
   next
   case False
   then show ?thesis 
    proof auto
     show "\<And>y. \<forall>x\<in>set xs. \<not> P x \<Longrightarrow>
           P x1a \<Longrightarrow> P y \<Longrightarrow> y \<in> set ys \<Longrightarrow> 
           filter P (xs \<ominus> ys) = filter P ys" 
      by (simp add: filter_intapp1) 
     show "\<forall>x\<in>set xs. \<not> P x \<Longrightarrow> P x1a \<Longrightarrow> \<exists>x\<in>set xs \<union> set ys. P x" 
      using Cons.prems(3) by blast
     show "\<And>y. \<forall>x\<in>set xs. \<not> P x \<Longrightarrow>
         \<not> P x1a \<Longrightarrow> P y \<Longrightarrow> y \<in> set ys \<Longrightarrow> 
           filter P (xs \<ominus> ys) = x1a \<odot> filter P ys" 
      using Cons.prems(2) by auto
     show "\<forall>x\<in>set xs. \<not> P x \<Longrightarrow> \<not> P x1a \<Longrightarrow> \<exists>x\<in>set xs \<union> set ys. P x " 
      using Cons.prems(2) by auto
    qed
 qed
qed

lemma filter_True:
assumes "\<forall>x \<in> set xs. P x"
shows   "filter P xs = xs"
using assms 
by (meson filter_id_conv length_filter_le nth_set)

lemma nfilter_map:
 assumes "\<exists> x \<in> set (map f xs). P x"
shows   " nfilter P (map f xs) n =  (nfilter (P\<circ>f) xs n)"
using assms 
by (induct xs arbitrary: n) auto


lemma filter_map:
assumes "\<exists> x \<in> set (map f xs). P x"
shows   " filter P (map f xs) = map f (filter (P\<circ>f) xs)"
using assms
by (induct xs) auto

lemma  length_nfilter_map[simp]:
assumes "\<exists> x \<in> set (map f xs). P x"
shows  "intlen (nfilter P (map f xs) n) = intlen(nfilter (P \<circ> f) xs n)" 
using assms by (simp add:nfilter_map)         

lemma length_filter_map[simp]:
assumes "\<exists> x \<in> set (map f xs). P x"
shows  "intlen (filter P (map f xs)) = intlen(filter (P \<circ> f) xs)"
using assms by (simp add:filter_map)

lemma nfilter_is_subset [simp]:
 assumes "\<exists> x \<in> set xs. P x"
 shows   "set (nfilter P xs n) \<le> {n+k|k. k\<le>intlen xs}"
using assms by auto

lemma filter_is_subset [simp]:
 assumes "\<exists> x \<in> set xs. P x"
 shows   "set (filter P xs) \<le> set xs"
using assms by auto

lemma length_nfilter_less:
assumes "\<exists>  x \<in> set xs. P x"
        "x \<in> set xs"
        "\<not> P x "
 shows "intlen(nfilter P xs n) < intlen xs"
using assms 
proof 
 (induct xs arbitrary: n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case using le_imp_less_Suc length_nfilter_le by (simp_all, blast)
qed

lemma length_filter_less:
assumes "\<exists>  x \<in> set xs. P x"
        " x \<in> set xs" 
        " \<not> P x "
 shows " intlen(filter P xs) < intlen xs"
using assms 
proof 
 (induct xs)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
       using length_filter_le le_imp_less_Suc by (simp_all, blast)
qed


lemma nfilter_set:
 assumes "\<exists>  x \<in> set xs. P x"
 shows   "set (nfilter P xs n) = {n+i|i. i\<le>intlen xs \<and> P(nth xs i)}"
using assms by auto

lemma State_eq_filterD:
assumes "\<exists> x \<in> set ys.  P x"
        " \<langle>x\<rangle> = filter P ys"
 shows "(\<exists> us vs . (ys = \<langle>x\<rangle> \<or>
         (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
         (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
         (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
           ) \<and> P x
        )" 
using assms
proof 
 (induct ys)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a ys)
 then show ?case 
  proof 
   (cases "P x1a")
   case True
   then show ?thesis
     by (metis Cons.prems(2) filter.simps(2) interval.distinct(1) interval.inject(1))
   next
   case False
   then show ?thesis 
    proof 
     (cases "x = x1a")
     case True
     then show ?thesis 
      using Cons.hyps Cons.prems(1) Cons.prems(2) False by auto
     next
     case False
     then show ?thesis 
      proof -
       have 1: "\<exists> x \<in> set ys.  P x"
            using Cons.prems(2) False interval.simps(15) by fastforce 
       have 2: "\<not> P x1a"
            using "1" Cons.prems(2) by auto
       have 3: "P x"
            using "1" "2" Cons.hyps Cons.prems(2) by auto
       have 4: "(\<langle>x\<rangle> = filter P (x1a \<odot> ys)) =
               (if (\<exists>y\<in>set ys. P y) then
                              (if P x1a then \<langle>x\<rangle> = x1a \<odot> filter P ys else \<langle>x\<rangle> =filter P ys)
                             else \<langle>x\<rangle> =\<langle>x1a\<rangle>)
             "
            using "2" by auto
        have 5: "\<langle>x\<rangle> =filter P ys "
            by (simp add: "1" "2" Cons.prems(2)) 
        have 6: "(\<exists> us vs . 
                  (ys = \<langle>x\<rangle> \<or>
                  (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
                  (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
                  (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
                  ) \<and> P x)"
            using "1" Cons.hyps Cons.prems(2) by auto
       obtain us vs where 61: "(ys = \<langle>x\<rangle> \<or>
                                (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
                                (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
                                (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
                                ) \<and> P x" 
            using  "6" by auto
        have 7: "ys=\<langle>x\<rangle> \<longrightarrow> 
              (x1a \<odot> ys = \<langle>x\<rangle> \<or>
              (x1a \<odot> ys = x \<odot> vs \<and> (\<forall>v\<in>set vs. \<not> P v)) \<or>
              (x1a \<odot> ys = \<langle>x1a\<rangle> \<ominus> \<langle>x\<rangle> \<and> (\<forall>u\<in>set \<langle>x1a\<rangle>. \<not> P u)) \<or>
              (x1a \<odot> ys = \<langle>x1a\<rangle> \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set \<langle>x1a\<rangle>. \<not> P u) \<and> (\<forall>v\<in>set vs. \<not> P v)))
              \<and> P x
             " 
            using "1" Cons.prems(2) by auto
        have 8: "(ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<longrightarrow>
               (x1a \<odot> ys = \<langle>x\<rangle> \<or>
               x1a \<odot> ys = x \<odot> vs \<and> (\<forall>v\<in>set vs. \<not> P v) \<or>
               x1a \<odot> ys = \<langle>x1a\<rangle> \<ominus> \<langle>x\<rangle> \<and> (\<forall>u\<in>set \<langle>x1a\<rangle>. \<not> P u) \<or>
               x1a \<odot> ys = \<langle>x1a\<rangle> \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set \<langle>x1a\<rangle>. \<not> P u) \<and> (\<forall>v\<in>set vs. \<not> P v))
                \<and> P x
             " 
            using "1" Cons.prems(2) by fastforce
        have 9: "(ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<longrightarrow>
                  (x1a \<odot> ys = \<langle>x\<rangle> \<or>
                   x1a \<odot> ys = x \<odot> vs \<and> (\<forall>v\<in>set vs. \<not> P v) \<or>
                   x1a \<odot> ys = (x1a \<odot>us) \<ominus> \<langle>x\<rangle> \<and> (\<forall>u\<in>set (x1a \<odot>us). \<not> P u) \<or>
                   x1a \<odot> ys = (x1a \<odot>us) \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set (x1a\<odot>us). \<not> P u) \<and> 
                    (\<forall>v\<in>set vs. \<not> P v))
                  \<and> P x" 
            using "2" "3" by auto
        have 10: "(ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v)) \<longrightarrow>
                   (x1a \<odot> ys = \<langle>x\<rangle> \<or>
                    x1a \<odot> ys = x \<odot> vs \<and> (\<forall>v\<in>set vs. \<not> P v) \<or>
                    x1a \<odot> ys = (x1a\<odot>us) \<ominus> \<langle>x\<rangle> \<and> (\<forall>u\<in>set (x1a\<odot>us). \<not> P u) \<or>
                    x1a \<odot> ys = (x1a \<odot>us) \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set (x1a\<odot>us). \<not> P u) \<and> 
                    (\<forall>v\<in>set vs. \<not> P v))
                   \<and> P x"
             using "2" "3" by auto
         have 11: "( (ys = \<langle>x\<rangle> \<or>
              (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
              (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
              (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
              ) \<and> P x) \<longrightarrow> (\<exists>us vs.
                (x1a \<odot> ys = \<langle>x\<rangle> \<or>
                 x1a \<odot> ys = x \<odot> vs \<and> (\<forall>v\<in>set vs. \<not> P v) \<or>
                 x1a \<odot> ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall>u\<in>set us. \<not> P u) \<or>
                 x1a \<odot> ys = us \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set us. \<not> P u) \<and> (\<forall>v\<in>set vs. \<not> P v))
                \<and> P x
              ) "
        using "3" 
        using "10" "7" "8" "9" False by blast
     show ?thesis using "11" "3" "61" by blast
    qed
   qed
  qed
qed


lemma filter_eq_StateD:
 assumes " \<exists> x \<in> set ys.  P x"
         " filter P ys = \<langle>x\<rangle>"
 shows   "(\<exists> us vs . (ys = \<langle>x\<rangle> \<or>
         (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
         (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
         (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
        ) \<and> P x)"
using assms State_eq_filterD[of "ys" "P" "x" ]  by simp 

lemma filter_eq_State_iff:
 assumes " \<exists> x \<in> set ys.  P x"
 shows   "(filter P ys = \<langle>x\<rangle>)   =
        (\<exists> us vs . (ys = \<langle>x\<rangle> \<or>
         (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
         (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
         (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
        ) \<and> P x)"
proof -
 have 1: "(filter P ys = \<langle>x\<rangle>) \<Longrightarrow>  
          (\<exists> us vs . (ys = \<langle>x\<rangle> \<or>
         (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
         (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
         (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
        ) \<and> P x)
        " 
      using assms by (rule State_eq_filterD)  simp
 have 2: "(\<exists> us vs . (ys = \<langle>x\<rangle> \<or>
         (ys = x \<odot> vs \<and> (\<forall> v \<in> set vs. \<not> P v)) \<or>
         (ys = us \<ominus> \<langle>x\<rangle> \<and> (\<forall> u \<in> set us. \<not> P u)) \<or>
         (ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u \<in> set us. \<not> P u) \<and> (\<forall> v \<in> set vs. \<not> P v))
        ) \<and> P x) \<Longrightarrow> (filter P ys = \<langle>x\<rangle>)"
       by (auto simp add: filter_intapp1) 
  from 1 2 show ?thesis by metis
qed

lemma Cons_eq_filterD:
assumes " \<exists> x \<in> set ys. P x"
        " (x\<odot> xs = filter P ys) "
 shows  " (\<exists> us vs. (ys = x\<odot> vs \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                    \<and> (\<exists> x \<in> set vs. P x) 
                    \<and> P x \<and> xs = filter P vs)"
using assms
proof 
 (induct ys)
 case (St x)
 then show ?case by auto
 next
 case (Cons x1a ys)
 then show ?case 
   proof (cases " P x1a")
    case True
    then show ?thesis
     by (metis Cons.prems(2) filter.simps(2) interval.inject(2) interval.simps(4))
    next
    case False
    then show ?thesis 
     proof (cases "x = x1a")
     case True
     then show ?thesis
      using Cons.hyps Cons.prems(2) False by force
     next
     case False
     then show ?thesis 
      proof -
       have 1: "\<exists> x \<in> set ys.  P x"
           by (metis Cons.prems(2) filter.simps(2) interval.simps(4)) 
       have 2: "\<not> P x1a"
           using "1" Cons.prems(2) False by auto 
       have 4: "(x\<odot>xs = filter P (x1a \<odot> ys)) =
               (if (\<exists>y\<in>set ys. P y) then
                              (if P x1a then x\<odot>xs = x1a \<odot> filter P ys else x\<odot>xs =filter P ys)
                             else x\<odot>xs =\<langle>x1a\<rangle>)"
           by simp
       have 5: "x\<odot>xs = filter P ys"
           by (simp add: "1" "2" Cons.prems(2)) 
       have 6: "(\<exists> us vs. (ys = x\<odot> vs   \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                    \<and> (\<exists> x \<in> set vs. P x)
                    \<and> P x \<and> xs = filter P vs)" 
           by (simp add: "1" "2" Cons.hyps Cons.prems(2))
       obtain us vs where 61: "(ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                   \<and> (\<exists> x \<in> set vs. P x)
                   \<and> P x \<and> xs = filter P vs"
            using "6" by auto
       have 62: "P x \<and> (\<exists> x \<in> set vs. P x)"
            using "61" by blast
       have 7: "ys = x\<odot> vs \<and> P x \<and> xs = filter P vs \<longrightarrow> 
             (x1a \<odot> ys = x \<odot> vs  \<or>
             x1a \<odot> ys = \<langle>x1a\<rangle> \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set \<langle>x1a\<rangle>. \<not> P u)) 
              \<and> (\<exists> x \<in> set vs. P x)
            \<and> P x \<and> xs = filter P vs
            "   
            by (simp add: "2" "62")
        have 8: "ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u) \<and> P x \<and> xs = filter P vs \<longrightarrow>
                (x1a \<odot> ys = x \<odot> vs  \<or>
                 x1a \<odot> ys = (x1a \<odot>us) \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set (x1a \<odot>us). \<not> P u)) 
                 \<and> (\<exists> x \<in> set vs. P x)
                 \<and> P x \<and> xs = filter P vs
                " 
            using "2" "62" by auto
        have 9: "(ys = x\<odot> vs   \<or>
                  ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) \<and> P x \<and> xs = filter P vs \<longrightarrow>
            (\<exists>us vs.
             (x1a \<odot> ys = x \<odot> vs  \<or>
              x1a \<odot> ys = us \<ominus> x \<odot> vs \<and> (\<forall>u\<in>set us. \<not> P u)) 
             \<and> (\<exists> x \<in> set vs. P x)
            \<and> P x \<and> xs = filter P vs) 
            "
        using "7" "8" by blast
       show ?thesis using "61" "9" by blast
      qed
   qed
  qed 
qed


lemma filter_eq_ConsD:
assumes " \<exists> x \<in> set ys. P x"
        " (filter P ys = x \<odot> xs)"
 shows  " (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                    \<and> (\<exists> x \<in> set vs. P x)
                    \<and> P x \<and> xs = filter P vs)"
using assms Cons_eq_filterD[of "ys"]   by simp


lemma filter_eq_Cons_iff: 
assumes "(\<exists> x \<in> set ys. P x)"
 shows " (filter P ys = x \<odot> xs)  =
         ( 
          (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                \<and> (\<exists> x \<in> set vs. P x)
                \<and> P x \<and> xs = filter P vs))"
proof -
 have 1: "(filter P ys = x \<odot> xs) \<Longrightarrow>
          ( 
          (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                \<and> (\<exists> x \<in> set vs. P x)
                \<and> P x \<and> xs = filter P vs))
          "
        using assms by (rule Cons_eq_filterD)  simp
 have 2: "( 
          (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                \<and> (\<exists> x \<in> set vs. P x)
                \<and> P x \<and> xs = filter P vs)) \<Longrightarrow>
           (filter P ys = x \<odot> xs)
         " 
     by (auto simp add: filter_intapp1)
 from 1 2 show ?thesis by metis
qed


lemma Cons_eq_filter_iff: 
assumes "(\<exists> x \<in> set ys. P x)"
 shows " (x \<odot> xs = filter P ys  )  =
         ( 
          (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                \<and> (\<exists> x \<in> set vs. P x)
                \<and> P x \<and> xs = filter P vs))"
proof -
 have 1: "(x \<odot> xs = filter P ys  ) = (filter P ys = x \<odot> xs)" 
   by auto
 have 2: "(filter P ys = x \<odot> xs) =
          ( 
          (\<exists> us vs. (ys = x\<odot> vs  \<or>
                    ys = us \<ominus> (x \<odot> vs) \<and> (\<forall> u  \<in> set us. \<not> P u)) 
                \<and> (\<exists> x \<in> set vs. P x)
                \<and> P x \<and> xs = filter P vs))
         " 
    using assms 
    by (simp add: filter_eq_Cons_iff)
 from 1 2 show ?thesis by auto
qed

lemma nfilter_cong[fundef_cong]:
 assumes "xs = ys"
         " (\<And>x. x \<in> set ys \<Longrightarrow> P x = Q x) "
 shows   " nfilter P xs n = nfilter Q ys n"
using assms by (induction xs arbitrary: ys n) auto


lemma filter_cong[fundef_cong]:
 assumes "xs = ys"
         " (\<And>x. x \<in> set ys \<Longrightarrow> P x = Q x) "
 shows   " filter P xs = filter Q ys"
using assms by (induct ys arbitrary: xs) auto

lemma remdups_filter: 
 assumes "\<exists> x \<in> set xs . P x"
shows "remdups(filter P xs) = filter P (remdups xs)"
using assms
by (induct xs) auto

lemma distinct_map_filter:
assumes "\<exists> x \<in> set xs . P x"
        " distinct (map f xs)"
shows   " distinct (map f (filter P xs))"
using assms by (induct xs) auto

lemma distinct_nfilter [simp]:
 "distinct (nfilter P xs n)"
by (induction xs arbitrary: n) auto

lemma distinct_filter [simp]: 
assumes "distinct xs"
shows   "distinct (filter P xs)"
using assms by (induct xs) auto

lemma distinct_length_filter: 
assumes "\<exists> x \<in> set xs . P x"
        " distinct xs "
shows " intlen (filter P xs) +1 = card ({x. P x} Int set xs)"
using assms by (induct xs) auto

lemma filter_mono_osfx:
assumes "\<exists> x \<in> set xs . P x"
        " osfx xs ys "
shows " osfx (filter P xs) (filter P ys)"
using assms 
proof (auto simp: osfx_def)
fix x :: 'a and zs :: "'a interval"
assume a1: "P x"
assume a2: "x \<in> set xs"
assume a3: "filter P (zs \<ominus> xs) \<noteq> filter P xs"
assume a4: "ys = zs \<ominus> xs"
have f5: "\<forall>i p a aa ia. 
          ((\<exists>a. (a::'a) \<in> set i \<and> p a) \<or> \<not> p a \<or> \<not> p aa \<or> aa \<notin> set ia) \<or> 
            filter p (i \<ominus> ia) = filter p ia"
by (metis (no_types) filter_intapp1)
obtain aa :: "('a \<Rightarrow> bool) \<Rightarrow> 'a interval \<Rightarrow> 'a" where
"\<forall>x3 x4. (\<exists>v5. v5 \<in> set x4 \<and> x3 v5) = (aa x3 x4 \<in> set x4 \<and> x3 (aa x3 x4))"
by moura 
then have "\<forall>i p a ab ia. aa p i \<in> set i \<and> p (aa p i) \<or> \<not> p a \<or> \<not> p ab \<or> 
           ab \<notin> set ia \<or> filter p (i \<ominus> ia) = filter p ia"
using f5 by presburger 
then have f6: "\<exists>a. a \<in> set zs \<and> P a"
using a3 a2 a1 by blast 
have "osfx xs (zs \<ominus> xs)"
using a4 assms(2) by blast 
then show "\<exists>i. filter P (zs \<ominus> xs) = i \<ominus> filter P xs"
using f6 a2 a1 by (meson filter_intapp set_mono_osfx subsetD) 
qed



lemma idx_nfilter_mono:
assumes "\<exists> x \<in> set xs . P x"
        "na<intlen (nfilter P xs n)" 
shows   "nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na) " 
using assms 
proof (induct xs arbitrary: n na)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof (cases "\<exists> x \<in> set xs . P x")
    show "(\<And>na n.
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            na < intlen (nfilter P xs n) \<Longrightarrow> 
            nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
          \<exists>x\<in>set xs. P x \<Longrightarrow> 
          nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na)"
    proof  (cases na)
     show "(\<And>na n.
             \<exists>a\<in>set xs. P a \<Longrightarrow>
             na < intlen (nfilter P xs n) \<Longrightarrow> 
             nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
           \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
           na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
           \<exists>x\<in>set xs. P x \<Longrightarrow>
           na = 0 \<Longrightarrow> 
           nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na)"
      proof (cases "P x1a")
       show "(\<And>na n.
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               na < intlen (nfilter P xs n) \<Longrightarrow> 
               nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
             \<exists>x\<in>set xs. P x \<Longrightarrow>
             na = 0 \<Longrightarrow> 
             P x1a \<Longrightarrow> 
             nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na)"
        by (simp add: Suc_le_lessD nfilter_lower_bound)
       show "(\<And>na n.
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               na < intlen (nfilter P xs n) \<Longrightarrow> 
               nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
             \<exists>x\<in>set xs. P x \<Longrightarrow>
             na = 0 \<Longrightarrow> 
             \<not> P x1a \<Longrightarrow> 
             nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na) "
        by simp 
      qed
     show "\<And>nat. 
             (\<And>na n.
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               na < intlen (nfilter P xs n) \<Longrightarrow>
               nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
           \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
           na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
           \<exists>x\<in>set xs. P x \<Longrightarrow>
           na = Suc nat \<Longrightarrow> 
           nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na) "
       by auto 
     qed
    show "(\<And>na n. 
             \<exists>a\<in>set xs. P a \<Longrightarrow>
             na < intlen (nfilter P xs n) \<Longrightarrow> 
             nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          na < intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
          \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow>
          nth (nfilter P (x1a \<odot> xs) n) na < nth (nfilter P (x1a \<odot> xs) n) (Suc na) "
     by auto 
     qed
qed


lemma idx_nfilter:
assumes "\<exists> x \<in> set xs . P x"
shows   "index_sequence (intfirst((nfilter P xs n))) (nfilter P xs n)"
using assms by (simp add: index_sequence_def idx_nfilter_mono)

lemma idx_nfilter_expand:
assumes "\<exists> x \<in> set xs . P x"
shows   "\<forall>na<intlen (nfilter P xs n). nth (nfilter P xs n) na < nth (nfilter P xs n) (Suc na)"
using assms idx_nfilter by (simp add: index_sequence_def)


lemma idx_nfilter_gr_eq:
 assumes "\<exists> x \<in> set xs . P x"
         "k \<le> j"
         "j \<le> intlen(nfilter P xs n)" 
  shows  "nth (nfilter P xs n) k \<le>  nth (nfilter P xs n) j"
using assms by (meson idx_nfilter interval_idx_less_eq)

lemma idx_nfilter_gr:
 assumes "\<exists> x \<in> set xs . P x"
  shows  "(\<forall> j .  k< j \<and> j \<le> intlen(nfilter P xs n)\<longrightarrow>
              nth (nfilter P xs n) k <  nth (nfilter P xs n) j)"
using assms 
by (meson Suc_leI dual_order.strict_trans1 idx_nfilter_expand idx_nfilter_gr_eq)


lemma idx_nfilter_less_eq:
 assumes "\<exists> x \<in> set xs . P x"
         " k\<le> intlen(nfilter P xs n)" 
  shows  " (\<forall> j \<le> k.  nth (nfilter P xs n) j \<le>  nth (nfilter P xs n) k)"
using assms by (simp add: idx_nfilter_gr_eq)


lemma idx_nfilter_less:
 assumes "\<exists> x \<in> set xs . P x"
         "k\<le> intlen(nfilter P xs n)" 
  shows  "(\<forall> j < k.  nth (nfilter P xs n) j <  nth (nfilter P xs n) k)"
using assms 
by (simp add: idx_nfilter_gr)



lemma nfilter_prefix_set_0:
assumes "k\<le> intlen xs"
        "\<exists> x \<in> set (prefix k xs) . P x"
 shows "intlen (nfilter P (prefix k xs) n) \<le> intlen (nfilter P xs n)" 
using assms 
proof (induct xs arbitrary: n k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases k)
    show "(\<And>k n. 
            k \<le> intlen xs \<Longrightarrow>
            \<exists>a\<in>set (prefix k xs). P a \<Longrightarrow> 
            intlen (nfilter P (prefix k xs) n) \<le> intlen (nfilter P xs n)) \<Longrightarrow>
          k \<le> intlen (x1a \<odot> xs) \<Longrightarrow>
          \<exists>a\<in>set (prefix k (x1a \<odot> xs)). P a \<Longrightarrow>
          k = 0 \<Longrightarrow> 
          intlen (nfilter P (prefix k (x1a \<odot> xs)) n) \<le> intlen (nfilter P (x1a \<odot> xs) n)"
     by simp
    show "\<And>nat. 
             (\<And>k n. 
                k \<le> intlen xs \<Longrightarrow>
                \<exists>a\<in>set (prefix k xs). P a \<Longrightarrow> 
                intlen (nfilter P (prefix k xs) n) \<le> intlen (nfilter P xs n)) \<Longrightarrow>
             k \<le> intlen (x1a \<odot> xs) \<Longrightarrow>
             \<exists>a\<in>set (prefix k (x1a \<odot> xs)). P a \<Longrightarrow>
             k = Suc nat \<Longrightarrow> 
             intlen (nfilter P (prefix k (x1a \<odot> xs)) n) \<le> intlen (nfilter P (x1a \<odot> xs) n)" 
     using prefix_subset by simp blast
   qed
qed


lemma nfilter_subset:
 assumes "\<exists> x \<in> set (prefix k xs) . P x"
         "k\<le> intlen xs" 
 shows " set (nfilter P (prefix k xs) n) \<le> set (nfilter P xs n)"
using assms 
proof (induct xs arbitrary:k n)
case (St x)
then show ?case by auto
next
case (Cons x1a xs)
then show ?case 
  proof (cases k)
   show "(\<And>k n. 
            \<exists>a\<in>set (prefix k xs). P a \<Longrightarrow>
            k \<le> intlen xs \<Longrightarrow> 
            set (nfilter P (prefix k xs) n) \<subseteq> set (nfilter P xs n)) \<Longrightarrow>
         \<exists>a\<in>set (prefix k (x1a \<odot> xs)). P a \<Longrightarrow>
         k \<le> intlen (x1a \<odot> xs) \<Longrightarrow>
         k = 0 \<Longrightarrow> 
         set (nfilter P (prefix k (x1a \<odot> xs)) n) \<subseteq> set (nfilter P (x1a \<odot> xs) n) " 
     by simp
   show "\<And>nat. 
           (\<And>k n. 
              \<exists>a\<in>set (prefix k xs). P a \<Longrightarrow>
              k \<le> intlen xs \<Longrightarrow> 
              set (nfilter P (prefix k xs) n) \<subseteq> set (nfilter P xs n)) \<Longrightarrow>
           \<exists>a\<in>set (prefix k (x1a \<odot> xs)). P a \<Longrightarrow>
           k \<le> intlen (x1a \<odot> xs) \<Longrightarrow>
           k = Suc nat \<Longrightarrow> 
           set (nfilter P (prefix k (x1a \<odot> xs)) n) \<subseteq> set (nfilter P (x1a \<odot> xs) n) "
       by (auto simp add: nth_set) 
  qed
qed



lemma nfilter_prefix_set:
 assumes "\<exists> x \<in> set (prefix k xs) . P x"
         "k\<le> intlen xs " 
 shows   "set (nfilter P (prefix k xs) n) = {n+i|i. i\<le> k \<and> P(nth xs i)}"
using assms
by auto

lemma nfilter_suffix_set:
 assumes "\<exists> x \<in> set (suffix k xs) . P x"
         " k\<le> intlen xs " 
 shows " set (nfilter P (suffix k xs) n) = {n+i|i. i \<le> intlen xs-k \<and> P(nth xs (k+i))}"
using assms by auto

lemma nfilter_suffix_set_a:
 assumes "\<exists> x \<in> set (suffix k xs) . P x"
         " k\<le> intlen xs " 
 shows   " set (nfilter P (suffix k xs) n) = {n+(j-k)|j. k\<le>j \<and>  j \<le> intlen xs \<and> P(nth xs j)}"
using assms  Nat.le_diff_conv2 by auto fastforce

lemma nfilter_suffix_set_b:
 assumes "\<exists> x \<in> set (suffix k xs) . P x"
         " k\<le> intlen xs " 
 shows " set (nfilter P (suffix k xs) n) = {n+i|i. i \<le> intlen xs-k \<and> P(nth xs (i+k))}"
using assms  nfilter_suffix_set by (auto simp add: add.commute)
  
lemma nfilter_card:
assumes "\<exists> x \<in> set xs . P x"
 shows " intlen (nfilter P xs n) +1 = card (set (nfilter P xs n)) " 
using assms 
by (induct xs arbitrary: n) auto

lemma nfilter_prefix_set_1:
assumes "k\<le> intlen xs"
        "\<exists> x \<in> set (prefix k xs) . P x"
 shows   "set (prefix (intlen(nfilter P (prefix k xs) n)) (nfilter P xs n)) =
            {(nth (nfilter P xs n) i) | i. i\<le> intlen(nfilter P (prefix k xs) n)}"
using  prefix_set[of "(intlen(nfilter P (prefix k xs) n))" "(nfilter P xs n)"]
by (simp add: assms(1) assms(2) nfilter_prefix_set_0)

lemma nfilter_prefix_subset:
assumes "\<exists> x \<in> set xs . P x"
        " k \<le> intlen (nfilter P xs n) " 
shows   " set (prefix k (nfilter P xs n)) \<le> set (nfilter P xs n)"
using assms 
using prefix_subset by blast


lemma intlen_filter_conv_card:
assumes "\<exists>  x \<in> set xs. P x"
 shows  "intlen (filter P xs) +  1 = card {i. i\<le> intlen xs \<and> P (nth xs i)}"
proof -
 have 1: " intlen (filter P xs) = intlen(nfilter P xs 0)" 
  by (simp add: assms nfilter_intlen)
 have 2: "  intlen (nfilter P xs 0) +1 = card (set (nfilter P xs 0))"  
   by (meson assms nfilter_card) 
 have 3: " set (nfilter P xs 0) = {i|i. i\<le> intlen xs \<and> P(nth xs i)}"
   using assms by auto
 show ?thesis 
 using "1" "2" "3" by auto
qed

lemma nfilter_all:
 assumes " \<exists>  x \<in> set xs. P x"
 shows   " ((nfilter P xs n) = [n..\<le>n+intlen xs]) = (\<forall>x \<in> set xs. P x)"
using assms 
proof (induction xs arbitrary: n)
case (St x)
then show ?case by (simp add: upt_same)
next
case (Cons x1a xs)
then show ?case 
   proof (cases " \<exists>  x \<in> set xs. P x")
    show "(\<And>n. 
            \<exists>a\<in>set xs. P a \<Longrightarrow> 
            (nfilter P xs n = [n..\<le>n + intlen xs]) = (\<forall>a\<in>set xs. P a)) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          \<exists>x\<in>set xs. P x \<Longrightarrow> 
          (nfilter P (x1a \<odot> xs) n = [n..\<le>n + intlen (x1a \<odot> xs)]) = (\<forall>a\<in>set (x1a \<odot> xs). P a)" 
   proof auto
    show "\<And>x xa.
            (\<And>n. (nfilter P xs n = [n..\<le>n + intlen xs]) = (\<forall>x\<in>set xs. P x)) \<Longrightarrow>
            x \<in> set xs \<Longrightarrow>
            P x \<Longrightarrow>
            P x1a \<Longrightarrow> 
            n \<odot> nfilter P xs (Suc n) = [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle> \<Longrightarrow> 
            xa \<in> set xs \<Longrightarrow> 
            P xa"
       by (metis add.right_neutral add_diff_cancel_left' add_right_imp_eq interval_intlen_cons 
           interval_intlen_intapp intlen.simps(1) length_nfilter_less nat_less_le upt_length)
    show " \<And>x. 
            (\<And>n. nfilter P xs n = [n..\<le>n + intlen xs]) \<Longrightarrow>
            x \<in> set xs \<Longrightarrow>
            P x1a \<Longrightarrow>
            \<forall>x\<in>set xs. P x \<Longrightarrow>
            Suc 0 \<le> intlen xs \<Longrightarrow>
            n \<odot> [Suc n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle> = 
            [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle>"
   using upt_rec by auto
     show "\<And>x. 
            (\<And>n. nfilter P xs n = [n..\<le>n + intlen xs]) \<Longrightarrow>
            x \<in> set xs \<Longrightarrow>
            P x1a \<Longrightarrow>
            \<forall>x\<in>set xs. P x \<Longrightarrow>
            \<not> Suc 0 \<le> intlen xs \<Longrightarrow> 
            \<langle>n, Suc (n + intlen xs)\<rangle> = [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle>" 
       by (simp add: upt_rec)
      show "\<And>x. 
             (\<And>n. (nfilter P xs n = [n..\<le>n + intlen xs]) = (\<forall>x\<in>set xs. P x)) \<Longrightarrow>
             x \<in> set xs \<Longrightarrow>
             P x \<Longrightarrow> 
             \<not> P x1a \<Longrightarrow> 
             nfilter P xs (Suc n) = [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle> \<Longrightarrow> False " 
       by (metis Suc_n_not_le_n interval_intfirst_intapp2 interval_intlen_gr_zero 
           interval_nth_zero_intfirst le_add1 nfilter_lower_bound upt_intfirst)
    qed   
    show "(\<And>n. \<exists>a\<in>set xs. P a \<Longrightarrow> (nfilter P xs n = [n..\<le>n + intlen xs]) = (\<forall>a\<in>set xs. P a)) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow>
          (nfilter P (x1a \<odot> xs) n = [n..\<le>n + intlen (x1a \<odot> xs)]) = (\<forall>a\<in>set (x1a \<odot> xs). P a) "
     proof auto 
      show "\<And>x. P x1a \<Longrightarrow>
             \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
             \<langle>n\<rangle> = [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle> \<Longrightarrow> 
             x \<in> set xs \<Longrightarrow>
             False"
       by (metis interval_intapp_not_state)
       show "P x1a \<Longrightarrow> 
             set xs = {} \<Longrightarrow> 
             \<langle>n\<rangle> = [n..\<le>n + intlen xs] \<ominus> \<langle>Suc (n + intlen xs)\<rangle> " 
      using interval_set_nonempty by blast
     qed
   qed 
qed

lemma filter_nfilter_prefix_intlen_0:
   assumes "P (nth xs   ( (nth (nfilter P xs n) k) -n))"
           "k \<le> intlen (filter P xs)" 
   shows   "(\<exists> x \<in> set xs. P x)"
using assms 
by (induction xs arbitrary: k) auto


lemma nfilter_intlen_n_zero:
  assumes "(\<exists> x \<in> set xs. P x)"
  shows   "intlen (nfilter P xs n) = intlen (nfilter P xs 0)"
using assms 
proof 
 (induction xs arbitrary: n)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases n)
     show "(\<And>n. \<exists>a\<in>set xs. P a \<Longrightarrow> intlen (nfilter P xs n) = intlen (nfilter P xs 0)) \<Longrightarrow>
           \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow> 
           n = 0 \<Longrightarrow>
           intlen (nfilter P (x1a \<odot> xs) n) = intlen (nfilter P (x1a \<odot> xs) 0)"
      by blast
     show "\<And>nat. 
            (\<And>n. \<exists>a\<in>set xs. P a \<Longrightarrow> intlen (nfilter P xs n) = intlen (nfilter P xs 0)) \<Longrightarrow>
            \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
            n = Suc nat \<Longrightarrow> 
            intlen (nfilter P (x1a \<odot> xs) n) = intlen (nfilter P (x1a \<odot> xs) 0) "
      by (simp add: nfilter_intlen)
    qed
qed

lemma nfilter_nth_n_zero_a:
  assumes "(\<exists> x \<in> set xs. P x)"
          " k \<le> intlen (nfilter P xs n)"
  shows   " n \<le> (nth (nfilter P xs n) k)"
using assms by (simp add: nfilter_lower_bound)

lemma nfilter_nth_n_zero:
  assumes "(\<exists> x \<in> set xs. P x)"
          " k \<le> intlen (nfilter P xs n)"
  shows   "(nth (nfilter P xs n) k) - n  = nth (nfilter P xs 0) k"
using assms 
proof 
 (induction xs arbitrary: n k)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
    proof (cases "(\<exists> x \<in> set xs. P x)")
     show "(\<And>k n. 
              \<exists>a\<in>set xs. P a \<Longrightarrow> 
              k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
              nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
           \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
           k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow> 
           \<exists>x\<in>set xs. P x \<Longrightarrow> 
           nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k"
      proof  (cases "P x1a")
       show "(\<And>k n. 
                \<exists>a\<in>set xs. P a \<Longrightarrow> 
                k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
             \<exists>x\<in>set xs. P x \<Longrightarrow> 
             P x1a \<Longrightarrow> 
             nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k"
      proof  (cases k)
       show "(\<And>k n. 
               \<exists>a\<in>set xs. P a \<Longrightarrow> 
               k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
               nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
             \<exists>x\<in>set xs. P x \<Longrightarrow> 
             P x1a \<Longrightarrow> 
             k = 0 \<Longrightarrow> 
             nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k" 
    by auto
       show "\<And>nat. 
              (\<And>k n. 
                 \<exists>a\<in>set xs. P a \<Longrightarrow> 
                 k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                 nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
              \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
              k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
              \<exists>x\<in>set xs. P x \<Longrightarrow> 
              P x1a \<Longrightarrow> 
              k = Suc nat \<Longrightarrow> 
              nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k "
          proof auto
            fix nat
            fix x
            show "(\<And>k n. 
                    k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                    nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
                  nat \<le> intlen (nfilter P xs (Suc n)) \<Longrightarrow>
                  P x1a \<Longrightarrow> 
                  k = Suc nat \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  nth (nfilter P xs (Suc n)) nat - n = nth (nfilter P xs (Suc 0)) nat"
             proof -
              assume a0: "(\<And>k n. 
                            k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                            nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k)"   
              assume a1: " nat \<le> intlen (nfilter P xs (Suc n))" 
              assume a2: "P x1a" 
              assume a3: " k = Suc nat"
              assume a4: "x \<in> set xs"
              assume a5: "P x"  
              show "nth (nfilter P xs (Suc n)) nat - n = nth (nfilter P xs (Suc 0)) nat"
                proof -
                 have 1: "nth (nfilter P xs (Suc n)) nat - (Suc n) = nth (nfilter P xs 0) nat"
                       using a0 a1 by blast
                 have 2: "nth (nfilter P xs (Suc 0)) nat - (Suc 0) = nth (nfilter P xs 0) nat"
                    by (metis a0 a1 a4 a5 nfilter_intlen_n_zero) 
                 show ?thesis 
                 by (metis "1" "2" One_nat_def Suc_diff_le a1 a4 a5 diff_Suc_Suc 
                     nfilter_intlen_n_zero nfilter_lower_bound 
                     ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
                qed
              qed
             qed
            qed    
       show "(\<And>k n. 
               \<exists>a\<in>set xs. P a \<Longrightarrow> 
               k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
               nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow>
             \<exists>x\<in>set xs. P x \<Longrightarrow> 
             \<not> P x1a \<Longrightarrow> 
             nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k "
        proof auto
         fix x
         show "(\<And>k n. 
                  k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                  nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
               k \<le> intlen (nfilter P xs (Suc n)) \<Longrightarrow>
               \<not> P x1a \<Longrightarrow> 
               x \<in> set xs \<Longrightarrow> 
               P x \<Longrightarrow> 
               nth (nfilter P xs (Suc n)) k - n = nth (nfilter P xs (Suc 0)) k "
         proof -
          assume b0: "(\<And>k n. 
                       k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
                       nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k)"
          assume b1: "k \<le> intlen (nfilter P xs (Suc n))"
          assume b2: "\<not> P x1a"
          assume b3: "x \<in> set xs"
          assume b4: "P x" 
          show "nth (nfilter P xs (Suc n)) k - n = nth (nfilter P xs (Suc 0)) k"
            proof -
             have 3: "nth (nfilter P xs (Suc n)) k - (Suc n) = nth (nfilter P xs 0) k"
              using b0 b1 by blast                   
             have 4: "nth (nfilter P xs (Suc 0)) k - (Suc 0) = nth (nfilter P xs 0) k"
                by (metis b0 b1 b3 b4 nfilter_intlen_n_zero)      
             show ?thesis 
             by (metis "3" "4" One_nat_def Suc_diff_le b1 b3 b4 diff_Suc_Suc nfilter_intlen_n_zero 
                 nfilter_lower_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse 
                 plus_1_eq_Suc)
            qed
         qed
       qed
    qed
    show "(\<And>k n. 
             \<exists>a\<in>set xs. P a \<Longrightarrow> 
             k \<le> intlen (nfilter P xs n) \<Longrightarrow> 
             nth (nfilter P xs n) k - n = nth (nfilter P xs 0) k) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          k \<le> intlen (nfilter P (x1a \<odot> xs) n) \<Longrightarrow> 
          \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow> 
          nth (nfilter P (x1a \<odot> xs) n) k - n = nth (nfilter P (x1a \<odot> xs) 0) k" 
    by auto
  qed
qed



lemma nfilter_n_zero:
  assumes "(\<exists> x \<in> set xs. P x)"
  shows   " (nfilter P xs n) = map (\<lambda>i. i+n) (nfilter P xs 0)"
using assms 
proof -
 have 1: "intlen (nfilter P xs n) = intlen (map (\<lambda>i. i+n) (nfilter P xs 0))" 
  using assms nfilter_intlen_n_zero by fastforce
 have 2: "\<And>k. k\<le> intlen (nfilter P xs n) \<longrightarrow>
               nth (nfilter P xs n) k = nth (map (\<lambda>i. i+n) (nfilter P xs 0)) k"
  using assms nfilter_nth_n_zero[of "xs" "P" _ "n"] 
  by (metis interval_nth_map le_add_diff_inverse2 nfilter_lower_bound) 
 show ?thesis by (simp add: "1" "2" interval_eq_nth_eq)
qed

lemma nfilter_n_zero_a:
  assumes "(\<exists> x \<in> set xs. P x)"
  shows   "  (nfilter P xs 0) = map (\<lambda>i. i-n) (nfilter P xs n)"
proof -
 have 1: "intlen(nfilter P xs 0) = intlen( map (\<lambda>i. i-n) (nfilter P xs n))"
     by (metis assms interval_intlen_map nfilter_intlen_n_zero) 
 have 2: "\<And>k. k\<le> intlen (nfilter P xs 0) \<longrightarrow>
              nth (nfilter P xs 0) k = nth (map (\<lambda>i. i-n) (nfilter P xs n)) k"
    using assms 
    by (simp add: "1" interval_nth_map nfilter_nth_n_zero) 
 show ?thesis 
 using "1" "2" interval_eq_nth_eq by blast
qed


lemma nfilter_count:
  assumes "(\<exists> x \<in> set xs. P x)"
  shows   " card {( nth (nfilter P xs n) k)| k. k \<le> intlen (nfilter P xs n)} = 
            intlen (nfilter P xs n) +1"
using assms nfilter_card[of "xs" "P" "n"]   set_nfilter[of "xs" "P" "n"]   
 interval_nth_and_set by (simp add: set_nth)

lemma nfilter_holds:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " (\<forall> x \<in> set (nfilter P xs n). P (nth xs (x-n)))"
using assms by auto


lemma nfilter_holds_not:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " (\<forall> x \<in> ({i+n| i. i\<le> intlen xs} - (set (nfilter P xs n))). \<not> P (nth xs (x-n)))"
using assms by auto

lemma nfilter_holds_a:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " (\<forall> i\<le>intlen xs.  (i+n) \<in> set (nfilter P xs n) \<longrightarrow> P (nth xs i))"
using assms by auto

lemma nfilter_holds_not_a:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " (\<forall> i\<le> intlen xs.  P (nth xs i) \<longrightarrow> (i+n) \<in> set (nfilter P xs n))"
using assms by auto

lemma nfilter_holds_b:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " (\<forall> i\<le>intlen xs.  (i+n) \<in> set (nfilter P xs n) = P (nth xs i))"
using assms by auto

lemma nfilter_holds_c:
 assumes "(\<exists> x \<in> set xs. P x)"
         " i\<le>intlen xs "
  shows  "   (i+n) \<in> set (nfilter P xs n) = P (nth xs i)"
by (simp add: assms(1) assms(2))

lemma nfilter_holds_d:
 assumes "(\<exists> x \<in> set xs. P x)"
         " n \<le> i "
         " i \<le>intlen xs +n "
  shows  "   i \<in> set (nfilter P xs n) = P (nth xs (i-n))"
using assms 
by (metis diff_add le_diff_conv nfilter_holds_b)

lemma nfilter_holds_not_b:
 assumes "(\<exists> x \<in> set xs. P x)"
         " n \<le> i "
         " i \<le>intlen xs +n "
  shows  "   i \<notin> set (nfilter P xs n) = (\<not> P (nth xs (i-n)))"
using assms by auto

lemma nfilter_disjoint_set_coset:
 assumes "(\<exists> x \<in> set xs. P x)"
  shows  " ({i+n| i. i\<le> intlen xs} - (set (nfilter P xs n))) \<inter> (set (nfilter P xs n)) = {}"
using assms by auto

lemma nfilter_not_before:
 assumes "(\<exists> x \<in> set xs. P x)"
         " i< (nth (nfilter P xs 0) 0)" 
 shows   " \<not> P (nth xs i)"
proof -
 have 0: "(nth (nfilter P xs 0) 0) \<le> intlen xs"
      by (metis add.left_neutral assms(1) interval_intlen_gr_zero nfilter_upper_bound)
 have 1: " i \<notin> set (nfilter P xs 0)"
   using assms 
   proof (induction xs arbitrary: i)
   case (St x)
   then show ?case by simp
   next
   case (Cons x1a xs)
   then show ?case  
     by (metis idx_nfilter interval_idx_greater interval_nth_and_set interval_nth_zero_intfirst leD)
   qed
 have 2: "i \<notin> set (nfilter P xs 0) \<and> i\<le> intlen xs \<longrightarrow> \<not> P (nth xs (i))" 
      by (metis add.right_neutral nfilter_holds_not_a nth_set)
 have 3: "i \<le> intlen xs"
      using "0" assms(2) by linarith
 from 0 1 2 3 show ?thesis by auto
qed    

lemma nfilter_n_not_before:
 assumes "(\<exists> x \<in> set (suffix n xs). P x)"
         "n \<le> intlen xs" 
         "n \<le>i"
         "i< (nth (nfilter P (suffix n xs) n) 0)"
 shows   "\<not> P (nth xs (i))"
proof -
 have 0: "(nth (nfilter P (suffix n xs) n) 0)  \<le> intlen xs "
     by (metis assms(1) assms(2) interval_intlen_gr_zero interval_suffix_length_good 
         le_add_diff_inverse nfilter_upper_bound)
 have 1: " i \<notin> set (nfilter P (suffix n xs) n)"
   using assms 
   proof (induction xs arbitrary: i)
   case (St x)
   then show ?case by simp
   next
   case (Cons x1a xs)
   then show ?case 
    by (metis idx_nfilter interval_idx_greater interval_nth_and_set interval_nth_zero_intfirst leD)
   qed
 have 2: "i \<notin> set (nfilter P (suffix n xs) n) \<and> n \<le> i \<and> i\<le> intlen xs  \<longrightarrow> \<not> P (nth xs (i))" 
     using assms nfilter_holds_not_a[of "(suffix n xs)" "P" "n"] 
     by (metis add_le_imp_le_right interval_nth_suffix interval_suffix_length le_add_diff_inverse2 
         ordered_cancel_comm_monoid_diff_class.add_diff_inverse)  
 have 3: "n \<le> i \<and> i\<le> intlen xs" 
     using "0" assms(3) assms(4) by linarith
 from 0 1 2 3 show ?thesis by auto
qed   

lemma nfilter_not_after:
 assumes "(\<exists> x \<in> set xs. P x)"
         "(nth (nfilter P xs 0) (intlen (nfilter P xs 0))) < i"
         "  i \<le> intlen xs " 
 shows   "  \<not> P (nth xs (i))"
proof -
 have 1: "i \<notin> set (nfilter P xs 0)" 
   using assms 
     proof (induction xs arbitrary: i)
     case (St x)
     then show ?case by auto
     next
     case (Cons x1a xs)
     then show ?case 
        by (metis idx_nfilter_gr_eq interval_nth_and_set leD le_refl)
   qed
 have 2: "i \<notin> set (nfilter P xs 0) \<and> i \<le> intlen xs \<longrightarrow>\<not> P (nth xs (i)) " 
   by (metis add.right_neutral nfilter_holds_b nth_set)
 have 3: "i \<le> intlen xs" 
    by (simp add: assms(3))
 from 1 2 3 show ?thesis by auto
qed

lemma nfilter_n_not_after:
 assumes "(\<exists> x \<in> set (suffix n xs). P x)"
         "n \<le> intlen xs" 
         "(nth (nfilter P (suffix n xs) n) (intlen (nfilter P (suffix n xs) n))) < i"
         " i \<le> intlen xs "
 shows   " \<not> P (nth xs (i))"
proof -
 have 1: "i \<notin> set (nfilter P (suffix n xs) n)" 
   using assms 
     proof (induction xs arbitrary: i)
     case (St x)
     then show ?case by auto
     next
     case (Cons x1a xs)
     then show ?case 
         by (metis idx_nfilter_gr_eq interval_nth_and_set leD le_eq_less_or_eq)
     qed
 have 2: "i \<notin> set (nfilter P (suffix n xs) n) \<and> n\<le> i \<and> i \<le> intlen xs \<longrightarrow>\<not> P (nth xs (i)) " 
       using assms nfilter_holds_not_a[of "suffix n xs" "P" "n"] 
       by (metis add_le_imp_le_left interval_nth_suffix interval_suffix_length le_add_diff_inverse2
        ordered_cancel_comm_monoid_diff_class.add_diff_inverse)   
 have 3: "n\<le> i \<and> i \<le> intlen xs"
     by (meson assms(1) assms(3) assms(4) dual_order.strict_implies_order dual_order.strict_trans2 
         nfilter_nth_n_zero_a order_refl)
 from 1 2 3 show ?thesis by auto
qed


lemma nfilter_not_between_help_a:
assumes "(\<And>k i. 
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            k < intlen (nfilter P xs 0) \<Longrightarrow>
            nth (nfilter P xs 0) k < i \<Longrightarrow>
            i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
            nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
            i \<notin> set (nfilter P xs 0))"
         "\<exists>a\<in>set (x1a \<odot> xs). P a"
         "k < intlen (nfilter P (x1a \<odot> xs) 0)"
         "nth (nfilter P (x1a \<odot> xs) 0) k < i "
         "i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) "
         "nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs)"
         "\<exists>x\<in>set xs. P x"
         "P x1a "
   shows     " i \<notin> set (nfilter P (x1a \<odot> xs) 0)"
proof -
 have 1: "k=0 \<Longrightarrow> i \<notin> set (nfilter P (x1a \<odot> xs) 0)"
     using assms by auto 
      (metis One_nat_def Suc_less_eq2 diff_Suc_1 interval_intlen_gr_zero nfilter_not_before 
     nfilter_nth_n_zero)
 have 2: "\<And>n. k= (Suc n) \<Longrightarrow>  i \<notin> set (nfilter P (x1a \<odot> xs) 0)"
   using assms 
   proof auto
    fix n
    fix x
    fix ka
    assume a0: "k= Suc n"
    assume a1: "(\<And>k i. k < intlen (nfilter P xs 0) \<Longrightarrow>
                  nth (nfilter P xs 0) k < i \<Longrightarrow>
                  i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow>
                  nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
                  \<not> P (nth xs i))" 
    assume a2: "n < intlen (nfilter P xs (Suc 0))"
    assume a3: " nth (nfilter P xs (Suc 0)) n < Suc ka"  
    assume a4: "Suc ka < nth (nfilter P xs (Suc 0)) (Suc n)" 
    assume a5: "nth (nfilter P xs (Suc 0)) (Suc n) \<le> Suc (intlen xs)"
    assume a6: "P x1a"
    assume a7: "x \<in> set xs"
    assume a8: "P x"
    assume a9: "i = Suc ka"   
    assume a10: "P (nth xs ka)"
    show "False"
     proof -
      have 3: "n < intlen (nfilter P xs 0) \<Longrightarrow>
               nth (nfilter P xs 0) n < ka \<Longrightarrow>
               ka < nth (nfilter P xs 0) (Suc n) \<Longrightarrow> 
               nth (nfilter P xs 0) (Suc n) \<le> intlen xs \<Longrightarrow> 
               \<not> P (nth xs ka)"
         using a1[of "n" "ka"] by auto
      have 4: "n < intlen (nfilter P xs 0)" 
        by (metis a2 a7 a8 nfilter_intlen_n_zero)
      have 5: "nth (nfilter P xs 0) (Suc n) \<le> intlen xs" 
        by (metis "4" Suc_leI add.commute add.right_neutral assms(7) nfilter_upper_bound)
      have 6: "\<exists>x\<in>set xs. P x" 
         using assms(7) by auto
      have 7: "Suc 0 \<le> intlen (nfilter P xs (Suc n)) " 
        by (metis One_nat_def Suc_le_mono Suc_mono a2 a7 a8 interval_intlen_gr_zero le_SucE 
            nfilter_intlen_n_zero not_add_less1 plus_1_eq_Suc)
      have 8: "ka < nth (nfilter P xs 0) (Suc n)" 
         using nfilter_nth_n_zero[of "xs" "P" "Suc n" "Suc 0"]  a4 "6" a2 by linarith
      have 9: "nth (nfilter P xs 0) n < ka"
         using  a2 a3 a7 a8 nfilter_nth_n_zero[of "xs" "P" "n" "Suc 0" ]
         by (metis One_nat_def add.commute dual_order.strict_implies_order less_diff_conv2 
             nfilter_lower_bound plus_1_eq_Suc)  
      have 10: "  \<not> P (nth xs ka) "
          using 3 4 9 8 5 by auto
      from a10 10 show ?thesis by auto
   qed
  qed
 show ?thesis using 1 2 
 using gr0_implies_Suc by blast 
qed

lemma nfilter_not_between_help_b:
assumes "(\<And>k i. 
     \<exists>a\<in>set xs. P a \<Longrightarrow>
     k < intlen (nfilter P xs 0) \<Longrightarrow>
     nth (nfilter P xs 0) k < i \<Longrightarrow>
     i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
     nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
     i \<notin> set (nfilter P xs 0))"
    "\<exists>a\<in>set (x1a \<odot> xs). P a"
    "k < intlen (nfilter P (x1a \<odot> xs) 0)"
    "nth (nfilter P (x1a \<odot> xs) 0) k < i "
    "i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) "  
    "nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs)"
    "\<exists>x\<in>set xs. P x"
    "\<not> P x1a "
 shows " i \<notin> set (nfilter P (x1a \<odot> xs) 0)" 
proof -
 have 1: "k=0 \<Longrightarrow> i \<notin> set (nfilter P (x1a \<odot> xs) 0)"
    using assms  
    proof auto
     fix x :: 'a and ka :: nat
     assume a1: "nth (nfilter P xs (Suc 0)) 0 < Suc ka"
     assume a2: "x \<in> set xs"
     assume a3: "P x"
     assume a4: "\<And>k i. \<lbrakk>k < intlen (nfilter P xs 0); nth (nfilter P xs 0) k < i; 
                        i < nth (nfilter P xs 0) (Suc k); nth (nfilter P xs 0) (Suc k) \<le> intlen xs\<rbrakk> 
                        \<Longrightarrow> \<not> P (nth xs i)"
     assume a5: "P (nth xs ka)"
     assume a6: "0 < intlen (nfilter P xs (Suc 0))"
     assume a7: "Suc ka < nth (nfilter P xs (Suc 0)) (Suc 0)"
     assume a8: "nth (nfilter P xs (Suc 0)) (Suc 0) \<le> Suc (intlen xs)"
     have f9: "\<exists>a. a \<in> set xs \<and> P a"
     using a3 a2 by blast 
     then have f10: "0 \<le> intlen (nfilter P xs (Suc 0)) \<longrightarrow> nth (nfilter P xs (Suc 0)) 0 \<noteq> 0"
     by (metis le_zero_eq nfilter_lower_bound not_less_eq_eq) 
     have f11: "intlen (nfilter P xs (Suc 0)) = intlen (nfilter P xs 0)"
     using f9 by (meson nfilter_intlen_n_zero) 
     obtain nn :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
     f12: "nth (nfilter P xs (Suc 0)) (Suc 0) = Suc (nn (nth (nfilter P xs (Suc 0)) (Suc 0)) ka) \<and> 
           ka < nn (nth (nfilter P xs (Suc 0)) (Suc 0)) ka"
     using a7 by (meson Suc_less_eq2) 
     then have "nth (nfilter P xs 0) (Suc 0) = nn (nth (nfilter P xs (Suc 0)) (Suc 0)) ka"
     using f9 a6 by (metis One_nat_def diff_Suc_1 le_zero_eq neq0_conv 
                     nfilter_nth_n_zero not_less_eq_eq) 
     then have "\<not> 0 \<le> intlen (nfilter P xs 0)"
     using f12 f11 f10 f9 a8 a6 a5 a4 a1 
      by (metis One_nat_def Suc_le_mono diff_Suc_1 less_Suc_eq_0_disj nfilter_nth_n_zero) 
     then show False 
     by blast 
     qed
  have 2: "\<And>n. k= (Suc n) \<Longrightarrow>  i \<notin> set (nfilter P (x1a \<odot> xs) 0)"    
    using assms
    proof auto
     fix n
     fix x
     fix ka
     assume a0: " k = Suc n"
     assume a1: "(\<And>k i. k < intlen (nfilter P xs 0) \<Longrightarrow>
                   nth (nfilter P xs 0) k < i \<Longrightarrow>
                   i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow>
                   nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
                   \<not> P (nth xs i)) "  
     assume a2: "Suc n < intlen (nfilter P xs (Suc 0))"
     assume a3: "nth (nfilter P xs (Suc 0)) (Suc n) < Suc ka"  
     assume a4: "Suc ka < nth (nfilter P xs (Suc 0)) (Suc (Suc n))"
     assume a5: "nth (nfilter P xs (Suc 0)) (Suc (Suc n)) \<le> Suc (intlen xs)"  
     assume a6: "\<not> P x1a"
     assume a7: "x \<in> set xs"
     assume a8: "P x"
     assume a9: "i = Suc ka"   
     assume a10: "P (nth xs ka)"
     show "False"
     proof -
      have 3: "Suc n < intlen (nfilter P xs 0) \<Longrightarrow>
               nth (nfilter P xs 0) (Suc n) < ka \<Longrightarrow>
               ka < nth (nfilter P xs 0) (Suc (Suc n)) \<Longrightarrow> 
               nth (nfilter P xs 0) (Suc (Suc n)) \<le> intlen xs \<Longrightarrow> 
               \<not> P (nth xs ka)"
         using a1[of "Suc n" "ka"] by auto
      have 4: "(Suc n) < intlen (nfilter P xs 0)"  
          by (metis a2 a7 a8 nfilter_intlen_n_zero)
      have 5:  "nth (nfilter P xs 0) (Suc (Suc n)) \<le> intlen xs" 
        by (metis "4" Suc_leI a7 a8 add.left_neutral nfilter_upper_bound) 
      have 6: "\<exists>x\<in>set xs. P x" 
        by (simp add: assms(7))
      have 7: "Suc 0 \<le> intlen (nfilter P xs (Suc n)) " 
        by (metis One_nat_def Suc_le_mono Suc_mono a2 a7 a8 interval_intlen_gr_zero le_SucE 
            nfilter_intlen_n_zero not_add_less1 plus_1_eq_Suc)
      have 8: "ka < nth (nfilter P xs 0) (Suc (Suc n))" 
          using nfilter_nth_n_zero[of "xs" "P" "Suc (Suc n)" "Suc 0"]  a4 6 a2 by linarith
      have 9: "nth (nfilter P xs 0) (Suc n) < ka"
          using  a2 a3 a7 a8 nfilter_nth_n_zero[of "xs" "P" "Suc n" "Suc 0" ]
          by (metis One_nat_def add.commute dual_order.strict_implies_order less_diff_conv2
              nfilter_lower_bound plus_1_eq_Suc)
      have 10: "\<not> P (nth xs ka)"
         using 3 4 9 8 5 by auto
      from a10 10 show ?thesis by auto
     qed
    qed
  show ?thesis using 1 2 
  using gr0_implies_Suc by blast 
qed  


lemma nfilter_not_between_help:
 assumes "(\<exists> x \<in> set xs. P x)"
         "k < intlen (nfilter P xs 0)"
         "(nth (nfilter P xs 0) k ) < i"  
         "i < (nth (nfilter P xs 0) (Suc k) )" 
         "(nth (nfilter P xs 0) (Suc k) ) \<le> intlen xs" 
 shows "i \<notin> set (nfilter P xs 0)"  
using assms 
proof (induction xs arbitrary: i k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
    proof (cases "(\<exists> x \<in> set xs. P x)")
     show "(\<And>k i. 
              \<exists>a\<in>set xs. P a \<Longrightarrow>
              k < intlen (nfilter P xs 0) \<Longrightarrow>
              nth (nfilter P xs 0) k < i \<Longrightarrow>
              i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
              nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
              i \<notin> set (nfilter P xs 0)) \<Longrightarrow>
           \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
           k < intlen (nfilter P (x1a \<odot> xs) 0) \<Longrightarrow>
           nth (nfilter P (x1a \<odot> xs) 0) k < i \<Longrightarrow>
           i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<Longrightarrow>
           nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs) \<Longrightarrow> 
           \<exists>x\<in>set xs. P x \<Longrightarrow> 
           i \<notin> set (nfilter P (x1a \<odot> xs) 0)" 
     proof  (cases "P x1a") 
      show "(\<And>k i. 
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               k < intlen (nfilter P xs 0) \<Longrightarrow>
               nth (nfilter P xs 0) k < i \<Longrightarrow>
               i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
               nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
               i \<notin> set (nfilter P xs 0)) \<Longrightarrow>
             \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
             k < intlen (nfilter P (x1a \<odot> xs) 0) \<Longrightarrow>
             nth (nfilter P (x1a \<odot> xs) 0) k < i \<Longrightarrow>
             i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<Longrightarrow>
             nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs) \<Longrightarrow> 
             \<exists>x\<in>set xs. P x \<Longrightarrow> 
             P x1a \<Longrightarrow> 
             i \<notin> set (nfilter P (x1a \<odot> xs) 0)" 
     using nfilter_not_between_help_a[of "xs" "P" "x1a" "k" "i" ]  by simp
      show "(\<And>k i. 
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               k < intlen (nfilter P xs 0) \<Longrightarrow>
               nth (nfilter P xs 0) k < i \<Longrightarrow>
               i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
               nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
               i \<notin> set (nfilter P xs 0)) \<Longrightarrow>
            \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
            k < intlen (nfilter P (x1a \<odot> xs) 0) \<Longrightarrow>
            nth (nfilter P (x1a \<odot> xs) 0) k < i \<Longrightarrow>
            i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<Longrightarrow>
            nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs) \<Longrightarrow> 
            \<exists>x\<in>set xs. P x \<Longrightarrow> 
            \<not> P x1a \<Longrightarrow> 
            i \<notin> set (nfilter P (x1a \<odot> xs) 0) "
        using nfilter_not_between_help_b[of "xs" "P" "x1a" "k" "i"] by simp
    qed
    show "(\<And>k i. 
             \<exists>a\<in>set xs. P a \<Longrightarrow>
             k < intlen (nfilter P xs 0) \<Longrightarrow>
             nth (nfilter P xs 0) k < i \<Longrightarrow>
             i < nth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
             nth (nfilter P xs 0) (Suc k) \<le> intlen xs \<Longrightarrow> 
             i \<notin> set (nfilter P xs 0)) \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          k < intlen (nfilter P (x1a \<odot> xs) 0) \<Longrightarrow>
          nth (nfilter P (x1a \<odot> xs) 0) k < i \<Longrightarrow>
          i < nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<Longrightarrow>
          nth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> intlen (x1a \<odot> xs) \<Longrightarrow> 
          \<not> (\<exists>x\<in>set xs. P x) \<Longrightarrow> 
          i \<notin> set (nfilter P (x1a \<odot> xs) 0) "
     by auto        
  qed
qed     


lemma nfilter_not_between:
 assumes "(\<exists> x \<in> set xs. P x)"
         " (nth (nfilter P xs 0) k ) < i "
         " i < (nth (nfilter P xs 0) (Suc k) ) " 
         " k < intlen (nfilter P xs 0)  " 
  shows  " \<not> P (nth xs (i))"
proof -
 have 0: "(nth (nfilter P xs 0) (Suc k) ) \<le> intlen xs "
      by (metis Suc_leI add_cancel_right_left assms(1) assms(4) nfilter_upper_bound)
 have 1: " i \<le> intlen xs" 
      using "0" assms(3) by linarith
 have 2: " k < intlen (nfilter P xs 0) \<and> (nth (nfilter P xs 0) k ) < i \<and>
            i < (nth (nfilter P xs 0) (Suc k) )  \<and> (nth (nfilter P xs 0) (Suc k) ) \<le> intlen xs\<longrightarrow>
           i \<notin> set (nfilter P xs 0)" 
   using assms(1) 
    proof (induction xs arbitrary: i k)
    case (St x)
    then show ?case by simp
    next
    case (Cons x1a xs)
    then show ?case using assms nfilter_not_between_help 
    by metis    
    qed  
 have 3 : "i \<notin> set (nfilter P xs 0) \<and>  i \<le> intlen xs \<longrightarrow>\<not> P (nth xs (i))" 
    by (metis add.right_neutral nfilter_holds_b nth_set)
 from 0 1 2 3  show ?thesis using assms by blast
qed



lemma idx_imp_distinct:
 assumes "index_sequence (nth xs 0) xs"
 shows   "distinct xs"
using assms 
proof (induction xs)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
 by (auto simp add: interval_idx_expand1)
    (metis interval_idx_expand1 interval_idx_greater_first interval_nth_and_set 
     interval_nth_zero_intfirst le_zero_eq not_less not_less_iff_gr_or_eq)
qed



lemma idx_set_eq:
 assumes "index_sequence (nth xs 0) xs"
         "index_sequence (nth ys 0) ys"
         "set xs = set ys"
 shows   "xs = ys"
using assms 
proof 
 (induction xs arbitrary: ys)
 case (St x)
 then show ?case 
   by (metis Suc_leI index_sequence_def interval.simps(15) interval_suffix_intlen 
       interval_suffix_zero le0 nat_neq_iff neq0_conv nth_set singletonD)
 next
 case (Cons x1a xs)
 then show ?case 
   proof (cases ys)
   case (St x1)
   then show ?thesis 
   using Cons.prems(1) Cons.prems(3) interval_idx_less_last_1 le_eq_less_or_eq nth_set by fastforce
   next
   case (Cons x21 x22)
   then show ?thesis  
    proof (cases "x1a = x21")
    case True
    then show ?thesis
     proof -
      have 1: "intlen (x1a\<odot>xs) = intlen (x21\<odot>x22)" 
           by (metis Cons.prems(1) Cons.prems(2) Cons.prems(3) distinct_card idx_imp_distinct 
               local.Cons nat.inject)
      have 2: "index_sequence (nth xs 0) xs"
           using Cons.prems(1) interval_idx_expand1 by auto
      have 3: "index_sequence (nth x22 0) x22"
           using Cons.prems(2) interval_idx_expand1 local.Cons by auto
      have 4: "distinct xs "
           using "2" idx_imp_distinct by auto
      have 5: "distinct x22"
           by (simp add: "3" idx_imp_distinct)
      have 6: " set (x1a\<odot>xs) = {x1a} \<union> set xs"
           by auto
      have 7: "x1a \<notin> set xs"
           by (meson Cons.prems(1) distinct.simps(2) idx_imp_distinct) 
      have 8: "set (x21\<odot>x22) = {x21} \<union> set x22 "
           by auto
      have 9: "x21 \<notin> set x22"
           using Cons.prems(2) idx_imp_distinct local.Cons by fastforce
      have 10: "set xs =set x22"
           using "7" "9" Cons.prems(3) True local.Cons by fastforce
      have 11: "xs = x22" 
           using "10" "2" "3" Cons.IH by blast
      have 12: "x1a\<odot> xs = x21\<odot>x22 "
           by (simp add: "11" True)
      show ?thesis by (simp add: "12" local.Cons)
     qed
    next
    case False
    then show ?thesis 
    by (metis Cons.prems(1) Cons.prems(2) Cons.prems(3) IntervalFilter.distinct.simps(2) 
        dual_order.strict_trans idx_imp_distinct interval_hd_in_set interval_idx_expand1 
        interval_nth_zero interval_set_ConsD local.Cons not_less_iff_gr_or_eq)
   qed
  qed
qed

lemma filter_nfilter_prefix_idx_a:
  assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
  shows   "  index_sequence (nth (prefix k (nfilter P xs 0)) 0) (prefix k (nfilter P xs 0))"
using assms by (auto simp add: index_sequence_def)
               (metis diff_zero filter_nfilter_prefix_intlen_0 idx_nfilter_mono)

lemma filter_nfilter_prefix_idx_a_1:
  assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
  shows   "  index_sequence (nth (prefix k (nfilter P xs 0)) 0) (prefix k (nfilter P xs 0))"
using assms by (auto simp add: index_sequence_def)
               (meson idx_nfilter_mono)

lemma filter_nfilter_suffix_idx_a:
  assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
  shows   "  index_sequence (nth (suffix k (nfilter P xs 0)) 0) (suffix k (nfilter P xs 0))"
using assms by (simp add: index_sequence_def)
 (metis add.commute diff_zero filter_nfilter_prefix_intlen_0 idx_nfilter_mono less_diff_conv)

lemma filter_nfilter_suffix_idx_a_1:
  assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
  shows   "  index_sequence (nth (suffix k (nfilter P xs 0)) 0) (suffix k (nfilter P xs 0))"
using assms
by (simp add: index_sequence_def idx_nfilter_mono nfilter_intlen)

lemma filter_nfilter_prefix_idx_b:
  assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
  shows   "index_sequence (nth (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) 0)
             (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
using assms idx_nfilter[of "(prefix (nth (nfilter P xs 0) k) xs)" "P" "0"] 
by (metis add.right_neutral interval_intlast_intfirst interval_nth_intlen_intlast 
    interval_nth_suffix interval_nth_zero_intfirst le0 le_refl nth_set)

lemma filter_nfilter_prefix_idx_b_1:
  assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
  shows   "index_sequence (nth (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) 0)
             (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
using assms 
by (metis filter_nfilter_prefix_idx_b nfilter_holds nfilter_intlen nfilter_nth_n_zero nth_set)


lemma filter_nfilter_suffix_idx_b:
 assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
 shows   "index_sequence (nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)
                              (nth (nfilter P xs 0) k)) 0)
             (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 
                        (nth (nfilter P xs 0) k))"
using assms 
by (auto simp add: index_sequence_def)
 (metis idx_nfilter_mono interval_intfirst_suffix interval_intlen_gr_zero interval_nth_zero_intfirst 
   interval_suffix_length_code length_nfilter_le less_le_trans not_less0 nth_set)

lemma filter_nfilter_suffix_idx_b_1:
  assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
  shows   "index_sequence (nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)
                                          (nth (nfilter P xs 0) k)) 0)
             (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k))"
using assms 
by (metis filter_nfilter_suffix_idx_b nfilter_holds nfilter_intlen nfilter_nth_n_zero nth_set)

lemma filter_nfilter_prefix_set_eq:
  assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
  shows   " set (prefix k (nfilter P xs 0)) =
            set (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
proof -
 have 1: "(nth (nfilter P xs 0) k)  \<le> intlen xs"
    by (metis assms(1) assms(2) diff_zero filter_nfilter_prefix_intlen_0 interval_intlen_gr_zero 
    nfilter_intlen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> set xs . P x "
    using "1" assms(1) nth_set by blast 
 have 3: "{ i. i\<le> intlen(prefix ((nth (nfilter P xs 0) k)) xs) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)} =

          { i. i\<le> (nth (nfilter P xs 0) k) \<and> P (nth xs i)}"
   using "1" by auto 
 have 4: "\<exists> x \<in> set(prefix ((nth (nfilter P xs 0) k))    xs). P x"
  by (metis "1" assms(1) interval_nth_prefix interval_prefix_length_good nth_set order_refl) 
 have 5: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and> P (nth xs i)} = 
           { i. i\<le> (nth (nfilter P xs 0) k) \<and> i\<in>set(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 6: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and> i\<in>set(nfilter P xs 0)} =
           { i. i\<le> (nth (nfilter P xs 0) k) \<and> 
                i\<in>{(nth (nfilter P xs 0) j)| j. j\<le> intlen(nfilter P xs 0)}   }"
      by (auto simp add: interval_nth_and_set)
 have 7: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and>
               i\<in>{(nth (nfilter P xs 0) j)| j. j\<le> intlen(nfilter P xs 0)}   } =
           { (nth (nfilter P xs 0) j) | j. j\<le> k }"
    using assms "2" by (auto simp add: nfilter_intlen, 
        metis dual_order.antisym idx_nfilter interval_idx_less_eq le_cases nfilter_intlen,
        metis idx_nfilter_less_eq nfilter_intlen)
 have 8: "k\<le> intlen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_intlen)
 have 9: "{ (nth (nfilter P xs 0) j) | j. j\<le> k } = set (prefix k (nfilter P xs 0))"
    using "8" "2" using prefix_set by force
 have 10: "set (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) =
            { i. i\<le> ((nth (nfilter P xs 0) k)) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)}"
   using "4" "1"  nfilter_prefix_set by auto
 have 11: "intlen(prefix ((nth (nfilter P xs 0) k)) xs) = ((nth (nfilter P xs 0) k))"
   using "1" interval_prefix_length_good by blast
 have 12: "{ i. i\<le> ((nth (nfilter P xs 0) k)) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)} =
            { i. i\<le> intlen(prefix ((nth (nfilter P xs 0) k)) xs) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)}"
    using "11" by auto
 show ?thesis 
 using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma filter_nfilter_prefix_set_eq_1:
  assumes "\<exists>x \<in> set xs . P x"
           "k \<le> intlen (filter P xs)" 
  shows   " set (prefix k (nfilter P xs 0)) =
            set (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
proof -
 have 1: "(nth (nfilter P xs 0) k)  \<le> intlen xs"
    by (metis assms(1) assms(2) diff_zero  interval_intlen_gr_zero 
    nfilter_intlen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> set xs . P x "
    using "1" assms(1) nth_set by blast 
 have 3: "{ i. i\<le> intlen(prefix ((nth (nfilter P xs 0) k)) xs) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)} =

          { i. i\<le> (nth (nfilter P xs 0) k) \<and> P (nth xs i)}"
   using "1" by auto 
 have 4: "\<exists> x \<in> set(prefix ((nth (nfilter P xs 0) k))    xs). P x"
    using assms "1" "2" nfilter_holds[of "xs" "P" "0"] 
    by (metis interval_intlast_prefix interval_nth_intlen_intlast le_refl nfilter_intlen 
        nfilter_nth_n_zero nth_set) 
 have 5: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and> P (nth xs i)} = 
           { i. i\<le> (nth (nfilter P xs 0) k) \<and> i\<in>set(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 6: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and> i\<in>set(nfilter P xs 0)} =
           { i. i\<le> (nth (nfilter P xs 0) k) \<and> 
                i\<in>{(nth (nfilter P xs 0) j)| j. j\<le> intlen(nfilter P xs 0)}   }"
    by (auto simp add: interval_nth_and_set)
 have 7: "{ i. i\<le> (nth (nfilter P xs 0) k) \<and> 
               i\<in>{(nth (nfilter P xs 0) j)| j. j\<le> intlen(nfilter P xs 0)}   } =
           { (nth (nfilter P xs 0) j) | j. j\<le> k }"
   using assms "2" by (auto simp add: nfilter_intlen, 
        metis dual_order.antisym idx_nfilter interval_idx_less_eq le_cases nfilter_intlen,
        metis idx_nfilter_less_eq nfilter_intlen)
 have 8: "k\<le> intlen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_intlen)
 have 9: "{ (nth (nfilter P xs 0) j) | j. j\<le> k } = set (prefix k (nfilter P xs 0))"
    using "8" "2" using prefix_set by force
 have 10: "set (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) =
            { i. i\<le> ((nth (nfilter P xs 0) k)) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)}"
   using "4" "1"  nfilter_prefix_set by auto
 have 11: "intlen(prefix ((nth (nfilter P xs 0) k)) xs) = ((nth (nfilter P xs 0) k))"
   using "1" interval_prefix_length_good by blast
 have 12: "{ i. i\<le> ((nth (nfilter P xs 0) k)) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)} =
            { i. i\<le> intlen(prefix ((nth (nfilter P xs 0) k)) xs) \<and> 
                    P (nth (prefix ((nth (nfilter P xs 0) k)) xs) i)}"
    using "11" by auto
 show ?thesis 
 using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma filter_nfilter_suffix_set_eq:
  assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
  shows   " set (nfilter P (suffix (nth (nfilter P xs 0) k)  xs) (nth (nfilter P xs 0) k)) =
            set (suffix k (nfilter P xs 0)) 
            "
proof -
 have 1: "(nth (nfilter P xs 0) k)  \<le> intlen xs"
    by (metis assms(1) assms(2) diff_zero filter_nfilter_prefix_intlen_0 interval_intlen_gr_zero 
    nfilter_intlen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> set xs . P x "
    using "1" assms(1) nth_set by blast 
 have 4: "\<exists> x \<in> set(suffix ((nth (nfilter P xs 0) k))    xs). P x"
    using "1" assms(1) interval_nth_and_set by force
 have 10: "set (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)) =
            { (nth (nfilter P xs 0) k)+i| i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
                    P (nth xs (i+(nth (nfilter P xs 0) k)) )}
          "
   using nfilter_suffix_set_b[of "((nth (nfilter P xs 0) k))" "xs" "P" "(nth (nfilter P xs 0) k)"] 
    using "1" "4" by blast 
 have 5: " { (nth (nfilter P xs 0) k)+i|i.  i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
              P (nth xs (i+(nth (nfilter P xs 0) k))  )}
            = 
           { (nth (nfilter P xs 0) k)+i|i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
              i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 51: "{ (nth (nfilter P xs 0) k)+i |i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
               i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)} =
           {  (nth (nfilter P xs 0) k)+i| i. 
                (nth (nfilter P xs 0) k) \<le> i + (nth (nfilter P xs 0) k) \<and>
                i + (nth (nfilter P xs 0) k)\<le> intlen xs \<and> 
                i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)}
          " 
          using "1" by auto
 have 52: "{ (nth (nfilter P xs 0) k)+i| i. 
                (nth (nfilter P xs 0) k) \<le> i + (nth (nfilter P xs 0) k) \<and>
                i + (nth (nfilter P xs 0) k)\<le> intlen xs \<and> 
                i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)} =
           { j. (nth (nfilter P xs 0) k) \<le> j \<and> j\<le> intlen xs \<and> j\<in>set(nfilter P xs 0)}
          " 
      by (metis (no_types, lifting) add_diff_cancel_right' le0 le_add_diff_inverse 
          le_add_same_cancel2)
 have 53: "{ j. (nth (nfilter P xs 0) k) \<le> j \<and> j\<le> intlen xs \<and> j\<in>set(nfilter P xs 0)} =
           { j. (nth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> intlen xs \<and> j\<in>{(nth (nfilter P xs 0) jj)| jj. jj\<le> intlen(nfilter P xs 0)}  }
          " 
     by (auto simp add: interval_nth_and_set)
 have 54: "{ j. (nth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> intlen xs \<and> j\<in>{(nth (nfilter P xs 0) jj)| jj. jj\<le> intlen(nfilter P xs 0)}  } =
            { (nth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> intlen (nfilter P xs 0) }
          " 
  using assms "2" by (auto,
      metis dual_order.antisym idx_nfilter_less_eq le_cases nfilter_intlen,
      metis idx_nfilter_gr_eq,
      metis add_cancel_right_left nfilter_upper_bound)
 have 8: "k\<le> intlen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_intlen)
 have 9: "{ (nth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> intlen (nfilter P xs 0) } = 
          set (suffix k (nfilter P xs 0))"
    using "8" "2" using suffix_set_a by blast
 show ?thesis 
 using "10"  "5" "51" "52" "53" "54" "9" by simp 
qed

lemma filter_nfilter_suffix_set_eq_1:
  assumes "\<exists>x \<in> set xs . P x"
           "k \<le> intlen (filter P xs)" 
  shows   " set (nfilter P (suffix (nth (nfilter P xs 0) k)  xs) (nth (nfilter P xs 0) k)) =
            set (suffix k (nfilter P xs 0)) 
            "
proof -
 have 1: "(nth (nfilter P xs 0) k)  \<le> intlen xs"
    by (metis assms(1) assms(2) diff_zero interval_intlen_gr_zero 
    nfilter_intlen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> set xs . P x "
    using  assms(1) by blast 
 have 4: "\<exists> x \<in> set(suffix ((nth (nfilter P xs 0) k))    xs). P x"
   using nfilter_holds[of "xs" "P" "0"] 
   by (metis "1" "2" assms(2) interval_intfirst_suffix interval_nth_zero_intfirst le0 
       nfilter_intlen nfilter_nth_n_zero nth_set) 
 have 10: "set (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)) =
            { (nth (nfilter P xs 0) k)+i| i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
                    P (nth xs (i+(nth (nfilter P xs 0) k)) )}
          "
   using nfilter_suffix_set_b[of "(nth (nfilter P xs 0) k)" "xs" "P" "(nth (nfilter P xs 0) k)"] 
     using "1" "4" by blast
 have 5: " { (nth (nfilter P xs 0) k)+i|i.  i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
              P (nth xs (i+(nth (nfilter P xs 0) k))  )}
            = 
           { (nth (nfilter P xs 0) k)+i|i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
              i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 51: "{ (nth (nfilter P xs 0) k)+i |i. i\<le> intlen xs - (nth (nfilter P xs 0) k) \<and> 
               i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)} =
           {  (nth (nfilter P xs 0) k)+i| i. 
                (nth (nfilter P xs 0) k) \<le> i + (nth (nfilter P xs 0) k) \<and>
                i + (nth (nfilter P xs 0) k)\<le> intlen xs \<and> 
                i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)}
          " 
          using "1" by auto
 have 52: "{ (nth (nfilter P xs 0) k)+i| i. 
                (nth (nfilter P xs 0) k) \<le> i + (nth (nfilter P xs 0) k) \<and>
                i + (nth (nfilter P xs 0) k)\<le> intlen xs \<and> 
                i+(nth (nfilter P xs 0) k)\<in>set(nfilter P xs 0)} =
           { j. (nth (nfilter P xs 0) k) \<le> j \<and> j\<le> intlen xs \<and> j\<in>set(nfilter P xs 0)}
          " 
      by (metis (no_types, lifting) add_diff_cancel_right' le0 le_add_diff_inverse 
          le_add_same_cancel2)
 have 53: "{ j. (nth (nfilter P xs 0) k) \<le> j \<and> j\<le> intlen xs \<and> j\<in>set(nfilter P xs 0)} =
           { j. (nth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> intlen xs \<and> j\<in>{(nth (nfilter P xs 0) jj)| jj. jj\<le> intlen(nfilter P xs 0)}  }
          " 
     by (auto simp add: interval_nth_and_set)
 have 54: "{ j. (nth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> intlen xs \<and> j\<in>{(nth (nfilter P xs 0) jj)| jj. jj\<le> intlen(nfilter P xs 0)}  } =
            { (nth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> intlen (nfilter P xs 0) }
          " 
    using assms "2" by (auto,
      metis  dual_order.antisym idx_nfilter_less_eq le_cases nfilter_intlen,
      simp add: "2" idx_nfilter_less_eq,
      metis add_cancel_right_left nfilter_upper_bound)
 have 8: "k\<le> intlen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_intlen)
 have 9: "{ (nth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> intlen (nfilter P xs 0) } = 
          set (suffix k (nfilter P xs 0))"
    using "8" "2" using suffix_set_a by blast
 show ?thesis 
 using "10"  "5" "51" "52" "53" "54" "9" by simp
qed

lemma nfilter_nfilter_prefix:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " (prefix k (nfilter P xs 0))   = 
             (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
using assms  
by (meson filter_nfilter_prefix_idx_a filter_nfilter_prefix_idx_b 
    filter_nfilter_prefix_set_eq idx_set_eq)


lemma nfilter_nfilter_prefix_1:
   assumes "\<exists>x \<in> set xs . P x "
           "k \<le> intlen (filter P xs)" 
   shows   " (prefix k (nfilter P xs 0))   = 
             (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
using assms  
by (meson filter_nfilter_prefix_idx_a_1 filter_nfilter_prefix_idx_b_1 
    filter_nfilter_prefix_set_eq_1 idx_set_eq)



lemma nfilter_nfilter_suffix:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " (suffix k (nfilter P xs 0))   = 
             (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)  (nth (nfilter P xs 0) k))"
using assms 
by (metis filter_nfilter_suffix_idx_a filter_nfilter_suffix_idx_b 
    filter_nfilter_suffix_set_eq idx_set_eq)

lemma nfilter_nfilter_suffix_1:
   assumes "\<exists>x \<in> set xs . P x "
           "k \<le> intlen (filter P xs)" 
   shows   " (suffix k (nfilter P xs 0))   = 
             (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)  (nth (nfilter P xs 0) k))"
proof -
 have 1: "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
   by (metis assms(1) assms(2) nfilter_holds nfilter_intlen nfilter_nth_n_zero nth_set) 
 show ?thesis using assms "1" nfilter_nfilter_suffix by auto
qed

lemma nfilter_map_filter:
 assumes "(\<exists> x \<in> set xs. P x)"
 shows   " map (\<lambda>n. (nth xs n)) (nfilter P xs 0) = filter P xs"
proof -
 have 1: " intlen (map (\<lambda>n. (nth xs n)) (nfilter P xs 0)) = intlen (filter P xs)"
   by (simp add: assms nfilter_intlen)
 have 2: "\<And>i. i\<le> intlen (map (\<lambda>n. (nth xs n)) (nfilter P xs 0)) \<longrightarrow>
           (nth (map (\<lambda>n. (nth xs n)) (nfilter P xs 0)) i) =
           (nth (filter P xs) i)"
    by (metis assms diff_zero interval_intlen_map interval_nth_map nfilter_filter)
  from 1 2 show ?thesis 
  using interval_eq_nth_eq by blast
qed

lemma filter_nfilter_prefix:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " (prefix k (filter P xs ))   = 
             (filter P (prefix ((nth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> set xs . P x "
   by (metis assms(1) assms(2) diff_zero filter_nfilter_prefix_intlen_0)
 have 2: "(filter P xs ) = map (\<lambda>s. nth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_map_filter)
 have 3: "(prefix k (filter P xs )) =
          (prefix k (map (\<lambda>s. nth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(prefix k (map (\<lambda>s. nth xs s) (nfilter P xs 0))) =
          map (\<lambda>s. nth xs s) (prefix k (nfilter P xs 0))"
    by (simp add: "1" assms(2) map_prefix nfilter_intlen)
 have 5: "\<exists> x \<in> set(prefix ((nth (nfilter P xs 0) k))    xs). P x"
   by (metis "1" add.left_neutral assms(1) assms(2) interval_intlast_prefix interval_nth_and_set 
       interval_nth_intlen_intlast nfilter_intlen nfilter_upper_bound order_refl)
 have 6: "(filter P (prefix ((nth (nfilter P xs 0) k) ) xs) ) =
          map (\<lambda>s. (nth (prefix ((nth (nfilter P xs 0) k) ) xs) s)) 
           (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)"
      by (simp add: "5" nfilter_map_filter)
 have 7: "map (\<lambda>s. nth xs s) (prefix k (nfilter P xs 0)) =
          map (\<lambda>s. nth xs s) (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0)" 
   by (simp add: assms(1) assms(2) nfilter_nfilter_prefix)
 have 8: "map (\<lambda>s. nth xs s) (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. (nth (prefix ((nth (nfilter P xs 0) k) ) xs) s)) 
           (nfilter P (prefix ((nth (nfilter P xs 0) k) ) xs) 0) 
         " 
     using "1" "5" by simp
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8")     
qed

lemma filter_nfilter_prefix_1:
   assumes "\<exists>x \<in> set xs . P x"
           "k \<le> intlen (filter P xs)" 
   shows   " (prefix k (filter P xs ))   = 
             (filter P (prefix ((nth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "P (nth xs   ( (nth (nfilter P xs 0) k)) )" 
    by (metis assms(1) assms(2) nfilter_holds nfilter_intlen nfilter_nth_n_zero nth_set)
 show ?thesis using assms "1" filter_nfilter_prefix by auto
qed

lemma filter_nfilter_prefix_intlen:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " intlen(prefix k (filter P xs))   = 
             intlen(filter P (prefix ((nth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: filter_nfilter_prefix)

lemma filter_nfilter_prefix_intlen_1:
   assumes "\<exists>x \<in> set xs . P x"
           "k \<le> intlen (filter P xs)" 
   shows   " intlen(prefix k (filter P xs))   = 
             intlen(filter P (prefix ((nth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: filter_nfilter_prefix_1)

lemma filter_nfilter_prefix_nth:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
           " j\<le>  intlen(prefix k (filter P xs))"
   shows   " nth (prefix k (filter P xs)) j  = 
             nth (filter P (prefix ((nth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: filter_nfilter_prefix)

lemma filter_nfilter_prefix_nth_1:
   assumes "\<exists>x \<in> set xs . P x"
           "k \<le> intlen (filter P xs)" 
           " j\<le>  intlen(prefix k (filter P xs)) "
   shows   " nth (prefix k (filter P xs)) j  = 
             nth (filter P (prefix ((nth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: filter_nfilter_prefix_1)

lemma nfilter_map_shift:
assumes "\<exists> x \<in> set xs. P x"
shows "map (\<lambda>s. nth xs (s+n)) (nfilter P xs 0) =
       map (\<lambda>s. nth xs s) (nfilter P xs n)"
proof -
 have 1: " intlen (map (\<lambda>s. nth xs (s+n)) (nfilter P xs 0)) =
           intlen (map (\<lambda>s. nth xs s) (nfilter P xs n))"
   by (metis assms(1) interval_intlen_map nfilter_intlen_n_zero)
 have 2: "\<And>i. nth (map (\<lambda>s. nth xs (s+n)) (nfilter P xs 0)) i =
              (nth xs ( (nth (nfilter P xs 0) i) +n))"
     by (simp add: interval_nth_map)
 have 3: "\<And>i. nth ( map (\<lambda>s. nth xs s) (nfilter P xs n) ) i =
              (nth xs ( (nth (nfilter P xs n) i)))"
    by (simp add: interval_nth_map)
 have 4: "\<And>i. (nth (nfilter P xs 0) i) +n = (nth (nfilter P xs n) i)"
    by (metis assms(1) interval_nth_map nfilter_n_zero)
 show ?thesis 
 by (metis "1" "2" "3" "4" interval_eq_nth_eq)
qed

lemma nfilter_map_shift_suffix:
 assumes " \<exists> x \<in> set(suffix ((nth (nfilter P xs 0) k))    xs). P x"
 shows   "map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k))"
proof -
 have 1: "intlen (map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)) =
          intlen (map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)))" 
    by (metis assms interval_intlen_map nfilter_intlen_n_zero)
 have 2: "\<And>i. nth (map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)) i =
           nth xs 
           ((nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) i) + (nth (nfilter P xs 0) k))
           " 
    using interval_nth_map by blast
 have 3: "\<And>i. 
          nth ( map (\<lambda>s. nth xs s) 
                    (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)
                               (nth (nfilter P xs 0) k))) i =
          nth xs (nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs)
                                 (nth (nfilter P xs 0) k)) i)
          " 
    using interval_nth_map by blast
 have 4: "\<And>i. (nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) i) 
               + (nth (nfilter P xs 0) k) =
               (nth (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)) i) "
  by (metis assms interval_nth_map nfilter_n_zero)
 show ?thesis 
 by (metis "1" "2" "3" "4" interval_eq_nth_eq)
qed

lemma filter_nfilter_suffix:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " (suffix k (filter P xs ))   = 
             (filter P (suffix ((nth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> set xs . P x "
   by (metis assms(1) assms(2) diff_zero filter_nfilter_prefix_intlen_0)
 have 2: "(filter P xs ) = map (\<lambda>s. nth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_map_filter)
 have 3: "(suffix k (filter P xs )) =
          (suffix k (map (\<lambda>s. nth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(suffix k (map (\<lambda>s. nth xs s) (nfilter P xs 0))) =
          map (\<lambda>s. nth xs s) (suffix k (nfilter P xs 0))"
  by (simp add: "1" assms(2) map_suffix nfilter_intlen)
 have 5: "\<exists> x \<in> set(suffix ((nth (nfilter P xs 0) k))    xs). P x"
   by (metis "1" add_cancel_right_left assms(1) assms(2) interval_intfirst_suffix 
       interval_intlen_gr_zero interval_nth_and_set interval_nth_zero_intfirst nfilter_intlen 
       nfilter_upper_bound)   
 have 6: "(filter P (suffix ((nth (nfilter P xs 0) k) ) xs) ) =
          map (\<lambda>s. nth (suffix ((nth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)
         "
     by (simp add: "5" nfilter_map_filter)
 have 7: "map (\<lambda>s. nth (suffix ((nth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)
         " 
 using "1" "5" by (simp add: add.commute)
 have 8: "map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k))
         " 
   using "5" nfilter_map_shift_suffix by metis
 have 9: "map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)) =
          map (\<lambda>s. nth xs s)
           (suffix k (nfilter P xs 0))
         " 
     by (simp add: assms(1) assms(2) nfilter_nfilter_suffix)
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8" "9")
qed


lemma filter_nfilter_suffix_1:
   assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
   shows   " (suffix k (filter P xs ))   = 
             (filter P (suffix ((nth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> set xs . P x "
   using assms by auto
 have 2: "(filter P xs ) = map (\<lambda>s. nth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_map_filter)
 have 3: "(suffix k (filter P xs )) =
          (suffix k (map (\<lambda>s. nth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(suffix k (map (\<lambda>s. nth xs s) (nfilter P xs 0))) =
          map (\<lambda>s. nth xs s) (suffix k (nfilter P xs 0))"
  by (simp add: "1" assms(2) map_suffix nfilter_intlen)
 have 5: "\<exists> x \<in> set(suffix ((nth (nfilter P xs 0) k))    xs). P x"
  using assms nth_set  "2"  
  by (metis add.right_neutral diff_zero interval_intlen_map interval_nth_suffix le0 nfilter_holds)
 have 6: "(filter P (suffix ((nth (nfilter P xs 0) k) ) xs) ) =
          map (\<lambda>s. nth (suffix ((nth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)
         "
     by (simp add: "5" nfilter_map_filter)
 have 7: "map (\<lambda>s. nth (suffix ((nth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0)
         " 
   using "1" "5" 
   by (metis add.right_neutral interval_nth_suffix interval_suffix_suffix le0)
 have 8: "map (\<lambda>s. nth xs (s+(nth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) 0) =
          map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k))
         " 
   using "5" nfilter_map_shift_suffix by metis
 have 9: "map (\<lambda>s. nth xs s) 
           (nfilter P (suffix ((nth (nfilter P xs 0) k) ) xs) (nth (nfilter P xs 0) k)) =
          map (\<lambda>s. nth xs s)
           (suffix k (nfilter P xs 0))
         " 
     by (simp add: assms(1) assms(2) nfilter_nfilter_suffix_1)
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8" "9")
qed

lemma filter_nfilter_suffix_intlen:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
   shows   " intlen(suffix k (filter P xs))   = 
             intlen(filter P (suffix ((nth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: filter_nfilter_suffix)

lemma filter_nfilter_suffix_intlen_1:
   assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
   shows   " intlen(suffix k (filter P xs))   = 
             intlen(filter P (suffix ((nth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: filter_nfilter_suffix_1)

lemma filter_nfilter_suffix_nth:
   assumes "P (nth xs   ( (nth (nfilter P xs 0) k)) )"
           "k \<le> intlen (filter P xs)" 
           " j\<le>  intlen(suffix k (filter P xs)) " 
   shows   " nth (suffix k (filter P xs)) j  = 
             nth (filter P (suffix ((nth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: filter_nfilter_suffix)

lemma filter_nfilter_suffix_nth_1:
   assumes "\<exists> x \<in> set xs. P x"
           "k \<le> intlen (filter P xs)" 
           " j\<le>  intlen(suffix k (filter P xs)) " 
   shows   " nth (suffix k (filter P xs)) j  = 
             nth (filter P (suffix ((nth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: filter_nfilter_suffix_1)

lemma nfilter_intlast:
 assumes " P (intlast (prefix k xs))"
         "k \<le> intlen xs" 
 shows   " (nth xs ((intlast (nfilter P (prefix k xs) n))-n)) = (intlast (prefix k xs))"
using assms
proof (induction xs arbitrary: n k)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (auto simp add: min.absorb1 split:nat.split)
    show "\<And>x2 x.
           (\<And>k n. P (nth xs k) \<Longrightarrow>
               k \<le> intlen xs \<Longrightarrow>
               nth xs (nth (nfilter P (prefix k xs) n) (intlen (nfilter P (prefix k xs) n)) - n) =
               nth xs k) \<Longrightarrow>
          P (nth xs x2) \<Longrightarrow>
          x \<in> set (prefix x2 xs) \<Longrightarrow>
          P x \<Longrightarrow>
          k = Suc x2 \<Longrightarrow>
          x2 \<le> intlen xs \<Longrightarrow>
          nth (nfilter P (prefix x2 xs) (Suc n)) (intlen (nfilter P (prefix x2 xs) (Suc n))) \<le> n \<Longrightarrow> 
          x1a = nth xs x2" 
     by (metis le_refl nfilter_nth_n_zero_a not_less_eq_eq)
   show "\<And>x2 x x2a.
          (\<And>k n. P (nth xs k) \<Longrightarrow>
               k \<le> intlen xs \<Longrightarrow>
               nth xs (nth (nfilter P (prefix k xs) n) (intlen (nfilter P (prefix k xs) n)) - n) =
               nth xs k) \<Longrightarrow>
         P (nth xs x2) \<Longrightarrow>
         x \<in> set (prefix x2 xs) \<Longrightarrow>
         P x \<Longrightarrow>
         k = Suc x2 \<Longrightarrow>
         x2 \<le> intlen xs \<Longrightarrow>
         nth (nfilter P (prefix x2 xs) (Suc n)) (intlen (nfilter P (prefix x2 xs) (Suc n))) - n = 
         Suc x2a \<Longrightarrow>
         nth xs x2a = nth xs x2" 
   by (metis Suc_eq_plus1 add_diff_cancel_left' diff_diff_left plus_1_eq_Suc)  
   show "\<And>x2. 
          (\<And>k n. P (nth xs k) \<Longrightarrow>
             k \<le> intlen xs \<Longrightarrow>
             nth xs (nth (nfilter P (prefix k xs) n) (intlen (nfilter P (prefix k xs) n)) - n) =
                  nth xs k) \<Longrightarrow>
          P (nth xs x2) \<Longrightarrow>
          \<forall>x\<in>set (prefix x2 xs). \<not> P x \<Longrightarrow> 
          k = Suc x2 \<Longrightarrow> 
          x2 \<le> intlen xs \<Longrightarrow> 
          x1a = nth xs x2" 
   using interval_intlast_prefix nth_set by fastforce
 qed
qed

lemma nfilter_intfirst:
 assumes " P (intfirst (suffix k xs)) "
         "k \<le> intlen xs" 
 shows   "intfirst (nfilter P (suffix k xs) n)= n "
using assms
proof (induction xs arbitrary: k n)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
     by (auto  split:nat.split)
qed

lemma filter_intlast:
 assumes " P (intlast (prefix n xs))"
         "n \<le> intlen xs"
 shows   " intlast (filter P (prefix n xs)) = (intlast (prefix n xs))"
using assms   
proof (induction n arbitrary: xs)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
   proof (cases xs)
   case (St x1)
   then show ?thesis 
   by simp
   next
   case (Cons x21 x22)
   then show ?thesis 
        proof auto
         show "\<And>x. xs = x21 \<odot> x22 \<Longrightarrow>
                x \<in> set (prefix n x22) \<Longrightarrow>
                P x \<Longrightarrow>
                nth (filter P (prefix n x22)) (intlen (filter P (prefix n x22))) =
                nth x22 (min n (intlen x22))" 
          using Suc.IH Suc.prems by auto
         show "xs = x21 \<odot> x22 \<Longrightarrow> 
               \<forall>x\<in>set (prefix n x22). \<not> P x \<Longrightarrow> 
               x21 = nth x22 (min n (intlen x22)) " 
          using Suc.prems nth_set by force
        qed
   qed
qed

lemma filter_intfirst:
 assumes " P (intfirst (suffix n xs))"
 shows   " intfirst (filter P (suffix n xs)) = (intfirst (suffix n xs))"
using assms 
proof (induction xs arbitrary: n)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case by (auto split:nat.split)
qed

lemma filter_nth_aa:
 assumes "(\<exists> x \<in> set xs. P x)"
         "n \<le> intlen (filter P xs)"
 shows   "P (nth (filter P xs) n)"
using assms set_filter nth_set by fastforce



lemma filter_length_zero_conv_a:
 assumes "(\<exists>x\<in>set xs. P x)"
         "intlen (filter P xs) =0"
 shows   " (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j))) "
proof -
 have 1: "P (nth xs  (nth (nfilter P xs 0) 0)) "
   by (metis nfilter_map_filter assms(1) assms(2) filter_nth_aa interval_nth_map 
       le_numeral_extra(3)) 
 have 2: "(nth (nfilter P xs 0) 0) \<le> intlen xs" 
   by (metis assms(1) diff_zero interval_intlen_gr_zero nfilter_upper_bound 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 3: "(\<forall> j \<le> intlen xs. j \<noteq> (nth (nfilter P xs 0) 0) \<longrightarrow> \<not> P (nth xs j)) " 
    using assms nfilter_not_after[of "xs" "P" ] nfilter_not_before[of "xs" "P"] 
    by (metis linorder_neqE_nat nfilter_intlen)
 show ?thesis
 using "1" "2" "3" by blast
qed

lemma filter_length_zero_conv_c:
 "(\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j))) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j)))
 " 
using antisym_conv3 by  auto

lemma filter_length_zero_conv_d:
 " (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j))) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (nth xs j)) \<and>
     (\<forall> j \<le> intlen xs. k<j  \<longrightarrow> \<not> P (nth xs j))
   )
 " 
by auto

lemma filter_length_zero_conv_b:
 assumes "(\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j)))"
 shows   "(\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0 "
proof -
 have 1: "(\<exists>x\<in>set xs. P x)" 
   using assms nth_set by auto
 obtain k where 2: "k \<le> intlen xs \<and>  P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j))" 
   using assms by auto
 have 3: "intlen (filter P xs) =0"
  using 1 2 
   proof (induct xs arbitrary: k)
   case (St x)
   then show ?case by simp
   next
   case (Cons x1a xs)
   then show ?case 
        proof (cases k)
         show "(\<And>k. \<exists>a\<in>set xs. P a \<Longrightarrow>
                     k \<le> intlen xs \<and> P (nth xs k) \<and> (\<forall>j\<le>intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j)) \<Longrightarrow> 
                     intlen (filter P xs) = 0) \<Longrightarrow>
               \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
               k \<le> intlen (x1a \<odot> xs) \<and> P (nth (x1a \<odot> xs) k) \<and> 
               (\<forall>j\<le>intlen (x1a \<odot> xs). j \<noteq> k \<longrightarrow> \<not> P (nth (x1a \<odot> xs) j)) \<Longrightarrow>
               k = 0 \<Longrightarrow> intlen (filter P (x1a \<odot> xs)) = 0"
       by (metis Suc_le_mono filter_intlen_d interval_nth_Suc interval_nth_and_set intlen.simps(2) 
           le0  not_less_eq_eq order_refl plus_1_eq_Suc)
    show "\<And>nat. 
              (\<And>k. \<exists>a\<in>set xs. P a \<Longrightarrow>
                 k \<le> intlen xs \<and> P (nth xs k) \<and>
                  (\<forall>j\<le>intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j)) \<Longrightarrow> 
                   intlen (filter P xs) = 0) \<Longrightarrow>
              \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
              k \<le> intlen (x1a \<odot> xs) \<and> P (nth (x1a \<odot> xs) k) \<and> 
              (\<forall>j\<le>intlen (x1a \<odot> xs). j \<noteq> k \<longrightarrow> \<not> P (nth (x1a \<odot> xs) j)) \<Longrightarrow>
              k = Suc nat \<Longrightarrow> 
              intlen (filter P (x1a \<odot> xs)) = 0 " 
      by fastforce 
    qed
   qed
 show ?thesis 
 using "1" "3" by blast
qed

lemma filter_length_zero_conv:
 " ((\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j)))"
using filter_length_zero_conv_a[of "xs" "P"]  filter_length_zero_conv_b[of "xs" "P"]
by blast

lemma filter_length_zero_conv_1:
 " ((\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j)))"
proof -
 have 1: "((\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0) = 
          (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j)))" 
    by (simp add: filter_length_zero_conv)
 have 2: "(\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j \<noteq> k \<longrightarrow> \<not> P (nth xs j))) =
          (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j)))" 
        by fastforce
 show ?thesis by (simp add: "1" "2")
qed

lemma filter_length_zero_conv_2:
 " ((\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (nth xs j)) \<and>
     (\<forall> j \<le> intlen xs. k<j  \<longrightarrow> \<not> P (nth xs j))
   )"
proof -
 have 1: "((\<exists>x\<in>set xs. P x) \<and>  intlen (filter P xs) =0) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j)))"
   by (simp add: filter_length_zero_conv_1) 
 have 2: "(\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
             (\<forall> j \<le> intlen xs. j<k \<or> k<j \<longrightarrow> \<not> P (nth xs j))) =
   (\<exists> k \<le> intlen xs. P (nth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (nth xs j)) \<and>
     (\<forall> j \<le> intlen xs. k<j  \<longrightarrow> \<not> P (nth xs j))
   ) "
   using dual_order.strict_trans1 by auto 
 from 1 2 show ?thesis   by auto
qed 

lemma filter_suffixes_map_help_0_0:
assumes "\<exists> x \<in> set (x1a\<odot>xs). P x" 
        "\<not> P x1a" 
shows   "(filter P (suffix (nth(nfilter P (x1a\<odot>xs) 0) 0) (x1a\<odot>xs))) = (filter P xs)"
using assms 
by (metis filter.simps(2) filter_nfilter_suffix_1 interval_set_ConsD interval_suffix_zero le0)

lemma filter_suffixes_map_help_0:
assumes "j \<le> nth(nfilter P xs 0) 0 " 
        "\<exists> x \<in> set xs. P x" 
shows   "(filter P (suffix (nth(nfilter P xs 0) 0) xs)) = (filter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases "\<exists>a\<in>set xs. P a")
   case True
   then show ?thesis 
     proof (cases "P x1a")
     case True
     then show ?thesis 
     by (metis Cons.prems(1) le_zero_eq nfilter_nth_cons)
     next
     case False
     then show ?thesis 
       proof (cases j)
       case 0
       then show ?thesis 
       by (metis Cons.prems(2) False filter.simps(2) True interval_suffix_zero 
           filter_suffixes_map_help_0_0)
       next
       case (Suc nat)
       then show ?thesis 
           proof (auto split: nat.split)
           show "\<And>x. j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                  P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  x1a \<odot> filter P xs = filter P (suffix nat xs)" 
           using False by blast
           show "j = Suc nat \<Longrightarrow> 
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 P x1a \<Longrightarrow> 
                 \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                 \<langle>x1a\<rangle> = filter P (suffix nat xs)"
           using False by blast
           show "\<And>x. j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                  \<not> P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P xs = filter P (suffix nat xs)" 
            by (metis Cons.prems(1) add_leD1 nfilter.simps(2) not_one_le_zero plus_1_eq_Suc)
           show "j = Suc nat \<Longrightarrow> 
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow> 
                 \<not> P x1a \<Longrightarrow> 
                 \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                 \<langle>x1a\<rangle> = filter P (suffix nat xs)" 
           using True by blast
           show "\<And>x2 x.
                   j = Suc nat \<Longrightarrow>
                   nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                   P x1a \<Longrightarrow> 
                   x \<in> set xs \<Longrightarrow> 
                   P x \<Longrightarrow> 
                   x1a \<odot> filter P xs = filter P (suffix nat xs)" 
           using False by blast
           show "\<And>x2. 
                   j = Suc nat \<Longrightarrow>
                   nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>  
                   P x1a \<Longrightarrow>
                   \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                   \<langle>x1a\<rangle> = filter P (suffix nat xs)" 
           using False by blast
           show "\<And>x2 x.
                   j = Suc nat \<Longrightarrow>
                   nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                   \<not> P x1a \<Longrightarrow> 
                   x \<in> set xs \<Longrightarrow> 
                   P x \<Longrightarrow> 
                   filter P (suffix x2 xs) = filter P (suffix nat xs)" 
              by (metis Cons.hyps Cons.prems(1) One_nat_def Suc_le_mono add_diff_cancel_left' le0 
                  nfilter.simps(2) nfilter_nth_n_zero plus_1_eq_Suc) 
           show "\<And>x2. 
                   j = Suc nat \<Longrightarrow>
                   nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow> 
                   \<not> P x1a \<Longrightarrow>
                   \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                   \<langle>x1a\<rangle> = filter P (suffix nat xs) " 
             using True by blast
       qed
     qed
    qed
   next
   case False
   then show ?thesis 
   using Cons.prems(1) by auto
   qed
qed
    
lemma filter_suffixes_map_help_0_a:
assumes "j \<le> nth(nfilter P (suffixes xs) 0) 0 " 
        "\<exists> x \<in> set (suffixes xs). P x" 
shows   "(filter P (suffixes (suffix (nth(nfilter P (suffixes xs) 0) 0) xs))) = 
         (filter P (suffixes (suffix j xs)))"
proof -
 have 1: "(suffix (nth(nfilter P (suffixes xs) 0) 0) (suffixes xs)) =
          (suffixes (suffix (nth(nfilter P (suffixes xs) 0) 0) xs))" 
  by (metis assms(2) diff_zero interval_intlen_gr_zero nfilter_upper_bound 
      ordered_cancel_comm_monoid_diff_class.add_diff_inverse suffix_suffixes)
 have 2: "nth(nfilter P (suffixes xs) 0) 0 \<le> intlen (suffixes xs)"
    by (metis add.left_neutral assms(2) interval_intlen_gr_zero nfilter_upper_bound) 
 have 3: "(suffix j (suffixes xs)) = (suffixes (suffix j xs))" 
    using 2 suffix_suffixes assms le_trans by blast
 show ?thesis 
 using "1" "3" assms filter_suffixes_map_help_0 by fastforce
qed

lemma filter_suffixes_map_help_1:
assumes "j \<le> nth(nfilter P xs 0) 1 " 
        " 0 < intlen(filter P xs)" 
        "  nth(nfilter P xs 0) 0 < j" 
        "\<exists> x \<in> set xs. P x" 
shows   "(filter P (suffix (nth(nfilter P xs 0) 1) xs)) = (filter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
  proof (cases "\<exists>a\<in>set xs. P a")
  case True
  then show ?thesis 
    proof (cases "P x1a")
    case True
    then show ?thesis
       proof (cases j)
       case 0
       then show ?thesis 
       using Cons.prems(3) by blast
       next
       case (Suc nat)
       then show ?thesis 
         proof (auto split: nat.split)
          show "\<And>x. 
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow>
                 P x1a \<Longrightarrow> 
                 x \<in> set xs \<Longrightarrow> 
                 P x \<Longrightarrow> 
                 x1a \<odot> filter P xs = filter P (suffix nat xs)"
             by (metis Cons.prems(1) Cons.prems(3) One_nat_def add_diff_cancel_left' leD 
                 nfilter_nth_cons plus_1_eq_Suc) 
          show "j = Suc nat \<Longrightarrow>
                nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow> 
                P x1a \<Longrightarrow> 
                \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using Cons.prems(2) by auto 
          show "\<And>x. 
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow>
                 \<not> P x1a \<Longrightarrow> 
                 x \<in> set xs \<Longrightarrow> 
                 P x \<Longrightarrow> 
                 filter P xs = filter P (suffix nat xs)"
            using True by blast 
          show "j = Suc nat \<Longrightarrow>
                nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow> 
                \<not> P x1a \<Longrightarrow> \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using True by blast 
          show "\<And>x2 x.
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow>
                 P x1a \<Longrightarrow> 
                 x \<in> set xs \<Longrightarrow> 
                 P x \<Longrightarrow> 
                 x1a \<odot> filter P xs = filter P (suffix nat xs)"
            by (metis Cons.prems(1) Cons.prems(3) One_nat_def add_diff_cancel_left' leD 
                nfilter_nth_cons plus_1_eq_Suc) 
          show "\<And>x2. 
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow> 
                 P x1a \<Longrightarrow>
                 \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using Cons.prems(2) by auto 
          show "\<And>x2 x.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow>
                  \<not> P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P (suffix x2 xs) = filter P (suffix nat xs)"
            using True by blast 
          show "\<And>x2. 
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = 0 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow> 
                 \<not> P x1a \<Longrightarrow> 
                 \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                 \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using True by blast 
          show "\<And>x2 x.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow>
                  P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P (suffix x2 xs) = filter P (suffix nat xs)"
            by (metis Cons.prems(1) One_nat_def Suc_le_mono add_diff_cancel_left' 
                interval_intlen_gr_zero nfilter_nth_cons nfilter_nth_n_zero plus_1_eq_Suc 
                filter_suffixes_map_help_0) 
          show "\<And>x2. 
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow> 
                  P x1a \<Longrightarrow> 
                  \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                  \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using Cons.prems(2) by auto 
          show "\<And>x2 x.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow>
                  \<not> P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P xs = filter P (suffix nat xs)"
            using True by blast 
          show "\<And>x2. 
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow> 
                  \<not> P x1a \<Longrightarrow> 
                  \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                  \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using True by blast 
          show "\<And>x2 x2a x.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2a \<Longrightarrow>
                  P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P (suffix x2 xs) = filter P (suffix nat xs)"
           using nfilter_nth_n_zero[of _ "P" _ "Suc 0"] nfilter_nth_cons[of "P" "x1a" "xs"] 
           by (metis Cons.prems(1) One_nat_def Suc_le_mono add_diff_cancel_left' add_leD1 
              interval_intlen_gr_zero not_one_le_zero plus_1_eq_Suc filter_suffixes_map_help_0) 
         show "\<And>x2 x2a.
                 j = Suc nat \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                 nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2a \<Longrightarrow> 
                 P x1a \<Longrightarrow> 
                 \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                 \<langle>x1a\<rangle> = filter P (suffix nat xs)"
            using Cons.prems(2) by auto 
          show "\<And>x2 x2a x.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2a \<Longrightarrow>
                  \<not> P x1a \<Longrightarrow> 
                  x \<in> set xs \<Longrightarrow> 
                  P x \<Longrightarrow> 
                  filter P (suffix x2a xs) = filter P (suffix nat xs)"
            using True by blast
          show "\<And>x2 x2a.
                  j = Suc nat \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) 0 = Suc x2 \<Longrightarrow>
                  nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2a \<Longrightarrow> 
                  \<not> P x1a \<Longrightarrow> 
                  \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
                  \<langle>x1a\<rangle> = filter P (suffix nat xs) "
         using True by blast 
       qed
      qed
    next
    case False
    then show ?thesis 
      proof (auto split: nat.split)
       show "\<And>x2 x.
              \<not> P x1a \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow>
              j = Suc x2 \<Longrightarrow> 
              x \<in> set xs \<Longrightarrow> 
              P x \<Longrightarrow> 
              filter P xs = filter P (suffix x2 xs)"
         by (metis Cons.prems(2) Cons.prems(4) One_nat_def Suc_leI filter_intlen_c 
             nfilter_intlen nfilter_lower_bound not_one_le_zero) 
       show "\<And>x2. 
              \<not> P x1a \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) (Suc 0) = 0 \<Longrightarrow> 
              j = Suc x2 \<Longrightarrow> 
              \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
              \<langle>x1a\<rangle> = filter P (suffix x2 xs)"
         using True by blast 
       show "\<And>x2 x.
              \<not> P x1a \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow>
              j = 0 \<Longrightarrow> 
              x \<in> set xs \<Longrightarrow> 
              P x \<Longrightarrow> 
              filter P (suffix x2 xs) = filter P xs"
          using Cons.prems(3) by blast 
        show "\<And>x2 x2a x.
               \<not> P x1a \<Longrightarrow>
               nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow>
               j = Suc x2a \<Longrightarrow> 
               x \<in> set xs \<Longrightarrow> 
               P x \<Longrightarrow> 
               filter P (suffix x2 xs) = filter P (suffix x2a xs)"
           proof -
            fix x2
            fix x2a
            fix x
            assume a0: "\<not> P x1a" 
            assume a1: "nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2" 
            assume a2: "j = Suc x2a"  
            assume a3: "x \<in> set xs "
            assume a4: " P x"  
             show " filter P (suffix x2 xs) = filter P (suffix x2a xs)"
             proof - 
              have 1: "j>0"
                using Cons.prems(3) gr_implies_not_zero by blast
              have 2: "x2a = j-1"
                by (simp add: a2)
              have 3: "x2 = nth (nfilter P xs (Suc 0)) (Suc 0) - (Suc 0)"
                by (simp add: a1) 
              have 4: "(nth (nfilter P xs (Suc 0)) (Suc 0) - (Suc 0)) = (nth (nfilter P xs 0) (Suc 0)) "
                by (metis Cons.prems(2) Cons.prems(4) Suc_leI True a0 filter_intlen_c 
                    nfilter_intlen nfilter_nth_n_zero) 
              have 5: "0 < intlen (filter P xs)"
               using Cons.prems(2) Cons.prems(4) a0 by auto
              have 6: " \<exists>a\<in>set xs. P a"
               using True by blast
              have 7: "filter P (suffix x2 xs) = filter P (suffix (nth (nfilter P xs 0) (Suc 0)) xs)" 
                by (simp add: "3" "4")
              have 8: "filter P (suffix x2a xs) = filter P (suffix (j-1) xs)"
                 using "2" by blast 
              have 9: "j-1 \<le> nth (nfilter P xs 0) 1" 
                by (metis "2" "3" "4" Cons.prems(1) One_nat_def Suc_le_mono True a0 a1 a2 
                   nfilter_nth_cons)
              have 10: "(nth (nfilter P xs (Suc 0)) 0)- (Suc 0) = (nth (nfilter P xs 0) 0) " 
                  by (meson a3 a4 interval_intlen_gr_zero nfilter_nth_n_zero)
              have 101: "(nth (nfilter P xs 0) 0)  < j - (Suc 0)"
                 using 10 2 a0 6 a2 Cons.prems nfilter_nth_cons[of "P" "x1a" "xs" _ _] 
                 filter_nth_aa[of "xs" "P" "j-1"]
                 nfilter_nth_n_zero[of "xs" "P"] nfilter_holds[of "x1a\<odot>xs" "P"] 
                  by simp_all 
                  (metis "2" One_nat_def diff_less_mono interval_intlen_gr_zero nfilter_lower_bound)
              have 11: "filter P (suffix (nth (nfilter P xs 0) 1) xs) = filter P (suffix (j-1) xs)"
                using "101" "5" "9" Cons.hyps True by simp
               show ?thesis using 3 4 5 6 Cons.hyps Cons.prems 
               by (metis "11" "2" One_nat_def)
             qed
            qed
       show "\<And>x2 x2a.
              \<not> P x1a \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2 \<Longrightarrow>
               j = Suc x2a \<Longrightarrow>
               \<forall>x\<in>set xs. \<not> P x \<Longrightarrow> 
               \<langle>x1a\<rangle> = filter P (suffix x2a xs)"
         using True by blast 
   qed
  qed
  next
  case False
  then show ?thesis 
  using Cons.prems(2) by auto
  qed
qed

lemma filter_suffixes_map_help_j_a:
assumes " (\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
               i < intlen (filter P xs) \<Longrightarrow>
               nth (nfilter P xs 0) i < j \<Longrightarrow>
               filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs))"
       " x2a \<le> x2 "
       "i < Suc (intlen (filter P xs)) "
       "(case i of 0 \<Rightarrow> 0 | Suc k \<Rightarrow> nth (nfilter P xs (Suc 0)) k) < Suc x2a "
       "P x1a "
       "x \<in> set xs "
       "P x "
       "nth (nfilter P xs (Suc 0)) i = Suc x2"
       " j = Suc x2a"
shows  " filter P (suffix x2 xs) = filter P (suffix x2a xs)"
proof -
 have 1: "i=0 \<longrightarrow> filter P (suffix x2 xs) = filter P (suffix x2a xs)"
    using assms 
    by (metis One_nat_def add_diff_cancel_left' interval_intlen_gr_zero nfilter_nth_n_zero 
        plus_1_eq_Suc filter_suffixes_map_help_0)
 have 2: "i>0\<longrightarrow> (nth (nfilter P xs (Suc 0)) (i-1)) - (Suc 0) = nth (nfilter P xs 0) (i-1)  "
   by (metis One_nat_def Suc_leI add_leD1 assms(3) assms(6) assms(7) le_add_diff_inverse2 
       less_Suc_eq_le nfilter_intlen nfilter_nth_n_zero)
 have 21: "i>0 \<longrightarrow>  nth (nfilter P xs (Suc 0)) (i-1) <Suc  x2a"
      using assms(4) 
      by (metis Nitpick.case_nat_unfold less_Suc0 not_less_eq)
 have 22: "i>0 \<longrightarrow> nth (nfilter P xs (Suc 0)) (i-1) >0" 
   by (metis One_nat_def Suc_leI add_leD1 assms(3) assms(6) assms(7) gr_zeroI le_add_diff_inverse2
        less_Suc_eq_le nfilter_intlen nfilter_lower_bound not_one_le_zero)
 have 23: "i>0 \<longrightarrow> nth (nfilter P xs 0) (i-1) < x2a"
   using "2" "21" "22" by linarith
 have 24: "i>0 \<longrightarrow> x2a \<le> nth (nfilter P xs 0) i"
   by (metis One_nat_def add_diff_cancel_left' assms(2) assms(3) assms(6) assms(7) assms(8) 
       less_Suc_eq_le nfilter_intlen nfilter_nth_n_zero plus_1_eq_Suc) 
 have 25: "i>0 \<longrightarrow> i - 1 < intlen (filter P xs) " 
 using assms(3) by linarith
 have 3: "i>0 \<longrightarrow> filter P (suffix (nth (nfilter P xs 0) i) xs ) = filter P (suffix x2a xs)"
    by (metis "23" "24" "25" One_nat_def Suc_pred assms(1))
 have 4: "i>0 \<longrightarrow>  (nth (nfilter P xs 0) i) = x2"
   by (metis "25" One_nat_def Suc_leI Suc_pred add_diff_cancel_left' assms(6) assms(7) assms(8) 
       nfilter_intlen nfilter_nth_n_zero plus_1_eq_Suc)
 from 1 2 3 4 show ?thesis by auto
qed

lemma filter_suffixes_map_help_j_b:
assumes "(\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
               i < intlen (filter P xs) \<Longrightarrow>
               nth (nfilter P xs 0) i < j \<Longrightarrow>
               filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs))"
       " x2a \<le> x2 "
       "i < intlen (filter P xs) "
       "nth (nfilter P xs (Suc 0)) i < Suc x2a "
       "\<not> P x1a "
       "x \<in> set xs "
       "P x "
       "nth (nfilter P xs (Suc 0)) (Suc i) = Suc x2 "
       "j = Suc x2a "
shows " filter P (suffix x2 xs) = filter P (suffix x2a xs)"
proof -
  have 0: "i=0 \<longrightarrow> (nth (nfilter P xs (Suc 0)) (Suc i)) - (Suc 0) = nth (nfilter P xs 0) (Suc i)"
   by (metis Suc_leI assms(3) assms(6) assms(7) nfilter_intlen nfilter_nth_n_zero)
  have 00: "i=0 \<longrightarrow> (nth (nfilter P xs 0) (Suc i)) = x2"
      using "0" assms(8) by linarith
  have 03: "i=0 \<longrightarrow>  nth (nfilter P xs 0) i < x2a"
    by (metis One_nat_def add_diff_cancel_left' assms(4) assms(6) assms(7) interval_intlen_gr_zero 
        less_Suc_eq_0_disj nfilter_lower_bound nfilter_nth_n_zero not_one_le_zero plus_1_eq_Suc)
  have 04: "i=0 \<longrightarrow> i < intlen (filter P xs)"
    using assms(3) by blast
  have 05: "i=0 \<longrightarrow> x2a \<le> nth (nfilter P xs 0) (Suc i)" 
      using "00" assms(2) by blast 
  have 1: "i=0 \<longrightarrow> filter P (suffix x2 xs) = filter P (suffix x2a xs)"
    using "00" "03" "05" assms(1) assms(3) by blast
   have 2: "i>0\<longrightarrow> (nth (nfilter P xs (Suc 0)) (i)) - (Suc 0) = nth (nfilter P xs 0) (i)  "
   by (metis  assms(3) assms(6) assms(7) less_imp_le_nat nfilter_intlen nfilter_nth_n_zero)  
  have 21: "i>0 \<longrightarrow>  nth (nfilter P xs (Suc 0)) i <Suc  x2a"
      using assms(4) by auto
  have 22: "i>0 \<longrightarrow> nth (nfilter P xs (Suc 0)) (i) >0"
    by (metis assms(3) assms(6) assms(7) gr_zeroI idx_nfilter_gr less_imp_le_nat 
       nfilter_intlen not_less0) 
  have 23: "i>0 \<longrightarrow> nth (nfilter P xs 0) (i) < x2a" 
   using "2" "22" assms(4) by linarith
  have 24: "i>0 \<longrightarrow> x2a \<le> nth (nfilter P xs 0) (Suc i)"
     using assms by (metis One_nat_def Suc_leI add_diff_cancel_left'  nfilter_intlen 
      nfilter_nth_n_zero plus_1_eq_Suc) 
  have 25: "i>0 \<longrightarrow> i  < intlen (filter P xs)"
    by (simp add: assms(3))
  have 3: "i>0 \<longrightarrow>  filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs ) = filter P (suffix x2a xs)"
    using "23" "24" assms(1) assms(3) by blast
  have 4: "i>0 \<longrightarrow> (nth (nfilter P xs 0) (Suc i)) = x2"
   using assms  by (metis One_nat_def Suc_leI add_diff_cancel_left' nfilter_intlen 
    nfilter_nth_n_zero plus_1_eq_Suc) 
  from 1 2 3 4 show ?thesis by auto
qed 

lemma filter_suffixes_map_help_j:
assumes "j \<le> nth(nfilter P xs 0) (Suc i) " 
        " i < intlen(filter P xs)" 
        "  nth(nfilter P xs 0) i < j" 
        "\<exists> x \<in> set xs. P x" 
shows   "(filter P (suffix (nth(nfilter P xs 0) (Suc i)) xs)) = (filter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j i)
case (St x)
then show ?case by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases "(\<exists>a\<in>set xs. P a)")
    show "(\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
            i < intlen (filter P xs) \<Longrightarrow>
            nth (nfilter P xs 0) i < j \<Longrightarrow>
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs)) \<Longrightarrow>
          j \<le> nth (nfilter P (x1a \<odot> xs) 0) (Suc i) \<Longrightarrow>
          i < intlen (filter P (x1a \<odot> xs)) \<Longrightarrow>
          nth (nfilter P (x1a \<odot> xs) 0) i < j \<Longrightarrow>
          \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
          \<exists>a\<in>set xs. P a \<Longrightarrow>
          filter P (suffix (nth (nfilter P (x1a \<odot> xs) 0) (Suc i)) (x1a \<odot> xs)) = 
          filter P (suffix j (x1a \<odot> xs))" 
     proof (cases "P x1a") 
      show "(\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
               i < intlen (filter P xs) \<Longrightarrow>
               nth (nfilter P xs 0) i < j \<Longrightarrow>
               \<exists>a\<in>set xs. P a \<Longrightarrow>
               filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs)) \<Longrightarrow>
            j \<le> nth (nfilter P (x1a \<odot> xs) 0) (Suc i) \<Longrightarrow>
            i < intlen (filter P (x1a \<odot> xs)) \<Longrightarrow>
            nth (nfilter P (x1a \<odot> xs) 0) i < j \<Longrightarrow>
            \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            P x1a \<Longrightarrow> 
            filter P (suffix (nth (nfilter P (x1a \<odot> xs) 0) (Suc i)) (x1a \<odot> xs)) = 
            filter P (suffix j (x1a \<odot> xs))"
      using filter_suffixes_map_help_j_a by (auto split: nat.split) fastforce
     show "(\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
                 i < intlen (filter P xs) \<Longrightarrow>
                 nth (nfilter P xs 0) i < j \<Longrightarrow>
                 \<exists>a\<in>set xs. P a \<Longrightarrow>
                 filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs)) \<Longrightarrow>
            j \<le> nth (nfilter P (x1a \<odot> xs) 0) (Suc i) \<Longrightarrow>
            i < intlen (filter P (x1a \<odot> xs)) \<Longrightarrow>
            nth (nfilter P (x1a \<odot> xs) 0) i < j \<Longrightarrow>
            \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
            \<exists>a\<in>set xs. P a \<Longrightarrow>
            \<not> P x1a \<Longrightarrow>
            filter P (suffix (nth (nfilter P (x1a \<odot> xs) 0) (Suc i)) (x1a \<odot> xs)) = 
            filter P (suffix j (x1a \<odot> xs)) "
      proof (auto split: nat.split)
       show "\<And>x x2 x2a.
              (\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
                   i < intlen (filter P xs) \<Longrightarrow>
                   nth (nfilter P xs 0) i < j \<Longrightarrow>
                   filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs)) \<Longrightarrow>
              x2a \<le> x2 \<Longrightarrow>
              i < intlen (filter P xs) \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) i < Suc x2a \<Longrightarrow>
              \<not> P x1a \<Longrightarrow>
              x \<in> set xs \<Longrightarrow>
              P x \<Longrightarrow>
              nth (nfilter P xs (Suc 0)) (Suc i) = Suc x2 \<Longrightarrow>
              j = Suc x2a \<Longrightarrow> 
              filter P (suffix x2 xs) = filter P (suffix x2a xs) "
         using filter_suffixes_map_help_j_b[of "P" "xs" ]  
   by blast
   qed
  qed
  show "(\<And>j i. j \<le> nth (nfilter P xs 0) (Suc i) \<Longrightarrow>
             i < intlen (filter P xs) \<Longrightarrow>
             nth (nfilter P xs 0) i < j \<Longrightarrow>
             \<exists>a\<in>set xs. P a \<Longrightarrow>
             filter P (suffix (nth (nfilter P xs 0) (Suc i)) xs) = filter P (suffix j xs)) \<Longrightarrow>
        j \<le> nth (nfilter P (x1a \<odot> xs) 0) (Suc i) \<Longrightarrow>
        i < intlen (filter P (x1a \<odot> xs)) \<Longrightarrow>
        nth (nfilter P (x1a \<odot> xs) 0) i < j \<Longrightarrow>
        \<exists>a\<in>set (x1a \<odot> xs). P a \<Longrightarrow>
        \<not> (\<exists>a\<in>set xs. P a) \<Longrightarrow>
        filter P (suffix (nth (nfilter P (x1a \<odot> xs) 0) (Suc i)) (x1a \<odot> xs)) = 
        filter P (suffix j (x1a \<odot> xs)) "
    by auto 
   qed
qed

lemma filter_suffixes_map_help_j_aa:
assumes "j \<le> nth(nfilter P (suffixes xs) 0) (Suc i) " 
        " i < intlen(filter P (suffixes xs))" 
        "  nth(nfilter P (suffixes xs) 0) i < j" 
        "\<exists> x \<in> set (suffixes xs). P x" 
shows   "(filter P (suffixes (suffix (nth(nfilter P (suffixes xs) 0) (Suc i)) xs))) = 
         (filter P (suffixes (suffix j xs)))"
proof -
 have 1: "nth(nfilter P (suffixes xs) 0) (Suc i) \<le> intlen (suffixes xs)"
   by (metis Suc_leI add_cancel_right_left assms(2) assms(4) nfilter_intlen nfilter_upper_bound)
 have 2: "(suffixes (suffix (nth(nfilter P (suffixes xs) 0) (Suc i)) xs)) =
          (suffix (nth(nfilter P (suffixes xs) 0) (Suc i)) (suffixes xs))"
   using "1" suffix_suffixes by fastforce
 have 3: "(suffixes (suffix j xs)) = (suffix j (suffixes xs))"
    using "1" assms(1) le_trans suffix_suffixes by fastforce 
 show ?thesis 
 by (simp add: "2" "3" assms(1) assms(2) assms(3) assms(4) filter_suffixes_map_help_j)
qed

lemma filter_suffixes_map:
assumes "(Suc i) \<le> intlen (filter f (suffixes \<sigma>)) " 
        "\<exists> x \<in> set(suffixes \<sigma>). f x" 
shows   "(suffix (Suc i) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) =
           (map (\<lambda>s. nth s 0)
                (filter f (suffixes (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) i)) \<sigma>))))"
proof -
 have 1: "(suffix (Suc i) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) =
          (map (\<lambda>s. nth s 0) (suffix (Suc i) (filter f (suffixes \<sigma>))))"
  by (simp add: assms(1) map_suffix) 
 have 2: "(Suc (nth (nfilter f (suffixes \<sigma>) 0) i)) \<le> intlen(suffixes \<sigma>)"
      by (metis length_filter_le Suc_leD assms(1) assms(2) diff_is_0_eq diff_zero 
          filter_nfilter_suffix_1 interval_suffix_length not_less_eq_eq)
 have 3: "(filter f (suffixes (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) i)) \<sigma>))) =
          (filter f (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) i)) (suffixes \<sigma>))) 
         " 
     using "2" suffix_suffixes by fastforce
 have 4: "(suffix (Suc i) (filter f (suffixes \<sigma>))) =
          (filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i))  (suffixes \<sigma>))) 
         " 
    by (simp add: assms(1) assms(2) filter_nfilter_suffix_1)
 have 5: "(Suc (nth (nfilter f (suffixes \<sigma>) 0) i)) \<le> nth(nfilter f (suffixes \<sigma>) 0) (Suc i)"
    by (simp add: Suc_leI Suc_le_lessD assms(1) assms(2) idx_nfilter_mono nfilter_intlen)
 have 6: " i < intlen(filter f (suffixes \<sigma>))"
    using Suc_le_eq assms(1) by blast 
 have 7: "nth (nfilter f (suffixes \<sigma>) 0) i  < (Suc (nth (nfilter f (suffixes \<sigma>) 0) i))"
    by simp
 have 8: "(filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i))  (suffixes \<sigma>))) = 
          (filter f (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) i))  (suffixes \<sigma>)))
         " 
  using 5 6 7 
  filter_suffixes_map_help_j[of "(Suc (nth (nfilter f (suffixes \<sigma>) 0) i))" "f" "suffixes \<sigma>" "i" ]
     assms(2) by blast 
 show ?thesis 
 by (simp add: "1" "3" "4" "8")
qed          
  
lemma sfxfilter_suffix_intlen:
 assumes "k\<le>intlen \<sigma>"
         "\<exists>x \<in> set (suffixes(suffix k \<sigma>)). f x"
 shows   "intlen (filter f (suffixes (suffix k \<sigma>))) \<le> intlen (filter f (suffixes \<sigma>))" 
using assms
proof (induct k arbitrary: \<sigma>)
case 0
then show ?case by  simp
next
case (Suc k)
then show ?case 
  proof (cases \<sigma>)
  case (St x1)
  then show ?thesis 
  by simp
  next
  case (Cons x21 x22)
  then show ?thesis 
     proof auto
      show "\<And>y. 
             \<sigma> = x21 \<odot> x22 \<Longrightarrow>
             f (x21 \<odot> x22) \<Longrightarrow>
             osfx y x22 \<Longrightarrow> 
             f y \<Longrightarrow> 
             intlen (filter f (suffixes (suffix k x22))) \<le> Suc (intlen (filter f (suffixes x22)))" 
        by (metis length_filter_le Suc.hyps Suc.prems(2) interval_suffix_length_code 
            interval_suffix_suc intlen_suffixes le_0_eq le_SucI less_imp_le_nat zero_less_Suc)
      show "\<sigma> = x21 \<odot> x22 \<Longrightarrow>
            f (x21 \<odot> x22) \<Longrightarrow> 
            \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
            intlen (filter f (suffixes (suffix k x22))) = 0"
            using Suc.prems osfx_suffix by fastforce 
      show "\<And>y. \<sigma> = x21 \<odot> x22 \<Longrightarrow>
             \<not> f (x21 \<odot> x22) \<Longrightarrow>
             osfx y x22 \<Longrightarrow> 
             f y \<Longrightarrow> 
             intlen (filter f (suffixes (suffix k x22))) \<le> intlen (filter f (suffixes x22))" 
        by (metis length_filter_le Suc.hyps Suc.prems(2) interval_intlen_gr_zero 
            interval_suffix_length_code interval_suffix_suc intlen_suffixes le_0_eq)
      show "\<sigma> = x21 \<odot> x22 \<Longrightarrow> 
           \<not> f (x21 \<odot> x22) \<Longrightarrow> 
           \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
           intlen (filter f (suffixes (suffix k x22))) = 0 "
       using Suc.prems osfx_suffix by fastforce 
     qed
  qed
qed

lemma sfxfilter_suffix_nth:
 assumes "k\<le>intlen \<sigma>"
         "\<exists>x \<in> set (suffixes(suffix k \<sigma>)). f x"
         "j\<le> intlen (filter f (suffixes (suffix k \<sigma>)))"
 shows   "(nth (filter f (suffixes (suffix k \<sigma>))) j) =
             (nth (suffix (intlen(filter f (suffixes \<sigma>))-intlen((filter f (suffixes (suffix k \<sigma>)))))
                   (filter f (suffixes \<sigma>))) j)" 
using assms 
proof (induct k arbitrary: \<sigma>)
case 0
then show ?case by simp
next
case (Suc k)
then show ?case 
   proof (cases \<sigma>)
   case (St x1)
   then show ?thesis 
   by simp
   next
   case (Cons x21 x22)
   then show ?thesis 
      proof (auto split: nat.split)
       show "\<And>y. 
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = 0 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) \<le> 
              intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
              f (x21 \<odot> x22) \<Longrightarrow> 
              osfx y x22 \<Longrightarrow> 
              f y \<Longrightarrow> 
              nth (filter f (suffixes (suffix k x22))) 0 = x21 \<odot> x22"
              using Suc.prems not_less_eq_eq sfxfilter_suffix_intlen by fastforce
       show "\<sigma> = x21 \<odot> x22 \<Longrightarrow>
             j = 0 \<Longrightarrow>
             Suc (intlen (filter f (suffixes x22))) \<le> 
             intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
             f (x21 \<odot> x22) \<Longrightarrow> 
             \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
             nth (filter f (suffixes (suffix k x22))) 0 = x21 \<odot> x22"
             using Suc.prems osfx_order.order.trans by fastforce     
     show "\<And>y. 
            \<sigma> = x21 \<odot> x22 \<Longrightarrow>
            j = 0 \<Longrightarrow>
            Suc (intlen (filter f (suffixes x22))) \<le>
            intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
            \<not> f (x21 \<odot> x22) \<Longrightarrow>
            osfx y x22 \<Longrightarrow> 
            f y \<Longrightarrow> 
            nth (filter f (suffixes (suffix k x22))) 0 = nth (filter f (suffixes x22)) 0"
          using Suc.hyps Suc.prems by auto
      show "\<sigma> = x21 \<odot> x22 \<Longrightarrow>
            j = 0 \<Longrightarrow>
            Suc (intlen (filter f (suffixes x22))) \<le>
            intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
            \<not> f (x21 \<odot> x22) \<Longrightarrow> 
            \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
            nth (filter f (suffixes (suffix k x22))) 0 = x21 \<odot> x22"
            using Suc.prems osfx_order.order.trans by fastforce
      show "\<And>x2 y.
             \<sigma> = x21 \<odot> x22 \<Longrightarrow>
             j = 0 \<Longrightarrow>
             Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) =
             Suc x2 \<Longrightarrow>
             f (x21 \<odot> x22) \<Longrightarrow>
             osfx y x22 \<Longrightarrow> 
             f y \<Longrightarrow> 
             nth (filter f (suffixes (suffix k x22))) 0 = nth (filter f (suffixes x22)) x2" 
         using Suc.hyps Suc.prems 
         by (metis Suc_diff_diff Suc_le_mono add.right_neutral diff_zero interval_nth_suffix 
             interval_suffix_suc intlen.simps(2) le0 plus_1_eq_Suc)
      show "\<And>x2. 
             \<sigma> = x21 \<odot> x22 \<Longrightarrow>
             j = 0 \<Longrightarrow>
             Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) = 
             Suc x2 \<Longrightarrow>
             f (x21 \<odot> x22) \<Longrightarrow> 
             \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
             nth (filter f (suffixes (suffix k x22))) 0 = x21 \<odot> x22" 
       using Suc.prems osfx_suffix by fastforce
       show "\<And>x2 y.
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = 0 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) = 
              Suc x2 \<Longrightarrow>
              \<not> f (x21 \<odot> x22) \<Longrightarrow>
              osfx y x22 \<Longrightarrow>
              f y \<Longrightarrow>
              nth (filter f (suffixes (suffix k x22))) 0 =
              nth (filter f (suffixes x22))
              (intlen (filter f (suffixes x22)) - intlen (filter f (suffixes (suffix k x22))))"
         using Suc.hyps Suc.prems by auto 
       show "\<And>x2. 
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = 0 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) =
              Suc x2 \<Longrightarrow>
              \<not> f (x21 \<odot> x22) \<Longrightarrow>
              \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
              nth (filter f (suffixes (suffix k x22))) 0 = x21 \<odot> x22"
         using Suc.prems osfx_suffix by fastforce 
       show "\<And>x2 y.
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = Suc x2 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) \<le> 
              intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
              f (x21 \<odot> x22) \<Longrightarrow>
              osfx y x22 \<Longrightarrow>
              f y \<Longrightarrow> 
              nth (filter f (suffixes (suffix k x22))) (Suc x2) = nth (filter f (suffixes x22)) x2"
              using Suc.prems not_less_eq_eq sfxfilter_suffix_intlen by fastforce
       show "\<And>x2. 
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = Suc x2 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) \<le> 
              intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
              f (x21 \<odot> x22) \<Longrightarrow> 
              \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
              nth (filter f (suffixes (suffix k x22))) (Suc x2) = x21 \<odot> x22"
              using Suc.prems osfx_order.order.trans by fastforce
        show "\<And>x2 y.
               \<sigma> = x21 \<odot> x22 \<Longrightarrow>
               j = Suc x2 \<Longrightarrow>
               Suc (intlen (filter f (suffixes x22))) \<le> 
               intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
               \<not> f (x21 \<odot> x22) \<Longrightarrow>
               osfx y x22 \<Longrightarrow>
               f y \<Longrightarrow> 
               nth (filter f (suffixes (suffix k x22))) (Suc x2) = 
               nth (filter f (suffixes x22)) (Suc x2)"
               using Suc.prems not_less_eq_eq sfxfilter_suffix_intlen by fastforce
        show "\<And>x2. 
               \<sigma> = x21 \<odot> x22 \<Longrightarrow>
               j = Suc x2 \<Longrightarrow>
               Suc (intlen (filter f (suffixes x22))) \<le> 
               intlen (filter f (suffixes (suffix k x22))) \<Longrightarrow>
               \<not> f (x21 \<odot> x22) \<Longrightarrow> \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
               nth (filter f (suffixes (suffix k x22))) (Suc x2) = x21 \<odot> x22"
               using Suc.prems osfx_order.order.trans by fastforce
       show "\<And>x2 x2a y.
               \<sigma> = x21 \<odot> x22 \<Longrightarrow>
               j = Suc x2 \<Longrightarrow>
               Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) = 
               Suc x2a \<Longrightarrow>
               f (x21 \<odot> x22) \<Longrightarrow>
               osfx y x22 \<Longrightarrow>
               f y \<Longrightarrow>
               nth (filter f (suffixes (suffix k x22))) (Suc x2) = 
               nth (suffix x2a (filter f (suffixes x22))) (Suc x2)" 
            using Suc.hyps Suc.prems
            by (metis  Suc_diff_diff diff_zero interval_suffix_suc intlen.simps(2) not_less_eq_eq 
                plus_1_eq_Suc)
       show "\<And>x2 x2a.
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = Suc x2 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) =
              Suc x2a \<Longrightarrow>
              f (x21 \<odot> x22) \<Longrightarrow> 
              \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
              nth (filter f (suffixes (suffix k x22))) (Suc x2) = x21 \<odot> x22"
              using Suc.prems osfx_suffix by fastforce 
       show "\<And>x2 x2a y.
              \<sigma> = x21 \<odot> x22 \<Longrightarrow>
              j = Suc x2 \<Longrightarrow>
              Suc (intlen (filter f (suffixes x22))) - intlen (filter f (suffixes (suffix k x22))) = 
              Suc x2a \<Longrightarrow>
             \<not> f (x21 \<odot> x22) \<Longrightarrow>
              osfx y x22 \<Longrightarrow>
              f y \<Longrightarrow>
              nth (filter f (suffixes (suffix k x22))) (Suc x2) =
              nth (suffix (intlen (filter f (suffixes x22)) - 
                           intlen (filter f (suffixes (suffix k x22))))
                  (filter f (suffixes x22)))
                  (Suc x2)" 
       using Suc.hyps Suc.prems by auto
     show "\<And>x2 x2a.
            \<sigma> = x21 \<odot> x22 \<Longrightarrow>
            j = Suc x2 \<Longrightarrow>
            Suc (intlen (filter f (suffixes x22))) - 
                 intlen (filter f (suffixes (suffix k x22))) = Suc x2a \<Longrightarrow>
            \<not> f (x21 \<odot> x22) \<Longrightarrow> 
            \<forall>x\<in>set (suffixes x22). \<not> f x \<Longrightarrow> 
            nth (filter f (suffixes (suffix k x22))) (Suc x2) = x21 \<odot> x22 " 
       using Suc.prems osfx_suffix by fastforce
     qed 
   qed
qed

lemma sfxfilter_suffix_suffix:
 assumes "k\<le>intlen \<sigma>"
         "\<exists>x \<in> set (suffixes(suffix k \<sigma>)). f x"
 shows   " (filter f (suffixes (suffix k \<sigma>))) = 
           (suffix (intlen(filter f (suffixes \<sigma>))-intlen((filter f (suffixes (suffix k \<sigma>)))))
                   (filter f (suffixes \<sigma>)))"
by (simp add: assms(1) assms(2) interval_eq_nth_eq sfxfilter_suffix_intlen sfxfilter_suffix_nth)

lemma sfxfilter_suffix_suffix_a:
assumes "jj\<le>intlen (filter f (suffixes xs))"
         "\<exists>x \<in> set (suffixes xs). f x"
shows "(suffix jj (filter f (suffixes xs)))  =
        filter f (suffixes (suffix ((intlen xs) - intlen (nth (filter f (suffixes xs)) jj))   xs))"

proof -
 have 1: "(suffix jj (filter f (suffixes xs))) = 
           filter f (suffix (nth (nfilter f (suffixes xs) 0) jj) (suffixes xs))" 
      using filter_nfilter_suffix_1 
      by (simp add: filter_nfilter_suffix_1 assms(1) assms(2))
 have 2: "(nth (nfilter f (suffixes xs) 0) jj) \<le> intlen (suffixes xs)"
   by (metis add_cancel_right_left assms(1) assms(2) nfilter_intlen nfilter_upper_bound) 
 have 3: "(suffix (nth (nfilter f (suffixes xs) 0) jj) (suffixes xs)) =
           (suffixes (suffix (nth (nfilter f (suffixes xs) 0) jj) xs))"
   using suffix_suffixes[of "(nth (nfilter f (suffixes xs) 0) jj)" "xs" ] "2" by blast
 have 4: "(nth (filter f (suffixes xs)) jj) = nth (suffixes xs) (nth (nfilter f (suffixes xs) 0) jj) "
    by (metis nfilter_map_filter assms(2) interval_nth_map)
 have 5: "intlen(suffixes xs) \<le>  intlen (xs)" 
    by simp
 have 6: " (nth (nfilter f (suffixes xs) 0) jj) \<le> intlen (xs)" 
  using 2 5 le_trans by blast
 have 7: "nth (suffixes xs) (nth (nfilter f (suffixes xs) 0) jj) =
          (suffix (nth (nfilter f (suffixes xs) 0) jj) xs)"
   by (simp add: "6" nth_suffixes)  
 have 8: "(nth (nfilter f (suffixes xs) 0) jj) =
          ((intlen xs) - intlen (nth (filter f (suffixes xs)) jj))" 
   by (simp add: "4" "6" "7")
 show ?thesis 
 using "1" "3" "8" by auto
qed
 
lemma sfx_suffix_upperbound:
 "(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))). 
              ((intlen \<sigma>) - intlen (nth (filter f (suffixes \<sigma>)) jj)) < k
            )" 
proof 
 fix jj
 show "jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
          intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k "  
 proof (induct k arbitrary: \<sigma> jj)
   show "\<And>\<sigma> jj.
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix 0 \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < 0"
    by simp
   show "\<And>k \<sigma> jj.
       (\<And>\<sigma> jj.
           jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
           intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k) \<Longrightarrow>
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix (Suc k) \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < Suc k "
    proof  (case_tac \<sigma>)
      show "\<And>k \<sigma> jj x1.
       (\<And>\<sigma> jj.
           jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
           intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k) \<Longrightarrow>
       \<sigma> = \<langle>x1\<rangle> \<Longrightarrow>
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix (Suc k) \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < Suc k" 
          by simp
       show "\<And>k \<sigma> jj x21 x22.
       (\<And>\<sigma> jj.
           jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
           intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k) \<Longrightarrow>
       \<sigma> = x21 \<odot> x22 \<Longrightarrow>
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix (Suc k) \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < Suc k "
       proof (case_tac jj)
        show "\<And>k \<sigma> jj x21 x22.
       (\<And>\<sigma> jj.
           jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
           intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k) \<Longrightarrow>
       \<sigma> = x21 \<odot> x22 \<Longrightarrow>
       jj = 0 \<Longrightarrow>
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix (Suc k) \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < Suc k" 
               by auto
                  (metis Suc_diff_le Suc_mono in_set_suffixes interval_intlen_gr_zero 
                   sfxfilter_nth_bound zero_less_diff)
         show "\<And>k \<sigma> jj x21 x22 nat.
       (\<And>\<sigma> jj.
           jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
           intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < k) \<Longrightarrow>
       \<sigma> = x21 \<odot> x22 \<Longrightarrow>
       jj = Suc nat \<Longrightarrow>
       jj < intlen (filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix (Suc k) \<sigma>))) \<longrightarrow>
       intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) jj) < Suc k " 
         using Suc_diff_le less_Suc_eq_0_disj by auto
     qed
   qed
  qed
qed


end
